/*
 *  kernel_unitests.cpp
 *  gpu_hierleap
 *
 *  Created by David Orendorff on 5/18/10.
 *  Copyright 2010 UC Irvine. All rights reserved.
 *
 *  This standalone executable aims to test the hierleap kernels via unit tests.
 */

#include <stdio.h>
#include <assert.h>
#include <sys/stat.h>
#include <math.h>
#include <vector.h>

#include <OpenCL/OpenCL.h>

#include "../utils/gpu_interface.h"

#define EPS 0.0001
#define MAX_SPECIES_RXN_CON 8 

bool aboutEqual(float a, float b) {
	float d = fabs(a-b);
	return d <= fabs(max(a,b)) / 1000.0;
}

char * load_program_source(const char *filename)
{ 
	
	struct stat statbuf;
	FILE *fh; 
	char *source; 
	
	fh = fopen(filename, "r");
	if (fh == 0)
		return 0; 
	
	stat(filename, &statbuf);
	source = (char *) malloc(statbuf.st_size + 1);
	fread(source, statbuf.st_size, 1, fh);
	source[statbuf.st_size] = '\0'; 
	
	return source; 
} 

unsigned int urand() {
	// a better way to randomize all bits in uint 
	//return 0xFFFFFFFF;
	return rand() ^ (rand() << 16) ^ (rand() >> 16); 
}

// Remember, delete returned vector of values.
unsigned int *generateRandomUInts(int nToGen) {
	assert(nToGen > 0);
	unsigned int* rands;
	rands = new unsigned int[nToGen];
	for (int i = 0; i < nToGen; i++) {
		rands[i] = urand();
	}
	return rands;
}

/* Reacall prototype for computePropensities
 __kernel
 void computePropensities(__global const int *state, // const
 int stateSize,
 __global const int *cellStartIdx, // tuple for each reaction set-1: (start index in D, end idx exclusive)
 __global const int2 *inpSpec, // Assume each rxn has at most N=2 inp species
 __global const float *rates,	// TODO: maybe cache to local
 __local int *locState;  // a local copy of state used for fast referencing
 __global float *D,
 __local float *Dloc)    // cache in which to compute D, will write to global at end
 { */
#pragma mark -
bool test_computePropensities(cl_device_id cid, cl_command_queue cmd_queue, cl_context context,
							  cl_program program, cl_kernel kernel){

	/*  Build Inputs to Kernel */
	
		// allocate program memory
		int state[] = {0, 1,2,3,4,5,6,7,8,9};
		
		// there are 3 rxnSets, etch with {1,2,3} base_rxns respectively 
		//D = {root, p0, p1, p2, r00, r10, r11, r20, r21, r22}
		int cellStartIdx[] = {4, 5, 7, 10}; // (10 is end of array)
		cl_int2 inpSpecies[] = {{0, 1}, {1,2},{3,4},{5,5},{9,-1}, {-1,-1}}; // start at r0
		cl_float rates[] = {1.0, 1.0, 1.0, .5, .5, .5};
		float correctD[] = {24, 0, 14, 10, 0,2,12,5,4.5,.5};//{root, pi0, pi1, pi2, 0, 1*2*1, 3*4*1, 5*4 * .5 * .5, 9*.5, .5};
		float answD[10];
		
		int nRxns = 10; 
		int nStates = 10;
		int nCells = 3;
		int nBaseRxns = nRxns - nCells -1;
	
	
	/* Allocate Memory and Set Arguments */
	
		// allocate mem, and initialize (if required)
		size_t buffer_size;
		cl_int err;
		
		// Input array state
		buffer_size = sizeof(state);
		cl_mem state_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, state_mem, CL_TRUE, 0, buffer_size,
								   (void*)state, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  0, sizeof(cl_mem), &state_mem);
		assert(err == CL_SUCCESS);
		
		// Input stateSize
		err |= clSetKernelArg(kernel,  1, sizeof(int), &nStates);
		
		// Input cellStartIdx
		buffer_size = sizeof(cellStartIdx);
		cl_mem cellStart_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err |= clEnqueueWriteBuffer(cmd_queue, cellStart_mem, CL_TRUE, 0, buffer_size,
								   (void*)cellStartIdx, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  2, sizeof(cl_mem), &cellStart_mem);
		assert(err == CL_SUCCESS);
		
		// Input inpSpec
		buffer_size = sizeof(cl_int2) * nBaseRxns;
		cl_mem inpSpec_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err |= clEnqueueWriteBuffer(cmd_queue, inpSpec_mem, CL_TRUE, 0, buffer_size,
									(void*)inpSpecies, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  3, sizeof(cl_mem), &inpSpec_mem);
		assert(err == CL_SUCCESS);
		
		// Input rates
		buffer_size = sizeof(float) * nBaseRxns;
		cl_mem rates_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err |= clEnqueueWriteBuffer(cmd_queue, rates_mem, CL_TRUE, 0, buffer_size,
									(void*)rates, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  4, sizeof(cl_mem), &rates_mem);
		assert(err == CL_SUCCESS);
		
		// Input (local) state
		err |= clSetKernelArg(kernel,  5, sizeof(state), NULL);
		assert(err == CL_SUCCESS);	
	
		// Input D
		buffer_size = sizeof(correctD);
		cl_mem D_mem = clCreateBuffer(context, CL_MEM_WRITE_ONLY, buffer_size, NULL, NULL);
		err |= clSetKernelArg(kernel,  6, sizeof(cl_mem), &D_mem);
		assert(err == CL_SUCCESS);	
		
		// Input Dlocal, same buffer_size as D
		err |= clSetKernelArg(kernel,  7, buffer_size, NULL);
	
	
	/* Run Kernel */
	
		size_t localSize = nCells;
		size_t globalSize= nCells;
		cl_event event;
		err = clEnqueueNDRangeKernel(cmd_queue, kernel, 1, NULL, 
									 &globalSize, &localSize, 0, NULL, &event);
		assert(err == CL_SUCCESS);
		
		clFinish(cmd_queue);
	
		// Read timing information.
		unsigned long endTime = 0, startTime = 0;
		err |= clGetEventProfilingInfo(event, CL_PROFILING_COMMAND_START, sizeof(endTime), &startTime, NULL);
		err |= clGetEventProfilingInfo(event, CL_PROFILING_COMMAND_END, sizeof(endTime), &endTime, NULL);
		assert(err == CL_SUCCESS);
		printf("testComputPropensities runtime %lld nanoseconds.\n", endTime - startTime);
		
		// read results
		err = clEnqueueReadBuffer(cmd_queue, D_mem, CL_TRUE, 0, sizeof(answD), 
								  &answD, 0, NULL, NULL);
		assert(err == CL_SUCCESS);
		clFinish(cmd_queue);
		
	
	/* Release Memory */
	
		clReleaseMemObject(state_mem);
		clReleaseMemObject(cellStart_mem);
		clReleaseMemObject(inpSpec_mem);  
		clReleaseMemObject(rates_mem);
		clReleaseMemObject(D_mem);
	
	/* Check Result Answers */
	
		// return false if error found
		for (int i=0; i < nRxns; i++){
			if(answD[i] != correctD[i])
				return false;
		}
		return true;
	
}

struct DeltaX{
	cl_ushort species;
	cl_ushort cell;
	cl_short delta;
	cl_short pad;
};

DeltaX initDeltaX(unsigned short s, unsigned short c, short d){
	DeltaX x;
	x.species = s; x.cell = c; x.delta = d;
	return x;
}

/*__kernel
void boundCellState(__global const int *state,		
					__local int *locCellBoundS,	// our __local storage of state
					__global const int *rxnEvents,	// how many events for each cell's rxnId 
					__local int *locRxnEvents,	// local version of locRxnEvents
					const int nCellRxns,
					__global const struct DeltaX *scDelta, // a list of DeltaX (speciesId, cellId, deltaX). 
					const int nSCD,			// number of deltaX
					__global const int *stateBoundL, // from the return of boundState. 
					__global int *cellBoundS)	// where the answer is to be stored
*/
bool test_boundCellState(cl_device_id cid, cl_command_queue cmd_queue, cl_context context,
						 cl_program program, cl_kernel kernel)
{
	/* Setup Input. There will be 3 cells. */
	
		int state [] = { 0, 1, 2 ,3,4,5,6,7,8,9};
		int rxnEvents [] = {3, 2, 0};
		DeltaX deltas[] = {initDeltaX(0, 0, 2), initDeltaX(0, 1, 1), initDeltaX(1,2, 2), initDeltaX(2,0,1), 
							initDeltaX(2, 1, 2), initDeltaX(3, 2, 1), initDeltaX(4,0,1), initDeltaX(5, 0, 1),
							initDeltaX(7, 0, 1), initDeltaX(7,2, 2), initDeltaX(9, 1, 2)};
		int stateBound [] = {9, 9, 9, 9, 9, 9, 9, 9, 9, 9};
		int correctBound [] = {8, 1, 9, 3, 7, 8, 6, 9, 8, 9};
		int gpuAnsw[10];
		cl_int nCells = 3;
		cl_int nDeltas = 11;
		int nSpecies = 10;
	
	/* Create GPU Memory and store relevant information. */
		size_t buffer_size;
		cl_int err;
		
		// Input array state
		buffer_size = sizeof(state);
		cl_mem state_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, state_mem, CL_TRUE, 0, buffer_size,
								   (void*)state, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  0, sizeof(cl_mem), &state_mem);
		assert(err == CL_SUCCESS);
		
		// Input locCellBoundS (local)
		err |= clSetKernelArg(kernel,  1, sizeof(state), NULL);
		
		// Input cell rxnEvents
		buffer_size = sizeof(rxnEvents);
		cl_mem cellEvents_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, cellEvents_mem, CL_TRUE, 0, buffer_size,
								   (void*)rxnEvents, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  2, sizeof(cl_mem), &cellEvents_mem);
		assert(err == CL_SUCCESS);
		
		// Input locRxnEvents
		err |= clSetKernelArg(kernel,  3, sizeof(rxnEvents), NULL);
		assert(err == CL_SUCCESS);
		
		// Input num cells (excluding root)
		err |= clSetKernelArg(kernel,  4, sizeof(cl_int), &nCells);
		assert(err == CL_SUCCESS);
		
		// Input scDelta
		buffer_size = sizeof(deltas);
		cl_mem delta_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, delta_mem, CL_TRUE, 0, buffer_size,
								   (void*)deltas, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  5, sizeof(cl_mem), &delta_mem);
		assert(err == CL_SUCCESS);
		
		// Input nSCD (len of scDelta)
		err |= clSetKernelArg(kernel,  6, sizeof(cl_int), &nDeltas);
		assert(err == CL_SUCCESS);
		
		// Input stateBoundL
		buffer_size = sizeof(stateBound);
		cl_mem stateBound_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, stateBound_mem, CL_TRUE, 0, buffer_size,
								   (void*)stateBound, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  7, sizeof(cl_mem), &stateBound_mem);
		assert(err == CL_SUCCESS);
		
		// Input cellBoundS, where answer will be written
		buffer_size = sizeof(stateBound);
		cl_mem cellBoundS_mem = clCreateBuffer(context, CL_MEM_WRITE_ONLY, buffer_size, NULL, NULL);
		err |= clSetKernelArg(kernel,  8, sizeof(cl_mem), &cellBoundS_mem);
		assert(err == CL_SUCCESS);
		
	/* Run kernel and get answer */
		size_t localSize = nSpecies;
		size_t globalSize= nSpecies;
		err = clEnqueueNDRangeKernel(cmd_queue, kernel, 1, NULL, 
									 &globalSize, &localSize, 0, NULL, NULL);
		assert(err == CL_SUCCESS);
		
		clFinish(cmd_queue);
		
		// read results
		err = clEnqueueReadBuffer(cmd_queue, cellBoundS_mem, CL_TRUE, 0, sizeof(gpuAnsw), 
								  &gpuAnsw, 0, NULL, NULL);
		assert(err == CL_SUCCESS);
		clFinish(cmd_queue);
	
	
	/* Release Memory */
		
		clReleaseMemObject(state_mem);
		clReleaseMemObject(cellEvents_mem);
		clReleaseMemObject(delta_mem);  
		clReleaseMemObject(stateBound_mem);
		clReleaseMemObject(cellBoundS_mem);
		
	
	/* Check answer against true; */
		for (int i=0; i < nSpecies; i++){
			//printf("%d ", gpuAnsw[i]);
			if (gpuAnsw[i] != correctBound[i]) return false;
		}
		
		return true;
}

//void computeFinalAcceptvoid computeFinalAccept(__global int *state,  // state/time updated if step is accepted
//__local int *locState,
//int nStates,
//__global float *time,
//__global const float *rates,
//__global const float *DboundCellHigh, // of size |base reactions|
//__global const float *taus,	// size of nEvents (L)
//__global const int *rxnEvents,
//__local int *locRxnEvents,
//int nEvents,
//float rootD, 
//float lowRootD,
//__global const int2 *inpSpec,
//__global const int2 *outSpec,
//float urand, // uniform random variable, used to accept/reject
//__global float *Pcoarse,
//__local float *Ddeltas,
//__local float locPfine)
#define ALTERNATE_COMPUTE_FINAL_ACCEPT_TEST
#define DCALC(r, s) ((r[0])*(s[0])+(r[1])*(s[1])+(r[2])*(s[1])*(s[2]))
bool test_computeFinalAccept(cl_device_id cid, cl_command_queue cmd_queue, cl_context context,
							 cl_program program, cl_kernel kernel)
{
#ifndef ALTERNATE_COMPUTE_FINAL_ACCEPT_TEST
	/*  Build Inputs to Kernel */
		cl_int totRxnSets = 6;
		int rxnStart = 3;
		cl_int state[] = {11, 21, 33}; //highcellstate = {11, 24, 35};
		cl_float time = 1.2345;
		cl_int nStates = 3;
		// there are 2 cells and a two rxns in first (s0->s1, s1->0) and 1 rxn in second (s1+s2->2s2) (3 base total)
		cl_float rates[] = {0.1, .2, .003};
		// we have L=5, (Lcell1=3, Lcell2=2)
		cl_float D[] = {7.379, 5.3, 2.07900 ,1.1, 4.2, 2.07900};
		cl_float boundCellHigh[] = {8.42 ,5.9 ,2.52 ,1.1, 4.8, 2.52}; //{11*.1, 24*.2, 35*24*.3}
		cl_float taus[] = {.025, .075, .05, .04, .06}; // len(taus)=5, tauSum should be .25
		cl_int rxnEvents[] = {5, 4, 4, 3, 5};//{2, 1, 1, 0 ,2};
		cl_int nEvents = 5;

		cl_float DboundLow[] = {5.0, 2.0, 3.0, 1.0, 1.0, 3.0}; // mostly arbitrary except [0]
		cl_int2 inpSpec[] = {{0,-1}, {1,-1}, {1,2}};
		cl_int2 outSpec[] = {{1,-1}, {-1,-1}, {2, 2}};
		cl_float urand = .00000001;
		cl_float Pcoarse = .95;
		
		// states before firing on nth step
		int state0[] = {11, 21, 33};//, D=7.379
		int state1[] =  {11, 20, 34};//, D=.1*11+.2*20+.003*20*34=7.14
		int state2[] =  {11, 19, 34};//, D=.1*11+.2*19+.003*19*34=6.838
		int state3[] =  {11, 18, 34};//, D=.1*11+.2*18+.003*18*34=6.536 
		int state4[] =  {10, 19, 34};//, D=.1*10+.2*18+.003*18*34=6.436 
		cl_float D0 = DCALC(rates, state0);
		cl_float D1 = DCALC(rates, state1);
		cl_float D2 = DCALC(rates, state2);
		cl_float D3 = DCALC(rates, state3);
		cl_float D4 = DCALC(rates, state4);
		
		float truePrxns = (rates[2]*(state0[1]*state0[2])/boundCellHigh[5]) * (rates[1]*state1[1]/boundCellHigh[4]) * (rates[1]*state2[1]/boundCellHigh[4]) * (rates[0]*state3[0]/boundCellHigh[3]) * rates[2]*state4[1]*state4[2]/boundCellHigh[5];
		float expSum = taus[0]*(D0-5) + taus[1]*(D1-5) + taus[2]*(D2-5) + taus[3]*(D3-5) + taus[4]*(D4-5);
		float truePtaus = exp(-expSum) ; // expsum = 
		float truePFinal = truePrxns*truePtaus*Pcoarse;
		int nBaseRxn = 3;
#endif
#ifdef ALTERNATE_COMPUTE_FINAL_ACCEPT_TEST
	/*  Build Inputs to Kernel */
	cl_int totRxnSets = 4;
	int rxnStart = 2;
	cl_int state[] = {91}; //highcellstate = {11, 24, 35};
	cl_float time = .555822;
	cl_int nStates = 1;
	// there are 2 cells and a two rxns in first (s0->s1, s1->0) and 1 rxn in second (s1+s2->2s2) (3 base total)
	cl_float rates[] = {.1, .11};
	// we have L=5, (Lcell1=3, Lcell2=2)
	cl_float D[] = {19.110001, 19.110001, 9.100000, 10.010000};
	cl_float boundCellHigh[] = {19.529999, 19.529999, 9.300000, 10.230000}; //{11*.1, 24*.2, 35*24*.3}
	cl_float taus[] = {0.11283, 0.00129, 0.00776}; // len(taus)=5, tauSum should be .25
	cl_int rxnEvents[] = {3, 3, 2};//{2, 1, 1, 0 ,2};
	cl_int nEvents = 3;
	
	cl_float DboundCellLow[] = {18.690001, 18.690001, 8.900001, 9.790000}; // mostly arbitrary except [0]
	cl_int2 inpSpec[] = {{0,-1}, {0,-1}};
	cl_int2 outSpec[] = {{0,0}, {-1,-1}};
	cl_float urand = .00000001;
	cl_float Pcoarse = 1.0;
	
	// states before firing on nth step
	int state0[] = {91};//, D=7.379
	int state1[] =  {90};//, D=.1*11+.2*20+.003*20*34=7.14
	int state2[] =  {89};//, D=.1*11+.2*19+.003*19*34=6.838
	cl_float D0 = state0[0] * (rates[0] + rates[1]);
	cl_float D1 = state1[0] * (rates[0] + rates[1]);
	cl_float D2 = state2[0] * (rates[0] + rates[1]);
	
	float truePrxns = (rates[1]*state0[0]/boundCellHigh[3]) * (rates[1]*state1[0]/boundCellHigh[3]) * (rates[0]*state2[0]/boundCellHigh[2]);
	float DrootLow = DboundCellLow[0];
	float expSum = taus[0]*(D0-DrootLow) + taus[1]*(D1-DrootLow) + taus[2]*(D2-DrootLow);
	float truePtaus = exp(-expSum) ; // expsum = 
	float truePFinal = truePrxns*truePtaus*Pcoarse;
	int nBaseRxn = 2;
#endif
	
	
	
		cl_float gpuPfine; // where the answer is stored from GPU (for debuggin)
	
	/* Allocate Memory and Set Arguments */
		
		// allocate mem, and initialize (if required)
		size_t buffer_size;
		cl_int err;
		int argNum = 0;
		
		// Input array state
		buffer_size = sizeof(state);
		cl_mem state_mem = clCreateBuffer(context, CL_MEM_READ_WRITE, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, state_mem, CL_TRUE, 0, buffer_size,
								   (void*)state, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  argNum++, sizeof(cl_mem), &state_mem);
		assert(err == CL_SUCCESS);
	
		// Input (local) state
		buffer_size = sizeof(state);
		err |= clSetKernelArg(kernel,  argNum++, buffer_size , NULL);
		assert(err == CL_SUCCESS);	
	
		// Input stateSize
		err |= clSetKernelArg(kernel,  argNum++, sizeof(cl_int), &nStates);
		
		// Input time
		buffer_size = sizeof(cl_float);
		cl_mem time_mem = clCreateBuffer(context, CL_MEM_READ_WRITE, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, time_mem, CL_TRUE, 0, buffer_size,
								   (void*)(&time), 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  argNum++, sizeof(cl_mem), &time_mem);
		assert(err == CL_SUCCESS);
		
		// Input rates
		buffer_size = sizeof(rates);
		cl_mem rates_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, rates_mem, CL_TRUE, 0, buffer_size,
								   (void*)rates, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  argNum++, sizeof(cl_mem), &rates_mem);
		assert(err == CL_SUCCESS);
		
		// Input DboundCellHigh
		buffer_size = sizeof(cl_float)*totRxnSets;
		cl_mem boundCellHigh_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, boundCellHigh_mem, CL_TRUE, 0, buffer_size,
								   (void*)&boundCellHigh[0], 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  argNum++, sizeof(cl_mem), &boundCellHigh_mem);
		assert(err == CL_SUCCESS);
		
		// Input taus
		buffer_size = sizeof(taus);
		cl_mem taus_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, taus_mem, CL_TRUE, 0, buffer_size,
								   (void*)taus, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  argNum++, sizeof(cl_mem), &taus_mem);
		assert(err == CL_SUCCESS);	
		
		
		// Input orderedRxnEvents,
		buffer_size = sizeof(rxnEvents);
		cl_mem rxnEvents_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, rxnEvents_mem, CL_TRUE, 0, buffer_size,
								   (void*)rxnEvents, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  argNum++, sizeof(cl_mem), &rxnEvents_mem);
		assert(err == CL_SUCCESS);	
		
		
		// Input locOrderedRxnEvents,
		err |= clSetKernelArg(kernel,  argNum++, sizeof(rxnEvents), NULL);
		
		// Input nEvents
		err |= clSetKernelArg(kernel,  argNum++, sizeof(cl_int), &nEvents);
		
		// Input D
		buffer_size = sizeof(cl_int) * totRxnSets;
		cl_mem D_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, D_mem, CL_TRUE, 0, buffer_size,
								   (void*)D, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  argNum++, sizeof(cl_mem), &D_mem);
		
		// Input DboundCellLow
		buffer_size = sizeof(cl_int) * totRxnSets;
		cl_mem DboundCellLow_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, DboundCellLow_mem, CL_TRUE, 0, buffer_size,
								   (void*)DboundCellLow, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  argNum++, sizeof(cl_mem), &DboundCellLow_mem);
		
		// Input inpSpec
		buffer_size = sizeof(inpSpec);
		cl_mem inpSpec_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, inpSpec_mem, CL_TRUE, 0, buffer_size,
								   (void*)inpSpec, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  argNum++, sizeof(cl_mem), &inpSpec_mem);
		assert(err == CL_SUCCESS);	
		
		// Input outSpec
		buffer_size = sizeof(outSpec);
		cl_mem outSpec_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, outSpec_mem, CL_TRUE, 0, buffer_size,
								   (void*)outSpec, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  argNum++, sizeof(cl_mem), &outSpec_mem);
		assert(err == CL_SUCCESS);	
		
		// Input urand
		err |= clSetKernelArg(kernel,  argNum++, sizeof(cl_float), &urand);
		
		// Input Pcoarse 
		buffer_size = sizeof(cl_float);
		cl_mem Pcoarse_mem = clCreateBuffer(context, CL_MEM_READ_WRITE, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, Pcoarse_mem, CL_TRUE, 0, buffer_size,
								   (void*)&Pcoarse, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  argNum++, sizeof(cl_mem), &Pcoarse_mem);
		assert(err == CL_SUCCESS);	
	

		// Input Ddeltas
		buffer_size = sizeof(cl_int) * MAX_SPECIES_RXN_CON;
		err |= clSetKernelArg(kernel,  argNum++,buffer_size, NULL);
		assert(err == CL_SUCCESS);
	
//		// Input locPfine
//		buffer_size = sizeof(cl_float);
//		err |= clSetKernelArg(kernel,  argNum++, buffer_size, NULL);
//		assert(err == CL_SUCCESS);
	
		// Input rxnStart
		err |= clSetKernelArg(kernel,  argNum++, sizeof(cl_int), &rxnStart);
	
		// This is where function output will be written to.
		buffer_size = sizeof(cl_float);
		cl_mem Paccept_mem = clCreateBuffer(context, CL_MEM_READ_WRITE, buffer_size, NULL, NULL);
		err |= clSetKernelArg(kernel,  argNum++, sizeof(cl_mem), &Paccept_mem);
		assert(err == CL_SUCCESS);	

	/* Run Kernel */
	
		size_t localSize = nBaseRxn;
		size_t globalSize= nBaseRxn;
		err = clEnqueueNDRangeKernel(cmd_queue, kernel, 1, NULL, 
									 &globalSize, &localSize, 0, NULL, NULL);
		assert(err == CL_SUCCESS);
		
		clFinish(cmd_queue);
		
		// read results
		err = clEnqueueReadBuffer(cmd_queue, Paccept_mem, CL_TRUE, 0, sizeof(cl_float), 
								  &gpuPfine, 0, NULL, NULL);
		assert(err == CL_SUCCESS);
		clFinish(cmd_queue);
	
	/* Release Memory */
	
		clReleaseMemObject(state_mem);
		clReleaseMemObject(time_mem);
		clReleaseMemObject(rates_mem);  
		clReleaseMemObject(boundCellHigh_mem);
		clReleaseMemObject(taus_mem);
		clReleaseMemObject(rxnEvents_mem);
		clReleaseMemObject(inpSpec_mem);  
		clReleaseMemObject(outSpec_mem);
		clReleaseMemObject(Pcoarse_mem);
	
	/* Check Result Answers */
	
	// return false if error found
	float diff = fabs(gpuPfine - truePFinal);
	assert(diff < EPS);
	return true;
}


bool test_computePCoarse(cl_device_id cid, cl_command_queue cmd_queue, cl_context context,
						 cl_program program, cl_kernel kernel)
{
	/*  Build Inputs to Kernel */
		int nCells = 3;
		int rxnEvents[] = {5, 2, 0, 3, 2, 0, 3}; // {root, c0, c1, c2, r00, r10, r20}
		float Dbh[] = {6.6, 1.1, 2.2, 3.3};
		float Dbch[] = {5.7 ,1.1 ,1.8 ,2.8};
		float Dbl[] = {4.5, 1.0, 1.5, 2.0};
		float Dbcl[] = {4.8, 1.1, 1.6, 2.1};
		cl_float tauSum = 0.5;
		float taus[] = {.2, .01, .15, .04, .1};
		int leap = 5;

		float correctPcoarse = 0.525760678;//0.915844074; // (1.1/1.1)^2 * (1.8/2.2)^0 * (2.8/3.3)^3 * exp(-.5 * (4.8, - 4.5))
		
		float gpuPcoarse = -1;

	
	/* Allocate Memory and Set Arguments */
	
		// allocate mem, and initialize (if required)
		size_t buffer_size;
		cl_int err;
		
		// Input array DboundStateHigh
		buffer_size = sizeof(Dbh);
		cl_mem dbh_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, dbh_mem, CL_TRUE, 0, buffer_size,
								   (void*)Dbh, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  0, sizeof(cl_mem), &dbh_mem);
		assert(err == CL_SUCCESS);
		
		
		// Input array DboundCellStateHigh
		buffer_size = sizeof(Dbch);
		cl_mem Dbch_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, Dbch_mem, CL_TRUE, 0, buffer_size,
								   (void*)Dbch, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  1, sizeof(cl_mem), &Dbch_mem);
		assert(err == CL_SUCCESS);
		
	
		// Input DboundStateLow
		buffer_size = sizeof(Dbl);
		cl_mem Dbl_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, Dbl_mem, CL_TRUE, 0, buffer_size,
								   (void*)Dbl, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  2, sizeof(cl_mem), &Dbl_mem);
		assert(err == CL_SUCCESS);
	
		// Input DboundCellStateLow
		buffer_size = sizeof(Dbcl);
		cl_mem Dbcl_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, Dbcl_mem, CL_TRUE, 0, buffer_size,
								   (void*)Dbcl, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  3, sizeof(cl_mem), &Dbcl_mem);
		assert(err == CL_SUCCESS);
	
		// Input array rxnEvents
		buffer_size = sizeof(rxnEvents);
		cl_mem rxnEvents_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, rxnEvents_mem, CL_TRUE, 0, buffer_size,
								   (void*)rxnEvents, 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  4, sizeof(cl_mem), &rxnEvents_mem);
		assert(err == CL_SUCCESS);
		
		// Input cellTaus
		buffer_size = sizeof(cl_float) * leap;
		cl_mem taus_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, taus_mem, CL_TRUE, 0, buffer_size,
							   (void*)(&taus), 0, NULL, NULL);
		err |= clSetKernelArg(kernel,  5, sizeof(cl_mem), &taus_mem);
		assert(err == CL_SUCCESS);
		
		// Input multTerms
		buffer_size = sizeof(cl_int) * rxnEvents[0];
		err |= clSetKernelArg(kernel,  6, buffer_size , NULL);
		assert(err == CL_SUCCESS);
		
		// Input Pcoarse term
		buffer_size = sizeof(cl_float);
		cl_mem Pcoarse_mem = clCreateBuffer(context, CL_MEM_WRITE_ONLY, buffer_size, NULL, NULL);
		err |= clSetKernelArg(kernel,  7, sizeof(cl_mem), &Pcoarse_mem);
		assert(err == CL_SUCCESS);
	
		err |= clSetKernelArg(kernel,  8, sizeof(cl_int), &rxnEvents[0]);
	
	/* Run Kernel */
	
		size_t localSize = nCells;
		size_t globalSize= nCells;
	
		err |= clFinish(cmd_queue);
		assert(err == CL_SUCCESS);
	
		err |= clEnqueueNDRangeKernel(cmd_queue, kernel, 1, NULL, 
									 &globalSize, &localSize, 0, NULL, NULL);
		
		assert(err == CL_SUCCESS);
		err |= clFinish(cmd_queue);
	
		assert(err == CL_SUCCESS);
		
		// read results
		err = clEnqueueReadBuffer(cmd_queue, Pcoarse_mem, CL_TRUE, 0, sizeof(cl_float), 
								  &gpuPcoarse, 0, NULL, NULL);
		assert(err == CL_SUCCESS);
		clFinish(cmd_queue);
	
	
	/* Release Memory */
	
		clReleaseMemObject(dbh_mem);
		clReleaseMemObject(Dbch_mem);
		clReleaseMemObject(rxnEvents_mem);  
		clReleaseMemObject(Pcoarse_mem);
		clReleaseMemObject(taus_mem);
	
	/* Check Result Answers */
	
	// return false if error found
	assert(fabs(correctPcoarse - gpuPcoarse) < EPS);
	return  true; 
}

#define NUM_CELLS 5
bool test_sampleCoarseEvents(cl_command_queue cmd_queue, cl_context context,
							 cl_kernel kernel){
	
	/*  Build Inputs to Kernel */
	
	// allocate program memory
	float DboundHigh[] = {100.0, 10.0, 20.0, 50.0, 5.0, 15.0}; // first entry will be ignored.
	float unormalizedProbSum = 100.0;
	int nCells = NUM_CELLS;
	int trialsPerSample = 20;
	int totSamples = 1000; // Number of times we'll draw the multinomial sample.
	
	
	/* Allocate Memory and Set Arguments */
	
	// allocate mem, and initialize (if required)
	size_t buffer_size;
	cl_int err;
	
	// Input array DboundHigh
	buffer_size = sizeof(float) * (1+nCells);
	cl_mem probs_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
	err = clEnqueueWriteBuffer(cmd_queue, probs_mem, CL_TRUE, 0, buffer_size,
							   (void*)DboundHigh, 0, NULL, NULL);
	err |= clSetKernelArg(kernel,  0, sizeof(cl_mem), &probs_mem);
	assert(err == CL_SUCCESS);
	
	// Input number of cells
	err |= clSetKernelArg(kernel,  1, sizeof(int), &nCells);
	
	// Allocate and assign memory for successes (results)
	buffer_size = sizeof(int) * (nCells+1);
	cl_mem results_mem = clCreateBuffer(context, CL_MEM_WRITE_ONLY, buffer_size, NULL, NULL);
	err |= clSetKernelArg(kernel,  2, sizeof(cl_mem), &results_mem);
	assert(err == CL_SUCCESS);
	
	// Local copy of successes
	err |= clSetKernelArg(kernel,  3, buffer_size, NULL);
	assert(err == CL_SUCCESS);
	
	// Number of trials
	err |= clSetKernelArg(kernel,  4, sizeof(cl_int), &trialsPerSample);
	assert(err == CL_SUCCESS);
	
	// Random unsigned integers used for sampling
	buffer_size = trialsPerSample * sizeof(cl_int);
	cl_mem random_ints_mem = clCreateBuffer(context, CL_MEM_READ_WRITE, buffer_size, NULL, NULL);
	err |= clSetKernelArg(kernel,  5, sizeof(cl_mem), &random_ints_mem);
	assert(err == CL_SUCCESS);		
	
	// Offset for random numbers (0 for this test)
	cl_int rnOffset = 0;
	err |= clSetKernelArg(kernel,  6, sizeof(cl_int), &rnOffset);
	
	/* Run Kernel */
	
	size_t localSize = 32;
	size_t globalSize= localSize;
	
	int cummulativeSamples[NUM_CELLS+1];
	int kthSample[NUM_CELLS+1];
	
	for (int i = 0; i < NUM_CELLS+1; i++) cummulativeSamples[i] = 0;
	
	for (int sample = 0; sample < totSamples; sample++) {
		buffer_size = sizeof(unsigned int) * trialsPerSample;
		unsigned int* randInts = generateRandomUInts(trialsPerSample);
		err = clEnqueueWriteBuffer(cmd_queue, random_ints_mem, CL_TRUE, 0, buffer_size,
								   (void*)randInts, 0, NULL, NULL);
		assert(err == CL_SUCCESS);		
		clFinish(cmd_queue);
		delete [] randInts;
		
		err = clEnqueueNDRangeKernel(cmd_queue, kernel, 1, NULL, 
									 &globalSize, &localSize, 0, NULL, NULL);
		assert(err == CL_SUCCESS);
		
		clFinish(cmd_queue);
		
		// read results
		err = clEnqueueReadBuffer(cmd_queue, results_mem, CL_TRUE, 0, sizeof(kthSample), 
								  &kthSample, 0, NULL, NULL);
		assert(err == CL_SUCCESS);
		clFinish(cmd_queue);
		
		for (int i = 0; i < NUM_CELLS+1; i++) 
			cummulativeSamples[i] += kthSample[i];
	}

	
	
	/* Release Memory */
	
	clReleaseMemObject(probs_mem);
	clReleaseMemObject(results_mem);
	clReleaseMemObject(random_ints_mem);  
	
	/* Check Result Answers */
	
	// return false if error found
	// std of binomial is sqrt(n*p*(1-p))

	int n = totSamples * trialsPerSample;
	for (int i=1; i < NUM_CELLS+1; i++){ // ignore first (root) of cummulative
		float top = i == 0 ? 0 : DboundHigh[i];
		float p = top / unormalizedProbSum;
		float std = sqrt(n * p * (1-p));
		
		assert(abs((float) cummulativeSamples[i] - p * n) <= std * 3);

	}
	return true;
	
}

bool test_fineSample(cl_command_queue cmd_queue, cl_context context, cl_kernel kernel){
	// Topography for this test:
	//		- root rxns, 3 cells, each cell has two rxns.
	int cellStartIdx[] = {4, 6, 8, 10}; // where in Dprobs the cell range will be.
	float c1a = 2.0; float c2a = 55; float c3a = 99.5;
	float dProbs[] = {300, 100, 100, 100, c1a, 100-c1a, c2a, 100-c2a, c3a, 100-c3a}; 
	
	int rxnEvents[] = {7, 1, 2, 3, 0, 0, 0, 0, 0, 0};
	int nTrials = 1000;
	unsigned int* randUints = generateRandomUInts(rxnEvents[0] * nTrials);
	int randStart = 0;
	
	
	/* Allocate Memory and Set Arguments */
	
	// allocate mem, and initialize (if required)
	size_t buffer_size;
	cl_int err;
	
	// Input array specifying cell containment 
	buffer_size = sizeof(cl_int) * 4; // (4=nCells + 1)
	cl_mem cell_start_idx_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
	err = clEnqueueWriteBuffer(cmd_queue, cell_start_idx_mem, CL_TRUE, 0, buffer_size,
							   (void*)cellStartIdx, 0, NULL, NULL);
	err |= clSetKernelArg(kernel,  0, sizeof(cl_mem), &cell_start_idx_mem);
	assert(err == CL_SUCCESS);
	
	// Dprobs input
	buffer_size = sizeof(cl_float) * 10;
	cl_mem d_probs_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
	
	err = clEnqueueWriteBuffer(cmd_queue, d_probs_mem, CL_TRUE, 0, buffer_size,
							   (void*)dProbs, 0, NULL, NULL);
	err |= clSetKernelArg(kernel,  1, sizeof(cl_mem), &d_probs_mem);
	assert(err == CL_SUCCESS);
	
	
	
	// local Dprobs
	err |= clSetKernelArg(kernel,  2, sizeof(cl_int) * 10, 0);
	assert(err == CL_SUCCESS);
	
	// rxn events
	buffer_size = sizeof(cl_int) * 10;
	cl_mem rxn_events_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
	err = clEnqueueWriteBuffer(cmd_queue, rxn_events_mem, CL_TRUE, 0, buffer_size,
							   (void*)rxnEvents, 0, NULL, NULL);
	err |= clSetKernelArg(kernel,  3, sizeof(cl_mem), &rxn_events_mem);
	assert(err == CL_SUCCESS);
	
	// loc rxn events
	err |= clSetKernelArg(kernel,  4, sizeof(cl_int) * 10, 0);
	assert(err == CL_SUCCESS);
	
	
	// randomUints
	buffer_size = sizeof(cl_uint) * rxnEvents[0] * nTrials;
	cl_mem randuints_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
	err = clEnqueueWriteBuffer(cmd_queue, randuints_mem, CL_TRUE, 0, buffer_size,
							   (void*)randUints, 0, NULL, NULL);
	err |= clSetKernelArg(kernel,  5, sizeof(cl_mem), &randuints_mem);
	assert(err == CL_SUCCESS);
	
	// randomUints start idx
	
	
	/* Now gather our samples. */
	int cummulativeSamples[] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
	int kthSample[10];
	size_t globalSize = 32;
	
	for (int trial = 0; trial < nTrials; trial++) {
		// Set Uints start index.
		err |= clSetKernelArg(kernel,  6, sizeof(cl_int), &randStart);
		randStart += rxnEvents[0];
		assert(err == CL_SUCCESS);
		
		// Run the trials.
		err = clEnqueueNDRangeKernel(cmd_queue, kernel, 1, NULL, 
									 &globalSize, &globalSize, 0, NULL, NULL);
		assert(err == CL_SUCCESS);
		
		clFinish(cmd_queue);
		
		// read results
		err = clEnqueueReadBuffer(cmd_queue, rxn_events_mem, CL_TRUE, 0, sizeof(kthSample), 
								  &kthSample, 0, NULL, NULL);
		assert(err == CL_SUCCESS);
		clFinish(cmd_queue);
		
		// Gather statistics.
		for (int i = 0; i < 10; i++) 
			cummulativeSamples[i] += kthSample[i];	
		
	}
	
	clReleaseMemObject(cell_start_idx_mem);
	clReleaseMemObject(d_probs_mem);
	clReleaseMemObject(rxn_events_mem);  
	clReleaseMemObject(randuints_mem);
	
	int expectedVals[] = {rxnEvents[0] * nTrials, rxnEvents[1] * nTrials, rxnEvents[2] * nTrials, rxnEvents[3] * nTrials,0, 0, 0, 0, 0, 0};
	int std[] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
	for (int i = 4; i < 10; i++) {
		int parentIdx = i/2 - 1;
		double p = dProbs[i] / dProbs[parentIdx];
		float n = expectedVals[parentIdx];
		expectedVals[i] = n * p;
		std[i] = sqrt(n * p * (1-p));
	}
	
	// determine if aquired stats closely match the expected values. 
	for (int i = 0; i < 10; i++) {
		if(abs((float) cummulativeSamples[i] - expectedVals[i]) > std[i] * 3) {
			return false;
		}
	}
	return true;
}


#define TTS_LEAP 12
bool test_tausSample(cl_command_queue cmd_queue, cl_context context, cl_kernel kernel){
	int leap = TTS_LEAP;
	float rate = 100.0;
	
	int trials = 1000;
	int randStartIdx = 0;
	unsigned int *randNums = generateRandomUInts(leap * trials);
	
	// Setup Memory
	
	size_t buffer_size;
	cl_int err = CL_SUCCESS;
	int idx = 0;
	// Leap.
	err |= clSetKernelArg(kernel,  idx++, sizeof(cl_int), &leap);
	
	
	// Rate.
	buffer_size = sizeof(cl_float);
	cl_mem rate_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
	err = clEnqueueWriteBuffer(cmd_queue, rate_mem, CL_TRUE, 0, buffer_size,
							   (void*)&rate, 0, NULL, NULL);
	err |= clSetKernelArg(kernel,  idx++, sizeof(cl_mem), &rate_mem);
	
	// Sampled Taus storage.
	buffer_size = sizeof(cl_float) * leap;
	cl_mem taus_mem = clCreateBuffer(context, CL_MEM_READ_WRITE, buffer_size, NULL, NULL);
	err |= clSetKernelArg(kernel,  idx++, sizeof(cl_mem), &taus_mem);
	
	
	// randNums
	buffer_size = sizeof(cl_uint) * leap * trials;
	cl_mem randNums_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
	err = clEnqueueWriteBuffer(cmd_queue, randNums_mem, CL_TRUE, 0, buffer_size,
							   (void*)randNums, 0, NULL, NULL);
	err |= clSetKernelArg(kernel,  idx++, sizeof(cl_mem), &randNums_mem);
	
	
	// Run the trials.
	float taus[TTS_LEAP];
	size_t globalSize = 16;
	double cumulativeTauSqrd = 0;
	double cumulativeTau = 0;
	
	for (int trial = 0; trial < trials; trial++) {
		err |= clSetKernelArg(kernel,  idx, sizeof(cl_int), &randStartIdx);
		randStartIdx += leap;
		assert(err == CL_SUCCESS);
		
		// Run the trials.
		err = clEnqueueNDRangeKernel(cmd_queue, kernel, 1, NULL, 
									 &globalSize, &globalSize, 0, NULL, NULL);
		assert(err == CL_SUCCESS);
		
		clFinish(cmd_queue);
		
		// read results
		// taus.
		err = clEnqueueReadBuffer(cmd_queue, taus_mem, CL_TRUE, 0, sizeof(taus), 
								  &taus, 0, NULL, NULL);
		clFinish(cmd_queue);
		assert(err == CL_SUCCESS);
		
		// Stats tests for correctness. 
		
		// Gather statistics.
		double sum = 0.0;
		for (int i = 0; i < leap; i++) {
			float t = taus[i];
			sum += t;
			cumulativeTau += t;
			cumulativeTauSqrd += t * t;
		}
	}
	
	// Check the mean of taus is close to expected.
	float expectedMean = 1.0 / rate; 
	float analyticStd = (1.0 / rate) / sqrt(trials * leap);
	assert(abs(cumulativeTau / (leap * trials) - expectedMean) < analyticStd * 3);
	
	// See if empirical variance is close enough to expected. 
	double var = abs(cumulativeTauSqrd - cumulativeTau * cumulativeTau);
	assert(abs(sqrt(var) - trials * leap / rate) / trials < analyticStd * 3);
	
	// We actually just throw asserts if the answer is wrong.
	clReleaseMemObject(taus_mem);
	clReleaseMemObject(randNums_mem);	
	return true;
}

#define TRP_LEAP 12
bool test_reactionPermutation(cl_command_queue cmd_queue, cl_context context, cl_kernel kernel){
	int reactionEvents[] = {12, 8, 4, 3, 5, 0, 4}; // For each rxnId the number of events the rxn will get {root, cell1, cell2, r1a, r1b, r2a, r2b}
	//int reactionEvents[] = {8, 0, 8, 0, 8, 0, 4};
	
	//int cellStartIdx[] = {3, 5, 7};
	int cellStartIdx[] = {3, 4, 7};
	//float taus[TRP_LEAP] = {.1, .1, .1, .1, .1, .1, .1, .1, .1, .1, .1, .1};
	float taus[TRP_LEAP] = {.01, .19, .05, .15, .001, .199, .04, .16, .03, .17, .02, .18};
	int nCells = cellStartIdx[0] - 1;
	
	int trials = 500;
	int nRand = trials * TRP_LEAP;
	unsigned int *rands = generateRandomUInts(nRand);
	
	/* Save values into GPU memory. */
	
	// Input rxnEvents.
	size_t buffer_size = sizeof(reactionEvents); 
	assert(buffer_size == sizeof(int)*7); // double check my logic.	
	cl_mem rxnEvents_mem = clCreateBuffer(context, CL_MEM_READ_WRITE, buffer_size, NULL, NULL);
	int err = clEnqueueWriteBuffer(cmd_queue, rxnEvents_mem, CL_TRUE, 0, buffer_size,
							   (void*)reactionEvents, 0, NULL, NULL);
	err |= clSetKernelArg(kernel,  0, sizeof(cl_mem), &rxnEvents_mem);
	
	// Input cellStartIdx
	buffer_size = sizeof(cellStartIdx);
	cl_mem cellStartIdx_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
	err |= clEnqueueWriteBuffer(cmd_queue, cellStartIdx_mem, CL_TRUE, 0, buffer_size,
							   (void*)cellStartIdx, 0, NULL, NULL);
	err |= clSetKernelArg(kernel,  1, sizeof(cl_mem), &cellStartIdx_mem);
	
	
//	// Input idxRxnStart
//	err |= clSetKernelArg(kernel,  1, sizeof(cl_int), &idxRxnStart);
//	
//	// Input numRxn.
//	err |= clSetKernelArg(kernel,  2, sizeof(cl_int), &numRxn);
	
	// locOrderedRxns
	err |= clSetKernelArg(kernel,  2, sizeof(cl_int) * TRP_LEAP, NULL);
	assert(err==CL_SUCCESS);
	
	// orderedRxns (this is where output occurs)
	buffer_size = sizeof(cl_int) * TRP_LEAP;
	cl_mem orderedRxns_mem = clCreateBuffer(context, CL_MEM_WRITE_ONLY, buffer_size, NULL, NULL);
	err |= clSetKernelArg(kernel,  3, sizeof(cl_mem), &orderedRxns_mem);
	assert(err==CL_SUCCESS);
	
	
	// Input randNums.
	buffer_size = nRand * sizeof(cl_uint);
	cl_mem randNums_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
	err = clEnqueueWriteBuffer(cmd_queue, randNums_mem, CL_TRUE, 0, buffer_size,
								   (void*)rands, 0, NULL, NULL);
	err |= clSetKernelArg(kernel,  4, sizeof(cl_mem), &randNums_mem);
	assert(err==CL_SUCCESS);
	
	/* Now run the trials. */
	// We'll keep track of two arbitrary statistics.
	//	1. the probability rxn 4 will be in slot 10.
	int fourIn10 = 0;
	//  2. the probability rxn 3 will be in slot 0.
	int threeIn0 = 0;
	const int trueSortedRxns[TRP_LEAP] = {3, 3, 3, 4, 4, 4, 4, 4, 6, 6, 6, 6};
	//const int trueSortedRxns[TRP_LEAP] = {4, 4, 4, 4, 4, 4, 4, 4, 6, 6, 6, 6};
	
	int orderedRxns[TRP_LEAP];
	int randStartIdx = 0;
	size_t globalSize = 1;
	for (int trial=0; trial < trials; trial++) {
		err |= clSetKernelArg(kernel,  5, sizeof(cl_int), &randStartIdx);
		randStartIdx += TRP_LEAP;
		assert(err == CL_SUCCESS);
		
		// Run the trials.
		err = clEnqueueNDRangeKernel(cmd_queue, kernel, 1, NULL, 
									 &globalSize, &globalSize, 0, NULL, NULL);
		assert(err == CL_SUCCESS);
		
		clFinish(cmd_queue);
		
		// read results
		// orderedRxns.
		err = clEnqueueReadBuffer(cmd_queue, orderedRxns_mem, CL_TRUE, 0, sizeof(orderedRxns), 
								  &orderedRxns, 0, NULL, NULL);
		clFinish(cmd_queue);
		assert(err == CL_SUCCESS);
		
		// Stats tests for correctness. 
		if (orderedRxns[10] == 4) {
			fourIn10++;
		}
		if (orderedRxns[0] == 3) {
			threeIn0++;
		}
		
		// Now make sure all elements are in the array.
			// sort.
		vector<int> v(orderedRxns, orderedRxns+TRP_LEAP);
		
		sort(v.begin(), v.end());
		for (int i = 0; i < TRP_LEAP; i++){
			assert(trueSortedRxns[i] == v[i]);
		}
		
	}
	
	// Now check statistics.
	// think of our counters, fourIn10 and threeIn0 as the results of binomial trials.
	int n = trials;
	double p = float(reactionEvents[4]) / TRP_LEAP;
	double var = n * p * (1 - p);
	assert(abs(fourIn10 - n * p) < sqrt(var) * 3); // Make sure our val is within 3 standard deviations.
	
	p = float(reactionEvents[3]) / TRP_LEAP;
	if (p == 0)
		assert(threeIn0 == 0);
	else{
		var = n * p * (1 - p);
		assert(abs(threeIn0 - n * p) < sqrt(var) * 3);
	}
	
	// sanity check
	assert(!(abs(fourIn10 - n * p) < sqrt(var) * 3));
	
	clReleaseMemObject(rxnEvents_mem);
	clReleaseMemObject(randNums_mem);
	clReleaseMemObject(orderedRxns_mem);
	return true;
}

#pragma mark -
#pragma mark Main OpenCL Routine
float runCL()
{
	cl_program program[2];
	cl_kernel kernel[15];
	
	cl_command_queue cmd_queue;
	cl_context   context;
	
	cl_device_id cpu = NULL, device = NULL;
	
	cl_int err = 0;
	size_t returned_size = 0;
	
	
	float ret = 0;
	
#pragma mark Device Information
	{
		// Find the CPU CL device, as a fallback
		err = clGetDeviceIDs(NULL, CL_DEVICE_TYPE_CPU, 1, &cpu, NULL);
		assert(err == CL_SUCCESS);
		
		// Find the GPU CL device, this is what we really want
		// If there is no GPU device is CL capable, fall back to CPU
		err = clGetDeviceIDs(NULL, CL_DEVICE_TYPE_GPU, 1, &device, NULL);
		if (err != CL_SUCCESS)  // TODO: remove true
			device = cpu; 
		assert(device);
		
		// Get some information about the returned device
		cl_char vendor_name[1024] = {0};
		cl_char device_name[1024] = {0};
		cl_int units[100] = {0};
		err = clGetDeviceInfo(device, CL_DEVICE_VENDOR, sizeof(vendor_name), 
							  vendor_name, &returned_size);
		err |= clGetDeviceInfo(device, CL_DEVICE_NAME, sizeof(device_name), 
							   device_name, &returned_size);
		err |= clGetDeviceInfo(device, CL_DEVICE_MAX_COMPUTE_UNITS, sizeof(units), 
							   units, &returned_size);
		
		assert(err == CL_SUCCESS);
		printf("Connecting to %s %s (%d cores)...\n", vendor_name, device_name, units[0]);
	}
	
#pragma mark Context and Command Queue
	{
		// Now create a context to perform our calculation with the 
		// specified device 
		context = clCreateContext(0, 1, &device, NULL, NULL, &err);
		assert(err == CL_SUCCESS);
		
		// And also a command queue for the context
		cmd_queue = clCreateCommandQueue(context, device, CL_QUEUE_PROFILING_ENABLE, NULL);
	}
	
#pragma mark Program and Kernel Creation
	{
		// Load the program source from disk
		// The kernel/program is the project directory and in Xcode the executable
		// is set to launch from that directory hence we use a relative path
		const char * filename = "../../../hier.cl";
		char *program_source = load_program_source(filename);
		program[0] = clCreateProgramWithSource(context, 1, (const char**)&program_source,
											   NULL, &err);
		assert(err == CL_SUCCESS);
		
		err = clBuildProgram(program[0], 0, NULL, NULL, NULL, NULL);
		
		if (err != CL_SUCCESS){
			// query build about the problem
			size_t logSize;
			
#define MAXLOGSIZE 50012
			char log[MAXLOGSIZE];
			
			printf("Error: Failed to build program executable!\n");
			
			// get the details on the error, and store it in buffer
			clGetProgramBuildInfo(
								  program[0],              // the program object being queried
								  device,            // the device for which the OpenCL code was built
								  CL_PROGRAM_BUILD_LOG, // specifies that we want the build log
								  sizeof(log),       // the size of the buffer
								  log,               // on return, holds the build log
								  &logSize);                // on return, the actual size in bytes of the
			//   data returned
			
			assert(logSize < MAXLOGSIZE);
			
			printf("%s\n", log);
			
			exit(21);
		}
		
		
		// Now create the kernel "objects" that we want to use in the file 
		kernel[0] = clCreateKernel(program[0], "computePropensities", &err);
		assert(err == CL_SUCCESS);
		kernel[1] = clCreateKernel(program[0], "boundCellState", &err); 
		assert(err == CL_SUCCESS);
		kernel[2] = clCreateKernel(program[0], "computePCoarse", &err);
		assert(err == CL_SUCCESS);
		kernel[3] = clCreateKernel(program[0], "computeFinalAccept", &err);
		assert(err == CL_SUCCESS);
		kernel[4] = clCreateKernel(program[0], "sampleCoarseEvents", &err);
		assert(err == CL_SUCCESS);
		kernel[5] = clCreateKernel(program[0], "fineSample", &err);
		assert(err == CL_SUCCESS);
		kernel[6] = clCreateKernel(program[0], "tausSample", &err);
		assert(err == CL_SUCCESS);
		kernel[7] = clCreateKernel(program[0], "reactionPermutation", &err);
		assert(err == CL_SUCCESS);
		
	}
#pragma mark Run  Unittests
	{
		bool ret;
		ret = test_computePropensities(device, cmd_queue, context,
											 program[0], kernel[0]);
		assert(ret);
		printf("\nPassed computePropensities.\n");
		
		ret = test_boundCellState(device, cmd_queue, context,
							program[0], kernel[1]);
		assert(ret);
		printf("Passed boundCellState.\n");
		
		ret = test_computePCoarse(device, cmd_queue, context,
								  program[0], kernel[2]);
		assert(ret);
		printf("Passed computePCoarse, return value %d.\n", ret);
		
		ret = test_computeFinalAccept(device, cmd_queue, context,
								  program[0], kernel[3]);
		assert(ret);
		printf("Passed computeFinalAccept.\n");
		
		ret = test_sampleCoarseEvents(cmd_queue, context, kernel[4]);
		assert(ret);
		printf("Passed sampleMultinomial.\n");
		
		ret = test_fineSample(cmd_queue, context, kernel[5]);
		assert(ret);
		printf("Passed fineSample.\n");
		
		ret = test_tausSample(cmd_queue, context, kernel[6]);
		assert(ret);
		printf("Passed tauSample.\n");
		
		ret = test_reactionPermutation(cmd_queue, context, kernel[7]);
		assert(ret);
		printf("Passed reactionPermutation.\n");
	}
	
#pragma mark Teardown
	{
		clReleaseCommandQueue(cmd_queue);
		clReleaseContext(context);
	}
	return ret;
}


//void hierLeap(global int *globalState, int stateSize,
//			  global float *time,
//			  int constant *cellStartIdx,
//			  float constant *rates,	
//			  int2 constant *inpSpec, // Assume each rxn has at most N=2 inp species
//			  int2 constant *outSpec,
//			  constant int *highDeltaS, constant int *lowDeltaS, 
//			  constant struct DeltaX *scDeltaHighCell, int nScdHighCell,
//			  constant struct DeltaX *scDeltaLowCell, int nScdLowCell,
//			  global unsigned int *randomInts, int randIdx,
//			  int leap,  
//			  local int *highBoundS, local int *highBoundCellS, local int *state, local int *lowBoundCellS, local int *lowBoundS, 
//			  local float *DboundHigh, local float *DboundCellHigh, local float *D, local float *DboundCellLow, local float *DboundLow,
//			  local int *rxnEvents, local int *orderedRxnEvents, local float *taus)
// This function is mainly to determine the source of a memory leak when calling
// hierLeap from a GPU.
void runHokTests() {
	cl_int deviceArgs = CL_DEVICE_TYPE_GPU;
	GpuInterface *gpuHok = new GpuInterface(deviceArgs, "/Users/dweo/research/gpu_hierleap/hier_one_kernel.cl");
	int nStates = 10;
	int totRxnSets = 31;
	int nCells = 10;
	cl_kernel kernel = gpuHok->extractKernel("hierLeap");
	cl_context context = gpuHok->getContext();
	
	int nConseqConstantArgs = 7; // number of constant args before nScdHighCell
	int nConseqLocalArgs = 13;
	int leap = 11;
	
	// setup kernel args, set the indecies
	int argIdx = 0;
	int bufferSize = sizeof(cl_int);// * sizeof(cl_float) * sizeof(DeltaX) * nStates;

	cl_mem state = clCreateBuffer(context, CL_MEM_READ_ONLY, bufferSize, NULL, NULL); // todo: change CL_MEM_READ_ONLY to CL_MEM_READ_WRITE
	int err = clSetKernelArg(kernel, argIdx++, sizeof(cl_mem), &state); // state
	err |= clSetKernelArg(kernel, argIdx++, sizeof(cl_int), &nStates); // nStates
	assert(err == CL_SUCCESS);
	cl_mem time = clCreateBuffer(context, CL_MEM_READ_WRITE, bufferSize, NULL, NULL);
	err |= clSetKernelArg(kernel, argIdx++, sizeof(cl_mem), &time); // time 
	assert(err == CL_SUCCESS);

	// Note this leaks by dangling mem.
	//nConseqConstantArgs = 0; // todo lined added for debug.
	for (int i = 0; i < nConseqConstantArgs; i++){
		cl_mem mem = clCreateBuffer(context, CL_MEM_READ_ONLY, bufferSize, NULL, NULL);
		err |= clSetKernelArg(kernel, argIdx++, sizeof(cl_mem), &mem);
	}
	assert(err == CL_SUCCESS);
	
	err |= clSetKernelArg(kernel, argIdx++, sizeof(cl_int), &nCells); //nScdHighCell
	cl_mem mem = clCreateBuffer(context, CL_MEM_READ_ONLY, bufferSize, NULL, NULL);
	err |= clSetKernelArg(kernel, argIdx++, sizeof(cl_mem), &mem); // scDeltaLowCell
	assert(err == CL_SUCCESS);
	err |= clSetKernelArg(kernel, argIdx++, sizeof(cl_int), &nCells); // nScdLowCell
	assert(err == CL_SUCCESS);
	
	mem = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(cl_int) * 5112, NULL, NULL); // randUints
	err |= clSetKernelArg(kernel, argIdx++, sizeof(cl_mem), &mem);
	assert(err == CL_SUCCESS);
	err |= clSetKernelArg(kernel, argIdx++, sizeof(cl_int), &nCells); // randIdx
	assert(err == CL_SUCCESS);
	err |= clSetKernelArg(kernel, argIdx++, sizeof(cl_int), &leap); // leap
	assert(err == CL_SUCCESS);
	
	//nConseqLocalArgs = 0; // todo changed for debug
	for (int i = 0; i < nConseqLocalArgs; i++) {
		err |= clSetKernelArg(kernel, argIdx++, sizeof(cl_int) * sizeof(cl_float) * totRxnSets, NULL); // leap
	}
	assert(err == CL_SUCCESS);
	
	int runs = 1000000;
	// run kernel.
	size_t localSize = 1; // Heuristic to determine group size.
	for (int run = 0; run < runs; run++) {

		err |= clEnqueueNDRangeKernel(gpuHok->getCmdQueue(), kernel, 1, NULL, 
									  &localSize, &localSize, 0, NULL, NULL);	
		assert(err == CL_SUCCESS);
		err |= clFinish(gpuHok->getCmdQueue());
		assert(err == CL_SUCCESS);
		for (int k = 0; k < 10000; k++) {float x = sqrt((float)k); } // just kill some cpu cycles between calls
	}
	
	// we should probably clean up... oh well!
}

int main(){
	srand(22);
//	printf("Starting openclhier-erleap kernel unittests.\n");
//	runCL();
	printf("Starting hier_one_kernel.cl unittests.\n");
	runHokTests();
	printf("Test(s) pass!\n");
	
	
}
