/*
David Orendorff

Initial attempt at hierchical kernel for GPU.
4/23/10


*/

//// use CL_MEM_ALLOC_HOST_PTR for async mem allocation (maybe ?)
#pragma OPENCL EXTENSION cl_khr_local_int32_base_atomics : enable


///////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////// FUNCTIONS CALLED BY KERNELS ///////////////////////////////////////////

/* Indicates the maximum/minimum an event in parent may change species. */
// Q: faster if we help align this? (ie: add short to end)
struct DeltaX{
	ushort species;
	ushort cell;
	short delta;
	short pad;
};


void updateState(__local int *state, int rxnId, const int2 ins, const int2 outs){
	if (ins.x >= 0) state[ins.x]--;
	if (ins.y >= 0) state[ins.y]--;
	if (outs.x >= 0) state[outs.x]++;
	if (outs.y >= 0) state[outs.y]++;
}

// Returns a uniform float [0, 1) based on the random uniform integer.
// This will never return 0.
float randomFloat(unsigned int randomInt) {
	if (randomInt == 0) randomInt++; 
	return (float) randomInt / UINT_MAX;
}


int uniformInt(int lowestPossible, int highestPossible, unsigned int randomUint) {
	float r = 1+highestPossible - lowestPossible;
	//printf("uint %d\t", randomUint);
	int x = (int) (r * randomFloat(randomUint));
	return lowestPossible + x;
}

// For a reaction in *state* with at most 2 inp species
//		what will the reaction propensity be?
// Note: this should be inlined
float propensity(__local int *state, int2 inpSpecies, float rate){
	float D = rate;

	// I wonder if there is a faster way to do this step? (ie, having a species that is always 1, etc)
	if (inpSpecies.x >= 0){
		D *= (float)state[inpSpecies.x];
		
		if (inpSpecies.y >= 0 &&inpSpecies.x==inpSpecies.y){ 
			D *= ((float)state[inpSpecies.y]-1.0)/2.0;
		}
		else if(inpSpecies.y >= 0){
			D *= (float)state[inpSpecies.y];
		}
	}
	return D;
}



///////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////// KERNEL FUNCTIONS ///////////////////////////////////////////////////


// Each thread represents a 2nd level reaction set ("cell")
// ASSERT global_size == numRxnSets == local_size
__kernel
void computePropensities(__global const int *state, // const
						int stateSize,
						__global const int *cellStartIdx, // where each rxnSet's base rxns start in D
						__global const int2 *inpSpec, // Assume each rxn has at most N=2 inp species
						__global const float *rates,	// TODO: maybe cache to local
						__local int *locState,  // a local copy of state used for fast referencing
						__global float *D,
						__local float *Dloc)    // cache in which to compute D, will write to global at end
{
	event_t loadState = async_work_group_copy(locState, state, stateSize, 0);
	int rxnSetId = get_global_id(0) + 1;

	float sum = 0;
	wait_group_events(1, &loadState);
	
	
	int start = cellStartIdx[get_global_id(0)];
	int end = cellStartIdx[get_global_id(0)+1];
	int ratesOffset = cellStartIdx[0];
	for (int i = start; i < end; i++){
			int2 is = inpSpec[i-ratesOffset];
			float p = propensity(locState, is, rates[i-ratesOffset]);
			Dloc[i] = p;
			sum += p;
	}
	Dloc[rxnSetId] = sum;
	
	// now set root's val
	if (get_global_id(0)==0){
		float cSum = 0;
		for (int i=1; i <= get_global_size(0); i++){
			cSum += Dloc[i];
		}
		Dloc[0] = cSum;
	}
	
	barrier(CLK_LOCAL_MEM_FENCE);
	int totRxnSets = cellStartIdx[get_global_size(0)]; 

	// save to global memory
	event_t saveD = async_work_group_copy(D, Dloc, totRxnSets, 0);
	wait_group_events(1, &saveD);
	
}

// One kernel per species id
// boundS[i] = MAX(state[i] + deltaS[i]*(leap-1), 0)
__kernel
void boundState(__global const int *state, // const
				__global const int *deltaS, // const
				int leap,
				__global int *boundS) // where we put answer
{
	int gid = get_global_id(0);
	int x = state[gid];
	int dx = deltaS[gid];
	int newX = max(x + dx*(leap-1), 0);
	boundS[gid] = newX;
}


// Assume we know how many rxn events each cell will get, what will bound be on species?
// Assume this is called for EACH speciesId in state.
// ASSERT: all deltaX have same sign
__kernel
void boundCellState(__global const int *state,		
					__local int *locCellBoundS,	// our __local storage of state
					__global const int *rxnEvents,	// how many events for each cell's rxnId 
					__local int *locRxnEvents,	// local version of locRxnEvents
					const int nCellRxns, // all cells and root rxnSets
					__global const struct DeltaX *scDelta, // a list of DeltaX (speciesId, cellId, deltaX). 
					const int nSCD,			// number of deltaX
					__global const int *stateBoundL, // from the return of boundState. 
					__global int *cellBoundS)	// where the answer is to be stored
{
	int gid = get_global_id(0);
	int nThreads = get_global_size(0);
	bool findHighBound = scDelta[0].delta > 0; // if true, want max cell bound, otherwise min
	
	// load into local mem
	event_t loadRxnEvents = async_work_group_copy(locRxnEvents, rxnEvents, nCellRxns, 0); 

	locCellBoundS[gid] = state[gid];
	
	wait_group_events(1, &loadRxnEvents);
	barrier(CLK_LOCAL_MEM_FENCE);
	
	for (int i=gid; i < nSCD; i += nThreads){ 
		// TODO: we need to take provision for the instance where the cell which cares about s will 
		// not give 'too much delta' to s.
		struct DeltaX dx = scDelta[i];
		atom_add(&locCellBoundS[dx.species], locRxnEvents[dx.cell] * dx.delta);
	}
	
	// Make sure we have the smallest bound possible
	// note: locCellBoundS[i] < 0 taken care of here
	if (findHighBound){ 
		locCellBoundS[gid] = min(locCellBoundS[gid], stateBoundL[gid]);
	}else{
		locCellBoundS[gid] = max(locCellBoundS[gid], stateBoundL[gid]);
	}
	
	barrier(CLK_LOCAL_MEM_FENCE);
	cellBoundS[gid] = locCellBoundS[gid];

}					


// For optimal performance we assume there are more states (numCells) than trials. 
// This will run in O(numCells).
// Note: DBoundHigh, where we obtain unormalized probabilities, has rootD=D..[0], so we skip that entry.
// Note that (trials) randomInts will be used up. 
// Running time will be optimal when trials == get_global_size.
__kernel
void sampleCoarseEvents(__global const float *DboundHigh, 
					   int numCells,
					   __global int *rxnEvents,  // where the answer will be stored.
					   __local int *locSuccesses, 
					   int trials, 
					   __global const unsigned int *randomInts, 
					   int randomNumOffset){
	int gid = get_global_id(0);
	int nThreads = get_global_size(0);
	// need to zero out successes (?)
	
	float probSum = 0;
	for (int i = 1; i <= numCells; i++) {
		probSum += DboundHigh[i];
		locSuccesses[i] = 0;
	}
	locSuccesses[0] = trials;
	
	for (int i = gid; i < trials; i += nThreads) {
	
		__private float r = randomFloat(randomInts[i + randomNumOffset]);
		__private float targetVal = probSum * r;
		__private float sum = 0;
		__private int j = 0;
		do {
			j++;
			sum += DboundHigh[j];
		} while(sum < targetVal && j <= numCells);
		
		atom_add(&locSuccesses[j], 1);
		//locSuccesses[i] = i; // todo remove
	}
	
	barrier(CLK_LOCAL_MEM_FENCE);
	async_work_group_copy(rxnEvents, locSuccesses, numCells+1, 0);
}


// ASSERT: This should be called for every cell (not including root)
// This may be quicker running serially 
// TODO: instead of barrier use mem_fence
__kernel
void computePCoarse(__global const float *DboundHigh,
					__global const float *DboundCellHigh,
					__global const float *DboundLow,
					__global const float *DboundCellLow,
					__global const int *rxnEvents,
					__global const float *taus,
					__local float *multTerms, // where store our non 1.0 terms to multiply. ASSERT: size==L
					__global float *Pcoarse,
					int leap)
{

	int gid = get_global_id(0);
	int cellId = gid+1;
	
	__local int multIdx; multIdx = 0;
		
	barrier(CLK_LOCAL_MEM_FENCE);
	
	float ds = DboundHigh[cellId];
	float dcs = DboundCellHigh[cellId];
	int e = rxnEvents[cellId];
	float myP = 1.0;
	
	if (e > 0) {
		myP = pown(dcs/ds, e);
		//myP *= exp(-cellTaus[cellId-1] * (DboundCellLow[cellId] - DboundLow[cellId]));
	}
	
	if (myP != 1){
		int saveIdx = atom_add(&multIdx, 1); 
		multTerms[saveIdx] = myP;
	}
	
	barrier(CLK_LOCAL_MEM_FENCE); 

	if (gid==0){
		float tauSum = 0;
		for (int i = 0; i < leap; i++) 
			tauSum += taus[i];
		// note: all threads but gid==0 are no-ops here
		float P = exp(-tauSum * (DboundCellLow[0] - DboundLow[0]));

		for (int i=0; i < multIdx; i++){
			// I would like some way to 'blow up' if prob is greater than 1.0
			P *= multTerms[i];
		}
		
		Pcoarse[0] = P;
	}
	
}


// Each thread will drawn one sample from the first cell it comes across that has more fine rxns
// to fire. 
__kernel
void fineSample(__global const int *cellStartIdx,
				__global const float *DboundCellHigh,
				__local float *locDboundCellHigh,
				__global int *rxnEvents,
				__local int *locRxnEvents,
				__global uint *randomUints, 
				int randUintsStart) 
{
	__local int randIdx; randIdx = randUintsStart;
	int totCells = cellStartIdx[0] - 1; // number of children Rxns that root immediatly contains.
	size_t nRxnSetsTotal = cellStartIdx[totCells];
	int rxnStartIdx = totCells + 1;
	int totalRxn = cellStartIdx[totCells] - totCells - 1;
	
	
	async_work_group_copy(locDboundCellHigh, DboundCellHigh, nRxnSetsTotal, 0);
	async_work_group_copy(locRxnEvents, rxnEvents, totCells+1, 0);
	
	for (int i = rxnStartIdx; i < nRxnSetsTotal; i++)
		locRxnEvents[i] = 0;
	
	barrier(CLK_LOCAL_MEM_FENCE); // this may not be required.
//	if (get_local_id(0) != get_global_id(0)){
//		locRxnEvents[get_global_id(0)] = -1000000;
//	}

	while (1) { 
		private int cellId = 1;
		private int rxnCount = 0;
		
		// If rxnCount > 0 we can proceed, but we need to set locRxnEvents[cellId] to its new value (locRxnEvents[cellId]--).
		while (cellId <= totCells){
			rxnCount = atom_xchg(&locRxnEvents[cellId], 0);
			if (rxnCount > 0) break;
			
			cellId++;
		}
			
		if (cellId > totCells) break; // assert(rxnCount == 0)
		
		// We basically locked this cell until we could set its locRxnEvents[cellId] to the proper value.
		atom_xchg(&locRxnEvents[cellId], rxnCount - 1);
		
		// Now take a sample from the fine rxns in cellId.
		private float targetVal = locDboundCellHigh[cellId] * randomFloat(randomUints[atom_inc(&randIdx)]);
		private float sum = 0; 
		private int j = cellStartIdx[cellId - 1] - 1;
		private int end = cellStartIdx[cellId];
		do {
			j++;
			sum += locDboundCellHigh[j];
		} while(sum < targetVal && j < end);
		
		atom_inc(&locRxnEvents[j]);
	}
	
	// All fine reactions are allocated, write to global mem.
	
	barrier(CLK_LOCAL_MEM_FENCE);
	
	async_work_group_copy(&rxnEvents[rxnStartIdx], &locRxnEvents[rxnStartIdx], totalRxn, 0);
}


// Samples L taus and TauSum. L random numbers are used.
// In math: drawns an Erland RV with parameters (leap, rate) and 
// then samples L taus from the simplex.
__kernel
void tausSample(int leap, 
				__global float *rate,
				__global float *taus, // (return val) of size L
				__global unsigned int *randNums,
				int randNumStartIdx)
{
	float lambda = rate[0];
	__local int randIdx; randIdx = randNumStartIdx;
	
	barrier(CLK_LOCAL_MEM_FENCE); 

	int gid = get_global_id(0);
	int nThreads = get_global_size(0);
	
	for (int i = gid; i < leap; i+=nThreads) {
		int idx = atom_inc(&randIdx);
		taus[i] = -log(randomFloat(randNums[idx])) / lambda;
	}
}


// Create an permuted list of reactions both by parent id and rxn id.
__kernel 
void reactionPermutation(__global const int *rxnEvents, // of size |ReactionSet|
						__global const int *cellStartIdx,
						__local int *locOrderedRxns,
						__global int *orderedRxns, // where the output (ordered rxn) will go
						__global const unsigned int *randNums, // we need L of these 
						int startRandIdx) {

	return; // todo remove
	startRandIdx = 0; // todo remove
	__local int orderedRxnIdx; orderedRxnIdx = 0; // may need to sync this.
	int totThreads = get_global_size(0);
	int gid = get_global_id(0);
	const int nCells = cellStartIdx[0] - 1;
	const int numRxn = cellStartIdx[nCells] - cellStartIdx[0];


	// First we put the reactions into an array, we then will randomly perturb the array.
	for (int thread = gid; thread < numRxn; thread += totThreads) {
		int rxnId = cellStartIdx[0] + thread;
		int e = rxnEvents[rxnId];
		while (e > 0) {
			int i = atom_add(&orderedRxnIdx, 1);
			locOrderedRxns[i] = rxnId;
			e--;
		}
	}
	
	barrier(CLK_LOCAL_MEM_FENCE); // Make sure locOrderedRxns is done being written to.
	// we now have a list of reactions, permute them!

	// This random shuffle is very suboptimal (serially done) but more complicated methods are prone to error
	// and may be added later if needed.
	
	int leap = orderedRxnIdx;
	// todo uncomment this area

	if (gid==0) {
		for (int i = 0; i < leap-1; i++) {
			int randIdx = uniformInt(i, leap-1, randNums[startRandIdx++]);
			if (randIdx >= leap) {
				locOrderedRxns[i] = randIdx * -100;
			}
			//printf("uniform(%d, %d)->%d\n",i, leap-1, randIdx); 
			int t = locOrderedRxns[i];
			locOrderedRxns[i] = locOrderedRxns[randIdx];
			locOrderedRxns[randIdx] = t;
		}
	}

	async_work_group_copy(orderedRxns, locOrderedRxns, leap, 0);
}


/*
	ASSERT: call this for each BASE REACTION.
	ASSERT: DboundCellHigh is adjusted to be exactly of size |base rxns|

	ON ACCEPT:
	Write our updated state/time if we accept the move proposed by the |reactions|==LEAP. 
	
	ON REJECT
	Do not write to new state/time.
	
	Every iteration of the loop we load in our new state and update 
*/
__kernel
void computeFinalAccept(__global int *state,  // state/time updated if step is accepted
						__local int *locState,
						int nStates,
						__global float *time,
						__global const float *rates,
						__global const float *DboundCellHigh, // of size |base reactions|
						__global const float *taus,	// size of nEvents (L)
						__global const int *orderedRxnEvents,
						__local int *locOrderedRxnEvents,
						int nEvents,
						__global float *D, 
						__global float *DboundCellLow, 
						__global const int2 *inpSpec,
						__global const int2 *outSpec,
						float urand, // uniform random variable, used to accept/reject
						__global const float *Pcoarse, // from Pcoarse calculation.
						__local float *Ddeltas, // used to compute a product.
						int rxnStart,// where Rxn begin in DboundCellHigh, etc
						__global float *probAccept) // where final calculated probability is stored, regardless of acceptance.
{
	__local float locPcoarse; locPcoarse = *Pcoarse;
	__local float locPfine; locPfine = 1.0f;
	
	int gid = get_global_id(0);
	float rootD = D[0];
	float lowRootD = DboundCellLow[0];
	float expSum = 0;
	__local float tauSum; tauSum = 0;
	__local int DdeltasIdx; DdeltasIdx = 0;
				
		
	event_t loadState = async_work_group_copy(locState, state, nStates, 0);
	event_t loadOrderedRxnEvents = async_work_group_copy(locOrderedRxnEvents, orderedRxnEvents, nEvents, 0);
	
	
	// These should all be coalesced reads
	int myRxnId = gid + rxnStart;
	int2 inpS = inpSpec[gid];
	int2 outS = outSpec[gid];
	float rate = rates[gid];
	float highDcell = DboundCellHigh[myRxnId];
	
	wait_group_events(1, &loadState);
	wait_group_events(1, &loadOrderedRxnEvents);
	

	float myD = propensity(locState, inpS, rate);
	float prevD = myD;
	barrier(CLK_LOCAL_MEM_FENCE); 

	for (int i = 0; i < nEvents; i++){
		int rxnId = locOrderedRxnEvents[i];

		if (myRxnId==rxnId){
			// should only be one thread at a time, so no lock needed
			float d = myD / highDcell;
			if (d > 1.00001 || d < 0) // kind of a lame assert
				locPfine = pown(100.0f, i+1) * i;
			else
				locPfine *= d; // TODO
		}
		
		if (myD != prevD){
			// find delta and apply to D_root
			float delta = myD - prevD;
			int idx = atom_add(&DdeltasIdx, 1); // causing problems!
			Ddeltas[idx] = delta;
		}

		barrier(CLK_LOCAL_MEM_FENCE); // make sure D_root is updated

		
		if (gid==0){
		
			// update D (from previous iteration)
			float delta = 0;
			while (DdeltasIdx > 0){
				DdeltasIdx--;
				delta += Ddeltas[DdeltasIdx];
			}
			rootD += delta;

			float t = taus[i];
			float s = (rootD - lowRootD) * t;
			expSum += s; 
			tauSum += t;
			
		}
		
		
		if (myRxnId==rxnId){

			//updateState(locState, rxnId, inpS, outS);
			if (inpS.x >= 0) atom_dec(&locState[inpS.x]);//locState[inpS.x]--;
			if (inpS.y >= 0) atom_dec(&locState[inpS.y]);
			if (outS.x >= 0) atom_inc(&locState[outS.x]);
			if (outS.y >= 0) atom_inc(&locState[outS.y]);

		}
		
		barrier(CLK_LOCAL_MEM_FENCE);
		

		prevD = myD;
		myD = propensity(locState, inpS, rate);
	}
	
	// Find the final acceptance prob and see if we accept
	if (gid==0){
		locPfine *= exp(-expSum); 
		*probAccept = locPfine * locPcoarse;
	}
	
	barrier(CLK_LOCAL_MEM_FENCE);
	bool accept = locPfine * locPcoarse > urand;
	
	// if accept do a write to state out and tau out. otherwise just quit.
	if (accept){
		event_t saveState = async_work_group_copy(state, locState, nStates, 0);
		
		if (gid==0){
			*time += tauSum;
		}
		
		// speed idea: could update D here
		wait_group_events(1, &saveState);
	}
}






