/*
David Orendorff

Second attempt at hierchical kernel for GPU.
1/28/11

For now we assume anything in MY_CONSTANT memory will have fast access and there is no 
need to create local copies. 

*/

#pragma OPENCL EXTENSION cl_khr_local_int32_base_atomics : enable

// some #defines to help with control flow. Actual values don't matter. Todo: could probably use an enum here later on.
#define ACCEPT 21
#define REJECT 7
#define UNKNOWN 14


// there seems to be a bug with the NVIDIA compiler, and we cannot use __constant memory (?). 
// For now we #define CONSTANT until a fix comes along.
#define MY_CONSTANT global
//#define MY_CONSTANT constant

//l#define DEBUG

#ifdef DEBUG
//#define assert(x) \
//		if (! (x)) \
//		{ \
//			int lmn = (int) (x); \
//		}
#define assert(x) \
		if (! (x)) \
		{ \
			printf("Assert(%s) failed in %s:%d", #x, __FILE__, __LINE__); \
			exit(-1); \
		}
#else
	#define assert(x) 
//	#define assert(x) \
//			if (! (x)) \
//			{ \
//				x; \
//			}
	#define printf(fmt, ...);
#endif


/* Indicates the maximum/minimum an event in parent may change species. */
// Q: faster if we help align this? (ie: add short to end)
struct DeltaX{
	ushort species;
	ushort cell;
	short delta;
	short pad;
};


///////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////// MEMORY OUTPUT FUNCTIONS /////////////////////////////////////////////

void printfArray(local int *a, int len) {
		for (int i = 0; i < len; i++) {
			printf("%d, ", a[i]);
		}
		printf("\n");
}

void printfInt2Array(local int2 *a, int len) {
		for (int i = 0; i < len; i++) {
			printf("(%d, %d), ", a[i].x, a[i].y);
		}
		printf("\n");
}

void printFloatArray(local float *a, int len) {
		for (int i = 0; i < len; i++) {
			printf("%f, ", a[i]);
		}
		printf("\n");
}

void printfDeltaCellArray(local struct DeltaX* deltas, int len) {
	for (int i = 0; i < len; i++){ 
		printf("<<species(%hu), cell(%hu), delta(%hd)>>, ", deltas[i].species, deltas[i].cell, deltas[i].delta);
	}
	printf("\n");
}

void constMemPrintf(local int *cellStartIdx, local float  *rates, local int2 *inpSpec, 
					local int2 *outSpec, local int *highDeltaS, local int *lowDeltaS,
					local struct DeltaX *scDeltaHighCell, int nScdHighCell, local struct DeltaX *scDeltaLowCell, int nScdLowCell, 
					int stateSize, int gid, int nThreads) {
	if (gid != 0) return;
	
	int nCells = cellStartIdx[0] - 1;
	int totRxns = cellStartIdx[nCells];
	
	int totRxn = totRxns - nCells - 1;
	
	printf("InpSpecies...\n");
	printfInt2Array(inpSpec, totRxn);
	
	printf("OutSpecies...\n");
	printfInt2Array(outSpec, totRxn);
	
	printf("HighDeltaS...\n");
	printfArray(highDeltaS, stateSize);
	
	printf("LowDeltaS...\n");
	printfArray(lowDeltaS, stateSize);
	
	printf("CellStartIdx... \n");
	printfArray(cellStartIdx, cellStartIdx[0]); 

	printf("Rates... \n");
	printFloatArray(rates, totRxn);

	printf("scDeltaHighCell... \n");
	printfDeltaCellArray(scDeltaHighCell, nScdHighCell);
	
	printf("scDeltaLowCell... \n");
	printfDeltaCellArray(scDeltaLowCell, nScdLowCell);
}

void locMemPrintf(local int *highBoundS, local int *highBoundCellS, local int *state, local int *lowBoundCellS, local int *lowBoundS,
				  local float *DboundHigh, local float *DboundCellHigh, local float *D, local float *DboundCellLow, local float *DboundLow,
			      local int *rxnEvents, local int *orderedRxnEvents, local float *taus, int __local *cellStartIdx, int stateSize, 
				  int leap, global unsigned int *randomInts, int randIdxOffset, 
				  local float Pcoarse, local float Pfine, int gid, int nThreads) {
	if (gid != 0) return;
	int nCells = cellStartIdx[0] - 1;
	int totRxns = cellStartIdx[nCells];
	int totRxn = totRxns - nCells - 1;

	
	for (int i = 0; i < 5; i++){
		local int *st;

		switch (i){
		case 0: printf("highBoundS"); st = highBoundS; break;
		case 1: printf("highBoundCellS"); st = highBoundCellS; break;
		case 2: printf("state"); st = state; break;
		case 3: printf("lowBoundCellS"); st = lowBoundCellS; break;
		case 4: printf("lowBoundS"); st = lowBoundS; break;
		default: break;
		}
		printf("...\n");
		printfArray(st, stateSize);
	}
	
	printf("rxnEvents: \n");
	printfArray(rxnEvents, totRxns);
	printf("OrderedRxnEvents: \n");
	printfArray(orderedRxnEvents, leap);
	printf("taus: \n");
	printFloatArray(taus, leap);
	
	for (int j = 0; j < 5; j++){
		local float *d;

		switch (j){
		case 0: printf("DboundHigh"); d = DboundHigh; break;
		case 1: printf("DboundCellHigh"); d = DboundCellHigh; break;
		case 2: printf("D"); d = D; break;
		case 3: printf("DboundCellLow"); d = DboundCellLow; break;
		case 4: printf("DboundLow"); d = DboundLow; break;
		default: break;
		}
		printf("...\n");
		printFloatArray(d, totRxns);
	}
	
	printf("RandomOffsetIdx: %d\n", randIdxOffset);
	printf("RandomUInts: \n");
	for (int i = 0; i < leap * 9; i++) { // todo,remove 9
		printf("%X, ", randomInts[randIdxOffset+i]);
	}
	printf("\n");

	
	printf("PCoarse: %f\nPFine: %f\n", Pcoarse, Pfine);
	
	printf("\n\n");
}


///////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////// FUNCTIONS CALLED BY KERNEL ///////////////////////////////////////////	

// Returns a uniform float (0, 1) based on the random uniform integer.
// This will never return 0 or 1.
private float randomFloat(global uint randomInt) {
	float f = (float) randomInt / UINT_MAX;
	if (f >= 1.0 || f <= 0) f = nextafter(f, 0.5f);
	assert(0 < f && f < 1);
	return f;
}

private float myFabs(float f) {
	return f >= 0 ? f : -f;
}

// Returns a random int between lowestPossible and highestPossible inclusive based
//	on randomUint.
int uniformInt(int lowestPossible, int highestPossible, unsigned int randomUint) {
	int range = 1 + highestPossible - lowestPossible;
	int x = randomUint % range; // todo: this is a little buggy if range is too high (modulo bias)
	x += lowestPossible;
	assert(x >= lowestPossible && x <= highestPossible);
	return x;
}

void copy_arr(local int *dst, local int *src, int len, int gid, int nThreads){
	for (int i = gid; i < len; i += nThreads) {
		dst[i] = src[i];
	}
}

void rxnPropensity(local int *state, private int xSpecies, private int ySpecies, private float rate, private float *D){
	//assert(rate >= 0);
	float d = rate;

	if (xSpecies >= 0){
		d *= (float) state[xSpecies];
		if (state[xSpecies] < 0) d = 0.0;
	}
	
	if (ySpecies >= 0)  {
		d *= xSpecies != ySpecies ? state[ySpecies] : state[ySpecies] - 1;
		if (state[ySpecies] < 0) d = 0.0;
	}

	//assert(d >= 0);
	*D = d;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////// KERNEL FUNCTION ///////////////////////////////////////////////////


// Given an initial state S, we assume each species i can change by at most 
// deltaS[i] upon every step and cannot go negative, and there are leap steps. 
// Results of the bound on state after leap steps are returned in boundS.
void boundState(int leap, __local int *state, int stateLen, __local int *deltaS, 
				__local int *boundS, int gid, int nThreads){
	assert(leap > 0);
	for (int i = gid; i < stateLen; i += nThreads) {
		int deltaX = abs(deltaS[i]);
		assert(deltaX <= 2); // we assume each rxn may change s by at most 2.  
		boundS[i] = max(state[i] + deltaS[i] * (leap - 1), 0);
	}
}

// Calculates the propensity for all rxnChanels, cells and root.
void propensity(local int *state, int __local *cellStartIdx, 
				int2 __local *inpSpec, __local float *rates, local float *D, int gid, int nThreads){
	int nCells = cellStartIdx[0] - 1;
	int totRxnSets = cellStartIdx[nCells];
	assert(nCells + 1 < totRxnSets);
	
	for (int r = cellStartIdx[0] + gid; r < totRxnSets; r += nThreads) {
		int rxnId = r - cellStartIdx[0];
			int x = inpSpec[rxnId].x;
			int y = inpSpec[rxnId].s1; 
			float rate = rates[rxnId];
			//int z = x + y - rate; 

			float d;
			rxnPropensity(state, x, y, rate, &d); // todo: commenting this line out allows us to compile on cpu
			D[r] = d;
	}
	
	barrier(CLK_LOCAL_MEM_FENCE);
	
	// sum up children rxn for each cell
	for (int i = gid; i < nCells; i += nThreads) {
		int cellId = i + 1;
		float sum = 0.0;

		for (int r = cellStartIdx[i]; r < cellStartIdx[i+1]; r++) {
			sum += D[r];
		}
		D[cellId] = sum; 
	}
	
	barrier(CLK_LOCAL_MEM_FENCE);
	
	// finally sum up for root
	if (gid == 0) {
		float sum = 0.0;
		for (int i = 1; i < cellStartIdx[0]; i++) {
			sum += D[i];
		}
		D[0] = sum;
	}
}

// Sample coarseEvents (cell level events)
// optimal when nThreads == leap 
void sampleCoarseEvents(local float *DboundHigh, __local int *cellStartIdx,
						local int *rxnEvents, global uint *randUInts, 
						local int *randUIntsOffset, int gid, int nThreads) {
	float probSum = DboundHigh[0];
	int numCells = cellStartIdx[0] - 1;

	int leap = rxnEvents[0];
	
	for (int i = gid; i < leap; i += nThreads) {
	
		private float r = randomFloat(randUInts[i + *randUIntsOffset]);
		private float targetVal = probSum * r;
		private float sum = 0;
		private int j = 0;
		do {
			j++;
			sum += DboundHigh[j];
			assert(sum <= probSum);
		} while(sum < targetVal && j <= numCells);
		
		atom_inc(&rxnEvents[j]);
	}			
	
	if (gid == 0) 
		*randUIntsOffset += leap;	
}

// boundCellState(state, stateSize, cellStartIdx, rxnEvents, scDeltaLowCell, nScdLowCell, lowBoundS, lowBoundCellS, gid, nThreads);
void boundCellState(__local int *state, int nStates, __local int *cellStartIdx, 
					__local int *rxnEvents, 
					__local struct DeltaX *scDelta, int nCellDeltas, local int *boundS, 
					local int *boundCellS, int gid, int nThreads) {
	assert((cellStartIdx[0] - 1) * nStates >=  nCellDeltas);				
					
	copy_arr(boundCellS, state, nStates, gid, nThreads);
	
	bool findHighBound = scDelta[0].delta > 0; // if true, want max cell bound, otherwise min
	
	barrier(CLK_LOCAL_MEM_FENCE);
	
	for (int i = gid; i < nCellDeltas; i += nThreads){ 
		// TODO: we need to take provision for the instance where the cell which cares about s will 
		// not give 'too much delta' to s.
		struct DeltaX dx = scDelta[i];
		assert(dx.delta >= 0 == findHighBound);
		assert(dx.delta <= 0 == !findHighBound);
		//if (rxnEvents[dx.cell] > 0) {
			int delta = rxnEvents[dx.cell] * dx.delta;
			
			if (!findHighBound && delta > 0) { // todo: for debuggin only
				printf("\nError, delta(%d) = rxnEvents[Cell(%d)](%d) * dx.delta(%d)", delta, dx.cell, rxnEvents[dx.cell], dx.delta);
			}
			if (findHighBound && delta < 0) { // todo: for debuggin only
				printf("\nError, delta(%d) = rxnEvents[Cell(%d)](%d) * dx.delta(%d)", delta, dx.cell, rxnEvents[dx.cell], dx.delta);
			}
			
			//if (dx.species!=10) { // todo: for debugging, error occurs here.
//				printf("\nDelta(%d) = rxnEvents[Cell(%d)](%d) * dx.delta(%d)", delta, dx.cell, rxnEvents[dx.cell], dx.delta);
//				printf("\nboundCellS[%d]=%d", dx.species, boundCellS[dx.species]);
			//}
			assert(findHighBound ? delta >= 0 : delta <= 0);
			if(!findHighBound && boundCellS[10] > state[10]){
				printf("boundCellS[10]=%d", boundCellS[10]);
				assert(false);
			}
			assert(findHighBound || boundCellS[10] <= state[10]);
			
			atom_add(&boundCellS[dx.species], delta);
			
			if(!(findHighBound || boundCellS[10] <= state[10])){
				printf("species(%d), delta(%d)\n", dx.species, delta);
				assert(false);
			}
			//printf("\nboundCellS[%d]=%d", dx.species, boundCellS[dx.species]);
		//}
		if(!findHighBound && boundCellS[10] > state[10]){
			printf("\nboundCellS[10]=%d", boundCellS[10]);
			assert(false);
		}
	}
	
	
	// Make sure we have the smallest (absolute difference) bound possible
	// note: cellBoundS[i] < 0 taken care of here
	for (int i = gid; i < nStates; i += nThreads) {	
		if (findHighBound){ 
			assert(boundCellS[i] >= state[i]);
			boundCellS[i] = min(boundCellS[i], boundS[i]);
			if(!(boundCellS[i] >= state[i])) {
				assert(false);
			}
		}else{
			
			if(!(boundCellS[i] <= state[i])) {
				printfArray(rxnEvents, 43);
				printfArray(state, nStates);
				printfArray(boundCellS, nStates);
				
				printf("\nFinding %s bound: ", findHighBound ? "upper" : "lower");
				printf("boundCellS(%d), state(%d)\n", boundCellS[i], state[i]);
				assert(false);
			}
			boundCellS[i] = max(boundCellS[i], boundS[i]);
		}
		assert(boundCellS[i] >= 0);
		if(!(findHighBound ? boundCellS[i] >= state[i] : boundCellS[i] <= state[i])) { // todo: for debuggin only
			printf("\nFinding %s bound: ", findHighBound ? "upper" : "lower");
			printf("boundCellS[%d] = %d, state[%d] = %d", i, boundCellS[i], i, state[i]);
			
		}
	}
}

// Samples fine events.
// When considering if this func is efficient,  this func does best when 
// cell rxn events are dispersed uniformly and pretty fully (for all cells)
// PS - have nThreds > nCells results in no speedup.
// Also appends baseRxn's to orderedRxns list of size leap.
void sampleFineEvents(__local int *cellStartIdx, local float *DboundCellHigh, 
					  local int *rxnEvents, local int *orderedRxns, global uint *randomUints, 
					  local int *randUIntsOffset, local int *orderedRxnIdx, int gid, int nThreads) 
					  // note: orderedRxnIdx is just a temp var to this func but local vars only declared in kernels
					  {
	*orderedRxnIdx = 0;
	int totCells = cellStartIdx[0] - 1; 
	
	for (int c = gid + 1; c <= totCells; c += nThreads) {
		// take rxnEvents[c] samples.
		for (int i = 0; i < rxnEvents[c]; i++) { 
			float targetVal = DboundCellHigh[c] * randomFloat(randomUints[atom_inc(randUIntsOffset)]);
			float sum = 0; 
			int j = cellStartIdx[c - 1] - 1;
			int end = cellStartIdx[c];
			do {
				j++;
				sum += DboundCellHigh[j];
			} while(sum < targetVal && j < end);
			
			rxnEvents[j]++; // this is the only thread that may access this particular 
							// rxn event .:. no atomic op needed.
			int k = atom_inc(orderedRxnIdx);
			assert(j >= cellStartIdx[0]);
			orderedRxns[k] = j;
		}
	}
}

void sampleFineEventsInSerial(__local int *cellStartIdx, local float *DboundCellHigh, 
					  local int *rxnEvents, local int *orderedRxns, global uint *randomUints, 
					  local int *randUIntsOffset, int gid, int nThreads) {
	if (gid != 0) return;
	
	// For each cell c sample rxnEvents[c] baseRxns in cell.
	int nCells = cellStartIdx[0] - 1;
	int orderedRxnIdx = 0;
	
	for (int cellId = 1; cellId <= nCells; cellId++) {
		int e = rxnEvents[cellId];
		while (e > 0) {
			float targetSum = DboundCellHigh[cellId] * randomFloat(randomUints[(*randUIntsOffset)++]);
	
			int ridx = cellStartIdx[cellId - 1];
			float sum = 0.0;
		
			for ( ;; ridx++) {
				sum += DboundCellHigh[ridx];
				if (sum >= targetSum) break;
			}
			
			assert(ridx < cellStartIdx[cellId]);
			
			orderedRxns[orderedRxnIdx] = ridx;
			orderedRxnIdx++;
			rxnEvents[ridx]++;
			
			e--;
		}
	}
}

void tausSample(int leap, float lambda, local float *taus, global uint *randomUints, 
				local int *randUIntsOffset, int gid, int nThreads) {
	for (int i = gid; i < leap; i+=nThreads) {
		int idx = *randUIntsOffset + i;
		taus[i] = -log((float)randomFloat(randomUints[idx])) / lambda;
	}		
	if (gid==0)
		*randUIntsOffset += leap;		  
}

void permuteList(local int *list, int len, global uint *randomUints, 
				local int *randUIntsOffset, int gid, int nThreads){
				
	if (gid==0) {
		for (int i = len-1; i >= 1; i--) {
			int idx = *randUIntsOffset;
			int randIdx = uniformInt(0, i, randomUints[idx]); // inclusive to 0, i
			*randUIntsOffset = idx + 1;
			int t = list[i];
			assert(randIdx < len && randIdx >= 0);
			list[i] = list[randIdx];
			list[randIdx] = t;
		}
	}
}

// temp is used for calculations s.t. |temp| >= nCells
void computePCoarse(int leap, int nCells, local float *temp,
					local float *DboundHigh, local float *DboundCellHigh, 
					local float *DboundLow, local float *DboundCellLow, 
					local int *rxnEvents, local float *taus, int gid, int nThreads, local float *pCoarse) {

	for (int i = gid; i < nCells; i += nThreads) {
		int cellId = i + 1;
		if (rxnEvents[cellId] > 0) {
			temp[i] = pown(DboundCellHigh[cellId] / DboundHigh[cellId], rxnEvents[cellId]);
		} else {
			temp[i] = 1.0;
		}
		
		if (temp[i] < 0 || temp[i] > 1.0) {
			printf("temp[%d] = %f", i, temp[i]);
			assert(false);
		}
	}
	
	barrier(CLK_LOCAL_MEM_FENCE);
	
	if (gid == 0) {
		float tauSum = 0;
		for (int i = 0; i < leap; i++) {
			tauSum += taus[i];
			assert(taus[i] > 0);
		}
		
		*pCoarse = exp(-tauSum * (DboundCellLow[0] - DboundLow[0]));
		
		assert(0 <= *pCoarse && *pCoarse <= 1.0);
		
		float maxRatio = 1.0;
		for (int i=0; i < nCells; i++){
			if (DboundHigh[i+1] > 0)
				maxRatio = fmax(DboundCellHigh[i+1] / DboundHigh[i+1], maxRatio); // redoing calcuation, not efficient
			assert(maxRatio >= 0 && maxRatio <= 1.0);
			*pCoarse *= temp[i];
		}
		assert(maxRatio > 0 && maxRatio <= 1.0);
		
		*pCoarse *= pown(maxRatio, -leap); // 1.0f / pown(maxRatio, leap);
		assert(0 <= *pCoarse && *pCoarse <= 1.0);

	}
}

// Computes a lower bound on Pfine, allowing for early acceptance.
// This function works in serial.
void lowBoundPFine(int leap, local int *orderedRxnEvents, local float *DboundCellHigh, 
					local float *DboundCellLow, local float *taus, local float *lowBoundP, int gid, int nThreads) {
	if (gid == 0) {
		float p = 1.0;
		for (int i = 0; i < leap; i++) {
			int rxnsId = orderedRxnEvents[i];
			float r = DboundCellLow[rxnsId] / DboundCellHigh[rxnsId];
			if (r > 1.0) {
				printf("\n%f = %f / %f", r, DboundCellLow[rxnsId], DboundCellHigh[rxnsId]);
			}
			p *= r;
		}
		float tauSum = 0; // todo: this computation is also done in computePCoarse. fix!
		for (int i = 0; i < leap; i++) 
			tauSum += taus[i];
			
		float r2 = exp(-tauSum * (DboundCellHigh[0] - DboundCellLow[0]));
		//printf("DboundCellHigh(%f) DboundCellLow(%f)/n", DboundCellHigh[0], DboundCellLow[0]);
		assert(r2 >= 0 && r2 <= 1.0);
		p *= r2;
		
		*lowBoundP = p;
	}
}

void computePFine(int leap, local int *state, int __local *cellStartIdx, __local int2 *inpSpec,
				   __local int2 *outSpec, __local float *rates, local int *orderedRxnEvents, 
				   local float *myD, local float *DboundCellLow, local float *DboundCellHigh, local float *taus,
				   int gid, int nThreads, local float *Pfine) { 
	*Pfine = 1.0;
	float expSum = 0.0;
	int nCells = cellStartIdx[0] - 1;
	int totRxnSets = cellStartIdx[nCells]; // for debugging
	
	for (int i = 0; i < leap && *Pfine > 0; i++) {
		propensity(state, cellStartIdx, inpSpec, rates, myD, gid, nThreads); 
		
		#ifdef DEBUG
		if (i < leap - 1) { // don't look at the results after the final reaction.
			for (int j = 0; j < totRxnSets; j++) {
				if(!(DboundCellLow[j] <= myD[j] &&  myD[j] <= DboundCellHigh[j])){
					printf("Error ReactionSet(%d): DboundCellLow(%f), D(%f), DboundCellHigh(%f)\n", j, DboundCellLow[j], myD[j], DboundCellHigh[j]);
					assert(false);
				}
			}
		}
		#endif
		
		barrier(CLK_LOCAL_MEM_FENCE); // need barrier here?
		// myD now contains the correct propensities given state.
	
		if (gid == 0) {			
			int r = orderedRxnEvents[i];
			*Pfine *= myD[r] / DboundCellHigh[r];
		
			expSum += taus[i] * (myD[0] - DboundCellLow[0]); 
			int rxnId = r - nCells - 1;
			
			// update state.
			int2 inpS = inpSpec[rxnId];
			int2 outS = outSpec[rxnId];
			if (inpS.x >= 0){
				state[inpS.x]--;
			}
			if (inpS.y >= 0) {
				state[inpS.y]--;
			}
			if (outS.x >= 0) {
				state[outS.x]++;
			}
			if (outS.y >= 0) {
				state[outS.y]++;
			}
			
			int xVal = inpS.x >= 0 ? state[inpS.x] : 0;
			int yVal = inpS.y >= 0 ? state[inpS.y] : 0;
			if (min(xVal, yVal) < 0) {
				*Pfine = 0.0; // todo: this could be a problem??? maybe change propensity to be 0 if any r(p) are (-)
				assert(false); // todo: this assert not needed
			}
		}
	}
	
	barrier(CLK_LOCAL_MEM_FENCE); // need barrier here?
	if (gid == 0){
		*Pfine *= exp(-expSum); 
	}
} 

void setStateFromRxnEvents(local int *state, local int *orderedRxnEvents, int leap, 
						   local int2 *inpSpec, local int2 *outSpec, int nCells, int gid, int nThreads) {
	// This may be faster if done in serial (?)
	for (int i = gid; i < leap; i += nThreads) {
		int rxnId = orderedRxnEvents[i] - nCells - 1; 
		assert(rxnId >= 0);
		int2 ins = inpSpec[rxnId];
		int2 outs = outSpec[rxnId];

		if (ins.x >= 0) 
			atom_dec(&state[ins.x]); 
		if (ins.y >= 0) 
			atom_dec(&state[ins.y]);
		if (outs.x >= 0) 
			atom_inc(&state[outs.x]);
		if (outs.y >= 0) 
			atom_inc(&state[outs.y]);
		
	}
}

void setStateFromRxnEventsSerially(local int *state, local int *orderedRxnEvents, int leap, 
								   local int2 *inpSpec, local int2 *outSpec, int nCells, int gid, int nThreads) {
	if (gid!=0) return;
	for (int i = 0; i < leap; i++) {
		int rxnId = orderedRxnEvents[i] - nCells - 1; 
		assert(rxnId >= 0);
		int2 ins = inpSpec[rxnId];
		int2 outs = outSpec[rxnId];

		if (ins.x >= 0) 
			state[ins.x]--; 
		if (ins.y >= 0) 
			state[ins.y]--;
		if (outs.x >= 0) 
			state[outs.x]++;
		if (outs.y >= 0) 
			state[outs.y]++;
	}
}



//////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////// Hier Leap using local memory //////////////////////////////////////////

// Execute multiple steps of hier-leap at a time
// This kernel will exit if either:
//		1. The simulated time nearly exceeds totalSimtime (but is guaranteed to not exceed)
//		2. There are not enough pre-generated random numbers to continue (lastValidRandIdx < randIdx). If this is the case then *randIdx is set to -1.
//
// This kernel should run on any number of threads: 1 to MAX_LOCAL_THREADS
// todo: (possible) move randUints from global to MY_CONSTANT space.
__kernel
void hierLeap(global int *globalState, int stateSize,
			  global float *totalSimtime, // we will run the simulator for slightly longer (within L reaction events of totalSimTime)
			  int MY_CONSTANT *gCellStartIdx,
			  float MY_CONSTANT *gRates,	
			  int2 MY_CONSTANT *gInpSpec, // Assume each rxn has at most N=2 inp species
			  int2 MY_CONSTANT *gOutSpec,
			  MY_CONSTANT int *gHighDeltaS, MY_CONSTANT int *gLowDeltaS, 
			  MY_CONSTANT struct DeltaX *gScDeltaHighCell, int nScdHighCell,
			  MY_CONSTANT struct DeltaX *gScDeltaLowCell, int nScdLowCell,
			  global unsigned int *randomInts, global int *randIdx,
			  global int lastValidRandIdx,
			  int leap,  
			  local int *highBoundS, local int *highBoundCellS, local int *state, local int *lowBoundCellS, local int *lowBoundS, 
			  local float *DboundHigh, local float *DboundCellHigh, local float *D, local float *DboundCellLow, local float *DboundLow,
			  local int *rxnEvents, local int *prevState, local int *orderedRxnEvents, local float *taus,
			  // local copies of globa const vars
			  local int *cellStartIdx, local float  *rates, local int2 *inpSpec, 
			  local int2 *outSpec, local int *highDeltaS, local int *lowDeltaS, 
			  local struct DeltaX *scDeltaHighCell, local struct DeltaX *scDeltaLowCell
			  )
{
// ideas to reduce local mem consumption
// 1. we only need one of: local int *highDeltaS, local int *lowDeltaS
// 2. only need one of: local struct DeltaX *scDeltaHighCell, local struct DeltaX *scDeltaLowCell
// 3. only needs DboundLow[0] and DboundCellLow[0]
//		- note: pFineLowBound needs all of DboundCellLow
// 4. lowBoundPFine only needs tauSum, we only need taus for computePFine
// 5. we only need outSpec for setStateFromRxnEvents and pFine
// 6. only need (highBoundS, highBoundCellS) or (lowBoundCellS, lowBoundS)
// 7. DboundHigh only needed for sampleCoarse and Pcoarse
	printf("%f\n", *totalSimtime);

	local float endTime = totalSimtime[0];
	local float time = 0.0f;
	
	local int randUIntsOffset = *randIdx;
	__private int gid = get_global_id(0);
	__private int nThreads = get_global_size(0);
	local int randIntIdx = *randIdx;
	int nCells = gCellStartIdx[0] - 1;
	int totRxnSets = gCellStartIdx[nCells];
	int totRxn = totRxnSets - nCells - 1;
	
	local float pFine;
	
	// setup local memory
	async_work_group_copy(state, globalState, stateSize, 0);
	
	async_work_group_copy(cellStartIdx, gCellStartIdx, nCells+1, 0);
	async_work_group_copy(rates, gRates, totRxn, 0);
	async_work_group_copy(inpSpec, gInpSpec, totRxn, 0);
	async_work_group_copy(outSpec, gOutSpec, totRxn, 0);
	async_work_group_copy(highDeltaS, gHighDeltaS, stateSize, 0);
	async_work_group_copy(lowDeltaS, gLowDeltaS, stateSize, 0);
	
	int ucharsPerDeltaX = sizeof(struct DeltaX) / sizeof(uchar);
	async_work_group_copy((local uchar*)scDeltaHighCell, (global uchar*)gScDeltaHighCell, nScdHighCell * ucharsPerDeltaX, 0);
	async_work_group_copy((local uchar*)scDeltaLowCell, (global uchar*)gScDeltaLowCell, nScdLowCell * ucharsPerDeltaX, 0);
	
	copy_arr(prevState, state, stateSize, gid, nThreads);
	
	// where rn's are needed: sampleCoarseEvents(L), sampleFineEvents(L), permuteList(L), tausSample(L), isAccept(1)
	local int nRandsPerIteration = leap * 4 + 1;
	local float prevDboundLow = 0.0f;
	local float prevD = 1e9;
	
	// We'll also add a 2 constant to average D[0] and DboundLow[0] to avoid div 0.
	const float stdsQuitBefore = 2; 
	
	barrier(CLK_LOCAL_MEM_FENCE);
	
//	if (time == 0) {
//		// only print const memory once
//		constMemPrintf(cellStartIdx, rates, inpSpec, 
//					outSpec, highDeltaS, lowDeltaS,
//					scDeltaHighCell, nScdHighCell, scDeltaLowCell, nScdLowCell, stateSize, gid, nThreads);
//	}
	
	local int iter = 0;
	while (time + 2 * stdsQuitBefore * leap / (prevD + prevDboundLow) < endTime && iter < 1000) {
		if (randUIntsOffset + nRandsPerIteration > lastValidRandIdx) {
			randUIntsOffset = -1;
			break;
		}
		
		if (gid == 0) {
			randUIntsOffset++;
		}
		
		local int acceptStatus = UNKNOWN;
		
		prefetch(&randomInts[randUIntsOffset], nRandsPerIteration);
		local float randUniformFloat = randomFloat(randomInts[randUIntsOffset]);
			
		// init rxnEvents to 0
		for (int i = gid; i < totRxnSets; i += nThreads) {
			rxnEvents[i] = 0;
		}
	
		// Begin computation
		boundState(leap, state, stateSize, highDeltaS, highBoundS, gid, nThreads);
		propensity(highBoundS, cellStartIdx, inpSpec, rates, DboundHigh, gid, nThreads);
		
		rxnEvents[0] = leap;
		sampleCoarseEvents(DboundHigh, cellStartIdx, rxnEvents, randomInts, &randUIntsOffset, gid, nThreads);
			
		boundCellState(state, stateSize, cellStartIdx, rxnEvents, scDeltaHighCell, nScdHighCell, highBoundS, highBoundCellS, gid, nThreads);
		barrier(CLK_LOCAL_MEM_FENCE);
		propensity(highBoundCellS, cellStartIdx, inpSpec, rates, DboundCellHigh, gid, nThreads);
		
		
		boundState(leap, state, stateSize, lowDeltaS, lowBoundS, gid, nThreads);
		barrier(CLK_LOCAL_MEM_FENCE);
		propensity(lowBoundS, cellStartIdx, inpSpec, rates, DboundLow, gid, nThreads);
			
		boundCellState(state, stateSize, cellStartIdx, rxnEvents, scDeltaLowCell, nScdLowCell, lowBoundS, lowBoundCellS, gid, nThreads);
		// maybe write results to global, then read again
		barrier(CLK_LOCAL_MEM_FENCE);
		propensity(lowBoundCellS, cellStartIdx, inpSpec, rates, DboundCellLow, gid, nThreads);


		if (lowBoundCellS[0] == 0) {
			// we cannot evolve anymore.
			break;
		}

		local int tmp;
		sampleFineEvents(cellStartIdx, DboundCellHigh, rxnEvents, orderedRxnEvents, 
						 randomInts, &randUIntsOffset, &tmp, gid, nThreads);
		
		if (gid == 0) {
			randUIntsOffset += leap;
		}
		
		tausSample(leap, DboundLow[0], taus, randomInts, &randUIntsOffset, gid, nThreads);
		
		// See if we can accept or reject early.
		local float pFineEarly; 
		lowBoundPFine(leap, orderedRxnEvents, DboundCellHigh, DboundCellLow, taus, &pFineEarly, gid, nThreads);
		if (pFineEarly < 0 || pFineEarly > 1.0) {
			printf("pFineEarly(%f) on iteration %d.\n", pFineEarly, iter);
			assert(false);
		}
		
		local float pCoarse;
		local float *temp = D;
				

		computePCoarse(leap, nCells, temp, DboundHigh, DboundCellHigh, DboundLow, 
									   DboundCellLow, rxnEvents, taus, gid, nThreads, &pCoarse);
		assert(pCoarse >= 0 && pCoarse <= 1.0);
		
		barrier(CLK_LOCAL_MEM_FENCE);

		
		if (pCoarse * pFineEarly >= randUniformFloat) {
			acceptStatus = ACCEPT;
			
			// set state acording to orderedRxnEvents.
			setStateFromRxnEventsSerially(state, orderedRxnEvents, leap, inpSpec, outSpec, nCells, gid, nThreads);
		}	
		
		if (pCoarse < randUniformFloat) {
			assert(acceptStatus == UNKNOWN);
			acceptStatus = REJECT;
		}
		
		if (acceptStatus == UNKNOWN) {
			permuteList(orderedRxnEvents, leap, randomInts, &randUIntsOffset, gid, nThreads); 
			
			if (gid == 0) 
				randUIntsOffset += leap;

			barrier(CLK_LOCAL_MEM_FENCE);
			
			computePFine(leap, state, cellStartIdx, inpSpec,  
					outSpec, rates, orderedRxnEvents, D, DboundCellLow, DboundCellHigh, taus, gid, nThreads, &pFine);

			if (pFine < pFineEarly) {
				printf("pFineEarly(%f) was larger than pFine(%f). ERROR!\n", pFineEarly, pFine);
					locMemPrintf(highBoundS, highBoundCellS, state, lowBoundCellS, lowBoundS, DboundHigh, DboundCellHigh, D, DboundCellLow, DboundLow,
						rxnEvents, orderedRxnEvents, taus, cellStartIdx, stateSize, leap, randomInts, 
						randUIntsOffset, pCoarse, pFine, gid, nThreads);
				assert(false);
			}

			// if accept then update state
			barrier(CLK_LOCAL_MEM_FENCE);
			
			acceptStatus = pCoarse * pFine >= randUniformFloat ? ACCEPT : REJECT;
		}
		
		if (acceptStatus == ACCEPT) {
			copy_arr(prevState, state, stateSize, gid, nThreads);
			
			if (gid == 0){
				float tau = 0; // todo: this operation (tau sum) is repeated in code.
				for (int i = 0; i < leap; i++) {
					tau += taus[i];
				}
				time += tau;
			}
			printf("a");
		} else {
			// reset state back to origional 
			copy_arr(state, prevState, stateSize, gid, nThreads);
			printf("r");
			
		}
		
		// TODO: for debugging only, an error occurs after this if is hit.
//		if (myFabs(*totalSimtime - 9.999990) < 0.00001 && myFabs(time - 0.248554) < 0.00001) {
//						constMemPrintf(cellStartIdx, rates, inpSpec, 
//							outSpec, highDeltaS, lowDeltaS,
//							scDeltaHighCell, nScdHighCell, scDeltaLowCell, nScdLowCell, stateSize, gid, nThreads);
//				locMemPrintf(highBoundS, highBoundCellS, state, lowBoundCellS, lowBoundS, DboundHigh, DboundCellHigh, D, DboundCellLow, DboundLow,
//					rxnEvents, orderedRxnEvents, taus, cellStartIdx, stateSize, leap, randomInts, 
//					randUIntsOffset, pCoarse, pFine, gid, nThreads);
//		}
		
//		locMemPrintf(highBoundS, highBoundCellS, state, lowBoundCellS, lowBoundS, DboundHigh, DboundCellHigh, D, DboundCellLow, DboundLow,
//					rxnEvents, orderedRxnEvents, taus, cellStartIdx, stateSize, leap, randomInts, 
//					randUIntsOffset, pCoarse, pFine, gid, nThreads);
		if (gid == 0) {
			iter++;
			randUIntsOffset++;
		}
		prevDboundLow = DboundLow[0];
		prevD = D[0];
		barrier(CLK_LOCAL_MEM_FENCE);
	}

	// Done simulating: write state and time to global memory
	async_work_group_copy(globalState, state, stateSize, 0);
	// assert(*totalSimtime >= time); // as long as this does not happen often we are ok.
	*totalSimtime = time;
	*randIdx = randUIntsOffset;
	printf("X");
}





