

#include <boost/program_options.hpp>

using namespace boost;
namespace po = boost::program_options;

#include <sys/time.h>
#include <iostream>
#include <iterator>

//#include <tbb/tick_count.h>

#include "reactions/setup_topology/build_reaction_topology.h"
#include "reactions/reactions.h"
#include "reactions/reaction.h"
#include "reactions/connection_grandmaster.h"
#include "state/species.h"
#include "sim/simulator.h"

#include "sim/full_simulator.h"
#include "sim/time_hier_ssa.h"
#include "sim/time_hierleap.h"
#include "sim/time_hierleap_seperate_accept.h"
#include "sim/tau_leap_ssa.h"
#include "sim/gpuhier_tau_leap.h"
#include "sim/gpuhier_tau_leap_one_kernel.h"
#include "sim/gpu_ssa.h"
#include "sim/tau_leap_independent_cells.h"
#include "sim/threshold_optimal_leap.h"
#include "sim/time_hierleap_separate_clean.h"

#include "utils/common.h"
#include "utils/common_unittests.h"
#include "utils/gpu_interface.h"
#include "utils/sbml_interface.h"
#include "utils/ca_randgen.h"
#include "utils/gpu_randgen.h"
#include "utils/gpu_randgen_from_cpu.h"
#include "utils/random_number_provider_thread.h"

#define UNKNOWN_ALG -1
#define HIERLEAP_ALG 0
#define HIERLEAP_SEPERATE_ACCEPT_ALG 2
#define SSA_ALG 1

#include "sim/historian.h"

using namespace hierleap;
using namespace std;
//using namespace tbb;

const char* _usage = 
"\nGPU-HIER-LEAP\nMost theory by Eric Mjolsness. Coded by David Orendorff.\n\n\tusage: gpu_hierleap sbml_input runs duration history_interval [output_dir] [leap]\n\nSpecifing a positive leap (L) will be held constant. Non-specify L or an L of negative will have leap be dynamically updated.\n"; 

void birthDeathProcess(vector<Reaction*> & reactions, vector<Species*> & species,
						double birthRate, double deathRate){
	Species *s = new Species("X", 100);
	Reaction *birth = new Reaction(NULL, birthRate);
	birth->addReactant(s, 1);
	birth->addProduct(s, 2);

	Reaction *death = new Reaction(NULL, deathRate);
	death->addReactant(s, 1);

	species.push_back(s);
	reactions.push_back(birth);
	reactions.push_back(death);
}

// Gives species an exp distribution with peak at maxNum and falling down towards zero.
// guarantee that all values Xi >= 0.
//	Example common vals are mu=1000, std=275.
vector<int> genExpDistSpecies(int totNumSpecies, int maxNum, double std){
	
	vector<int> sVals;
	for (int i=0; i < totNumSpecies; i++){
		//double x = maxNum - erand() * std;
		double x = erand() * std;
		//double x = urand() * maxNum;
		if (x<=0)
			x = 1;
		sVals.push_back(int(x));
	}
	return sVals;
}

// Randomize species counts in all cells BUT the 1st
void randomizeSpecies(State *s, int maxNum, double std){
	int nUniqueSpecies = s->X().size();
	int nSpeciesTypes = s->speciesTypes();
	vector<int> x = s->X();
	vector<int> newX = genExpDistSpecies(nUniqueSpecies - nSpeciesTypes, maxNum, std);
	
	for (int i=nSpeciesTypes; i < nUniqueSpecies; i++){
		x[i] = newX[i-nSpeciesTypes];
	}
	s->setXdirectly(x);
}


void setReactionsToExpRates(ReactionSet *root){
	if (root->isReaction()){
		double r = 10 - erand();
		//double r = urand();
		r = max(0.001, r);
		root->asReaction()->_setRate(r);
	}else{
		BOOST_FOREACH(ReactionSet *r, root->getDependents()){
			setReactionsToExpRates(r);
		}
	}
}

void reportDtPropensitiesROC(State *s, Reactions *root){
	setReactionsToExpRates(root);
	ReactionTauLeaper *simAlgROC = new TimeHierarchySSA(root, s); // Get ROC like statistics
	exit(0);
}

GPUHierLeapSim* setUpGpu(State *s, Reactions *root, bool useGpu, int algorithmId) {
	cl_int deviceArgs = useGpu ? CL_DEVICE_TYPE_GPU : CL_DEVICE_TYPE_CPU;
	
	GpuInterface *cagpu = new GpuInterface(deviceArgs, "/Users/dweo/research/ca_randgen/ca_randgen/ca_randgen.cl", false); // todo: disable profiling for faster 
	
	unsigned int seed = urand_int(0, UINT_MAX);
	//GpuRandgen *gpuRandgen = new GpuRandGenFromCpu(seed, cagpu->getCmdQueue());
	GpuRandgen *gpuRandgen = new CaRandgen(seed, cagpu->getContext(), cagpu->getCmdQueue(), cagpu->extractKernel("ca_rng_get_ints"), cagpu->maxWorkgroupSize());
	
	GpuInterface *hiergpu = new GpuInterface(cagpu->getCmdQueue(), "/Users/dweo/research/gpusvm/hier.cl");
	
	GPUHierLeapSim* gpuhierleapsim;
	switch(algorithmId) {
		case HIERLEAP_ALG : 
			gpuhierleapsim = new GpuhierTauLeapOneKernel(s, root, hiergpu, gpuRandgen);
			printf("Simulating with HierLeap.\n");
			break;
		case SSA_ALG :
			gpuhierleapsim = new GpuSSA(s, root, hiergpu, gpuRandgen);
			printf("Simulating with SSA.\n");
			break;
		default:
			printf("\nUnknown algorithm specified.");
			assert(false);
	}

	return gpuhierleapsim;
}

void parseArgs(int argc, char * argv[], string &sbml_file, int &runs, double &update_period, double &duration, 
			    int &leap, bool &useOpenCl, bool &openClUseGpu, int &maxThreads, int &algorithmId, double &diffusionRate, int &cellsPerMetaCell, int &x, int &y, int &z, bool &output_to_file, string &out_dir, int &seed) {
	
	string computeDevice, algorithm;
	int otf, help;

	try {
	po::options_description desc("Allowed options");
	desc.add_options()
		("help,h", po::value<int>(&help)->default_value(0), "produce help message, follow by 1")
		("sbml_file,f", po::value<string>(&sbml_file), "sbml file containing cell network")
		("runs,r", po::value<int>(&runs)->default_value(1), "number of times to simulate the network")
		("update_period,p", po::value<double>(&update_period)->default_value(1.0),"results of simulation written at these intervals")
		("simulation_time,t", po::value<double>(&duration),"each run simulated for num seconds ")
		("initial_leap,l", po::value<int>(&leap)->default_value(4), "leap simulation will use, negative indicates automatic updating")
		("opencl_compute_device,c", po::value<string>(&computeDevice), "valid options are CPU, GPU or None")
		("max_threads,M", po::value<int>(&maxThreads)->default_value(-1), "maximum number of threads, -1 for program defined.")
		("algorithm,a", po::value<string>(&algorithm)->default_value("HIER"), "algorithm to use for simulation, valid values are HIER or SSA")
		("diffusion_rate,u", po::value<double>(&diffusionRate)->default_value(0.0), "is diffusion added to the network, set to zero if no diffusion")
		("cells_per_metacell,m", po::value<int>(&cellsPerMetaCell)->default_value(10), "number of cells replicated in x direction")
		("x,x", po::value<int>(&x)->default_value(1), "number of cells replicated in x direction")
		("y,y", po::value<int>(&y)->default_value(1), "number of cells replicated in y direction")
		("z,z", po::value<int>(&z)->default_value(1), "number of cells replicated in z direction")
		("file_output,F", po::value<int>(&otf)->default_value(1), "do not output to file")
		("output_directory,o", po::value<string>(&out_dir)->default_value("."), "directory where output will be written")
		("seed,s", po::value<int>(&seed)->default_value(time(NULL)), "random seed, uses clock if not specified")
		;
		
		po::variables_map vm;  
		po::store(po::parse_command_line(argc, argv, desc), vm);
		po::notify(vm); 
	
		if (help) {
			cout << "\nThere seems to be a problem with your input arguments.\n" << 
				"Unfortionatly there is a bug in OS X and I cannot automatically output the help string." << "\n";
			//cout << desc; // this should work but doesnt (bug in os X 10.6?), probably has to do with _GLIBCXX_FULLY_DYNAMIC_STRING
			exit(1);
		}

	
    }
    catch(std::exception& e) {
        cerr << "error: " << e.what() << "\n";
		assert(false);
    }
    catch(...) {
        cerr << "Exception of unknown type!\n";
		assert(false);
    }
	
	output_to_file = otf; // dumb hack because of bug in parse_arguments
	
	// todo: should be more assertions of input validity here.
	if (computeDevice.compare("CPU") == 0) {
		useOpenCl = true;
		openClUseGpu = false;
	} else if(computeDevice.compare("GPU") == 0) {
		useOpenCl = true;
		openClUseGpu = true;
	}else if (computeDevice.compare("None") == 0) {
		useOpenCl = false;
		openClUseGpu = false;
	} else {
		printf("\nA valid OpenCl compute device was not specified. Options are: CPU, GPU or None.\n");
		exit(1);
	}
	
	if (algorithm.compare("HIER") == 0) {
		algorithmId = HIERLEAP_ALG;
	} else if (algorithm.compare("SSA") == 0) {
		algorithmId = SSA_ALG;
	} else if (algorithm.compare("HIER_SEP") == 0) {
		algorithmId = HIERLEAP_SEPERATE_ACCEPT_ALG;
	}else {
		algorithmId = UNKNOWN_ALG;
	}

	assert(duration > 0);
	assert(update_period > 0 && update_period <= duration);
	assert(runs > 0 && diffusionRate >= 0 && x > 0 && y > 0 && z > 0);
	assert(leap > 0);
}

void run_tests() {
	ReactionPermutation<int>::test();
	CommonTests().test();
	exit(99);
}

// definition of neighbor. B is As neighbord:
//			if A changes X AND B needs X for propensity.
//			if B ''			   A ''
void connectNeighbors(ConnectionGrandmaster *cg, Reactions *root, State *s){
	vector<ReactionSet*> & deps = root->getDependents();
	BOOST_FOREACH(ReactionSet *rs, deps) {
		rs->clearNeighbors();
	}
	for (int sid = 0; sid < s->X().size(); sid++) {
		
		BOOST_FOREACH(Reaction *rA, *cg->reactionsChangingSpecies(s->species(sid))) {
			BOOST_FOREACH(Reaction *rB, *cg->changesPropensityFor(s->species(sid))) {
				rA->getParent()->addNeighbor(rB->getParent()->asReactionSet());
				rB->getParent()->addNeighbor(rA->getParent()->asReactionSet());
			}
		}
		
//		vector<Reactions*> inCommon;
//		BOOST_FOREACH(ConnectingRxns *cr, *cg->increasingConnectingGroups(sid)) {
//			if (!contains(inCommon, cr->_parentConnecting))
//				inCommon.push_back(cr->_parentConnecting);
//		}
//		BOOST_FOREACH(ConnectingRxns *cr, *cg->decreasingConnectingGroups(sid)) {
//			if (!contains(inCommon, cr->_parentConnecting))
//				inCommon.push_back(cr->_parentConnecting);
//		}
//		
//		// The cells in inCommon are all nbrs.
//		BOOST_FOREACH(Reactions *c1, inCommon) {
//			BOOST_FOREACH(Reactions *c2, inCommon) {
//				c1->addNeighbor(c2);
//				c2->addNeighbor(c1);
//			}
//		}
	}
}

int main (int argc, char * argv[]) {	
	int seed;
	int runs, leap = -1;
	double update_period;
	double duration;
	int cellsPerMetacell;
	int x; int y; int z; // dimensions of cell simultion
	bool hasDiffusion;
	bool useOpenCl;
	bool openClUseGpu;
	bool outputToFile;
	int algorithmId;
	int maxThreads;
	double diffusionRate; 
	string sbmlFile;
	string outputDir;

	parseArgs(argc, argv, sbmlFile, runs, update_period, duration, leap, useOpenCl, openClUseGpu, maxThreads, algorithmId, diffusionRate, cellsPerMetacell, x, y, z, outputToFile, outputDir, seed);
	hasDiffusion = diffusionRate > 0;
	
	init_random(seed);
	printf("seed: %d\n", seed);
	
	// run_tests(); // xxx: only have this for testing
	
    // First build the typical parent cell with the 'template' reaction list
	vector<Reaction*> cellRxns;
	vector<Species*> cellSpecies;
	
	// Duplicate the cell over a 3D grid and add diffusion
	State *state; Reactions *root; 

	int levels = 3; // including root
	assert(levels ==3); // Current simulation method requires 3 levels exactly

	SBMLInterface si;
	vector<int> initialState = si.sbml_interface((char*)sbmlFile.c_str(), cellRxns, cellSpecies);
	
	BuildReactionTopology brt;
	brt.build(cellRxns, cellSpecies,levels, x,y,z, hasDiffusion, diffusionRate, cellsPerMetacell, state, root );
	
	//brt.buildFlat(cellRxns, cellSpecies,  levels, x,  y,  z, hasDiffusion, diffusionRate, state, root);

	state->setXbySpeciesId(initialState);
	root->setupInteractingRxn(state);
	ConnectionGrandmaster *cg = ConnectionGrandmaster::getConnectionGrandmaster(root, state);
	connectNeighbors(cg, root, state);

	brt.assertConsistency(root, state);
	
	printf("\nTotal cells: %d\nTotal ReactionSets: %d\nTree height: %d\n", 
		   root->getDependents().size(), root->numRxnSets(), root->height());
	//root->print(state, "disp");
	root->print(state, "dicpCXxsmnN");
	//print("reactoin channels %d\n", root->descendantReactionList().size());
	//exit(-1);
	
	Historian *hist = NULL;
	if (outputToFile) {
		hist = new Historian(state, (char*)outputDir.c_str(), duration, update_period);
	} else {
		printf("Not writing to file.\n");
		hist = new DumbHistorian(duration); 
	}
	assert(hist);

	State *savedState = new State(state); 
	
	struct timeval start, end;
    long mtime, seconds, useconds;   
	root->_reconnectNbrs();
	
    
	/* If simulating on the GPU */
	if (useOpenCl) {
		GPUHierLeapSim *gpuSim = setUpGpu(state, root, openClUseGpu, algorithmId);
		
		// simulate trajectories and record
		for (int i=0; i < runs; i++){
			state->setXdirectly(savedState->X());
			gettimeofday(&start, NULL);
			gpuSim->run(state, duration, leap, hist);
			
			//sleep(4);
			gettimeofday(&end, NULL);

			double deltaTime = elapsedSeconds(start, end);
			printf("Run%d of duration %f sec\n", i, deltaTime);
			hist->nextRun();
		}
		delete gpuSim;
	}else {
		FullSimulator sim(root);
		//TauLeapSSA *simAlg = new TauLeapSSA(root, state); //new TimeHierLeap(root, state);
		//ReactionTauLeaper *simAlg = new TauLeapIndependentCells(root, state);
		ReactionTauLeaper *simAlg = NULL;
		switch(algorithmId) {
			case SSA_ALG:
				simAlg = new TauLeapSSA(root, state);
				break;
			case HIERLEAP_ALG:
				//simAlg = new TimeHierLeap(root, state);
				simAlg = new TimeHierLeapSepClean(root, state, maxThreads);
				
				break;
			case HIERLEAP_SEPERATE_ACCEPT_ALG:
				simAlg = new TimeHierLeapSeperateAccept(root, state);
				break;
			case UNKNOWN_ALG: assert(false);
		}
		
		root->preCache();
#ifdef CALCULATE_P_ACCEPT_ONLY
		// find optimal L
		double pAcceptThresh = 0.5;
		printf("finding optimal L that is greater than %f.\n", pAcceptThresh);
		ThresholdOptimalLeap optimaL(simAlg, root, state);
		double lAvg = optimaL.optimalLeap(pAcceptThresh);
		printf("Avg_L_of_marginal_accept %f\n", lAvg);
		exit(0);
#endif
		
		// simulate trajectories and record
		printf("Simulation started with %s.\n", simAlg->name());
		for (int i=0; i < runs; i++){
			state->setXdirectly(savedState->X());
			gettimeofday(&start, NULL);
			//tick_count t0 = tick_count::now();
			sim.run(simAlg, hist, root, duration, leap, state); 
			//tick_count t1 = tick_count::now();
			gettimeofday(&end, NULL);
			
			
			double deltaTime = elapsedSeconds(start, end);
			printf("Run%d of duration %f sec\n", i, deltaTime);
			//printf("TBB Runtime reported as %f sec.\n", (t1-t0).seconds());
			hist->nextRun();
		}
	}

	delete hist;
	delete cg;
	
    return 0;
}