/*
 *  connection_grandmaster.cpp
 *  gpu_hierleap
 *
 *  Created by David Orendorff on 1/9/12.
 *  Copyright 2012 UC Irvine. All rights reserved.
 *
 */

#include "connection_grandmaster.h"

#include <vector.h>

#include "../state/species.h"
#include "reaction_set.h"
#include "reaction.h"
#include "reactions.h"
#include "../state/state.h"
#include "reaction_set_map.h"
#include "../utils/common.h"

using namespace hierleap;
using namespace std;

Reactions *ConnectionGrandmaster::_rootUsedForInit = NULL;
State * ConnectionGrandmaster::_stateUsedForInit = NULL;
ConnectionGrandmaster * ConnectionGrandmaster::_lastInitedGrandmaster = NULL;

ConnectionGrandmaster * ConnectionGrandmaster::getConnectionGrandmaster(Reactions *root, State *s) {
	checkNotNull(s);
	checkNotNull(root); 
	Reactions *& lastRoot = ConnectionGrandmaster::_rootUsedForInit;
	State *& lastState = ConnectionGrandmaster::_stateUsedForInit;
	ConnectionGrandmaster *& lastCG = ConnectionGrandmaster::_lastInitedGrandmaster;
	
	
	if (s == lastState && root == lastRoot && lastCG) {
		return lastCG;
	}
	lastRoot = root;
	lastState = s;
	lastCG = new ConnectionGrandmaster(lastRoot, lastState);
	return lastCG;
}

// same as above, but ASSUMES state is the origional value. 
// NOTE: getConnectionGrandmaster(Reactions *root, State *s) must have already been called.
ConnectionGrandmaster * ConnectionGrandmaster::getConnectionGrandmaster(Reactions *root) {
	assert(root==ConnectionGrandmaster::_rootUsedForInit);
	return checkNotNull(ConnectionGrandmaster::_lastInitedGrandmaster);
}

// note: all work is done upon initialization, so beware if time is important.
// we aim to examine all Reaction and Species O(c) number of times.
ConnectionGrandmaster::ConnectionGrandmaster(Reactions *root, State *state) {
	//ReactionSetMap<Species*> s;
	//s[root] = NULL;
	int nrs = root->numRxnSets();
	
	_effectedRxnPropensities = new ReactionSetMap<vector<Reaction*>* >(nrs);
	_effectedSpecies = new ReactionSetMap<vector<Species*>* >(nrs);
	_increaseSpecies = new ReactionSetMap<vector<DeltaSpecies*>* >(nrs);
	_decreaseSpecies = new ReactionSetMap<vector<DeltaSpecies*>* >(nrs);
	
	BOOST_FOREACH(ReactionSet *r, root->getOrderedProgeny()) {
		_increaseSpecies->set(r, new vector<DeltaSpecies*>());
		_decreaseSpecies->set(r, new vector<DeltaSpecies*>());
		_effectedSpecies->set(r, new vector<Species*>());
		_effectedRxnPropensities->set(r, new vector<Reaction*>());

	}
	
	int ns = state->speciesSize();
	_reactionsChangingSpecies = new vector<vector<hierleap::Reaction*> *>(ns, NULL);
	_propensityChangedBySpecies = new vector<vector<hierleap::Reaction*> *>(ns, NULL);
	_increasingConnectingGroups = new vector<vector<ConnectingRxns*> *>(ns, NULL);
	_decreasingConnectingGroups = new vector<vector<ConnectingRxns*> *>(ns, NULL);
	BOOST_FOREACH(Species *s, state->species()) {
		(*_reactionsChangingSpecies)[s->getUniqueId()] = new vector<Reaction*>();
		(*_propensityChangedBySpecies)[s->getUniqueId()] = new vector<Reaction*>();
	}
	
	// Setup for each Reaction
	BOOST_FOREACH(Reaction *r, root->descendantReactionList()) {
		_increaseSpecies->set(r, new vector<DeltaSpecies*>());
		_decreaseSpecies->set(r, new vector<DeltaSpecies*>());

		
		r->updateSpeciesInfo(state); // could cause small mem leak
				
		BOOST_FOREACH(int speciesId, r->speciesWhichChange()) {
			int delta = r->delta(speciesId);
			assert(delta);
			(*_reactionsChangingSpecies)[speciesId]->push_back(r);
			
			_effectedSpecies->get(r)->push_back(state->species(speciesId));
			_effectedSpecies->get(checkNotNull(r->getParent()))->push_back(state->species(speciesId));
			DeltaSpecies *d = new DeltaSpecies(speciesId, delta);
			if (delta > 0) {
				_increaseSpecies->get(r)->push_back(d);
				_increaseSpecies->get(r->getParent())->push_back(d); // add to parent cell
			}
			if (delta < 0) {
				_decreaseSpecies->get(r)->push_back(d);
				_decreaseSpecies->get(r->getParent())->push_back(d); // add to parent cell
			}
		}
	
		// link reaction propensity term with species
		for (int i = 0; i < r->numReactants(); i++) {
			(*_propensityChangedBySpecies)[r->getReactantId(i)]->push_back(r);
		}
	}
	
	BOOST_FOREACH(ReactionSet *rs, root->getDependents()) {
		remove_duplicates_and_sort_pointers(*_increaseSpecies->get(rs));
		remove_duplicates_and_sort_pointers(*_decreaseSpecies->get(rs));
		remove_duplicates_and_sort_pointers(*_effectedSpecies->get(rs));
	}
	
	BOOST_FOREACH(Reaction *fires, root->descendantReactionList()) {
		BOOST_FOREACH(Species *s, *_effectedSpecies->get(fires)) {
			BOOST_FOREACH(Reaction *rChanges, *changesPropensityFor(s)) {
				_effectedRxnPropensities->get(fires)->push_back(rChanges);
			}
		}
		remove_duplicates_and_sort_pointers(*_effectedRxnPropensities->get(fires));
	}
	
	// Set this up
	// vector<vector<ConnectingRxns*> *> *_increasingConnectingGroups;
	BOOST_FOREACH(Species *s, state->species()) {
		(*_increasingConnectingGroups)[s->getUniqueId()] = new vector<ConnectingRxns*>();
		(*_decreasingConnectingGroups)[s->getUniqueId()] = new vector<ConnectingRxns*>();
		BOOST_FOREACH(Reaction *r, *reactionsChangingSpecies(s)) {
			int delta = r->delta(s->getUniqueId());
			assert(delta);
			
			ConnectingRxns *cr = new ConnectingRxns(r->getParent(), delta);
			if (delta > 0) {
				(*_increasingConnectingGroups)[s->getUniqueId()]->push_back(cr);
			} else {
				(*_decreasingConnectingGroups)[s->getUniqueId()]->push_back(cr);
			}

		}
		// TODO, XXX: this leaks a little bit, don't worry about it for now.
		remove_duplicates_and_sort_pointers(*(*_increasingConnectingGroups)[s->getUniqueId()]);
		remove_duplicates_and_sort_pointers(*(*_decreasingConnectingGroups)[s->getUniqueId()]);
	}
	
	// now remove dups and take max delta.
	
	_setSpeciesParents(state);
}


void ConnectionGrandmaster::_setSpeciesParents(State *state) {
	BOOST_FOREACH(Species *s, state->species()) {
		vector<Reaction*> * invovledRxn = changesPropensityFor(s);
		if (invovledRxn->size() == 0) {
			invovledRxn = (*_reactionsChangingSpecies)[s->getUniqueId()];
		}
		assert(invovledRxn->size() > 0);
		s->setParent((*invovledRxn)[0]->getParent());
	}
}

// If r fires, these Species change.
vector<hierleap::Species*> * ConnectionGrandmaster::effectedSpecies(hierleap::Reaction *r) {
	return _effectedSpecies->get(checkNotNull(r));
}

vector<Reaction*> * ConnectionGrandmaster::effectedRxn(Reaction *r) { 
	return checkNotNull(_effectedRxnPropensities->get(r));
}

vector<Reaction*> * ConnectionGrandmaster::changesPropensityFor(Species *s) { 
	return (*_propensityChangedBySpecies)[s->getUniqueId()]; 
}

vector<hierleap::Reaction*> * ConnectionGrandmaster::reactionsChangingSpecies(Species *s) {
	return (*_reactionsChangingSpecies)[s->getUniqueId()];
}

vector<DeltaSpecies*> * ConnectionGrandmaster::possiblyIncreasedByCellEvent(Reactions *cell) {
	return _increaseSpecies->get(checkNotNull(cell));
}

vector<DeltaSpecies*> * ConnectionGrandmaster::possiblyDecreasedByCellEvent(Reactions *cell) {
	return _decreaseSpecies->get(checkNotNull(cell));
}

// Which 'cells' may effect this species when a rxn occurs within ConnectingRxns.
vector<ConnectingRxns*> * ConnectionGrandmaster::increasingConnectingGroups(int speciesId) {
	assert(speciesId >= 0);
	assert(speciesId < _stateUsedForInit->speciesSize());
	return checkNotNull((*_increasingConnectingGroups)[speciesId]);
}

vector<ConnectingRxns*> * ConnectionGrandmaster::decreasingConnectingGroups(int speciesId) {
	assert(speciesId >= 0);
	assert(speciesId < _stateUsedForInit->speciesSize());
	return checkNotNull((*_decreasingConnectingGroups)[speciesId]);
}
