/*
 *  connection_grandmaster.h
 *  gpu_hierleap
 *
 *  Created by David Orendorff on 1/6/12.
 *  Copyright 2012 UC Irvine. All rights reserved.
 *
 */

#ifndef CONNECTION_GRANDMASTER_H_
#define CONNECTION_GRANDMASTER_H_

#include <vector.h>

#include "../utils/common.h"

class State;
class Reactions;

namespace hierleap{
	class ConnectingRxns;
	class Species;
	class Reaction;
}
template <class T> class ReactionSetMap;
class DeltaSpecies;

using namespace std;


/*
 * This class deals with species <--> reaction interactions and reaction <--> reaction
 * interactions via species.
 * As a side effect, species->parent are all setup.
 *
 * Once everything is set up, all methods will take O(c) time.
 */
class ConnectionGrandmaster {
public:
	
	static ConnectionGrandmaster * getConnectionGrandmaster(Reactions *root, State *s);
	
	// same as above, but ASSUMES state is the origional value. 
	// NOTE: getConnectionGrandmaster(Reactions *root, State *s) must have already been called.
	static ConnectionGrandmaster * getConnectionGrandmaster(Reactions *root);
	
	// When r fires, which rxn's WILL have their propensities changed.
	vector<hierleap::Reaction*> * effectedRxn(hierleap::Reaction *r);
	
	// If r fires, these Species change.
	vector<hierleap::Species*> * effectedSpecies(hierleap::Reaction *r);
	
	// The species s is involved in the propensity term for the following vector<Reaction*>.
	vector<hierleap::Reaction*> * changesPropensityFor(hierleap::Species *s);
	
	vector<hierleap::Reaction*> * reactionsChangingSpecies(hierleap::Species *s);
	
	vector<DeltaSpecies*> * possiblyIncreasedByCellEvent(Reactions *cell);
	
	vector<DeltaSpecies*> * possiblyDecreasedByCellEvent(Reactions *cell);
	
	// Which parent reaction sets *may* increase the *reactants* of this species if they are fired. 
	// The some reaction in the returned reaction group MAY increase species by delta amount.
//	vector<ConnectingRxns*> * increasingCells(Species *s) {
//	}
	
	// Which 'cells' may effect this species when a rxn occurs within ConnectingRxns.
	vector<hierleap::ConnectingRxns*> * increasingConnectingGroups(int speciesId);
	
	vector<hierleap::ConnectingRxns*> * decreasingConnectingGroups(int speciesId);
	
	~ConnectionGrandmaster() {
		_stateUsedForInit = NULL;
		_rootUsedForInit = NULL;
		_lastInitedGrandmaster = NULL;
		delete _effectedRxnPropensities; delete _effectedSpecies; 
		delete _reactionsChangingSpecies; delete _propensityChangedBySpecies;
		delete _increasingConnectingGroups; delete _decreasingConnectingGroups;
	}
	
private:
	
	// note: all work is done upon initialization, so beware if time is important.
	// we aim to examine all Reaction and Species O(c) number of times.
	ConnectionGrandmaster(Reactions *root, State *s);
	
	// All species in state are set to one of their (possible overdefined) parents.
	void _setSpeciesParents(State *state);
	
	// Vars for some code to return most recently created ConnectionGrandmaster iff state and root match values 
	// of previous invocation.
	static Reactions *_rootUsedForInit;
	static State *_stateUsedForInit;
	static ConnectionGrandmaster *_lastInitedGrandmaster;
	
	ReactionSetMap<vector<hierleap::Reaction*>* > *_effectedRxnPropensities;
	ReactionSetMap<vector<hierleap::Species*>* > *_effectedSpecies;
	
	// The species, and amounts, each ReactionSet may increase/decrease.
	ReactionSetMap<vector<DeltaSpecies*>* > *_increaseSpecies;
	ReactionSetMap<vector<DeltaSpecies*>* > *_decreaseSpecies;
	
	// for each _reactionsChangingSpecies[speciesId]: a list of Reaction which change speciesId
	vector<vector<hierleap::Reaction*> *> *_reactionsChangingSpecies;
	
	// species is in the lhs of Reaction
	vector<vector<hierleap::Reaction*> *> *_propensityChangedBySpecies;
	
	// of size |species|
	vector<vector<hierleap::ConnectingRxns*> *> *_increasingConnectingGroups;
	vector<vector<hierleap::ConnectingRxns*> *> *_decreasingConnectingGroups;
	
};

/*
 * Note: the overload operators are a little wacky, in order for us to easily use stl
 *     built in functions. 
 * DeltaSpecies species are equal if their speciesId are equal. 
 * However, d1 is SMALLER than d2 if abs(d1->delta) > abs(d2->delta)
 */
class DeltaSpecies {
public:
	DeltaSpecies(int species, int deltaS) : speciesId(species), delta(deltaS) {}
	
	// compare speciesId, if those are equal, compare delta. 
	bool operator<(DeltaSpecies d) {
		return speciesId != d.speciesId ? 
		speciesId < d.speciesId : 
		abs(delta) < abs(d.delta);
	}
	
	bool operator==(DeltaSpecies d) {
		return speciesId == d.speciesId;
	}
	
	const int speciesId;
	const int delta;
};

#endif 