/*
 *  DeltaPropensityBounds.cpp
 *  gpu_hierleap
 *
 *  Created by David Orendorff on 9/29/11.
 *  Copyright 2011 UC Irvine. All rights reserved.
 *
 */

#include "delta_propensity_bounds.h"

#include "reaction.h"
#include "reactions.h"
#include "../utils/common.h"
#include "../utils/dynamic_priority_queue.h"
#include "reaction_permutation.h"
#include "reaction_set_map.h"
#include "connection_grandmaster.h"

using namespace hierleap;


DeltaPropensityBounds::DeltaPropensityBounds(Reactions *rxns) : _dpqUpper(NULL) {
	BOOST_FOREACH(Reaction *r, rxns->getRoot()->descendantReactionList()) {
//		if (r->getParent() != rxns)
//			continue;
		_deltaCalculate[r] = new DeltaEquation(rxns, r);
		
		if (_deltaCalculate[r]->isAlwaysZero()) {
			delete _deltaCalculate[r];
			_deltaCalculate[r] = NULL;
			continue;
		}
		bool slotted = false;
		if (_deltaCalculate[r]->isCanIncreaseD()) {
			_positiveChangeRxn.push_back(r);
			slotted = true;
		}
		if (_deltaCalculate[r]->isCanDecreaseD()) {
			_negativeChangeRxn.push_back(r);
			slotted = true;
		}
		assert(slotted);
	}
	// todo xxx: change args 
	//vector<Reaction*> & baseRxn = rxns->descendantReactionList();
	int n = _positiveChangeRxn.size();
	//DynamicPriorityQueue<Reaction*, double> *dpq = new DynamicPriorityQueue<Reaction*, double>(baseRxn, vector<double>(rxns->descendantReactionList().size(), 0));
	//vector<double> deltas = vector<double>(_positiveChangeRxn.size(), 0);
	//DynamicPriorityQueue<Reaction*, double> *dpq = DynamicPriorityQueue<Reaction*, double>::buildMaxPriorityQueue(_positiveChangeRxn, deltas);
}

// for any rxn touching a state mentioned in rxnsPerm, re-calculate delta for that rxn.
// Note: this is inneficient if rxnsPerm is NULL (it has to rebuild entire queue)
void DeltaPropensityBounds::updateUpperBoundDeltaDs(ReactionPermutation<Reaction*> *rxnsPerm, vector<int> & updatedState) {
	if (!_dpqUpper || !rxnsPerm) {
		vector<double> delta = _upperBoundDeltaD(updatedState, _positiveChangeRxn);
		if (_dpqUpper) delete _dpqUpper;
		//_dpqUpper = DynamicPriorityQueue<Reaction*,double>::buildMaxPriorityQueue(_positiveChangeRxn, delta);
		_dpqUpper = new DynamicPriorityQueue(_positiveChangeRxn, delta, true);
		return;
	}
	
	// For each state that has changed, change the delta for corresponding reactions
	// we'd like to not repeat species or reactions, 
	//		-todo: we could build a cached list of the rxn's that need to change for each rxn which may fire
	BOOST_FOREACH(Reaction *r, rxnsPerm->_myReactionList()) {
//		if (!isRelevantToThisCell(r)) 
//			continue;
		
		BOOST_FOREACH(Reaction *r2, *r->changesPropensityFor()) {
			if (r2->getParent() != r->getParent() || !_deltaCalculate[r2]) continue; // todo: may need to revisit this.
			double newDelta = _getDeltaCalculator(r2)->evalPositive(updatedState);
			if (newDelta == 0) continue;
			_dpqUpper->changePriority(r2, newDelta);
		}
		//			BOOST_FOREACH(Species *s, r->effectedSpecies()) {
		//				BOOST_FOREACH(Reaction* r2, s->changesPropensityFor()) {
		//					
		//				}
		//			}
	}
}

// Calculates upper bounding delta if each rxn fires given state. Reaturns a 
// vector of rxn.size() with respective values.
vector<double> DeltaPropensityBounds::_upperBoundDeltaD(vector<int> & state, vector<Reaction*> & rxn) {
	vector<double> bounds(rxn.size(), 0);
	for (int i = 0; i < rxn.size(); i++) {
		bounds[i] = _getDeltaCalculator(rxn[i])->evalPositive(state);
	}
	return bounds;
}

// Given state, return the maximum change in propensity given that one rxn may occur
double DeltaPropensityBounds::upperBoundDeltaD(vector<int> & state) {
	double maxDelta = 0.0;
	
	if (_dpqUpper) {
		maxDelta = _dpqUpper->topKey();
	} else {
		// for each reaction that can occur, see maximum total propensity may change.
		BOOST_FOREACH(Reaction *rxn, _positiveChangeRxn) { 
			//maxDelta = max(maxDelta, _getDeltaCalculator(rxn)->eval(state));
			maxDelta = max(maxDelta, _getDeltaCalculator(rxn)->evalPositive(state));
		}
	}
	return maxDelta;
}

double DeltaPropensityBounds::lowerBoundDeltaD(vector<int> & state) {
	double minDelta = 0.0;
	
	// for each reaction that can occur, see maximum total propensity may change.
	BOOST_FOREACH(Reaction *rxn, _negativeChangeRxn) { 
		//minDelta = min(minDelta, _getDeltaCalculator[rxn]->eval(state));
		minDelta = min(minDelta, _getDeltaCalculator(rxn)->evalNegative(state));
	}
	return minDelta;
}

DeltaEquation::DeltaEquation(Reactions *cell, Reaction *r) {
	buildCoefficients(r, cell);
	// todo: maybe seperate into positive/negative eqs.
}

// Build coefficients related to firing r, only consider rxn under cell.
void DeltaEquation::buildCoefficients(Reaction *rFired, Reactions *cell) {
	checkNotNull(cell);
	ConnectionGrandmaster *cg = ConnectionGrandmaster::getConnectionGrandmaster(rFired->getRoot());
	vector<Reaction*> * er = cg->effectedRxn(rFired);
	BOOST_FOREACH(Reaction *rOfCell, *(cg->effectedRxn(checkNotNull(rFired)))) { 
		if (!rOfCell->grandchildOf(cell)) 
			continue;
		
		Coefficient *c = Coefficient::buildRelevantCoefficient(rFired, rOfCell);
		
		BOOST_FOREACH(Coefficient *cExist, _coeffs) {
			if (!c) break;
			
			if (cExist->isCombinable(c)){
				cExist->combine(c);
				delete c;
				c = NULL;
			}
		}
		if (c) {
			_coeffs.push_back(c);
		}
	}
}

/*
 * Can the total propensity of the system be increased when the reaction associated with
 * this is fired?
 */
bool DeltaEquation::isCanIncreaseD() {
	//BOOST_FOREACH(Coefficient *c, _coeffs) {
	for (int i = 0; i < _coeffs.size(); i++) {
		Coefficient *c = _coeffs[i];
		if (c->isCanIncreaseD())
			return true;
	}
	return false;
}

bool DeltaEquation::isCanDecreaseD() {
	//BOOST_FOREACH(Coefficient *c, _coeffs) {
	for (int i = 0; i < _coeffs.size(); i++) {
		Coefficient *c = _coeffs[i];
		if (c->isCanDecreaseD())
			return true;
	}
	return false;
}

/*
 * how much D will change after rxn is fired
 */
double DeltaEquation::eval(vector<int> & state) {
	double d = 0.0;
	//BOOST_FOREACH(Coefficient *c, _coeffs) {
	for (int i = 0; i < _coeffs.size(); i++) {
		Coefficient *c = _coeffs[i];
		double diff = c->eval(state);
		d += diff;
	}
	return d;
}

/*
 * how much D will change after rxn is fired; only considers positive terms in equation.
 */
double DeltaEquation::evalPositive(vector<int> & state) {
	double d = 0.0;
	//BOOST_FOREACH(Coefficient *c, _coeffs) {
	for (int i = 0; i < _coeffs.size(); i++) {
		Coefficient *c = _coeffs[i];
		double term = c->eval(state);
		if (term > 0)
			d += term;
	}
	return d;
}

/*
 * how much D will change after rxn is fired; only considers negative terms in equation.
 */
double DeltaEquation::evalNegative(vector<int> & state) {
	double d = 0.0;
	//BOOST_FOREACH(Coefficient *c, _coeffs) {
	for (int i = 0; i < _coeffs.size(); i++) {
		Coefficient *c = _coeffs[i];
		double term = c->eval(state);
		if (term < 0)
			d += term;
	}
	return d;
}

// if firedRxn occurs, the change of propensity of r can be described with the 
// returned coefficient
Coefficient* Coefficient::buildRelevantCoefficient(Reaction* firedRxn, Reaction *r) {
	bool possiblyEffects = false;
	BOOST_FOREACH(int sid, firedRxn->speciesWhichChange()) {
		for (int i = 0; i < r->numReactants(); i++) {
			possiblyEffects |= (r->getReactantId(i) == sid);
		}
	}
	if (!possiblyEffects) 
		return NULL;
	
	int nSpec = r->numReactants();
	
	int spec0 = nSpec >= 1 ? r->getReactantId(0) : -1;
	int spec1 = nSpec == 2 ? r->getReactantId(1) : (nSpec > 0) && r->getReactantStoic(0) == 2 ? spec0 : -1;
	
	int specDelta[2] = {firedRxn->delta(spec0), firedRxn->delta(spec1)};
	
	if (!(specDelta[0] || specDelta[1])) // no matches
		return NULL;
	
	if (!specDelta[0]) {
		swap(spec0, spec1);
		swap(specDelta[0], specDelta[1]); 
	}
	return new Coefficient(spec0, spec1, r->rate(), specDelta);
}