#ifndef DELTA_PROPENSITY_BOUNDS_H_
#define DELTA_PROPENSITY_BOUNDS_H_

/*
 *  DeltaPropensityBounds.h
 *  gpu_hierleap
 *
 *  Created by David Orendorff on 9/29/11.
 *  Copyright 2011 UC Irvine. All rights reserved.
 *
 */

#include <vector>
#include <assert.h>

#include "reaction_permutation.h"
#include "../utils/common.h"
#include "reaction_set_map.h"

using namespace std;

class Coefficient;
class DeltaEquation;
class Reactions;

class DynamicPriorityQueue;
template <class R>
class ReactionPermutation;

//template <class int>
//class ReactionMap;

namespace hierleap{
	class Reaction;
	class Species;
}
using namespace hierleap;


/*
 * For Reaction {r} contained in Reactions rxns, ~efficiently calculate the max/min change
 * in D after exactly one reaction event. ﻿﻿
 */
class DeltaPropensityBounds {
public:
	
	DeltaPropensityBounds(Reactions *rxns);
	
	// for any rxn touching a state mentioned in rxnsPerm, re-calculate delta for that rxn.
	void updateUpperBoundDeltaDs(ReactionPermutation<Reaction*> *rxnsPerm, vector<int> & updatedState);
	
	// Given state, return the maximum change in propensity for *this* cellParent, given that any rxn may occur
	double upperBoundDeltaD(vector<int> & state);
	
	double lowerBoundDeltaD(vector<int> & state);
	
private:
	// Calculates upper bounding delta if each rxn fires given state. Reaturns a 
	// vector of rxn.size() with respective values.
	vector<double> _upperBoundDeltaD(vector<int> & state, vector<Reaction*> & rxn);
	
	DeltaEquation *_getDeltaCalculator(Reaction *r) { 
		return checkNotNull(_deltaCalculate[r]);
	}
	
	
	//DynamicPriorityQueue<Reaction*, double, tr1::unordered_map<Reaction*, int> > * _dpqUpper;
	DynamicPriorityQueue * _dpqUpper;
	vector<Reaction*> _positiveChangeRxn; // these rxn *may* increase D after being fired
	vector<Reaction*> _negativeChangeRxn;
	ReactionSetMap<DeltaEquation*> _deltaCalculate;
};

// If Reaction fires given state, how will total 
// propensity of cell change? A DeltaEquation contains a list of coefficients
// that are essentially dot-multiplied with state.
class DeltaEquation {
public:
	// If r fires, compute in functional form the change of propensity of cell.
	DeltaEquation(Reactions *cell, Reaction *r);
	
	/*
	 * Can the total propensity of the cell system be increased when the reaction associated with
	 * this is fired?
	 */
	bool isCanIncreaseD();
	
	bool isCanDecreaseD();
	
	bool isAlwaysZero() {	
		return _coeffs.size() == 0;
	}
	
	/*
	 * how much D will change after rxn is fired
	 */
	double eval(vector<int> & state);
	double evalPositive(vector<int> & state);
	double evalNegative(vector<int> & state);
	
private:
	void buildCoefficients(Reaction *r, Reactions * cell);
	
	vector<Coefficient*> _coeffs;
};

class Coefficient {
private:
	const int _speciesId0;
	const int _speciesId1;
	int _deltaState[2]; // how much speciedId0/1 changes 
	double _rateSum;
	
	Coefficient(int species0, int species1, double rateSum, int deltaState[2]) : 
			_speciesId0(species0), _speciesId1(species1), 
			_rateSum(rateSum) {
		_deltaState[0] = deltaState[0];
		_deltaState[1] = deltaState[1];
		if (_speciesId0 < 0 ) 
			assert(_speciesId0 == -1);
		if (_speciesId1 < 0 ) 
			assert(_speciesId1 == -1);
		assert(_speciesId0 >= 0); // think here a bit more.
		assert(abs(_deltaState[0]) + abs(_deltaState[1]) <= 4);
	}
	
	// returns state if stateId is valid, otherwise 1
	inline int _identityState(vector<int> & state, int stateId) {
		return stateId >= 0 ? state[stateId] : 1;
	}
	
public:
	
	static Coefficient* buildRelevantCoefficient(Reaction* firedRxn, Reaction *r);
	
	/*
	 * returns the difference in propensity of this reaction given state when firedRxn occurs.
	 */
	double eval(vector<int> & state) {
		if (isAlwaysEvalZero()) return 0.0;
		if (_speciesId0 >= 0 && state[_speciesId0] + _deltaState[0] < 0) return 0.0;
		if (_speciesId1 >= 0 && state[_speciesId1] + _deltaState[1] < 0) return 0.0;
		
		// dimerisation reaction, special case
		if (_speciesId0 == _speciesId1) {
			double d = _deltaState[0];
			assert(d == _deltaState[1]);
			int s = _identityState(state, _speciesId0);
			int diff = combinations(s+d, 2) - combinations(s, 2);
			if (s+d-1 < 0) diff = 0;
			return _rateSum * diff;
		}
			
		int s0 = _identityState(state, _speciesId0);
		int s1 = _identityState(state, _speciesId1);
		int diff = (s0 + _deltaState[0]) * (s1 + _deltaState[1]) - s0 * s1;
		return _rateSum * diff; 
	}
	
	double multiplierConst() {
		return _rateSum;
	}
	
	bool isAlwaysEvalZero() {
		return !_deltaState || _rateSum == 0;
	}
	
	bool isCanIncreaseD() {
		return _rateSum * _deltaState[0] > 0 || _rateSum * _deltaState[1] > 0;
	}
	
	bool isCanDecreaseD() {
		return _rateSum * _deltaState[0] < 0 || _rateSum * _deltaState[1] < 0;
	}
	
	bool isCombinable(Coefficient *c) {
		if (!c) return false;
		return (_speciesId0 == c->_speciesId0 && _speciesId1 == c->_speciesId1) && 
		_deltaState[0] == c->_deltaState[0] && _deltaState[1] == c->_deltaState[1];
	}
	
	// s.t: this.eval = this_old.eval + c.eval
	void combine(Coefficient *c) {
		assert(isCombinable(c));
		assert(_deltaState[0] == c->_deltaState[0] && _deltaState[1] == c->_deltaState[1]);
		_rateSum += c->_rateSum;
	}
};

#endif