/*
 * PropensityBounds.cpp
 *
 *  Created on: Sep 10, 2009
 *      Author: dweo
 */

#include "propensity_bounds.h"

#include <math.h>
#include <vector.h>
#include <boost/foreach.hpp>

#include "../globals.h"

#include "../state/state.h"
#include "../state/species.h"
#include "reactions.h"
#include "reaction.h"
#include "reaction_set.h"
#include "connection_grandmaster.h"

using namespace hierleap;

//class State;
//#include "../state/state.h"
//#include "reaction.h"

void PropensityBounds::_setupDeltaN(Reactions *rootOfAllReaction, int numSpeciesTypes, bool diffusion){
	int initVal = diffusion ? 1:0; // all species types may be changed by 1+ if there is diffusion
	
	_maxDeltaN.resize(0);
	_maxDeltaN.resize(numSpeciesTypes, initVal);
	_lowDeltaN.resize(numSpeciesTypes, initVal);
	
	_setupMaxDeltaN(rootOfAllReaction);
	_setupLowDeltaN(rootOfAllReaction);
}

PropensityBounds::PropensityBounds(Reactions *root, int numSpeciesTypes, bool diffusion) : _root(checkNotNull(root)){
//	int initVal = diffusion ? 1:0; // all species types may be changed by 1+ if there is diffusion
//	
//	_maxDeltaN.resize(0);
//	_maxDeltaN.resize(numSpeciesTypes, initVal);
//	_lowDeltaN.resize(numSpeciesTypes, initVal);
//	assert(!root->isReaction());
//	
//	// move down until we find a cell parent
//	//vector<ReactionSet*> 
//	Reactions *rxns = _sufficientReactionsParent(root);
//	
////	ReactionSet *r = root ; 
////	for (; !(r->getDependents()[0]->isReaction()); 
////			r = (r->getDependents())[0]){
////		assert(!r->isReaction());
////	}
////	
//	//Reactions *rxns = r->asReactions();
//	_setupMaxDeltaN(rxns);
//	_setupLowDeltaN(rxns);
	
	Reactions *rxns = _sufficientReactionsParent(root);
	_setupDeltaN(rxns, numSpeciesTypes, diffusion);
	
	// also setup _cellReactions, need to use cell with most diffusion neighbors, too
	assert(rxns->onlyContainReaction());
	_diffusionRate = 0;
	BOOST_FOREACH(ReactionSet *rs, rxns->getDependents()){
		Reaction *rn = rs->asReaction();
		// copy reaction to _cellReactions (avoid diffusion rxn)
		if (strcmp("diffusion", rn->myType())==0){
			_diffusionRate = rn->rate();
			continue;
		}
		Reaction *rNew = new Reaction(rn);
		_cellReactions.push_back(rNew);
	}
	
	assert(_diffusionRate >= 0);
	_maxNumNeighbors = root->maxBaseCellNumNeighbors(); 
}


double PropensityBounds::upperBoundDrxn(Reaction* r, State *state, int leap){
	vector<int> & stateByType =  state->maximalSpeciesByType(r); // base case returns species values in cell
	vector<int> & highState = _highBoundState(stateByType, leap);// TODO confirm this logic
	return r->DbyType(highState);
}

// Assume we are in a 'closed world' from the parent of r down
double PropensityBounds::upperBoundDrxns(Reactions* root, State *state, int leap){
	
	// we want to calculate delta D for the highest state, and add D to this
	vector<int> & highStateByType= state->maximalSpeciesByType(root); // indexed by type, what is maximal value of species for each type
										// in ANY constituent cells (independent of leap, etc)

	// TODO XXX Off by one (for diffusion)
	vector<int> & highBoundHighState = _highBoundState(highStateByType, leap); // this should account for diffusion too, we only need to input


	double Dc = 0.0;
	double Dchigh = 0.0;
	for (unsigned int i=0; i < _cellReactions.size(); i++){
		Reaction* r = _cellReactions[i];
		Dc += r->DbyType(highStateByType);
		Dchigh += r->DbyType(highBoundHighState);
	}
	
	double deltaD = Dchigh - Dc;
	
	// also add in points for diffusion
	for (int i=0; i < highStateByType.size(); i++){
		deltaD += (highBoundHighState[i] - highStateByType[i]) * _diffusionRate * _maxNumNeighbors;
	}

	return root->D(state) + deltaD;
}

double PropensityBounds::lowerBoundDrxns(Reactions* r, State *state, int leap){
	vector<int> & highStateByType= state->maximalSpeciesByType(r);
	vector<int> & lowBoundHighState = _lowBoundState(highStateByType, leap); 
	
	double Dc = 0.0;
	double Dclow= 0.0;
	for (unsigned int i=0; i < _cellReactions.size(); i++){
		Reaction* r = _cellReactions[i];
		Dc += r->DbyType(highStateByType);
		Dclow += r->DbyType(lowBoundHighState);
	}

	double deltaD = Dc - Dclow;
	
	// Consider change in local diffusion reactions also
	for (int i=0; i < highStateByType.size(); i++){
		deltaD += (highStateByType[i] - lowBoundHighState[i] ) * _diffusionRate * _maxNumNeighbors;
	}

	return r->D(state) - deltaD; 
}

// Grab some reaction parent under root (?)
// Probabaly to get an example of a 'cell'
Reactions *PropensityBounds::_sufficientReactionsParent(Reactions *root){
		ReactionSet *r = (ReactionSet*)root ; 
		for (; !(r->getDependents()[0]->isReaction()); 
				r = (r->getDependents())[0]){
			assert(!r->isReaction());
		}
		
		return r->asReactions();
}

//vector<Reaction*> PropensityBounds::_getSufficientStatisticDependents(Reactions *root){
//	vector<Reaction*> rxnDeps;
//	vector<ReactionSet*> & rxnsDeps = root->getDependents();
//	for (int i=0; i < rxnsDeps.size(); i++){
//		rxnDeps.push_back(rxnsDeps[i]->asReaction());
//	}
//	
//	return rxnDeps;
//}

// all children of cellParent are of type Reaction*
// assume _maxDeltaN is initialized
// Note: as in ER-leap cellParent may be root
void PropensityBounds::_setupMaxDeltaN(Reactions *cellParent){

	// make a vector of species
	BOOST_FOREACH(Reaction *r, cellParent->descendantReactionList()){
		for (int i=0; i < r->numProducts(); i++){
			
			// R is not related to cellParent
			if (!r->inpParent()->grandchildOf(cellParent) && !r->outParent()->grandchildOf(cellParent)){ 
				continue;
			}
			
			int type = _speciesType(r, false, i); 
			int delta = deltaN(r, type);
			_maxDeltaN[type] = max(_maxDeltaN[type], delta);
		}
	}
}

void PropensityBounds::_setupLowDeltaN(Reactions *cellParent){
	// make a vector of species
	BOOST_FOREACH(Reaction *r, cellParent->descendantReactionList()){

		for (int i=0; i < r->numReactants(); i++){
			
			if (!r->inpParent()->grandchildOf(cellParent) && !r->outParent()->grandchildOf(cellParent)){ 
				continue;
			}
			
			int type = _speciesType(r, true, i); 
			int delta = deltaN(r, type);
			if (delta > 0) continue; // we see an increase in X, so ignore reaction.
			
			_lowDeltaN[type] = max((int)_lowDeltaN[type], abs(delta));
		}
	}
}

// assume reactant/product in same cell
int PropensityBounds::deltaN(Reaction *r, int typeId){
	int delta = 0;
	for (int i=0; i < r->numReactants(); i++){
		if (_speciesType(r, true, i) == typeId)
			delta -= r->getReactantStoic(i);
	}
	for (int i=0; i < r->numProducts(); i++){
		if (_speciesType(r, false, i)==typeId)
			delta += r->getProductStoic(i);
	}
	return delta;
}

int PropensityBounds::_speciesType(Reaction *r, bool isReactant, int listIdx){
	return isReactant ? r->getReactantType(listIdx) : r->getProductType(listIdx);
}

///////////////////////////////  PropensityBoundsNoCells  ///////////////////////////////

double PropensityBoundsNoCells::upperBoundDrxn(Reaction* r, State *state, int leap){
	vector<int> & stateVec =  state->X(); // base case returns species values in cell
	vector<int> & highState = _highBoundState(stateVec, leap);// TODO confirm this logic, also cache these results!
	return r->DbyUniqueId(highState); 
}

double PropensityBoundsNoCells::lowerBoundDrxn(Reaction* r, State *state, int leap){
	vector<int> & stateVec =  state->X(); // base case returns species values in cell
	vector<int> & lowState = _lowBoundState(stateVec, leap);// TODO confirm this logic, also cache these results!
	return r->DbyUniqueId(lowState);
}
 

// XXX Note: this implementation isn't very near optimal
double PropensityBoundsNoCells::upperBoundDrxns(Reactions* r, State *state, int leap){
	// after L steps find the max we can move
	// just sum up propensities for all internal reactions with bordering state
	vector<Reaction*> & rxns = r->descendantReactionList();
	vector<int> & stateVec =  state->X(); 
	vector<int> & highState = _highBoundState(stateVec, leap);
	
	double Dhigh = 0.0;
	for (int i=0; i < rxns.size(); i++){
		Dhigh += rxns[i]->DbyUniqueId(highState);
	}
	return Dhigh;
}

// XXX Note: this implementation isn't very near optimal
double PropensityBoundsNoCells::lowerBoundDrxns(Reactions* r, State *state, int leap){
	// after L steps find the max we can move
	// just sum up propensities for all internal reactions
	vector<Reaction*> & rxns = r->descendantReactionList();
	vector<int> & stateVec =  state->X(); 
	vector<int> & lowState = _lowBoundState(stateVec, leap);
	
	double Dlow = 0.0;
	for (int i=0; i < rxns.size(); i++){
		Dlow += rxns[i]->DbyUniqueId(lowState);
	}
	return Dlow;
}

int PropensityBoundsNoCells::_speciesType(Reaction *r, bool isReactant, int listIdx){
	return isReactant ? r->getReactantId(listIdx) : r->getProductId(listIdx);
}

// Bounds state wrt cell s.t. (cell.events-1) are assumed and all other events are as is.
// Note: Values for Species not considered by cell will be left untouched and undefined.
void PropensityBoundsNoCells::upperBoundSpeciesParentEventsGivenWithCellPov(Reactions *cell, State *s, vector<int> & returnBounds) {
	BOOST_FOREACH(int sid, cell->propensityEffectedByTheseSpecies()) {
		returnBounds[sid] = s->X()[sid];
		BOOST_FOREACH(ConnectingRxns *cr, *(ConnectionGrandmaster::getConnectionGrandmaster(cell->getRoot(), s)->increasingConnectingGroups(sid))) { 
			int events = s->numEvents(cr->_parentConnecting);

			if (cr->_parentConnecting == cell) 
				events--; // this should never be called on a cell with no events.

			assert(events >= 0);
			returnBounds[sid] += cr->_maxDeltaXPerEvent * events;
		}
	}
}

void PropensityBoundsNoCells::lowerBoundSpeciesParentEventsGivenWithCellPov(Reactions *cell, State *s, vector<int> & returnBounds) {
	assert(s->numEvents(cell) > 0);
	
	BOOST_FOREACH(int sid, cell->propensityEffectedByTheseSpecies()) {
		returnBounds[sid] = s->X()[sid];
		BOOST_FOREACH(ConnectingRxns *cr, *(ConnectionGrandmaster::getConnectionGrandmaster(cell->getRoot(), s)->decreasingConnectingGroups(sid))) { 
			int events = s->numEvents(cr->_parentConnecting);
			if (cr->_parentConnecting == cell) 
				events--; // this should never be called on a cell with no events.
			assert(events >= 0 && cr->_maxDeltaXPerEvent <= 0);
			returnBounds[sid] += cr->_maxDeltaXPerEvent * events;
		}
		returnBounds[sid] = max(returnBounds[sid], 0);
		assert(returnBounds[sid] <= s->X()[sid]);
	}
}

// bound the state and stores answer in returnBounds
void PropensityBoundsNoCells::upperBoundSpeciesParentEventsGiven(State *s, vector<int> & returnBounds){
	assert(s->X().size() == returnBounds.size());
	std::copy(s->X().begin(), s->X().end(), &returnBounds[0]);
	
	BOOST_FOREACH(ReactionSet *rs, _root->getDependents()) {
		Reactions * cell = rs->asReactions();
		int e = s->numEvents(rs);
		if (!e)
			continue;
		BOOST_FOREACH(DeltaSpecies *ds, *(ConnectionGrandmaster::getConnectionGrandmaster(_root, s)->possiblyIncreasedByCellEvent(cell))) {
			returnBounds[ds->speciesId] += e * ds->delta;
		}
	}
	
//	for (int i=0; i < returnBounds.size(); i++){
//		BOOST_FOREACH( ConnectingRxns *cr, ConnectionGrandmaster::getConnectionGrandmaster(_root, s)->increasingConnectingGroups {// s->species(i)->increasingConnectingGroups()){
//			// TODO: if parent event subtract 1 from events, not clear how to do this here. Maybe not worth it?
//			int events = s->numEvents(cr->_parentConnecting); // TODO this is calling a heap-lost Reactions *parent
//			returnBounds[i] += cr->_maxDeltaXPerEvent * events; 
//		}
//	}
}

vector<int> PropensityBoundsNoCells::upperBoundSpeciesParentEventsGiven(State *s){
	vector<int> & orig = s->X(); // XXX: this should be a copy of s->X (?)
	//vector<int> bound2;
	vector<int> bound(orig.begin(), orig.end()); 
	upperBoundSpeciesParentEventsGiven(s, bound);
	return bound;

//	bound.resize(orig.size());
//	for (int i=0; i < orig.size(); i++){ 
//		bound[i] = orig[i];
//		printf("%d->%d ", orig[i], bound[i]); fflush(0);
//	} // lame: couldn't get copy to work 

	
	// For each species find 'connecting' reaction->rxns and max delta of s[i] 
	// 		for each parent nbr ReactionSet.
//	for (int i=0; i < bound.size(); i++){
//		BOOST_FOREACH( ConnectingRxns *cr, s->species(i)->increasingConnectingGroups()){
//			// TODO: if parent event subtract 1 from events, not clear how to do this here. Maybe not worth it?
//			int events = s->numEvents(cr->_parentConnecting); // TODO this is calling a heap-lost Reactions *parent
//			bound[i] += cr->_maxDeltaXPerEvent * events; 
//		}
//	}
//	vector<int> ret = bound; // XXX: remove this
//	return ret;
}

void PropensityBoundsNoCells::lowerBoundSpeciesParentEventsGiven(State *s, vector<int> & returnBounds){
	assert(s->X().size() == returnBounds.size());
	std::copy(s->X().begin(), s->X().end(), &returnBounds[0]);
	
	// For each species find 'connecting' reaction->rxns and max delta of s[i] 
	// 		for each parent nbr ReactionSet.

	BOOST_FOREACH(ReactionSet *rs, _root->getDependents()) {
		int e = s->numEvents(rs);
		if (!e) 
			continue;
		BOOST_FOREACH(DeltaSpecies *ds, (*ConnectionGrandmaster::getConnectionGrandmaster(_root, s)->possiblyDecreasedByCellEvent(rs->asReactions()))) {
			assert(ds->delta < 0);
			returnBounds[ds->speciesId] += e * ds->delta;
			if (returnBounds[ds->speciesId] < 0)
				returnBounds[ds->speciesId] = 0;
		}
	}
	
	// todo: for efficiency, instead loop over cellWithEvent->speciesEffects
//	for (int i=0; i < returnBounds.size(); i++){
//		vector<ConnectingRxns*> & crxns = s->species(i)->decreasingConnectingGroups();
//		BOOST_FOREACH( ConnectingRxns *cr, crxns){
//			returnBounds[i] -= cr->_maxDeltaXPerEvent * s->numEvents(cr->_parentConnecting);
//			returnBounds[i] = max(returnBounds[i], 0);
//			assert(s->X()[i] >= returnBounds[i]);
//		}
//	}
}

vector<int> PropensityBoundsNoCells::lowerBoundSpeciesParentEventsGiven(State *s){
	vector<int> & orig = s->X(); // XXX: this should be a copy of s->X (?)
	vector<int> state = orig;	
	lowerBoundSpeciesParentEventsGiven(s, state);
	
	// For each species find 'connecting' reaction->rxns and max delta of s[i] 
	// 		for each parent nbr ReactionSet.
//	for (int i=0; i < state.size(); i++){
//		vector<ConnectingRxns*> & crxns = s->species(i)->decreasingConnectingGroups();
//		BOOST_FOREACH( ConnectingRxns *cr, crxns){
//			state[i] -= cr->_maxDeltaXPerEvent * s->numEvents(cr->_parentConnecting);
//			state[i] = max(state[i], 0);
//			assert(orig[i] >= state[i]);
//		}
//	}
	return state;
}

//double PropensityBoundsNoCells::upperBoundDrxnParentEventsGiven(vector<int> & highState, Reaction *r){
//	return r->DbyIniqueId(highState);
//}
//
//double PropensityBoundsNoCells::upperBoundDrxnsParentEventsGiven(vector<int> & highState, Reactions *rxn){
//	double highD = 0.0;
//
//	BOOST_FOREACH(ReactionSet *r, rxn->getDependents()){
//		highD += r->isReaction ? upperBoundDrxnWithParentEventsGiven(highState, r->asReaction()) :
//			upperBoundDrxnsWithParentEventsGiven(highState, r->asReactions());
//	}
//	return highD;
//}
