/*
 * PropensityBounds.h
 *
 *  Created on: Sep 10, 2009
 *      Author: dweo
 */

#ifndef PROPENSITYBOUNDS_H_
#define PROPENSITYBOUNDS_H_

#include <queue>
#include <vector.h>
#include <boost/foreach.hpp>
#include <assert.h>

#include "../utils/common.h"

//class State;
//#include "../state/state.h"
//#include "reaction.h"
//class Reaction;
class ReactionSet;
class Reactions;
class State;

namespace hierleap{
class Reaction;
}
using namespace hierleap;

// Used to calculated bounds on D for reaction sets
// 		- KEY ASSUMPTION: all blocks contain the same basic reaction set
class PropensityBounds {
public:


	
	// Root: root of all reactions, assumes all cells are identical 
	// Typically diffusion is true if we only consider one cell (Hier-leap)
	// 	and false if we consider multiple cells at a time (ie: ER-leap)
	// TODO clarify diffusionFromToOutside
	PropensityBounds(Reactions *root, int numSpeciesTypes, bool diffusionFromToOutside);

	void _setupDeltaN(Reactions *rootOfAllReaction, int numSpeciesType, bool diffusion);
	
	virtual double upperBoundDrxn(Reaction* r, State *state, int leap);

	// Assume we are in a 'closed world' from the (possible) parent of r down
	virtual double upperBoundDrxns(Reactions* r, State *state, int leap);

	// Assume we are in a 'closed world' from the (possible) parent of r down
	virtual double lowerBoundDrxns(Reactions* r, State *state, int leap);
	
	virtual ~PropensityBounds(){
		BOOST_FOREACH(Reaction *r, _cellReactions){
			delete r;
		}
	}
	
	void highBoundState(vector<int> & state, int leap, vector<int> &retBoundedState){
		assert(state.size() == _maxDeltaN.size());
		assert(retBoundedState.size() == state.size());
		
		for (unsigned int i=0; i < state.size(); i++){
			retBoundedState[i] = state[i] + _maxDeltaN[i] * (leap-1);
			// TODO add leap-1 optimization 
			// removed for l-1 'bug' in Dbound with cells _retState[i] = state[i] + _maxDeltaN[i] * leap;
		}
	}
	
	void lowBoundState(vector<int> & state, int leap, vector<int> & retBoundedState){
		assert(state.size() == _lowDeltaN.size());
		assert(state.size() == retBoundedState.size());

		for (unsigned int i=0; i < state.size(); i++){
			// TODO add leap-1 optimization _retState[i] = max((int)(state[i] - _lowDeltaN[i] * (leap-1)), 0);
			retBoundedState[i] = max((int)(state[i] - _lowDeltaN[i] * (leap-1)), 0);
		}
	}
	
	vector<int> & _highBoundState(vector<int> & state, int leap){
		assert(state.size() == _maxDeltaN.size());
		if (_retState.size() != state.size())
			_retState.resize(state.size());
		
		highBoundState(state, leap, _retState);
		
		return _retState;
	}
	
	vector<int> & _lowBoundState(vector<int> & state, int leap){
		assert(state.size() == _lowDeltaN.size());
		if (state.size()!=_retState.size()) 
			_retState.resize(state.size());
		
		lowBoundState(state, leap, _retState);
		
		return _retState;
	}
	
	// Returns vector where each entry is the maximim each species may
	// change after one reaction event occurs.
	vector<int> & highSpeciesDelta() {
		return _maxDeltaN;
	}
	
	// Return lowest stoichiometry for each species after one even.
	// Note: the sign is changed to always be positive.
	vector<unsigned int> & lowSpeciesDelta() {
		return _lowDeltaN;
	}
	
protected:
	// Setup _cellReactions and _retState
	//PropensityBounds(){ assert(false);} commented because inheriting class calls this
	//PropensityBounds(){ }
	PropensityBounds(Reactions *root) : _root(checkNotNull(root)) { }

	// TODO override this function
	virtual int _speciesType(Reaction *r, bool isReactant, int listIdx);
	
	// all children of cellParent are of type Reaction*
	// assume _maxDeltaN is initialized
	void _setupMaxDeltaN(Reactions *cellParent);
	
	// find the most that a species can be reduced by in one event
	void _setupLowDeltaN(Reactions *cellParent);
	
	// assume reactant/product in same cell
	int deltaN(Reaction *r, int typeId);
	
	
	// Find the Reactions* which containes all the reactions relevant to 
	// calculate sufficient statistics.
	// In this case we move down the hierarchy until we find a base-cell
	// which is assumed to be copied throughout the system.
	Reactions *_sufficientReactionsParent(Reactions *root);
	
	// Get the depedents which (minimally) enbale us to calculate quantities of interest.
	//vector<Reaction*> _getSufficientStatisticDependents(Reactions *root);

	// Store list of Reaction of a cell in here
	vector<Reaction*> _cellReactions;
	vector<int> _retState; // (temp) so we don't have to allocate many, many times
	vector<int> _maxDeltaN; // Maximum each species type may change after one event
	vector<unsigned int> _lowDeltaN; // Maximum a species may be decreased after one event
	
	int _maxNumNeighbors;
	double _diffusionRate;
	Reactions * _root;

};

// Very similiar to PropensityBounds but does not assume cells are in a 
//	grid and thus all references are through species' UniqueIds (not speciesId)
class PropensityBoundsNoCells : public PropensityBounds{
public:
	// call parent... NOT!
	//PropensityBoundsNoCells() : PropensityBounds(){assert(false);}
	
	// Note: DO NOT call parent constructor, interference will occur
	PropensityBoundsNoCells(Reactions *root, int numSpeciesTypes, bool diffusion) : PropensityBounds(root) {
		_setupDeltaN(root, numSpeciesTypes, diffusion);
	}
	
	// Note sure if this is needed
	~PropensityBoundsNoCells(){
		BOOST_FOREACH(Reaction *r, _cellReactions){
			delete r;
		}
	}
	
	virtual double upperBoundDrxns(Reactions* r, State *state, int leap);
	
	virtual double lowerBoundDrxns(Reactions* r, State *state, int leap);
	
	virtual double upperBoundDrxn(Reaction* r, State *state, int leap);
	
	virtual double lowerBoundDrxn(Reaction* r, State *state, int leap);
	
	void upperBoundSpeciesParentEventsGivenWithCellPov(Reactions *cell, State *s, vector<int> & returnBounds); 
	
	void lowerBoundSpeciesParentEventsGivenWithCellPov(Reactions *cell, State *s, vector<int> & returnBounds);
	
	// These would idealy be called with something like r->D(retSpecies)
	void upperBoundSpeciesParentEventsGiven(State *s, vector<int> & returnBounds);
	vector<int> upperBoundSpeciesParentEventsGiven(State *s);
	
	void lowerBoundSpeciesParentEventsGiven(State *s, vector<int> & returnBounds);
	vector<int> lowerBoundSpeciesParentEventsGiven(State *s);	
	
protected:	
	virtual int _speciesType(Reaction *r, bool isReactant, int listIdx);
	
	// In this case cells are not assumed, so we must consider all reactions
	Reactions *_sufficientReactionsParent(Reactions *root){
			return root;
	}
	
};

#endif /* PROPENSITYBOUNDS_H_ */
