/*
 *  reaction.cpp
 *  hierarchical_erleap
 *
 *  Created by David  Orendorff on 9/8/08.
 *  Copyright 2008 __MyCompanyName__. All rights reserved.
 *
 */

#include <boost/foreach.hpp>

#include "reaction.h"
#include "reactions.h"
#include "reaction_member.h"
#include "../state/species.h"
#include "../state/state.h"
#include "../utils/common.h"

#define UNDEFINED_DELTA 0x33

using namespace hierleap;

Reaction::Reaction(Reactions *parent, double rate) : 
	_rate(rate), ReactionSet(parent, true), _changesPropensityFor(NULL) {	// may need to sub ReactionSet for super (?)
	_reactantList = vector<ReactionMember*>();
} 

Reaction::~Reaction(){
	delete[] _myId;
}

double Reaction::D(State * state){
	double F = 1.0;
	for (unsigned int i=0; i < _reactantList.size(); i++){
		F *= MAX(combinations(state->x(_reactantList[i]), _reactantList[i]->getStoichiometry()), 0);
	}
	return _rate * F;
}

// calculates D by looking up the species type for each reactant
double Reaction::DbyType(vector<int> & state){
	double F = 1.0;
	BOOST_FOREACH( ReactionMember *rm, _reactantList){
		int idx = rm->getTypeId();
		F *= combinations(state[idx], rm->getStoichiometry());
	}
	return F * rate();
}

// calculates D by looking up the species type for each reactant
double Reaction::DbyUniqueId(vector<int> & state){
	double F = 1.0;
	BOOST_FOREACH( ReactionMember *rm, _reactantList){
		int idx = rm->getUniqueId();
		F *= combinations(state[idx], rm->getStoichiometry());
	}
	return F * rate();
}

// TODO probably not the best name :)
// Calculate D_bar except if parent fires find Dmax for L-1 reactions. given a state s which represents the high
// State of system after L events
double Reaction::highDByUniqueIdNoParentRxn(vector<int> &s, int nParentRxns){
	if (nParentRxns==0) return DbyUniqueId(s); // TODO Is D illedefined in this case?

	double F = 1.0;
	BOOST_FOREACH(ReactionMember *rm, _reactantList){
		int idx = rm->getUniqueId();
		int x = s[idx] - getParent()->maxStoicDelta(idx);  // TODO this isn't correct(?) Could be lower stoic rxn, or from a connecting group
		F *= MAX(combinations(x, rm->getStoichiometry()), 0);
	}
	return F * _rate;
}

int Reaction::delta(int speciesUniquId){
	/* what we'd like to code, most efficient soln
	switch(speciesUniquId) {
	 case s1 : return d1;
	 case s2: return d2;
	 default: return 0;
	 }
	 */
	if (speciesUniquId < 0) return 0;
	vector<int> & changingSpecies = speciesWhichChange();
	for (int i = 0; i < changingSpecies.size(); i++) {
		if (changingSpecies[i] == speciesUniquId)
			return _deltas[i];
	}
	return 0;
		 
//	
//	
//	if (_deltas.size() <= speciesUniquId)
//		_deltas.resize(getRoot()->descendantReactionList().size(), UNDEFINED_DELTA);
//	
//	if (_deltas[speciesUniquId] == UNDEFINED_DELTA){
//		int dn = 0;
//		BOOST_FOREACH(ReactionMember *rm, _reactantList)
//			if (rm->getUniqueId()==speciesUniquId)
//				dn -= rm->getStoichiometry();
//		
//		BOOST_FOREACH(ReactionMember *rm, _productList)
//			if (rm->getUniqueId()==speciesUniquId)
//				dn += rm->getStoichiometry();
//		_deltas[speciesUniquId] = dn;
//		assert(_deltas[speciesUniquId] != UNDEFINED_DELTA);
//	}
//	return _deltas[speciesUniquId];
}



vector<Reaction*> & Reaction::maybeEffectedRxn() {
	assert(_maybeEffectingRxn.size() > 0);
	return _maybeEffectingRxn;
}

int Reaction::_uncachedDelta(int speciesId) {
	if (speciesId < 0) return 0;
	
	int dn = 0;
	BOOST_FOREACH(ReactionMember *rm, _reactantList)
		if (rm->getUniqueId()==speciesId)
			dn -= rm->getStoichiometry();
			
	BOOST_FOREACH(ReactionMember *rm, _productList)
		if (rm->getUniqueId()==speciesId)
			dn += rm->getStoichiometry();
	return dn;
}

vector<int> & Reaction::speciesWhichChange() {
	if (_speciesWhichChange.size() == 0) {
		BOOST_FOREACH(ReactionMember *s, _reactantList) {
			int sid = s->getUniqueId();
			int d = _uncachedDelta(sid);
			if (d && !contains(_speciesWhichChange, sid)) {
				_speciesWhichChange.push_back(sid);
				_deltas.push_back(d);
			}
		}
		BOOST_FOREACH(ReactionMember *s, _productList) {
			int sid = s->getUniqueId();
			int d = _uncachedDelta(sid);
			if (d && !contains(_speciesWhichChange, sid)) {
				_speciesWhichChange.push_back(sid);
				_deltas.push_back(d);
			}
		}
		assert(_speciesWhichChange.size() > 0);
	}
	return _speciesWhichChange;
}

bool Reaction::isViable(State *s) {
	return isViable(s->X());
}

bool Reaction::isViable(vector<int> & s) {
	switch (numReactants()) {
		case 0:
			return true;
		case 1: 
			return s[getReactantId(0)] - getReactantStoic(0) >= 0;
		case 2:
			assert(getReactantId(0) != getReactantId(1));
			return s[getReactantId(0)] - getReactantStoic(0) >= 0 && s[getReactantId(1)] - getReactantStoic(1) >= 0;
		default:
			assert(false); // only 2 reactants allowed
	}
	return false;
}

// Update state supposing this reaction has fired once
bool Reaction::fire(State *state){
	fire(state->X());
//	BOOST_FOREACH(ReactionMember *r, _productList){
//		state->changeX(r->getUniqueId(), r->getStoichiometry());
//	}
//	
//	BOOST_FOREACH(ReactionMember *r, _reactantList){
//		state->changeX(r->getUniqueId(), -r->getStoichiometry());
//	}
}

// Update state as if this reaction had fired once.
// returns true iff negative species do not result.
bool Reaction::fire(vector<int> & state) {
	bool isPositive = true;
	BOOST_FOREACH(ReactionMember *r, _productList){
		state[r->getUniqueId()] += r->getStoichiometry();
	}
	
	BOOST_FOREACH(ReactionMember *r, _reactantList){
		state[r->getUniqueId()] -= r->getStoichiometry();
		isPositive &= state[r->getUniqueId()] >= 0;
	}
	return isPositive;
}

//  reverse reactant/product and update state accordingly
void Reaction::unFire(State *state){
	unFire(state->X());
//	BOOST_FOREACH(ReactionMember *r, _reactantList){
//		state->changeX(r->getUniqueId(), r->getStoichiometry());
//	}
//	BOOST_FOREACH(ReactionMember *r, _productList){
//		state->changeX(r->getUniqueId(), -r->getStoichiometry());
//	}
}

void Reaction::unFire(vector<int> & state) {
	BOOST_FOREACH(ReactionMember *r, _productList){
		state[r->getUniqueId()] -= r->getStoichiometry();
	}
	
	BOOST_FOREACH(ReactionMember *r, _reactantList){
		state[r->getUniqueId()] += r->getStoichiometry();
	}
}

void Reaction::fireReactantsOnly(State *s){
	BOOST_FOREACH(ReactionMember *r, _reactantList){
		s->changeX(r->getUniqueId(), r->getStoichiometry());
	}
}

// Are the input and output of this reaction fully contained in root?
bool Reaction::selfFullyContainedIn(Reactions *root){
	// XXX: cache these results for speed in a ~hash or something
	
	// go through reactants and products home cells to see if they belong to root
	
	BOOST_FOREACH(ReactionMember *rm, _reactantList){
		if (!rm->getParent()->grandchildOf(root)){ 
			return false;
		}
	}
	
	BOOST_FOREACH(ReactionMember *rm, _productList){
		if (!rm->getParent()->grandchildOf(root)) return false;
	}
	
	return true;
}


void Reaction::addReactant(Species *s, unsigned int stoichiometry) {
	ReactionMember *rm = new ReactionMember(s, stoichiometry);
	_reactantList.push_back(rm);


}

void Reaction::addProduct(Species *s, unsigned int stoichiometry) {
	ReactionMember *rm = new ReactionMember(s, stoichiometry);
	_productList.push_back(rm);
}


// the reactionset parent of the gridcell with the LHS diffusion reactant
Reactions *Reaction::inpParent(){
	//assert(strcmp(myType(),"diffusion")==0); // TODO not sure how this is relavant
	return getParent();
}

// the reactionset parent of the gridcell with the RHS diffusion reactant
//Reactions *Reaction::outParent(){
//	assert(strcmp(myType(), "diffusion")==0);
//	return _productList[0]->parentReactionset();
//}

void Reaction::_specialPrint(State *s, char format){
		switch (format){
			//case 'X' :printf(" X: %d", (_reactantList.size > 0 ? _reactantList[0]->getTypeId()) : NULL)); break;
			//case 'x' :printf(" |x|: %d", s->x(_reactantList[0]));
			case 's' :
						for (unsigned int i=0; i < _reactantList.size() ; i++) {
							printf("(%d %s:%x)", _reactantList[i]->getStoichiometry(),
									_reactantList[i]->name(), _reactantList[i]->getUniqueId());

						}  printf("-->");
						for (unsigned int i=0; i < _productList.size(); i++) {
							printf("(%d %s:%x)", _productList[i]->getStoichiometry(),
									_productList[i]->name(), _productList[i]->getUniqueId());
						}
				printf(", [rho=%f]", rate());
						
		}
}

int Reaction::getReactantStoic(int rxnidx){
	assert(rxnidx < _reactantList.size());
	return _reactantList[rxnidx]->getStoichiometry();
}

int Reaction::getProductStoic(int rxnidx){
	assert(rxnidx < _productList.size());
	return _productList[rxnidx]->getStoichiometry();
}

int Reaction::getReactantId(int rxnIdx){
	assert(rxnIdx < _reactantList.size());
	return _reactantList[rxnIdx]->getUniqueId(); 
}

int Reaction::getProductId(int ridx){
	assert(ridx < _productList.size());
	return _productList[ridx]->getUniqueId(); 
}

int Reaction::getReactantType(int rxnidx){
	assert(rxnidx < _reactantList.size());
	return _reactantList[rxnidx]->getTypeId();
}

int Reaction::getProductType(int rxnidx){
	assert(rxnidx < _productList.size());
	return _productList[rxnidx]->getTypeId();
}

// the reactionset parent of the gridcell with the RHS diffusion reactant
Reactions *Reaction::outParent(){
	if (_productList.size()==0)
		return inpParent(); // TODO is this a good default behavior?

	ReactionMember *rm = _productList[0];
	return rm->getParent();
}

char* Reaction::myType(){
	// TODO will blow up for different types of reactions
	if (_productList.size()==0){ return "death"; }
	if (_reactantList.size()==0 && _productList.size()>0) {return "immegration";}
	
	if (_productList.size()==1 && _reactantList.size()==1) { // && _reactantList[0]->getParent() != _productList[0]->getParent()){
		return "conversion";
	}
	
	switch(_productList[0]->getStoichiometry()){
		case 1: return "type1";
		case 2: return "type2";
		default: return "na";
	}
}

// Does this reaction involve r?
bool Reaction::involves(ReactionMember *r){
	BOOST_FOREACH(ReactionMember *rMe, _reactantList){\
		if (r->getUniqueId()==rMe->getUniqueId())
			return true;
	}
	BOOST_FOREACH(ReactionMember *rMe, _productList){\
		if (r->getUniqueId()==rMe->getUniqueId())
			return true;
	}
	return false;
}

void Reaction::updateSpeciesInfo(State *s) {
	for (int i = 0; i < _reactantList.size(); i++) {
		ReactionMember *origRm = _reactantList[i];
		_reactantList[i] = new ReactionMember(s->species(origRm->getUniqueId()), getReactantStoic(i)); 
		delete origRm;
	}
	for (int i = 0; i < _productList.size(); i++) {
		ReactionMember *origRm = _productList[i];
		_productList[i] = new ReactionMember(s->species(origRm->getUniqueId()), getProductStoic(i)); 
		delete origRm;
	}
}

vector<Reaction*> * Reaction::changesPropensityFor() {
	bool debugFlag = false;
	if (!_changesPropensityFor) {
		debugFlag = true;
		_changesPropensityFor = new vector<Reaction*>();
		vector<ReactionMember*> involved = _reactantList;
		involved.insert(involved.end(), _productList.begin(), _productList.end());
		BOOST_FOREACH(ReactionMember *rm, involved) {
			if (!delta(rm->getUniqueId()))
				continue;
			vector<Reaction*> cpf = rm->changesPropensityFor();
			BOOST_FOREACH(Reaction *r, cpf) {
				if (!contains(*_changesPropensityFor, r))
					_changesPropensityFor->push_back(r);
			}
		}
	}
	//return vector<Reaction*>(_changesPropensityFor->begin(), _changesPropensityFor->end());
	return _changesPropensityFor;
}

void Reaction::setupInteractingRxn(State *s) {
	BOOST_FOREACH(int i, speciesWhichChange()) {
		BOOST_FOREACH(Reaction *r, s->touchingRxn(i)) {
			if (!contains(_maybeEffectingRxn, r)) {
				_maybeEffectingRxn.push_back(r);
			}
		}
	}
	assert(_maybeEffectingRxn.size() > 0);
}
