/*
 * ReactionPermutation.h
 *
 *  Created on: Sep 7, 2009
 *      Author: dweo
 */

#include <assert.h>
#include <algorithm>

#ifndef REACTIONPERMUTATION_H_
#define REACTIONPERMUTATION_H_

#include <vector.h>
#include <boost/thread.hpp>

#include "reaction.h"
#include "reactions.h"
#include "../utils/common.h"

using namespace hierleap;
using namespace std;

// T should be a reaction type or reaction ID
// This is also what reaction-leaping functions will return.
// 		- so it needs to be efficient, and memory conscious
//		- we cannot have a .cpp file because this is a templated class
template <class T> class ReactionPermutation {
	// Internally, the first reaction is put into _reactions[0] and so on
	// reactions are popped from the back until _reactions[0] is reached
public:
	
	static void test() {
		ReactionPermutation<int> rp;
		vector<ReactionPermutation<int> *> toJoin;
		int nTests = 100000;
		int nEq61 = 0;
		int nEq13 = 0;
		// should join
		// 1 2 3
		// 4 5 6
		// P(1==a[0]) = 0.5 || P(6==a[-1]) = 0.5.
		
		for (int test = 0; test < nTests; test++) {
			ReactionPermutation<int> * r1 = new ReactionPermutation<int>();
			ReactionPermutation<int> * r2 = new ReactionPermutation<int>();
			r1->addReaction(1, 1); r1->addReaction(2, 1); r1->addReaction(3, 1); 
			r2->addReaction(4, 1); r2->addReaction(5, 1); r2->addReaction(6, 1);
			r1->lockOrder(); r2->lockOrder();
			assert(r1->next()==3); r1->previous();
			assert(r2->next()==6); r2->previous();
			toJoin.push_back(r1); toJoin.push_back(r2);
			
			rp.joinPermutations(toJoin);

			if (rp.next()==3) {
				nEq61++;
			}
			rp.next(); rp.next(); rp.next(); rp.next(); 
			if (rp.next()==1) {
				nEq13++;
			}
			delete r1; delete r2;
			rp.clear();
			toJoin.clear();
		}
		printf("Should be binomial vals with p=0.5\n\t%d/%d a[0]==1, %d/%d a[5]==3", nEq61, nTests, nEq13, nTests);
	}
	
	ReactionPermutation(vector<unsigned int> & rp):_locked(false){
		T x;
		isValidType(x); // this should cause a compile-time error if T is of the wrong type
		
		// TODO somehow make sure T is int type (?)
		for (int i=0; i < rp.size(); i++){
			if (rp[i] > 0){
				addReaction(i, rp[i]);
			}
		}
		_nextPopIdx = 0;
		_checkpoint = -1;
		_inInsertFiredReactionMode = false;
	}
	
	ReactionPermutation(ReactionPermutation<T> *rp):_locked(false){
		_reactions = rp->_reactions; // deep copy?
		_locked = rp->_locked;
		_nextPopIdx = rp->_nextPopIdx;
		_checkpoint = -1;
		_inInsertFiredReactionMode = false;
	}
	
	/**
	 * Joins the reaction permutations into a single permutation, where the ordering
	 * in each rp[i] is respected but each entry is selected at random from each of the rps
	 * This also locks the reaction permutation.
	 */ 
	void joinPermutations(vector<ReactionPermutation<T>*> & rps){
		assert(!_inInsertFiredReactionMode);
		assert(rps.size() > 0);
		
		// an optimization step
		move_non_null_to_front(rps);
		int nLeft = 0;
		for (int i = 0; i < rps.size(); i++) {
			if (!rps[i]) continue;
			nLeft += rps[i]->numEvents();
			rps[i]->lockOrder();
		}
		_reactions.resize(nLeft);
		while (nLeft) {
			unsigned int rndElem = urand_int(1, nLeft+1); // should be randInt 1 to nLeft inclusive
			assert(rndElem <= nLeft && rndElem > 0);
			
			int cumul = 0;
			int i = 0;
			for (; i < rps.size(); i++) {
				cumul += rps[i] == NULL ? 0 : rps[i]->numEvents();
				if (cumul >= rndElem) break;
			}

			_reactions[nLeft-1] = rps[i]->next();
			if (i > 0 && !rps[i]->hasNext()) // a little optimization
				swap(rps[i], rps[i-1]);
			nLeft--;
		}
		lockOrder();
	}
	
	// create a new ReactionPermutation<Reaction*> by looking up the int entries in rp and converting them to rxns
	ReactionPermutation(ReactionPermutation<unsigned int> *rp, vector<Reaction*> & myReactionList):_locked(false){
		vector<unsigned int> & rxnList = rp->_myReactionList();
		for (unsigned int i=0; i < rxnList.size(); i++){
			int ridx = rxnList[i];
			addReaction(myReactionList[ridx], 1);
		}
		_nextPopIdx =0;
		_inInsertFiredReactionMode = false;
	}
	
	ReactionPermutation():_locked(false){
		T x;
		isValidType(x); // this should cause a compile-time error if T is of the wrong type
		_nextPopIdx =0;
		_inInsertFiredReactionMode = false;
	}

	~ReactionPermutation() {
		// Nothing to do. No memory allocated on heap.
	}
	
	bool isLocked() {
		return _locked;
	}
	
	bool hasNext(){ 
		assert(!_inInsertFiredReactionMode);
		assert(_locked);
		return _nextPopIdx < _reactions.size() && _nextPopIdx >= 0;
	}

	T next(){
		assert(!_inInsertFiredReactionMode);
		assert(_locked && hasNext());
		return _reactions[_nextPopIdx--];
	}
	
	bool hasPrevious() {
		assert(!_inInsertFiredReactionMode);
		assert(_locked);
		return _nextPopIdx + 1 < _reactions.size() && _nextPopIdx+1 >= 0;
	}
	
	T previous(){
		assert(!_inInsertFiredReactionMode);
		assert(_locked && hasPrevious());
		return _reactions[++_nextPopIdx];
	}
	
	T nextFromBack(){
		assert(!_inInsertFiredReactionMode);
		assert(false); // this function is depricated to next(), see code if you need to have this exact behavior.
		assert(_locked && hasNext());
		T x = _reactions.back();
		_reactions.pop_back();
		return x;
	}

	void addReaction(T reactionId, int stoic){
		assert(!_inInsertFiredReactionMode);
		if (stoic==0) return;
		assert(validRxn(reactionId));
		assert(!_locked);
		addBack(reactionId); // TODO: find more efficient way
		addReaction(reactionId, stoic-1);
	}

	// Assuming only this function is called to add reactions, 
	//		this is a thread safe function.
	void parallelAddReaction(T reactionId, int stoic){
		assert(!_inInsertFiredReactionMode);
		if (stoic==0) return;
		assert(validRxn(reactionId));
		assert(!_locked);
		
		_parallelAddReactionMutex.lock();
		addReaction(reactionId, stoic);
		_parallelAddReactionMutex.unlock();
	}
	
	// Shuffles reactions, order is locked after this procedure
	void permute(){
		assert(!_inInsertFiredReactionMode);
		assert(!_locked);
		random_shuffle(_reactions.begin(), _reactions.end());
		lockOrder();
	}
	
	// Permutes even if *this* is locked.
	void forcePermute() {
		assert(!_inInsertFiredReactionMode);
		random_shuffle(_reactions.begin(), _reactions.end());
		lockOrder();
	}

	// Remove k reactions from end of queue
	void removeFinalKReactions(int k){
		assert(!_inInsertFiredReactionMode);
		assert(k <= numEvents());
		int newFinalIdx = _reactions.size() - k -1;
		assert(newFinalIdx < 0 || _nextPopIdx <= newFinalIdx); // this assert was from when _nextPopIdx went 0+, may need to change later if asserting
		_reactions.resize(newFinalIdx + 1);
	}
	
	int numEvents(){
		assert(!_inInsertFiredReactionMode);
		return _locked ? _nextPopIdx + 1 : _reactions.size();
	}
	
	// The order of reactions can no longer be adjusted
	void lockOrder(){
		assert(!_inInsertFiredReactionMode);
		//assert(!_locked);
		_locked = true;
		_nextPopIdx = _reactions.size() - 1;
	}

	////////////////////////////////////////////////////////////////////
	//////////////////////////////// Section for beginInsertFiredRxnMode 
	void beginInsertFiredReactionMode() {
		assert(!_inInsertFiredReactionMode);
		assert(numEvents()==0); 
		assert(!(hasNext() || hasPrevious()));
		assert(_reactions.size() == 0);
		_checkpoint = -1;
		_inInsertFiredReactionMode = true;
	}
	
	void addFiredReaction(T reactionId) {
		assert(_inInsertFiredReactionMode);
		_reactions.push_back(reactionId);
	}
	
	void endInsertFiredReactionMode() {
		assert(_inInsertFiredReactionMode);
		_inInsertFiredReactionMode = false;
		std::reverse(_reactions.begin(), _reactions.end()); // dumb 'hack'
		lockOrder();
		_nextPopIdx = -1;
	}
	
	//////////////////////////////// End Section for beginInsertFiredRxnMode 
	////////////////////////////////////////////////////////////////////////
	
	// Puts exactly one (1) reaction at end of events, events MAY be order-locked for this (?)
	void addBack(T reactionId){
		_reactions.push_back(reactionId); 
	}
	
	// It probably isn't a good idea to use this once if popIdx is not at the begining.
	// As ALL rxns will be reversed.
	void reverse() {
		
	}
	
	void checkpoint() {
		_checkpoint = _nextPopIdx;
	}
	
	void resumeCheckpoint() {
		assert(_checkpoint >= 0);
		_nextPopIdx = _checkpoint;
	}

	// After clear() returns, this obj state will be as if 
	// it was just created, freshly.
	void clear() {
		_nextPopIdx = 0;
		_checkpoint = -1;
		_locked = false;
		if (_reactions.size())
			_reactions.clear();
	}
	
	// All reaction that were in the list when lockOrder was called.
	vector<T> & _myReactionList(){ return _reactions;}

	void print() {
		printf("\nHave popped %d of %d reactions.\n", _nextPopIdx, _reactions.size());
		BOOST_FOREACH(T r, _reactions) {
			_printReaction(r);
		}
	}
	
private:
	int _checkpoint;
	boost::mutex _parallelAddReactionMutex;
	int _nextPopIdx; // the most ready to pop item, back to front
	
	vector<T> _reactions;
	bool _locked;
	bool _inInsertFiredReactionMode;
	
	void _printReaction(Reaction *r) {
		printf("Id=%d(parent=%d), ", r->getUniqueId(), r->getParent()->getUniqueId());
	}
	
	bool validRxn(int x){ return x >= 0; }
	bool validRxn(unsigned int x){ return true; }
	bool validRxn(Reaction *r){ return r != NULL; }
	bool validRxn(Reactions *r){ return r!= NULL;}
	void isValidType(int & ) const {}
	void isValidType(unsigned int & ) const {}
	void isValidType(Reaction* &r) const {} // XXX this may be a problem with pointers
	void isValidType(Reactions* &r) const {}
	
};

#endif /* REACTIONPERMUTATION_H_ */
