/*
 *  reaction_set.cpp
 *  hierarchical_erleap
 *
 *  Created by David  Orendorff on 8/15/08.
 *  Copyright 2008 __MyCompanyName__. All rights reserved.
 *
 */

#include "reaction_set.h"
#include "reactions.h"
#include "reaction.h"
#include "../state/species.h"
#include "../state/state.h"

using namespace hierleap;



/////////////////////////////////////////
////////// SETUP FUNCTIONS //////////////

ReactionSet::ReactionSet(Reactions* parent, bool isReaction) : _isReaction(isReaction) {
	if (parent) 
		assert(!parent->isReaction());
	
	_parent = parent;
	_x=_y=_z = -1;
	_myUniqueId = -1;
	_idxInParent = -1;
	_smallestChildId = -1;
}

void ReactionSet::addDependent(ReactionSet *r){
	assert(!isReaction());
	// Ascertain r isn't in _dependents
	for (unsigned int i=0; i < _dependents.size(); i++)
		if (_dependents[i]==r) return;

	_dependents.push_back(r);
}

// automatically looks for duplicates
void ReactionSet::_addNbrMCA(ReactionSet *s){
	if (s==(ReactionSet*)(getParent()))
		return;

	for(unsigned int i = 0;i < _nbrsMCA.size();i++){
		if(_nbrsMCA[i] == s)
			return; // duplicate found

	}
	_nbrsMCA.push_back(s);
}

void ReactionSet::setCoordinate(int x, int y, int z)
{
	_x = x;
	_y = y;
	_z = z;
}

void ReactionSet::getCoordinate(int &x, int &y, int &z){
	if (isReaction() || _x >= 0){
		x = _x; y = _y; z = _z; return;
	}
	vector<ReactionSet*> & deps = getDependents();
	if (deps[0]->isReaction()){
		deps[0]->getCoordinate(_x, _y, _z);
		x = _x; y = _y; z = _z;
		return;
	}
	assert(false); // This has no relevant 3D coordinate
}

void ReactionSet::idSystem(char * myAddressId){
	strcpy(_myId, myAddressId);
	
	// a single Reaction
	if (isReaction()){ 
		sprintf(_myId, "%sR", _myId);
		// XXX: this is a little fragile
		Reaction* r = asReaction();
		if (strcmp(r->myType(),"diffusion")==0){
			// put in coords of diffusion rxn
			ReactionSet *p1 = r->inpParent();
			ReactionSet *p2 = r->outParent();
			sprintf(_myId, "%s(%d,%d,%d)to(%d,%d,%d)",_myId,
				p1->_x, p1->_y, p1->_z, p2->_x, p2->_y, p2->_z);
		}else{
			sprintf(_myId, "%s%s", _myId, r->myType());
		}
		
	} else{// put an R at the end of reactions
	
		char szTemp[100];
		for (unsigned int i=0; i < _dependents.size(); i++){
			 sprintf(szTemp, "%s_%d", _myId, i);
			 assert(i >= 0 && i < _dependents.size());
			 _dependents[i]->idSystem(szTemp);
		}
	}
	
	if (getParent()==NULL)
		makeIdsTopological();
	//registerMe();
}


void ReactionSet::_bubbleUpBubbleId(int bId){
	bubbleId = bId;
	if (!getParent()) return;
	getParent()->_bubbleUpBubbleId(bId);
}

void ReactionSet::buildNbrsMCA(){
	
	// bubble up myId
	_bubbleUpBubbleId(21);
	
	// have neighbors bubble until they reach myid
	for (unsigned int i = 0; i < _nbrs.size(); i++){
		ReactionSet * nbr = _nbrs[i];
		ReactionSet *p;
		for (p = nbr; p->getParent()->bubbleId != bubbleId ; p = p->getParent());
		
		// XXX: may be a bug here for computing MCA, maybe:(!isReaction() && p->getParent() != _parent)?
		if (!(isReaction() && p->getParent() == getParent())){ // make sure two rxns are not in same gridcell
			assert(p->bubbleId != bubbleId);
			_addNbrMCA(p);
		}
	}
	
	// bubble up zero to reset ids
	_bubbleUpBubbleId(0);
	
	// now have children set their MCAs
	if(!isReaction()){
		for (unsigned int i=0; i < _dependents.size(); i++){
			_dependents[i]->buildNbrsMCA();
		}
	}
}

void ReactionSet::setupInteractingRxn(State *s) {
	BOOST_FOREACH(ReactionSet *r, getDependents()) {
		r->setupInteractingRxn(s);
	}
}

int ReactionSet::numRxnSets(){
	int tot = 1;
	if (isReaction()) return tot;

	for (unsigned int i=0; i < _dependents.size(); i++)
		tot += _dependents[i]->numRxnSets();

	return tot;
}

// Returns height (including self) for the tree. 
// O(n)
int ReactionSet::height(){
	if (isReaction()) return 1;
	
	int maxH = 0;
	BOOST_FOREACH(ReactionSet *r, getDependents()){
		maxH = max(r->height(), maxH);
	}
	return maxH + 1;
}

//////// END SETUP FUNCTIONS ////////////
/////////////////////////////////////////


/////////////////////////////////////////
//////// DEBUGGING FUNCTIONS ////////////

// format specifies which properties to print. Each node gets exactly 1 line.
// key: e -> num events. c -> numchildren. d -> D. f -> Dfloor. h -> Dceil.
void ReactionSet::print(State *state, char* format){
	// first print myself out
	printf("%s:\t", _myId);
	bool printReactionsOnly = false;
	for (int i=0; format[i]; i++){
		switch (format[i]){
			case 'n' : printf(" neighbors: %d", (int)_nbrs.size()); break;
			case 'd' : printf(" D: %f", D(state)); break;
			case 'i' : printf(" ID: %d", _myUniqueId);break;
			case 'e' : printf(" Events: %d", state->numEvents(this)); break;
			case 'p' : printf(" Par: %X", getParent()); break;
			case 'm' : printf(" myAddr: %X", this); break;
			case 'C' : printReactionsOnly = true; break;
			case 'N' : printf(" nbrs: (");
				BOOST_FOREACH(ReactionSet *rs, getNeighbors()) {
					printf(" %X", rs);
				} printf(") ");
				break;
			
//			case 'p' : if (lowBoundD()/highBoundD() > 1)
//							printf(" Dfloor/Dceil: %f", lowBoundD()/highBoundD());
//						break;
//			case 'x' : if(_x>=0&&_y>=0&&_z>=0){
//							printf("Coord(%d,%d,%d)", _x,_y,_z);}
//						break;
			default : break; // _specialPrint(state, format[i]); break;
		}
		_specialPrint(state, format[i]);  // added
	}
	printf("\n");
	
	// now apply to children
	if (isReaction()) return;
	for (unsigned int i=0; i < _dependents.size(); i++){
		if (!_dependents[i]->isReaction() || !printReactionsOnly)
			_dependents[i]->print(state, format);
	}
}


//////// END DEBUGGING FUNCTIONS ////////////
/////////////////////////////////////////////

Reactions* ReactionSet::getRoot() {
	if (getParent() == NULL) {
		assert(!isReaction());
		return this->asReactions();
	}
	return getParent()->getRoot();
}

// Return true if *this* is a direct decendent of root
bool ReactionSet::grandchildOf(Reactions *root){
	if (this==root) return true;
	if (!getParent()) return false;
	return getParent()->grandchildOf(root);
}

ReactionSet* ReactionSet::highestParentUnder(Reactions *root){
	assert(getParent());
	if (getParent()==root) return this;
	return getParent()->highestParentUnder(root);
}

Reactions *ReactionSet::getParent(){
	if (_parent) 
		assert(!_parent->isReaction());
	return _parent;
}

void ReactionSet::setParent(Reactions *r){
	assert(!_parent);
	if (r) 
		assert(!r->isReaction());
	_parent = r;
}

void ReactionSet::changeParent(Reactions *newParent) {
	getParent()->removeDependent(this);
	voidParent();
	setParent(newParent);
	newParent->addDependent(this);
}

void ReactionSet::removeDependent(ReactionSet *rxnSet) {
	for (vector<ReactionSet*>::iterator r = _dependents.begin(); r != _dependents.end(); r++) {
		if (*r == rxnSet) {
			_dependents.erase(r);
			return;
		}
	}
	assert(false); // value not found.
}

void ReactionSet::_reconnectNbrs() {
	assert(getRoot()==this);
	vector<ReactionSet*> & op = getOrderedProgeny();
	BOOST_FOREACH(ReactionSet *rs, op) {
		vector<ReactionSet*> & nbrs = rs->_nbrs;
		for (int i = 0; i < nbrs.size(); i++){ 
			nbrs[i] = op[nbrs[i]->getUniqueId()];
		}
	}
}

vector<ReactionSet*> & ReactionSet::getParentNeighbors(){
	if (getParent()){
		return getParent()->getNeighbors();
	}
	
	// means we're the root, so shouldn't have any neighbors
	assert(_nbrs.empty());
	return _nbrs;
}

Reactions* ReactionSet::asReactions(){
	assert(!isReaction());
	return (Reactions*)this;
}

Reaction* ReactionSet::asReaction(){
	assert(isReaction());
	return (Reaction*)this;
}

int ReactionSet::indexInParent(){
	if (_idxInParent>0) return _idxInParent;
	assert(getParent());
	_idxInParent = getParent()->_whatIndexChild(this);
	return _idxInParent;
}

void ReactionSet::_buildDRL(vector<Reaction*> & listSoFar){
	if (isReaction()) {
		listSoFar.push_back((Reaction*)this);
		return;
	}
	BOOST_FOREACH(ReactionSet *r, _dependents){
		r->_buildDRL(listSoFar);
	}
}