/*
 *  reactions.cpp
 *  hierarchical_erleap
 *
 *  Created by David  Orendorff on 9/8/08.
 *  Copyright 2008 __MyCompanyName__. All rights reserved.
 *
 */

#include <set>
#include <assert.h>
#include <boost/foreach.hpp>

using namespace std;

#include "reactions.h"
#include "reaction.h"
#include "reaction_permutation.h"
#include "reaction_set.h"
#include "../state/state.h"
#include "../state/species.h"

using namespace hierleap;

void Reactions::connectedCellsWithEvents(vector<Reactions*> & explored, State *s) {
	if (!contains(explored, this) && s->numEvents(this)) {
		explored.push_back(this);
	} else {
		return;
	}
	BOOST_FOREACH(ReactionSet *rs, getNeighbors()) {
		vector<ReactionSet*> & orderedProgeny = getRoot()->getOrderedProgeny();
		assert(orderedProgeny[rs->getUniqueId()] == rs);
		if (!s->numEvents(rs))
			continue;
		rs->asReactions()->connectedCellsWithEvents(explored, s);
	}
}

void Reactions::addDependent(ReactionSet *r){
	//assert(_dependents);
	_dependents.push_back(r);
}

double Reactions::D(State *s){
	double myD = 0;
	for (unsigned int idx = 0; idx < _dependents.size(); idx++){
		myD += _dependents[idx]->D(s);
	}
	return myD;
}

double Reactions::DbyUniqueId(vector<int> & s){
	double d = 0.0;

	BOOST_FOREACH(ReactionSet *r, getDependents()){
		d += r->DbyUniqueId(s);
	}
	return d;
}

void Reactions::preCache() {
	BOOST_FOREACH(Reaction *r, descendantReactionList()) {
		r->speciesWhichChange();
		r->changesPropensityFor();
	}
}

// Does this reactionSet have any species in common with r?
bool Reactions::isNeighbor(Reactions *r){
	// Check all dependents, see if any O(nm) children are dependents of this AND r
	BOOST_FOREACH(ReactionSet *r2, r->getDependents()){
		BOOST_FOREACH(ReactionSet *r1, this->getDependents()){
//		for (int i=0; i < getDependents().size(); i++){
//			ReactionSet *r1 = getDependents()[i];
		
			// For now assumes r1, r2 of same type. Assert thrown if it is not.
			if (r1->isReaction()){
				if ((r1->asReaction())->isNeighbor(r2->asReaction()))
					return true;
			}else{
				if (r1->asReactions()->isNeighbor(r2->asReactions()))
					return true;
			}
		}
	}
	return false;
}

ReactionPermutation<Reaction*> * Reactions::reactionPermutationSandbox() {
	if (!_rpCache) {
		_rpCache = new ReactionPermutation<Reaction*>();
	}
	return _rpCache;
}

vector<int> & Reactions::speciesXByType(State *s){
		vector<int> & species = s->cacheMSBT(this);
		for (unsigned int i=0; i < _childSpeciesUniqueIds.size(); i++){
			species[i] = s->x(_childSpeciesUniqueIds[i]);
		}
		return species;
}

vector<int> & Reactions::propensityEffectedByTheseSpecies() {
	if (_pebts.size() == 0) {
		BOOST_FOREACH(Reaction *r, descendantReactionList()) {
			for (int i = 0; i < r->numReactants(); i++) {
				int sid = r->getReactantId(i);
				if (!contains(_pebts, sid))
					_pebts.push_back(sid);
			}
		}
	}
	return _pebts;
}

vector<Reaction*> & Reactions::descendantReactionList(){
	if (_drl.size()==0)
		this->_buildDRL(_drl);
	return _drl;
}

void Reactions::addContainedSpecies(Species *s){
	// assert we don't have a duplicate?
	assert(!contains(_childSpeciesUniqueIds, s->getUniqueId()));
			//find(_childSpeciesUniqueIds.begin(), _childSpeciesUniqueIds.end(), s) == _childSpeciesUniqueIds.end());
	_childSpeciesUniqueIds.push_back(s->getUniqueId());
}

// pessimistically modifies state under the assumption that 'events' diffusion events 
// happens to *this* cell. 
// Should only be called on a base level cell
void Reactions::assumeDiffusionInto(State *state, int events){
	// this should be a base level cell Reactions*
	assert(onlyContainReaction());
	
	// for each contained cell species, add num events into it
	BOOST_FOREACH(int sIdx, _childSpeciesUniqueIds){
		state->changeX(sIdx, events);
	}
}


// Returns all bottom-level cells under this->parent but not 
//		under *this* which *this* events may diffuse into
// Build on demand.
vector<Reactions*> & Reactions::baseNeighbors(){
	if (_baseNeighbors.size()>0){
		return _baseNeighbors;
	}

	// build bottom up
	// Base case: if we are a cell reaction then return imediate neighbors
	if (onlyContainReaction()){
		for (unsigned int i=0; i < _nbrs.size(); i++){
			_baseNeighbors.push_back(_nbrs[i]->asReactions());
		}
		//_baseNeighbors.insert(_baseNeighbors.end(), _nbrs.begin(), _nbrs.end());
		return _baseNeighbors;
	}
	
	// if we know baseNeighbors of children, we take union of these and remove any belonging to *this*
	
	// take union of child baseNeighbors
	set<Reactions*> nbrs;
	
	BOOST_FOREACH(ReactionSet *child, getDependents()){
		if (child->isReaction()) continue;
		
		vector<Reactions*> & childNbrs = ((Reactions*)child)->baseNeighbors();
		nbrs.insert(childNbrs.begin(), childNbrs.end());
	}
	
	// now remove those which are under *this*
	BOOST_FOREACH(Reactions *r, nbrs){
		if (r->grandchildOf(this)){ //problem here
			nbrs.erase(r); // TODO make sure this works with iterator
		}
	}
	
	// now copy to class vector
	_baseNeighbors.insert(_baseNeighbors.end(), nbrs.begin(), nbrs.end());
	
	return _baseNeighbors;
}

// What is maximum a reaction under Reactions may change 
int Reactions::maxStoicDelta(int speciesUniqueId){
	// For all reactions determine maximum speciesUniqueId may increase in one event
	// TODO cache the results
	int maxDelta = 0;
	BOOST_FOREACH(Reaction *r, descendantReactionList()){
		maxDelta = max((int) r->delta(speciesUniqueId), maxDelta);
	}
	return maxDelta; 
}