/*
 * BuildReactionTopology.cpp
 *
 *  Created on: Sep 11, 2009
 *      Author: dweo
 */

#include <math.h>
#include <boost/foreach.hpp>

#include "build_reaction_topology.h"
#include "cell.h"
#include "../../utils/common.h"
#include "../../state/state.h"

using namespace hierleap;

// Root is one level, r is one level. No return value because r is added to root as (sub) dependent.
void BuildReactionTopology::_connectFlatRxnToRoot(Reactions *root, ReactionSet *r, int levelsToAdd){
	if (levelsToAdd<=0){
		root->addDependent(r);
		r->setParent(root);
		return;
	}
	
	// Create new reactions above r and connect up things.
	Reactions *rNew = new Reactions(NULL);
	rNew->addDependent(r);
	r->setParent(rNew);
	_connectFlatRxnToRoot(root, rNew, levelsToAdd-1);
}


void BuildReactionTopology::_reassignSpeciesReactionsRelationships(vector<Species*> species, Reactions *root){
	// Clear all species/reaction relationships
	// for all reactions and species
	root->clearContainedSpecies(); // a recursive function
	root->clearNeighbors();
	
	
	BOOST_FOREACH(Species *s, species){
		s->setParent(NULL);
	}
	
	
	// For each Species determine parent. Do this by going through all reactions.
	BOOST_FOREACH(Reaction *r, root->descendantReactionList()){
		// for each reactant
		for (int i=0; i < r->numReactants(); i++){
			int sId = r->getReactantId(i);
			assert(sId == species[sId]->getUniqueId());

			if (!species[sId]->getParent()){
				species[sId]->setParent(r->getParent());
			}
		}
		
	}
	
	// Look at products, this prioritizes reactants over products rxnSet
	BOOST_FOREACH(Reaction *r, root->descendantReactionList()){
		// for each product
		for (int i=0; i < r->numProducts(); i++){
			int sId = r->getProductId(i);
			assert(sId == species[sId]->getUniqueId());

			if (!species[sId]->getParent()){
				species[sId]->setParent(r->getParent());
			}
		}
	}

	
	// Double check that every species got assigned
	BOOST_FOREACH(Species *s, species){
		assert(s->getParent());
	}

	
	// Connect Neighbors
	_connectRxnNeighbors(root->descendantReactionList());
	_connectRxnsNeighbors(root->getDependents()); 
}

// Assume all RxnSets are on the same level
void BuildReactionTopology::_connectRxnNeighbors(vector<Reaction*> & rs){
	// Connection reactionSet Neighbors
	// this is inefficient, go through O(n^2) neighbor combinations and see who is related
	// Note: r1->isNeighbor(r2) <=> r2->isNeighbor(r1) 
	BOOST_FOREACH(Reaction *r1, rs){
		BOOST_FOREACH(Reaction *r2, rs){
			// May need to cast
			if (r1->isNeighbor(r2)){
				r1->addNeighbor(r2);
				r2->addNeighbor(r1);
			}
		}
	}
}

// I hate duplicating code...
// Note: rs must be ALL of type Reactions*
void BuildReactionTopology::_connectRxnsNeighbors(vector<ReactionSet*> & rs){
	// Connection reactionSet Neighbors
	// this is inefficient, go through O(n^2) neighbor combinations and see who is related
	// Note: r1->isNeighbor(r2) <=> r2->isNeighbor(r1) 
	BOOST_FOREACH(ReactionSet *r1, rs){
		BOOST_FOREACH(ReactionSet *r2, rs){
			Reactions *rr1 = r1->asReactions();
			Reactions *rr2 = r2->asReactions();

			if (rr1->isNeighbor(rr2)){
				rr1->addNeighbor(rr2);
				rr2->addNeighbor(rr1);
			}
		}
	}
}

// Convert vector of ReactionSet* to vector of Reactions*
vector<Reactions*> BuildReactionTopology::asReactionsVector(vector<ReactionSet*> v){
	vector<Reactions*> vNew;
	vNew.resize(v.size());
	for (int i=0; i < v.size(); i++){
		vNew[i] = v[i]->asReactions();
	}
	return vNew;
}

// Builds structure such that each 'cell' contains exactly one reaction. 
// The root cell will contain all reactions (|R| dependents)
void BuildReactionTopology::buildFlat(vector<Reaction*> & oneCell, vector<Species*> & cellSpecies, int levels,
												int x, int y, int z, bool diffusion, double diffusionRate,
												State * &stateReturnPtr, Reactions * &rootReturnPtr){
	// Create grid of reactions and species
	vector<Species*> gridSpecies;
	vector<ReactionSet*> rxnSets;
	
	// This is a list of cells, but each cell contains many reactions.
	// We need to rework the reactions.
	_myReactionsCreator(oneCell, cellSpecies, x, y, z, diffusion, diffusionRate, rxnSets, gridSpecies);
	
	
	rootReturnPtr = new Reactions(NULL);
	
	BOOST_FOREACH(ReactionSet *rs, rxnSets){
		BOOST_FOREACH(ReactionSet *r, rs->getDependents()){
			assert(r->isReaction());
			r->voidParent();
			_connectFlatRxnToRoot(rootReturnPtr, r, levels-2);
		}
		//delete rs; // TODO: hopefully this is iterator safe (?)
	}
	
	_reassignSpeciesReactionsRelationships(gridSpecies, rootReturnPtr);
	
	
	// Quality control check
	int h = rootReturnPtr->height();
	assert(rootReturnPtr->height() == levels);
	
	// Give each reaction a unique and meaningful name
	rootReturnPtr->idSystem("root");
	
	// Create state from species, etc
	stateReturnPtr = new State(gridSpecies, rootReturnPtr); 
	BOOST_FOREACH(Species *s, stateReturnPtr->species()) {
		s->setParent(rootReturnPtr);
	}
}


// Primary function of this class. "Return values" are the pass by reference
//	parameters at the end of function signature.
void BuildReactionTopology::build(vector<Reaction*> & oneCell, vector<Species*> & cellSpecies, int levels,
												int x, int y, int z, bool diffusion, double diffusionRate, int cellsPerMetaCell,
												State * &stateReturnPtr, Reactions * &rootReturnPtr){

	// Create grid of reactions and species
	vector<Species*> gridSpecies;
	vector<ReactionSet*> rxnSets;
	_myReactionsCreator(oneCell, cellSpecies, x, y, z, diffusion, diffusionRate, rxnSets, gridSpecies);
	
//	BOOST_FOREACH(Species *s, gridSpecies){ 
//		s->getParent();
//	}
	
	rxnSets = _mergeSomeCells(rxnSets, gridSpecies, cellsPerMetaCell);
//	BOOST_FOREACH(Species *s, gridSpecies){ 
//		s->getParent();
//	}
	
	// build bottom up
	rootReturnPtr = _hierchyCreator(rxnSets, levels-2); // -2 because we need to include bottom layer of Reaction and cell layer of rxns
//	BOOST_FOREACH(Species *s, gridSpecies){ 
//		s->getParent();
//	}
	
	assert(rootReturnPtr->height() == levels);

	// Give each reaction a unique and meaningful name
	rootReturnPtr->idSystem("root");
	
//	BOOST_FOREACH(Species *s, gridSpecies){ 
//		s->getParent();
//	}
	
	// Create state from species, etc
	stateReturnPtr = new State(gridSpecies, rootReturnPtr); 
	
	BOOST_FOREACH(Reaction *r, rootReturnPtr->descendantReactionList()) {
		r->updateSpeciesInfo(stateReturnPtr);
	}
	
	BOOST_FOREACH(Species *s, stateReturnPtr->species()) {
		s->setParent(rootReturnPtr);
	}

	//rootReturnPtr->print(stateReturnPtr, "ni");
}

void BuildReactionTopology::assertConsistency(Reactions* root, State *state) {
	vector<ReactionSet*> & cells = root->getDependents();
	BOOST_FOREACH(Reaction *r, root->descendantReactionList()) {
		assert(r);
		assert(r->isReaction());
		assert(r->getRoot() == root);
		assert(r->getParent()->getDependents()[r->indexInParent()] == r);
		assert(contains(cells, (ReactionSet*) r->getParent()));
		assert(r->getUniqueId() >= 0);
	}
	
	BOOST_FOREACH(Species *s, state->species()) {
		assert(s);
		//assert(contains(cells, (ReactionSet*) s->getParent()));
		assert(s->getUniqueId() >= 0);
	}
}

// Merge every |cellsMergedPerGroup| reactionsets into a single reaction set, deleting 
// extra 'parent' cells in the process.
vector<ReactionSet*> BuildReactionTopology::_mergeSomeCells(vector<ReactionSet*> & cells, 
															vector<Species*> & species, 
															int cellsMergedPerGroup){
	vector<ReactionSet*> mergedCells;
	for (int i = 0; i < cells.size(); i += cellsMergedPerGroup) {
		Reactions* newParent = cells[i]->asReactions();
		for (int j = i+1; j < min(i + cellsMergedPerGroup, (int)cells.size()); j++) {
			BOOST_FOREACH(Reaction *r, cells[j]->asReactions()->descendantReactionList()) {
				r->changeParent(newParent);
			}
			
			// Set parents for species. TODO: this is O(n^2) when it could be O(c) (?).
//			BOOST_FOREACH(Species *s, species) {
//				if (s->getParent() == cells[j])
//					s->setParent(newParent);
//			}
			delete cells[j];
		}
		
		mergedCells.push_back(newParent);
	}
//	BOOST_FOREACH(Species *s, species) {
//		s->setParent(NULL);
//	}
	return mergedCells;
}

// Note: this macro is fragile and depends on x,y,z being in scope and being the max dim sizes.
#define IDX(a, b, c) (c) * x * y + (b) * x + (a)

// Creates a list of reactions and species
// uses cellReactions/cellSpecies as templates for each cell to contain
void BuildReactionTopology::_myReactionsCreator(vector<Reaction*> & cellReactions,
			vector<Species*> & cellSpecies, int x, int y, int z, bool diffusion,
			float diffusionRate, vector<ReactionSet*> & rxnSets, vector<Species*> & species) {

	Cell **cellgrid = new Cell*[x * y * z]; 
	for (int i = 0; i < x * y * z; i++){
		cellgrid[i] = NULL;
	}
	

	// 'meta' id our cell species
	int tid = 0;
	for (vector<Species*>::iterator i = cellSpecies.begin();
		i != cellSpecies.end(); i++, tid++){
		(*i)->setTypeId(tid);
	}

	// create a lot of cells in a 3d-grid
	for (int ix = 0; ix < x; ix++){
		for (int iy=0; iy < y; iy++){
			for (int iz=0; iz < z; iz++){
				cellgrid[IDX(ix, iy, iz)] =  new Cell(cellReactions, cellSpecies, ix,iy,iz);
			}
		}
	}

	// Add diffusion reactions to the system
	// for each cell, for each nbr, add incoming diffusion to this cell
	for (int ix = 0; ix < x; ix++){
		for (int iy=0; iy < y; iy++){
			for (int iz=0; iz < z; iz++){

				if (diffusion){

				// for each neighbor direction
				for (int tx=-1; tx <= 1; tx++){ for (int ty=-1; ty <= 1; ty++){for (int tz=-1; tz <= 1; tz++){

					// is legit neighbor?
					if (abs(tx) + abs(ty) + abs(tz) != 1) continue; // no diffusion to self, or diagonals

					// calc absolute index of neighbor
					int mx = ix+tx; int my = iy+ty; int mz = iz+tz;
					if (mx < 0 || x <= mx || my < 0 || y <= my || mz < 0 || z <= mz) continue; // out of bounds

					cellgrid[IDX(ix, iy, iz)]->addDiffusionReactions(cellgrid[IDX(mx, my, mz)], diffusionRate);
					cellgrid[IDX(ix, iy, iz)]->addNeighbor(cellgrid[IDX(mx, my, mz)]);

				}}}

				}

				Reactions *r = cellgrid[IDX(ix, iy, iz)]->parentRxnset();
				
				// used to build hierarchy of reactions
				extend_vector(species, cellgrid[IDX(ix, iy, iz)]->getSpecies());

				if (contains(rxnSets, (ReactionSet*)r)) {
					r->printCoordinates();
					assert(false);
				}
				rxnSets.push_back(r);

			}
		}
	}

	// Cleanup
	for (int i = 0; i < x * y * z; i++) {
		assert(cellgrid[i]);
		delete cellgrid[i];
	}
	delete [] cellgrid;
}


// Try to make it so that each Reactions->parent has aveChildRxns children
// works from leaf nodes up
// switch ARG(levels) : 
//		L(0): assert(rxns.size()==1); return rxns[0];
//		L(1): root=new Reactions(); root->addDependents(rxns);
// 		L(n): int aveChildRxns = ...; for(i=1..aveChildRxns) rxnsNew->append((new Parent())->addDependents(rxns[:aveChildRxns:]); 
//										return L(n-1, rxnsNew);
// Returns: root node
Reactions* BuildReactionTopology::_hierchyCreator(vector<ReactionSet*> & rxns, int levels){
	assert(rxns.size() > 0); assert(levels >= 0);
	
	if (levels==0){
		assert(rxns.size()==1); 
		return rxns[0]->asReactions();
	}
	
	
	int aveChildRxns;
	if (levels==1){
		aveChildRxns = INT_MAX; // add all reactions to one parent
	}else{
		int numCells = rxns.size();
		aveChildRxns = (int)pow(double(numCells), 1.0 / levels);
	}
	

	// just step through reactions and build them that way
	vector<ReactionSet*> outputRxns;
	//int childrenPerParent = 1 + rxns.size() / aveChildRxns;
	int childrenPerParent = max(aveChildRxns, 2);
	while(!rxns.empty()){

		Reactions *newParent = new Reactions(NULL);
		for (int i=0; i < childrenPerParent && !rxns.empty(); i++){
			ReactionSet *child = rxns.back();
			rxns.pop_back();
			child->setParent(newParent);
			newParent->addDependent(child);
		}

		outputRxns.push_back(newParent);
	}

	// now connect neighbors
	//for (vector<ReactionSet*>::iterator parentR = outputRxns.begin(); parentR != outputRxns.end(); parentR++){
//		vector<ReactionSet*> & parentNbrs = (*parentR)->getNeighbors();
//
//		// look at children and see to whom they connect
//		vector<ReactionSet*> & children = (*parentR)->getDependents();
//		for (unsigned int i=0; i < children.size(); i++){
//
//			vector<ReactionSet*> & childNbrs = children[i]->getNeighbors();
//			for (unsigned int j = 0; j < childNbrs.size(); j++){
//
//				if (childNbrs[j]->getParent() != (*parentR)){
//					// see if child's nieghbors parent's are not already a neighbor of parent
//					if (find(parentNbrs.begin(), parentNbrs.end(), childNbrs[j]->getParent()) == parentNbrs.end()){
//						(*parentR)->addNeighbor(childNbrs[j]->getParent());
//					}
//				}
//			}
//		}
//	} todo: add for more layers.

	// recursive call to do higher up layers
	return _hierchyCreator(outputRxns, levels-1);
}


