# David Orendorff
#
# generate data for a bunch of algorithm run times (SSA/HIERLEAP, CPU/GPU, etc)

import subprocess
import os
import sys
import pickle # save optimal leap discoveries for future use.
from numpy import polyfit, polyval, poly1d

from copy import copy

#  /Users/dweo/research/gpusvm/build/Release/gpu_hierleap
#'-f /Users/dweo/research/er_leap/tests/cases/dsmts-001-01.xml -r 2000 -t 50 -p 1 -o . -l 3 -c GPU -a SSA -x 35 --seed 213'
# todo: change output dir from . to trash

duration = .8

# sbmlFile = '/Users/dweo/research/er_leap/tests/cases/dsmts-001-01.xml'
sbmlFile = '/Users/dweo/research/er_leap/out/GeysChaos.xml'
runProgram = '../build/Release/gpu_hierleap -f %(sbmlFile)s -r %(runs)d -t %(duration)f -p %(updatePeriod)f -o .trash/ -l %(leap)d -c %(architecture)s -a %(algorithm)s -x %(nCells)d'
optimalLeapFileName = 'bestHierLeap.txt'

# note: this does not use paralization
def getOptimalLeap(sbmlFile, nCells, alg, arch):
	'This requires file I/O on every call, other than that not too slow.'
	assert(alg!='SSA')
	fn = sbmlFile.split(os.sep)[-1]
	key = (fn, nCells)
	try: 
		f = file(optimalLeapFileName, 'r')
		lookup = pickle.load(f)
		f.close()
	except IOError:
		print 'unable to load saved values'
		lookup = {}
	
	if key in lookup:
		#print 'Loading optimal for %s from file\n' % str(key)
		return lookup[key]
	
	# we need to determine the optimal leap, then save to disk
	# print 'determining the optimal leap for ' + str(key)
	bestLeap = findOptimalLeap({'sbmlFile':sbmlFile, 'duration':duration, 'updatePeriod':duration/10.0, 'runs':6, 'architecture':arch, 'algorithm':alg, 'nCells':nCells})
	lookup[key] = bestLeap
	f = file(optimalLeapFileName, 'w')
	pickle.dump(lookup,f)
	f.close()
	
	return bestLeap

def averageRuntime(cmd, failuresAccepted=0):
	print "Running command %s..." % cmd
	p = subprocess.Popen(cmd, stdout = subprocess.PIPE, shell=True)

	p.wait() # note: this could possible lock if stdout buffer is too big (?)
	out,errs = p.communicate()
	# print out
	assert(errs==None)
	
	# stdout is now stored in out, parse to get vals. Note: this is fragile.
	times = out.split('\n')
	times = filter(lambda s: 'duration' in s, times)
	times = map(lambda s: float(s.split()[3]), times)
	times = times[1:] # skip the first entry
	if (not times):
		if failuresAccepted > 0:
			return averageRuntime(cmd, failuresAccepted-1)
		else:
			print cmd
			print out
			sys.stdout.flush()
			assert(False)
	return sum(times) / len(times) # we may want to truncate first few readings.
	
def mean(x):
	assert(x)
	return sum(x) / len(x) 
	
# assumes the runtime is monotonic and runtimes are not 'too' noisy.
# This is a heuristic finite difference gradient decent, roughly based on Nelder-Mead
def findOptimalLeap(runArgs):
	args = copy(runArgs)
	runtimesSoFar=[]
	
	def leapRuntime(leap, maxSamples=1):
		"sample or looks up the leapRunningtime for a particular value of L. if maxSamples > 1 we'll take another sample and average over all"
		
		if len(runtimesSoFar) <= leap:
			runtimesSoFar.extend([()] * (leap * 2))
			
		if len(runtimesSoFar[leap]) < maxSamples:
			args['leap'] = leap
			cmd = runProgram % args
			avg = averageRuntime(cmd, 4)
			print 'leap(%f) avg(%f)' % (leap, avg)
			sys.stdout.flush()
			runtimesSoFar[leap] += (avg,)
			
		return mean(runtimesSoFar[leap])
		
	leap = 2
	minRuntime = leapRuntime(2)
	 # our leap advancement basically grows as log(leap)
	while minRuntime * 2 >= leapRuntime(leap) and leap < 150:
		leap += int(leap / 5.0) + 2
		minRuntime = min(minRuntime, leapRuntime(leap))
	
	# we should now have a nice sampling of points over the important range
	# regress on an order 2 polynomial, removing 'worst' point at each iteration until we have answer 2 lowest points and answer
	def sampled():
		return filter(lambda i: runtimesSoFar[i], range(len(runtimesSoFar)))
	
	# not too efficient but it shouldn't really matter
	maxRedundant = 3
	numToConsider = len(sampled())
	while numToConsider >= 4:
		testedLeaps = sampled()
		testedLeaps.sort(key=lambda i: leapRuntime(i))
		p = polyfit(testedLeaps[:numToConsider], map(leapRuntime, testedLeaps[:numToConsider]), 2)
		if not (p[0] > 0 and p[1] < 0):
			"We didn't find a 'nice' polynomial to fit. coeffs: " + str(p)
			zeroX = mean(testedLeaps[:numToConsider])
		else:
			zeroX = int(round(-p[1] / (2 * p[0]))) # in y = ax^2+bx+c, this is -b/(2a)
		leapRuntime(zeroX, maxRedundant)
		numToConsider -= 1
		
	# now just have a few leaps to consider and want to find the minimum vals
	best = sampled()
	best.sort(key=lambda i: leapRuntime(i))
	best = best[:min(len(best), 3)]
	r = map(leapRuntime, best)
	i = r.index(min(r))
	return best[i]
	
def printMethodRuntimes(nSamplesPerCase):
	#algorithms = ['HIER', 'SSA'];
	algorithms = ['HIER_SEP'];
	#architecture = ['GPU', 'CPU'];
	architecture = ['None'];
	# nCells = [2, 3, 5, 8, 13, 17, 22, 27, 31, 34] # map(lambda x: x**2, range(1, 7))
	nCells = map(lambda x: x**2, range(1, 14))
	
	for alg in algorithms:
		for arch in architecture:
			print alg + arch + '= ['
			for cells in nCells:
				name = alg + arch + str(cells)
				if alg == 'SSA':
					bestL = 1
				else:
					bestL = getOptimalLeap(sbmlFile, cells, alg, arch)
				bestL = 3 # todo remove
				r = averageRuntime(runProgram % {'sbmlFile':sbmlFile, 'runs':nSamplesPerCase, 'duration':duration, 'updatePeriod':duration/10.0, 'architecture':arch, 'algorithm':alg, 'nCells':cells, 'leap':bestL})
				print '%d %f' % (cells, r)
			print ']'

if __name__=='__main__':
	printMethodRuntimes(8)
#	for cells in range(3,36):
#		b = getOptimalLeap(sbmlFile, cells)
#		print 'Best leap %d for %d cells' % (b, cells)
	
	
