/*
 *  cell_computation_task.h
 *  gpu_hierleap
 *
 *  Created by David Orendorff on 12/8/11.
 *  Copyright 2011 UC Irvine. All rights reserved.
 *
 */

#include <boost/thread.hpp>
#include <utmpx.h>
//#define _GNU_SOURCE
//#include <sched.h>

#ifndef CELL_COMPUTATION_TASK
#define CELL_COMPUTATION_TASK

template <class T, class C>
class Task {
public:
	Task(T *_ce) : ce(_ce) {assert(_ce); }
	
	vector<C> objs;
	T *ce;
};

// Executes an object of type CellExecuter on each cell.
// For each CellExecuter instance we require two functions:
//	   CellExecuter::copy() and CellExecuter::operator (Reactions* cell)
template <class Executer, class ExecuteOnMeObject>
class ComputationTask {
private:
	//vector<ReactionSet*> & _cells;
	//vector<boost::thread> _threadTask;
	boost::thread_group _threads; 
	vector<Task<Executer, ExecuteOnMeObject> *> _tasks;
	boost::barrier *_startCalculation;
	boost::barrier *_endCalculation;
	
	// Executes the task on the current thread.
	static void _executeTask(Task<Executer, ExecuteOnMeObject> *t) {
		BOOST_FOREACH(ExecuteOnMeObject obj, t->objs) {
			(*t->ce)(obj);
		}
	}
	
	static void _startThread(ComputationTask *_this, int _threadIdx) {
		//sched_getcpu();
		while (1) {
			_this->_startCalculation->wait();
			
			ComputationTask::_executeTask(checkNotNull(_this->_tasks[_threadIdx]));
			
			_this->_endCalculation->wait();
		}
	}
	
public:
	// nThreads if -1, will result in the defauult number of threads, otherwise we 
	// will try to use the given number of threads.
	ComputationTask(Executer *ce, vector<ExecuteOnMeObject> & objs, int nThreads) {
		double nTasks = min(maxConcurrentThreads(), (unsigned int) objs.size());
		nTasks = nThreads > 0 ? nThreads : nTasks;
		assert(nTasks >= 1);
		//nTasks = 2;
		_tasks.resize(nTasks, NULL);
		
		for (int i = 0; i < objs.size(); i++) {
			double r = (double) i / objs.size();
			int tid = (int) (r * nTasks);
			assert(tid < nTasks);
			if (!_tasks[tid]) 
				_tasks[tid] = new Task<Executer, ExecuteOnMeObject>(ce->copy());
			_tasks[tid]->objs.push_back(objs[i]);
		}
		
		_startCalculation = new boost::barrier(_tasks.size());
		_endCalculation = new boost::barrier(_tasks.size());
		
		for (int i = 1; i < nTasks; i++) {
			boost::thread t = boost::thread(&_startThread, this, i);
			_threads.add_thread(&t);
		}
	}
	
	// This code should not be replicated.
	ComputationTask(Executer *ce, vector<ExecuteOnMeObject> & objs) {
		assert(false); // make sure you don't want to call this without specifying nThreads.
		double nTasks = min(maxConcurrentThreads(), (unsigned int) objs.size());
		assert(nTasks >= 1);

		_tasks.resize(nTasks, NULL);
		
		for (int i = 0; i < objs.size(); i++) {
			double r = (double) i / objs.size();
			int tid = (int) (r * nTasks);
			assert(tid < nTasks);
			if (!_tasks[tid]) 
				_tasks[tid] = new Task<Executer, ExecuteOnMeObject>(ce->copy());
			_tasks[tid]->objs.push_back(objs[i]);
		}
		
		_startCalculation = new boost::barrier(_tasks.size());
		_endCalculation = new boost::barrier(_tasks.size());
		
		for (int i = 1; i < nTasks; i++) {
			boost::thread t = boost::thread(&_startThread, this, i);
			_threads.add_thread(&t);
		}
	}
	
	// Executes the tasks on the threads owned by *this*.
	void run() {
		_startCalculation->wait();
		_executeTask(_tasks[0]);
		_endCalculation->wait();
	}
};

#endif