/*
 *  functor_add_reaction_calculate_low_bound.h
 *  gpu_hierleap
 *
 *  Created by David Orendorff on 5/5/12.
 *  Copyright 2012 UC Irvine. All rights reserved.
 *
 */

#include "time_hierleap_separate_clean.h"

#include <cmath>
#include <numeric>
#include <functional>

#include "tbb/blocked_range.h"

#include "../../globals.h"

#include "../time_hierleap_separate_clean.h"

#include "../../reactions/connection_grandmaster.h"
#include "../../reactions/reaction_permutation.h"
#include "../../reactions/reaction_set.h"
#include "../../reactions/reactions.h"

#include "functors.h"

#include "../../state/state.h"

using namespace tbb;
using namespace std;

#ifndef functor_add_reaction_calculate_low_bound
#define functor_add_reaction_calculate_low_bound

// Class to calculate cell bounds with POV in parallel.
class ApplyCalculateCellBoundsWithPov {
		TimeHierLeapSepClean * _this;
public:
	ApplyCalculateCellBoundsWithPov(TimeHierLeapSepClean * _thiis) : _this(_thiis) {}
	
	ApplyCalculateCellBoundsWithPov * copy() {
		return new ApplyCalculateCellBoundsWithPov(_this);
	}
	
	void operator()( const blocked_range<std::vector<ReactionSet*>::iterator> & cells ) const { 
        for( std::vector<ReactionSet*>::iterator c=cells.begin(); c != cells.end(); c++) { 
			(*this)((*c)->asReactions());
		}
    }
	
	virtual void operator () (Reactions *cell) const {
		int events = _this->_state->numEvents(cell);
		assert(events == _this->_metaCellEvents[cell->getUniqueId()]);
		BoundingArrays<double> * cellD = _this->_cellD;
		State * s = _this->_state;
		if (!events) {
			cellD->highCellBound[cell->getUniqueId()] = cellD->x[cell->getUniqueId()];
			cellD->lowCellBound[cell->getUniqueId()] = cellD->x[cell->getUniqueId()];
			return;
		}
		
		// Compute bounds.
		double cellDhighCellSum = 0.0;
		double cellDlowCellSum = 0.0;
		
		vector<int> & exactUpperBound = cell->speciesSandbox(s->speciesSize());
		_this->_pb->upperBoundSpeciesParentEventsGivenWithCellPov(cell, s, exactUpperBound);
		BOOST_FOREACH(Reaction *r, cell->descendantReactionList()){
			cellD->highCellBound[r->getUniqueId()] = r->DbyUniqueId(exactUpperBound);
			cellDhighCellSum += cellD->highCellBound[r->getUniqueId()];
		}
		
		vector<int> & exactLowerBound = cell->speciesSandbox(s->speciesSize());
		_this->_pb->lowerBoundSpeciesParentEventsGivenWithCellPov(cell, s, exactLowerBound);
		BOOST_FOREACH(Reaction *r, cell->descendantReactionList()){
			cellD->lowCellBound[r->getUniqueId()] = r->DbyUniqueId(exactLowerBound);
			cellDlowCellSum += cellD->lowCellBound[r->getUniqueId()];
		}
		
		cellD->highCellBound[cell->getUniqueId()] = min(cellDhighCellSum, cellD->highBound[cell->getUniqueId()]);
		cellD->lowCellBound[cell->getUniqueId()] = max(cellDlowCellSum, cellD->lowBound[cell->getUniqueId()]);		
	}
};

//////////////////////////////////
///CLASS TO SAMPLE BLOCK IN PARALLEL 
//
// After calling this on a cell state->numEvents(cell) will be 0. but _this->_metacellEvents[cellId] will not be.
class ApplyCalculateLowBoundsSampleBlock {
private:
	static const int MAX_ITER = 8;
	TimeHierLeapSepClean * _this;
	bool _needPermutation;
	bool & _failure;
public:
	ApplyCalculateLowBoundsSampleBlock(TimeHierLeapSepClean * _thiis, bool needPermutation, bool & onFailure) : 
	_this(_thiis), _needPermutation(needPermutation), _failure(onFailure) {}
	
	void operator()( const blocked_range<std::vector<ReactionSet*>::iterator> & cells ) const { 
        for( std::vector<ReactionSet*>::iterator c=cells.begin(); c != cells.end(); c++) { 
			(*this)((*c)->asReactions());
		}
    }
	
	ApplyCalculateLowBoundsSampleBlock * copy() {
		return new ApplyCalculateLowBoundsSampleBlock(_this,_needPermutation, _failure);
	}
	
	// samples reactions channels insided of cell. The sampled reaction permutation
	// may be accesed from the cell->reactionPermutationSandbox().
	virtual void operator () (Reactions *cell) const {
		int events = _this->_state->numEvents(cell);
		BoundingArrays<double> * cellD = _this->_cellD;
		State * s = _this->_state;
		ReactionPermutation<Reaction*> *rp = cell->reactionPermutationSandbox();
		rp->clear();
		if (!events) {
			return;
		}
		
		assert(events == _this->_metaCellEvents[cell->getUniqueId()]);
		
		// determine our "connected components"
		vector<Reactions*> connectedCells;
		cell->connectedCellsWithEvents(connectedCells, _this->_state);
		
		// only proceed if *cell* has the highest id value.
		int myId = cell->getUniqueId();
		assert(myId != 0);
		BOOST_FOREACH(Reactions * nbrC, connectedCells) {
			if (nbrC->getUniqueId() < myId) {
				return;
			}
			
		}
		
		// indirect way to indicate that we have sampled from these cells.
//		BOOST_FOREACH(Reactions *cell, connectedCells) { // may speed up code but NOT thread safe (eg, if we split a connected component in the middle)
//			_this->_state->setEvents(cell, 0);
//		}
		
		ApplyAddReactionCalculateLowBound aarclb(rp, cell->getRoot(), cellD); 
		for (int i = 0; i < MAX_ITER; i++){
			
			BOOST_FOREACH(Reactions *c, connectedCells) {
				aarclb.setParentCell(c);
				multirnd_callback(_this->_metaCellEvents[c->getUniqueId()], &cellD->highCellBound[c->smallestChildId()], 
								  c->numDependents(), aarclb); 
			}
			
			if (_needPermutation)
				rp->permute(); 
			
			double u = urand();
			//  _acceptanceBlockProb will not get called if early accept.
			if (u <= aarclb.getLowerAcceptanceBound()) { 
				return;
			}
			
			vector<int> & localState = cell->speciesSandbox(s->speciesSize());
			BOOST_FOREACH(Reactions * c, connectedCells) {
				_copyRelatedSpecies(s->X(), localState, c); 
			}
			
			if( u <= _this->_acceptanceBlockProb(rp, localState, cellD->highCellBound)) {
				return;
			}
			
			aarclb.resetLowerAcceptanceBound();
			rp->clear();
		}
		_failure = true;
	}
	
	void _copyRelatedSpecies(vector<int> & fromState, vector<int> & toState, Reactions * cellReference) const {
		ConnectionGrandmaster *cg = ConnectionGrandmaster::getConnectionGrandmaster(_this->_root);
		BOOST_FOREACH(DeltaSpecies * ds, *cg->possiblyIncreasedByCellEvent(cellReference)) {
			int sid = ds->speciesId;
			toState[sid] = fromState[sid];
		}
		BOOST_FOREACH(DeltaSpecies * ds, *cg->possiblyDecreasedByCellEvent(cellReference)) { // may do at most 2x work here.
			int sid = ds->speciesId;
			toState[sid] = fromState[sid];
		}
	}
};

#endif