/*
 *  functors.h
 *  gpu_hierleap
 *
 *  Created by David Orendorff on 2/21/12.
 *  Copyright 2012 UC Irvine. All rights reserved.
 *
 */

#ifndef _FUNCTORS_H_
#define _FUNCTORS_H_

#include <vector>

#include "tbb/blocked_range.h"
#include "tbb/parallel_for.h"

#include "bounding_arrays.h"
#include "../time_hierleap.h"
#include "../../reactions/reaction_permutation.h"
#include "../../reactions/reaction_set.h"
#include "../../reactions/reactions.h"

#include "../../utils/common.h"

using namespace std;
using namespace tbb;

// Note: this class is a friend to TimeHierLeap
// Adjusts _cellD->lowBound, _cellD->highBound to reflect the system after rxns have fired.
// Assumes _this_->_states->highBound is up to date.
class ApplyCellBoundUpdate { 
	boost::function<DeltaPropensityBounds* (Reactions *)> _getDeltaPropensityComputer;
	int _leap;
	ReactionPermutation<Reaction*> * _rxns;
	BoundingArrays<int> * _state; 
	BoundingArrays<double> * _D;
	ReactionPermutation<Reaction*> **_firedRxns;
	
public: 
	ApplyCellBoundUpdate(boost::function<DeltaPropensityBounds* (Reactions *)> getDeltaPropensityComputer, 
						 BoundingArrays<int> * state, BoundingArrays<double> * D, ReactionPermutation<Reaction*> **rxns, int leap) : 
	_leap(leap), _state(state), _D(D), _firedRxns(rxns) {
		_getDeltaPropensityComputer = boost::ref(getDeltaPropensityComputer);
	}

	
	ApplyCellBoundUpdate * copy() {
		return new ApplyCellBoundUpdate(_getDeltaPropensityComputer, _state, _D, _firedRxns, _leap);
	}
	
	void operator()( Reactions *cell) const {
		vector<int> & stateHighBound = _state->highBound;
		int cid = cell->getUniqueId();
		_getDeltaPropensityComputer(cell)->updateUpperBoundDeltaDs(*_firedRxns, stateHighBound);
		
		double maxPosDelta = _getDeltaPropensityComputer(cell)->upperBoundDeltaD(stateHighBound);
		double maxNegDelta = _getDeltaPropensityComputer(cell)->lowerBoundDeltaD(stateHighBound);
		assert(maxNegDelta <= 0);
		
		_D->highBound[cell->getUniqueId()] = _D->x[cid] + maxPosDelta * (_leap - 1); 
		double low = _D->x[cid] + maxNegDelta * (_leap - 1); 
		if (low < 0) low = 0.0;
		_D->lowBound[cell->getUniqueId()] = low;
	}
	
    void operator()( const blocked_range<std::vector<ReactionSet*>::iterator> & cells ) const { 
        for( std::vector<ReactionSet*>::iterator c=cells.begin(); c != cells.end(); c++) { 
			(*this)((*c)->asReactions());
		}
    }
};

/*
 * Stores the propensity in D after computing from state.
 * Note: does not compute total propensity (root), which is the sum of cell propensities.
 */
class ApplyCalculateD {
public:
	ApplyCalculateD(vector<int> * state, vector<double> * D) : _state(state), _D(D) {
	}
	
	ApplyCalculateD * copy() {
		return new ApplyCalculateD(_state, _D);
	}
	
	void operator()( Reactions *cell) const {
		double sum = 0.0;
		BOOST_FOREACH(Reaction *r, cell->descendantReactionList()) {
			double d = r->DbyUniqueId(*_state);
			(*_D)[r->getUniqueId()] = d;
			sum += d;
		}
		(*_D)[cell->getUniqueId()] = sum;
	}
	
    void operator()( const blocked_range<std::vector<ReactionSet*>::iterator> & cells ) const { 
        for( std::vector<ReactionSet*>::iterator c=cells.begin(); c != cells.end(); c++) { 
			(*this)((*c)->asReactions());
		}
    }
private:
	vector<int> * _state;
	vector<double> * _D;
};

class ApplyParallelAddReaction : public ApplyMultirndCallbackFunc{
	ReactionPermutation<Reaction*> *_rp;
	vector<ReactionSet*> & _rxns;
public:
	ApplyParallelAddReaction(ReactionPermutation<Reaction*> * rp, vector<ReactionSet*> & rxns) : _rp(rp),  _rxns(rxns) {}
	
	virtual void operator () (int idx, int k){
		_rp->parallelAddReaction(_rxns[idx]->asReaction(), k);
	}
};


/**
 * When operator () is called, add the corresponding reaction in cell->descendantReactionList() to the ReactionPermutation.
 * Additionally, compute LowerAcceptanceBound as we get samples.
 */
class ApplyAddReactionCalculateLowBound : public ApplyMultirndCallbackFunc {
public:
	ApplyAddReactionCalculateLowBound(ReactionPermutation<Reaction*> * rp, 
									  Reactions * root, 
									  BoundingArrays<double> * cellD) : _rp(rp), _root(root), _cellD(cellD), 
		_lowBoundAcceptance(1.0), _cellRxn(NULL) { }
	
	void setParentCell(Reactions *cell) {
		_cellRxn = &cell->descendantReactionList();
	}
	
	virtual void operator () (int idx, int k){
		assert(k >= 0 && k < _cellRxn->size());
		Reaction *r = _cellRxn->at(idx);
		_rp->addReaction(r, k);
		int ridx = r->getUniqueId();
		_lowBoundAcceptance *= ipow(_cellD->lowCellBound[ridx] / _cellD->highCellBound[ridx], k);
	}
	
	double getLowerAcceptanceBound() {
		return _lowBoundAcceptance;
	}
	
	void resetLowerAcceptanceBound() {
		_lowBoundAcceptance = 1.0;
	}
	
private:
	Reactions * _root;
	ReactionPermutation<Reaction*> *_rp;
	BoundingArrays<double> * _cellD;
	
	double _lowBoundAcceptance;
	vector<Reaction*> * _cellRxn;
};

#endif