/*
 *  full_simulator.cpp
 *  gpu_hierleap
 *
 *  Created by David Orendorff on 12/22/11.
 *  Copyright 2011 UC Irvine. All rights reserved.
 *
 */


#include <math.h>
#include <string.h>

#include "full_simulator.h"

#include "../state/state.h"
#include "../reactions/reactions.h"
#include "reaction_leaper.h"
#include "ssa.h"
#include "er_leap.h"
#include "hier_leap.h"
#include "reaction_tally.h"
#include "historian.h"
#include "../globals.h"
#include "reaction_tau_leaper.h"
#include "../sim/tau_leap_ssa.h"
#include "simulator.h"

using namespace hierleap;


FullSimulator::FullSimulator(Reactions *root) : _rtlEvents(0), _SSAEvents(0), _totIterations(0), _missedHistWrites(0){
	_ssaSim = new TauLeapSSA(root);
}
	
	
	// updates state to reflect SSA equivalent simulation
	//		of reactions locate under *root* for duration *time*
void FullSimulator::run(ReactionTauLeaper *sim, Historian *hist, Reactions *root, double time, int leap, State *state){
		assert(leap > 0 && leap < MAX_EVENTS);
		_sim = sim;
		
		//Simulator dtSim(root, state); // TODO added for debuggin
		
		double t = 0.0;
		int nChangeRuntimeParams = 1;
		sim->changeRuntimeParameters(state, leap);
		double Dsum = 1.0;
		while (t <= time){
			if (Dsum == 0){
				double dt = INFINITY; 
				hist->jot(state, t, dt);
				t += dt;
				break;
			}
			
			// If it seems likely that we will pass the next write-time in hist, 
			// create a temp state until we update. Then delete this state and continue from
			// point we left off initially.
			double stepsAhead = (double)leap*3.5; // assuming D doesn't change, how many steps until we reach update-time
			// if SSA needed soon
			//if (false){
			//if (t >= 20.0){ // XXX this disables SSA update, so sims could be a bit off
			//			if (t + stepsAhead/D > hist->nextUpdate()){ 
			//			//if (true){
			//				// TODO keep track of how much work is done by SSA
			//				// copy state
			//				State *tState = state; //new State(state);
			//				
			//				// SSA steps until hist update
			//				double deltatime = 0; 
			//				double & myTime = t; // XXX debugging
			//				while(t <= time && !hist->jot(tState, myTime, deltatime)) {
			//					// also check for D=0
			//					myTime += deltatime;
			//					deltatime = 0; // shouldn't really be needed
			//					ReactionPermutation<Reaction*> *ssaRxns = _ssaSim->ReactionTauLeap(1, tState, deltatime); //_ssaSim
			//					delete ssaRxns;
			//					_SSAEvents++;
			//				}
			//				myTime += deltatime;
			//				//delete tState;
			//				sim->changeRuntimeParameters(state, leap);
			//				_totIterations++;
			//				printf("ran SSA %f\n", t);
			//				continue; 
			//			}
			
			// sample reactions
			double dt=0.0;
			//State *stateCpy = new State(state); 
			//ReactionPermutation<Reaction*> *rxns = sim->ReactionTauLeap(leap, stateCpy, dt);
//			if (t > 10) {	// experimental hack to see if we exactly simulate first 10 sec, problems still persist.
//				Dsum = sim->ReactionTauLeap(leap, state, dt);
//			} else {
//				dt = 0; 
//			}
			Dsum = sim->ReactionTauLeap(leap, state, dt);
			
			
			// Method could not advance
			// TODO: may want to use L=1 whenever we are close to a write-time
			if (dt==0){
				printf("f ");
				_ssaSim->changeRuntimeParameters(state, leap);
				Dsum = _ssaSim->ReactionTauLeap(leap, state, dt);
				_SSAEvents += leap;
				sim->changeRuntimeParameters(state, leap);
				nChangeRuntimeParams++;
			}else{
				_rtlEvents += leap;
			}
			
			// update time and state
			if(hist->jot(state, t, dt)){// write to file
				_missedHistWrites++;// for debugging, this shouldn't be called often at all
				//printf("missed write over %f-%f\n", t, t + dt);
			}	
			
			t += dt;
			
			_totIterations++;
		}
		printf("Updated Runtime parameters %d times.\n", nChangeRuntimeParams);
		printRunStats();
	}
	
void FullSimulator::printRunStats(){
	//fprintf(stderr, "Iterations:%d, %sEvents: %d, SSAevents: %d\n", _totIterations, _sim->name(), _rtlEvents, _SSAEvents);
	printf("Iterations:%d, %sEvents: %d, SSAevents: %d, HistoryFaults:%d\n", _totIterations, _sim->name(), _rtlEvents, _SSAEvents, _missedHistWrites);
}

// Calculates the probability of reactions occurring over nSamples worth
//		of reaction leaps. rxnEventsPerSample is the number of steps each
//		reactionLeap(leap) will at least simulate.
void FullSimulator::calcProbReactions(ReactionTauLeaper *sim, Reactions *root, State *initialState, int leap, int nSamples, int rxnEventsPerSample){
	assert(rxnEventsPerSample % leap == 0); // ascertain that we can achieve exactly rxnEventsPerSample
	
	ReactionTally *rxnHist = new ReactionTally(root);
	double dt;
	
	for (int sample=0; sample < nSamples; sample++){
		State *state = new State(initialState);
		for (int nEvents = 0; nEvents < rxnEventsPerSample; nEvents += leap){
			//ReactionPermutation<Reaction*> *rxns = sim->ReactionTauLeap(leap, state, dt);
			//if (!rxns) continue;
			//rxnHist->tally(rxns);
			assert(false); // if we need this func in future will need to somehow get which rxns fired.
			//delete rxns;
		}
		delete state;
	}
	
	rxnHist->printResults();
	delete rxnHist;
	delete sim;
}

// Applies all of the reactions in the ReactionPermutation to state
void FullSimulator::updateState(State *state, ReactionPermutation<Reaction*> *rxns){
	while (rxns->hasNext()){
		Reaction *r = rxns->next();
		r->fire(state);
	}
}

	
