/*
 *  gpu_ssa.h
 *  gpu_hierleap
 *
 *  Created by David Orendorff on 4/1/11.
 *  Copyright 2011 UC Irvine. All rights reserved.
 *
 */

#include "gpuhier_tau_leap.h"
#include "../utils/gpu_interface.h"

#ifndef GPU_SSA_H_
#define GPU_SSA_H_

class GpuSSA : public GPUHierLeapSim {
public:
	GpuSSA(State *s, Reactions *root, GpuInterface *gpuInterfaceHierLeap, GpuRandgen *gpuRandgen) 
			: GPUHierLeapSim(s, root, gpuInterfaceHierLeap, gpuRandgen) {
	
		_gpuSSA = new GpuInterface(gpuInterfaceHierLeap->getCmdQueue(), "/Users/dweo/research/gpusvm/ssa.cl");
		_ssa = _gpuSSA->extractKernel("ssa");
		
		_randIdx_mem = clCreateBuffer(gpuInterfaceHierLeap->getContext(), CL_MEM_READ_WRITE, sizeof(cl_int), NULL, NULL);
		
		// Note: this is duplicated code from gpuhier_one_kernel
		int preferMultipleOf = 32;
		int bestIfUnlimitedWgSize = _nCells <= preferMultipleOf ? _nCells : floor((float)_nCells / preferMultipleOf) * preferMultipleOf;
		_localWorkSize = min(gpuInterfaceHierLeap->maxWorkgroupSize(), bestIfUnlimitedWgSize);
		printf("GpuHierLeap workgroup size %d\n", _localWorkSize);
				
		_setKernelArgs();
	}
	
	virtual ~GpuSSA() {
		printf("In ~GpuSSA destructor.\n");
		cl_int err = clReleaseMemObject(_randIdx_mem);
		err |= clReleaseKernel(_ssa);
		assert(err == CL_SUCCESS);
		delete _gpuSSA;
	}
	
	virtual void run(State *s, double simTime, int initialLeap, Historian *hist) {
		int leap=initialLeap;
		
		double time = 0.0; 
		long totalRunningTime = 0.0;
		
		setGpuState(s);
		
		hist->jot(s, time, 0);
		int err = CL_SUCCESS;
		int iter = -1;
		while (time < simTime){
			iter++;
			cl_int rnOffset = _prepareRandNums(leap*10);
			setRandIdx(rnOffset);
			
			// Set run duration argument.
			float timeToRunSim = hist->nextUpdate() - time;
			setGpuTime(timeToRunSim);
			
			// start kernel.
			//cl_event event;
//			err |= clEnqueueNDRangeKernel(_cmd_queue, _ssa, 1, NULL, 
//										  &_localWorkSize, &_localWorkSize, 0, NULL, &event);	

			err |= clEnqueueNDRangeKernel(_cmd_queue, _ssa, 1, NULL, 
										  &_localWorkSize, &_localWorkSize, 0, NULL, NULL);	
			assert(err == CL_SUCCESS);
			err |= clFinish(_cmd_queue);
			
			// long runtime = kernelRunTime(event);
			//totalRunningTime += runtime;
			//err |= clReleaseEvent(event);	
			
			assert(err == CL_SUCCESS);
			
			updateStateFromGpu(s); 
			
			double dt = getSimulatedTime();
			
			bool writes = hist->jot(s,time, 0); // This is to bring jot history up to present: kinda hacky!
			writes |= hist->jot(s,time, dt); // to be thorough
			//assert(writes); // todo: figure out if this is needed.
			
			time += dt;
			
			_currentRnIndex = getRandIdx();
		}
	}
	
private:
	void _setKernelArgs() {
		int argIdx = 0;
		
		cl_int err = clSetKernelArg(_ssa, argIdx++, sizeof(cl_mem), &_state_mem);
		err |= clSetKernelArg(_ssa, argIdx++, sizeof(int), &_nSpecies);
		err |= clSetKernelArg(_ssa, argIdx++, sizeof(cl_mem), &_time_mem);
		err |= clSetKernelArg(_ssa, argIdx++, sizeof(cl_mem), &_rates_mem);
		err |= clSetKernelArg(_ssa, argIdx++, sizeof(cl_mem), &_inpSpec_mem);
		err |= clSetKernelArg(_ssa, argIdx++, sizeof(cl_mem), &_outSpec_mem);
		assert(err == CL_SUCCESS);

		err |= clSetKernelArg(_ssa, argIdx++, sizeof(cl_mem), &_randomUint_mem); 
		err |= clSetKernelArg(_ssa, argIdx++, sizeof(cl_mem), &_randIdx_mem); 
		err |= clSetKernelArg(_ssa, argIdx++, sizeof(cl_int), &_totalRandNums); 
		err |= clSetKernelArg(_ssa, argIdx++, sizeof(cl_int), &_nBaseRxn);
		assert(err == CL_SUCCESS);
		
		err |= clSetKernelArg(_ssa, argIdx++, sizeof(cl_int) * _nSpecies, NULL); // state
		err |= clSetKernelArg(_ssa, argIdx++, sizeof(cl_float) * _nBaseRxn, NULL); // D
		assert(err == CL_SUCCESS);
		err |= clSetKernelArg(_ssa, argIdx++, sizeof(cl_float) * _nBaseRxn, NULL); // rates
		err |= clSetKernelArg(_ssa, argIdx++, sizeof(cl_int2) * _nBaseRxn, NULL); // inpSpec
		err |= clSetKernelArg(_ssa, argIdx++, sizeof(cl_int2) * _nBaseRxn, NULL); // outSpec
		err |= clSetKernelArg(_ssa, argIdx++, sizeof(cl_float) * _localWorkSize, NULL); // DrxnSubgroups
		
		assert(err == CL_SUCCESS);
	}
	
	// Returns the value stored in global memory for totalSimtime
	float getSimulatedTime() {
		float t;
		cl_int err = clEnqueueReadBuffer(_cmd_queue, _time_mem, CL_TRUE, 0, sizeof(cl_float),
										 (void*)&t, 0, NULL, NULL);
		err |= clFlush(_cmd_queue);
		assert(err == CL_SUCCESS);
		return t;
	}
	
	// Stores this value in _randIdx_mem.
	void setRandIdx(cl_int rndIdx) {
		assert(rndIdx >= 0);
		cl_int err = clEnqueueWriteBuffer(_cmd_queue, _randIdx_mem, CL_TRUE, 0, sizeof(cl_int),
										  (void*)&(rndIdx), 0, NULL, NULL);
		clFinish(_cmd_queue);
		assert(err == CL_SUCCESS);
	}
	
	// Returns the value stored in _randIdx_mem
	int getRandIdx() {
		int i;
		cl_int err = clEnqueueReadBuffer(_cmd_queue, _randIdx_mem, CL_TRUE, 0, sizeof(cl_int),
										 (void*)&i, 0, NULL, NULL);
		err |= clFlush(_cmd_queue);
		assert(err == CL_SUCCESS);
		return i;
	}
	
	
	size_t _localWorkSize;
	
	cl_kernel _ssa;
	
	cl_mem _randIdx_mem;
	
	GpuInterface *_gpuSSA;
};


#endif