/*
 *  gpuhier_tau_leap.cpp
 *  gpu_hierleap
 *
 *  Created by David Orendorff on 6/3/10.
 *  Copyright 2010 UC Irvine. All rights reserved.
 *
 */

#include "gpuhier_tau_leap.h"

#include <boost/foreach.hpp>

#include "hierleap_calculations.h"
#include "../state/state.h"
#include "../reactions/reaction.h"
#include "../reactions/reactions.h"
#include "../reactions/reaction_set.h"
#include "../utils/gpu_randgen.h"
#include "../utils/gpu_interface.h"
#include "../reactions/propensity_bounds.h"
#include "tau_leap_ssa.h"

using namespace hierleap;

// recall 512 RN's will be drawn on every iteration. So this is about 100 iterations of RN generator.
#define RANDNUMS_PER_GENSTEP 5120000 // see MAX_BUFFER_SIZE under device.
//#define CL_CREATEMEM_FLAGS (CL_MEM_ALLOC_HOST_PTR)
#define CL_CREATEMEM_FLAGS (NULL)

GPUHierLeapSim::GPUHierLeapSim(State *s, Reactions *root, GpuInterface *gpuInterface, GpuRandgen *gpuRandgen) :
			_totalRandNums(RANDNUMS_PER_GENSTEP), _gpuRandgen(gpuRandgen), _nSpecies(s->X().size()), _currentRnIndex(RANDNUMS_PER_GENSTEP),
			_nCells(root->getDependents().size()), _nBaseRxn(root->descendantReactionList().size()),
			_state_mem(NULL), _cellStartIdx_mem(NULL), _inpSpec_mem(NULL), _rates_mem(NULL), _D_mem(NULL), 
			_nHighDeltaCell(0), _nLowDeltaCell(0), 

			// Upper bounding memory.
			_highDeltaS_mem(NULL), _highDeltaCell_mem(NULL), _highBoundS_mem(NULL), _highCellBoundS_mem(NULL),
			_DboundHigh_mem(NULL), _DboundCellHigh_mem(NULL), 
			// Lower bounding state/propensities with global/local constraints.
			_lowDeltaS_mem(NULL), _lowDeltaCell_mem(NULL), _lowBoundS_mem(NULL), _lowCellBoundS_mem(NULL),
			_DboundLow_mem(NULL), _DboundCellLow_mem(NULL), 

			_rxnEvents_mem(NULL), _Pcoarse_mem(NULL), _Paccept_mem(NULL), _time_mem(NULL), _taus_mem(NULL), _outSpec_mem(NULL), 
			_orderedRxnEvents_mem(NULL), _randomUint_mem(NULL)
{
	
	// 512 Uints get generated per cycle of CA. So this is like 51200/512=100 cycles of CA.
	_cmd_queue = gpuInterface->getCmdQueue();
	_createKernels(gpuInterface);
	
	_setupInputGPUMem(gpuInterface->getContext(), s, root);
	
	int nDeltas = root->asReactions()->descendantReactionList().size();
	_inputConstKernelArgs(s->speciesSize(), root->getDependents().size(), nDeltas, nDeltas, root->getOrderedProgeny().size()); 
	
	//_ssaSim = new TauLeapSSA(root);
	_calculationDoublechecker = new HierleapCalculations(_nSpecies, _nCells);
	
	rxnEventsHistory = vector<int>(_nCells + _nBaseRxn + 1); // zero's initially
}

void GPUHierLeapSim::setGpuTime(float time){
	cl_int err = clEnqueueWriteBuffer(_cmd_queue, _time_mem, CL_TRUE, 0, sizeof(cl_float),
							   (void*)&time, 0, NULL, NULL);
	err |= clFinish(_cmd_queue);
	assert(err==CL_SUCCESS);
}

void GPUHierLeapSim::startComputePropensities(cl_mem state_mem, cl_mem result_mem, cl_event *e){
	int err = clSetKernelArg(_computePropensities,  0, sizeof(cl_mem), &state_mem);
	err |= clSetKernelArg(_computePropensities,  6, sizeof(cl_mem), &result_mem);
	err |= clEnqueueNDRangeKernel(_cmd_queue, _computePropensities, 1, NULL, 
								  &_nCells, &_nCells, 0, NULL, e);
	assert(err == CL_SUCCESS);
	
}

void GPUHierLeapSim::startHighBoundState(int leap, cl_event *e){
	
	int err = clSetKernelArg(_boundState,  1, sizeof(cl_mem), &_highDeltaS_mem);
	err |= clSetKernelArg(_boundState,  2, sizeof(cl_int), &leap);
	err |= clSetKernelArg(_boundState,  3, sizeof(cl_mem), &_highBoundS_mem);
	err |= clEnqueueNDRangeKernel(_cmd_queue, _boundState, 1, NULL, 
								  &_nSpecies, &_nSpecies, 0, NULL, e);
	assert(err == CL_SUCCESS);
	
}

void GPUHierLeapSim::startHighBoundCellState(cl_event *e){
	
	int err = clSetKernelArg(_boundCellState, 5, sizeof(cl_mem), &_highDeltaCell_mem);
	err |= clSetKernelArg(_boundCellState, 6, sizeof(cl_int), &_nHighDeltaCell);
	err |= clSetKernelArg(_boundCellState, 7, sizeof(cl_mem), &_highBoundS_mem);
	err |= clSetKernelArg(_boundCellState, 8, sizeof(cl_mem), &_highCellBoundS_mem);
	err |= clEnqueueNDRangeKernel(_cmd_queue, _boundCellState, 1, NULL, 
									 &_nSpecies, &_nSpecies, 0, NULL, e);
	assert(err == CL_SUCCESS);
	
}

void GPUHierLeapSim::startLowBoundState(int leap, cl_event *e){
	
	int err = clSetKernelArg(_boundState,  1, sizeof(cl_mem), &_lowDeltaS_mem);
	err |= clSetKernelArg(_boundState,  2, sizeof(cl_int), &leap);
	err |= clSetKernelArg(_boundState,  3, sizeof(cl_mem), &_lowBoundS_mem);
	err |= clEnqueueNDRangeKernel(_cmd_queue, _boundState, 1, NULL, 
								  &_nSpecies, &_nSpecies, 0, NULL, e);
	assert(err == CL_SUCCESS);
	
}

void GPUHierLeapSim::startLowBoundCellState(cl_event *e){
	
	int err = clSetKernelArg(_boundCellState, 5, sizeof(cl_mem), &_lowDeltaCell_mem);
	err |= clSetKernelArg(_boundCellState, 6, sizeof(cl_int), &_nLowDeltaCell);
	err |= clSetKernelArg(_boundCellState, 7, sizeof(cl_mem), &_lowBoundS_mem);
	err |= clSetKernelArg(_boundCellState, 8, sizeof(cl_mem), &_lowCellBoundS_mem);
	err |= clEnqueueNDRangeKernel(_cmd_queue, _boundCellState, 1, NULL, 
								  &_nSpecies, &_nSpecies, 0, NULL, e);
	assert(err == CL_SUCCESS);
	
}

void GPUHierLeapSim::startComputePCoarse(int leap, cl_event *e){
	
	cl_int err = clSetKernelArg(_computePCoarse, 6, sizeof(cl_float)*leap, NULL);
	err |= clSetKernelArg(_computePCoarse, 8, sizeof(cl_int), &leap);
	err |= clEnqueueNDRangeKernel(_cmd_queue, _computePCoarse, 1, NULL, 
									 &_nCells, &_nCells, 0, NULL, e);
	assert(err == CL_SUCCESS);
	
}

void GPUHierLeapSim::sampleCoarseEvents(int leap, cl_event *e) {
	
	
	// Number of trials.
	cl_int err = clSetKernelArg(_sampleCoarseEvents,  4, sizeof(cl_int), &leap);
	
	// Set up random numbers. For now use a RN start index
	cl_int rnOffset = _prepareRandNums(leap);
	err = clSetKernelArg(_sampleCoarseEvents,  6, sizeof(cl_int), &rnOffset);
	assert(err == CL_SUCCESS);
	
	// Run kernel.
	size_t localSize = leap > 8 ? 32 : 1; // Ad-hoc heuristic to determine group size.
	err |= clEnqueueNDRangeKernel(_cmd_queue, _sampleCoarseEvents, 1, NULL, 
								  &localSize, &localSize, 0, NULL, e);		
	assert(err == CL_SUCCESS);
	
}

void GPUHierLeapSim::sampleFineEvents(int leap, cl_event *e) {
	
	// Set the random start position
	int rnOffset = _prepareRandNums(leap);
	int err = clSetKernelArg(_fineSample,  6, sizeof(cl_int), &rnOffset);
	assert(err == CL_SUCCESS);
	
	// Run kernel
	size_t localSize = leap > 8 ? 1 : 32; // Heuristic to determine group size.
	err |= clEnqueueNDRangeKernel(_cmd_queue, _fineSample, 1, NULL, 
								  &localSize, &localSize, 0, NULL, e);	
	assert(err == CL_SUCCESS);
	
}

// Sample leap taus from the simplex
// ie: |leap| exponentially distribution RV's with lambda=rate[0]. rate is dBoundLow
void GPUHierLeapSim::startTausSample(int leap, cl_event *e){
	
	assert(leap > 0);
	cl_int err = clSetKernelArg(_tausSample,  0, sizeof(cl_int), &leap);
	int roff = _prepareRandNums(leap);
	err |= clSetKernelArg(_tausSample,  4, sizeof(cl_int), &roff);
	assert(err == CL_SUCCESS);
	
	// Run kernel
	size_t localSize = leap > 8 ? 1 : 32; // Heuristic to determine group size.
	err |= clEnqueueNDRangeKernel(_cmd_queue, _tausSample, 1, NULL, 
								  &localSize, &localSize, 0, NULL, e);	
	assert(err == CL_SUCCESS);
	
}

void GPUHierLeapSim::startReactionPermutation(int leap, cl_event *e){
	
	cl_int err = clSetKernelArg(_reactionPermutation, 2, sizeof(cl_int) * leap, NULL);
	int roff = _prepareRandNums(leap);
	err |= clSetKernelArg(_reactionPermutation,  5, sizeof(cl_int), &roff);
	assert(err==CL_SUCCESS);
	
	// Run kernel
	size_t localSize = 4; // Heuristic to determine group size. todo
	err |= clEnqueueNDRangeKernel(_cmd_queue, _reactionPermutation, 1, NULL, 
								  &localSize, &localSize, 0, NULL, e);	
	assert(err == CL_SUCCESS);
	
}

void GPUHierLeapSim::startComputeFinalAccept(int leap, cl_event *e) {

	// local _orderedRxnEvents_mem
	cl_int err = clSetKernelArg(_computeFinalAccept,  8, sizeof(cl_int) * leap, NULL);
	
	// nEvents
	err |= clSetKernelArg(_computeFinalAccept,  9, sizeof(cl_int), &leap);
	
	// rand uniform normal float.
	float randFloat = urand();
	err |= clSetKernelArg(_computeFinalAccept,  14, sizeof(cl_float), &randFloat);
	
	// Ddeltas mem
	err |= clSetKernelArg(_computeFinalAccept,  16, sizeof(cl_float)*leap, NULL);
	
	
	size_t localSize = _nBaseRxn; // per fun spec.
	err |= clEnqueueNDRangeKernel(_cmd_queue, _computeFinalAccept, 1, NULL, 
								  &localSize, &localSize, 0, NULL, e);	
	assert(err==CL_SUCCESS);
	
}

// ASSUME: list of reactions from root->allReactions() are "topologically consistent"
// ASSUME: descendantReactionList() results are in order!
// Setup *most* memory for GPU, excluding mem blocks of var size (eg related to L).
void GPUHierLeapSim::_setupInputGPUMem(cl_context context, State *s, Reactions *root){
	vector<ReactionSet*> & allRxns = root->getOrderedProgeny();
	cl_int err = CL_SUCCESS;
	int nCells = root->getDependents().size();
	int nBaseRxns = root->descendantReactionList().size();
	int totRxnSets = 1 + nCells + nBaseRxns; // +1 for root
	int nStates = s->speciesSize();
	
	assert(totRxnSets == allRxns.size());
	
	size_t bufferSize;
	
	// Input State
	bufferSize = s->speciesSize() * sizeof(cl_int);
	_state_mem = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	err = clEnqueueWriteBuffer(_cmd_queue, _state_mem, CL_TRUE, 0, bufferSize,
								(void*)&(s->X()[0]), 0, NULL, NULL);
	
	// Input _cellStartIdx, (a list of where children start in the list of rxns), length nCells+1
	vector<int> cellStartIdx(nCells+1, -1);
	BOOST_FOREACH(ReactionSet *r, root->getDependents()){
		int cellId = r->getUniqueId();
		int firstChildId = r->getDependents()[0]->getUniqueId();
		cellStartIdx[cellId-1] = firstChildId; // remember to disregard root, so start -1
	}
	cellStartIdx[nCells] = totRxnSets;
	assert(!contains(cellStartIdx, -1));
	bufferSize = cellStartIdx.size() * sizeof(cl_int);
	_cellStartIdx_mem = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	err |= clEnqueueWriteBuffer(_cmd_queue, _cellStartIdx_mem, CL_TRUE, 0, bufferSize,
								(void*)&cellStartIdx[0], 0, NULL, NULL);
	
	// Input inpSpec/outSpec (length of BaseReactions), list of 2-tuples
	cl_int2 * inpSpec = (cl_int2*) new int [nBaseRxns * 2];
	cl_int2 * outSpec = (cl_int2*) new int [nBaseRxns * 2];

	vector<Reaction*> & drl = root->descendantReactionList(); 
	for (int i=0; i < drl.size(); i++){
		Reaction *r = drl[i];
		int nInp = r->numReactants();
		int nOut = r->numProducts();
		
		assert(nInp < 3 && nOut < 3);
		
		inpSpec[i][0] = nInp > 0 ? r->getReactantId(0) : -1;
		inpSpec[i][1] = nInp > 1 ? r->getReactantId(1) : -1;
		outSpec[i][0] = nOut > 0 ? r->getProductId(0) : -1;
		outSpec[i][1] = nOut > 1 ? r->getProductId(1) : -1;
		
		// if there was double stoichiometry. This method may have problems if we allow more than 2 reactants/products.
		if (nInp > 0 && r->getReactantStoic(0) > 1){
			assert(inpSpec[i][1]==-1);
			inpSpec[i][1] = inpSpec[i][0];
		}
		if (nOut > 0 && r->getProductStoic(0) > 1){
			assert(outSpec[i][1]==-1);
			outSpec[i][1] = outSpec[i][0];
		}
	}
	
	bufferSize = nBaseRxns * sizeof(cl_int2);
	assert(bufferSize == sizeof(int) * 2 * nBaseRxns);
	_inpSpec_mem = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	_outSpec_mem = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	err |= clEnqueueWriteBuffer(_cmd_queue, _inpSpec_mem, CL_TRUE, 0, bufferSize,
								(void*)&inpSpec[0], 0, NULL, NULL);
	err |= clEnqueueWriteBuffer(_cmd_queue, _outSpec_mem, CL_TRUE, 0, bufferSize,
								(void*)&outSpec[0], 0, NULL, NULL);
	
	// Input rates
	vector<cl_float> rates(nBaseRxns,-1);
	for (int i=0; i < nBaseRxns; i++){
		rates[i] = drl[i]->rate();
	}
	bufferSize = nBaseRxns * sizeof(cl_float);
	_rates_mem = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	err |= clEnqueueWriteBuffer(_cmd_queue, _rates_mem, CL_TRUE, 0, bufferSize,
								(void*)&rates[0], 0, NULL, NULL);

	// Input D_mem. Don't write to it yet, just allocate.
	bufferSize = totRxnSets * sizeof(cl_float);
	_D_mem = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	
	// Input _highDeltaS_mem (len state, each entry is the max a species may change in 1 event)
	// use propensity bounds for this
	PropensityBounds *pb = new PropensityBoundsNoCells(root, s->speciesSize(), false);
	vector<int> & highDeltaS = pb->highSpeciesDelta();
	bufferSize = sizeof(cl_int) * nStates;
	_highDeltaS_mem = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	err |= clEnqueueWriteBuffer(_cmd_queue, _highDeltaS_mem, CL_TRUE, 0, bufferSize,
								(void*)&highDeltaS[0], 0, NULL, NULL);
	
	// Input highBoundS (where results for state bounding are stored)
	bufferSize = nStates * sizeof(cl_int);
	_highBoundS_mem = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	
	// Input _highCellBoundS_mem (results stored here)
	bufferSize = nStates * sizeof(cl_int);
	_highCellBoundS_mem = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	
	// Input time memory
	bufferSize = sizeof(cl_float);
	_time_mem = clCreateBuffer(context, CL_MEM_READ_WRITE, bufferSize, NULL, &err);  // todo: this may need to change.
	assert(err == CL_SUCCESS);
	//_time_mem = clCreateBuffer(context, CL_MEM_USE_HOST_PTR, bufferSize, &_time, NULL);
	
	// High cell Deltas
	vector<DeltaX> deltas = calculateDeltaX(root, nStates, true);
	_nHighDeltaCell = deltas.size();
	assert(_nHighDeltaCell >= nCells);
	bufferSize = sizeof(DeltaX) * _nHighDeltaCell;
	_highDeltaCell_mem = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	err |= clEnqueueWriteBuffer(_cmd_queue, _highDeltaCell_mem, CL_TRUE, 0, bufferSize,
								(void*)&deltas[0], 0, NULL, NULL);
	
	//_lowDeltaS_mem
	vector<unsigned int> & lowDeltaS = pb->lowSpeciesDelta(); 
	vector<int> negLowDeltas(nStates); // negative of lowDeltaS
	for (int i = 0; i < nStates; i++)
		negLowDeltas[i] = -lowDeltaS[i];
	
	bufferSize = sizeof(cl_uint) * nStates;
	_lowDeltaS_mem = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	err |= clEnqueueWriteBuffer(_cmd_queue, _lowDeltaS_mem, CL_TRUE, 0, bufferSize,
								(void*)&negLowDeltas[0], 0, NULL, NULL);
	
	// Input lowBoundS (where results for state bounding are stored)
	bufferSize = nStates * sizeof(cl_int);
	_lowBoundS_mem = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	
	// Input _lowCellBoundS_mem (results stored here)
	bufferSize = nStates * sizeof(cl_int);
	_lowCellBoundS_mem = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	
	// Low cell Deltas
	deltas = calculateDeltaX(root, nStates, false);
	_nLowDeltaCell = deltas.size();
	bufferSize = sizeof(DeltaX) * _nLowDeltaCell;
	_lowDeltaCell_mem = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	err |= clEnqueueWriteBuffer(_cmd_queue, _lowDeltaCell_mem, CL_TRUE, 0, bufferSize,
								(void*)&deltas[0], 0, NULL, NULL);
	
	// _rxnEvents_mem
	//  this should be of size num rxnSets
	bufferSize = sizeof(cl_int) * totRxnSets;
	_rxnEvents_mem = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	
	// _DboundCellHigh_mem
	bufferSize = sizeof(cl_float) * totRxnSets;
	_DboundCellHigh_mem = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	
	// _DboundHigh_mem
	bufferSize = sizeof(cl_float) * totRxnSets;
	_DboundHigh_mem = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	
	// _DboundLow_mem
	bufferSize = sizeof(cl_float) * totRxnSets;
	_DboundLow_mem = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	
	// _DBoundCellLow_mem
	bufferSize = sizeof(cl_float) * totRxnSets;
	_DboundCellLow_mem = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	
	// Input Pcoarse memory
	bufferSize = sizeof(cl_float);
	_Pcoarse_mem = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	
	// Paccept memory (Pfine * Pcoarse)
	bufferSize = sizeof(cl_float);
	_Paccept_mem = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	
	// _taus_mem
	bufferSize = sizeof(cl_float) * MAX_LEAP;
	_taus_mem =  clCreateBuffer(context, CL_MEM_READ_WRITE | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	
	// _orderedRxnEvents_mem
	bufferSize = sizeof(cl_float) * MAX_LEAP;
	_orderedRxnEvents_mem = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	
	// _randomUint_mem
	assert(_totalRandNums > 0);
	bufferSize = sizeof(cl_uint) * _totalRandNums;
	_randomUint_mem = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_CREATEMEM_FLAGS, bufferSize, NULL, &err);
	assert(err == CL_SUCCESS);
	
	delete []inpSpec;
	delete []outSpec;
}

// These arguments to the kernels do not change throughout simulation
void GPUHierLeapSim::_inputConstKernelArgs(int nStates, int nCells, int nDeltas, int nRates, int totRxnSets){
	cl_int startIdxFirstRxn = nCells+1;
	cl_int numRxn = totRxnSets - nCells - 1;
	
	// computePropensities args...
	int err = clSetKernelArg(_computePropensities,  1, sizeof(cl_int), &nStates);
	err |= clSetKernelArg(_computePropensities,  2, sizeof(cl_mem), &_cellStartIdx_mem);
	err |= clSetKernelArg(_computePropensities,  3, sizeof(cl_mem), &_inpSpec_mem);
	err |= clSetKernelArg(_computePropensities,  4, sizeof(cl_mem), &_rates_mem);
	err |= clSetKernelArg(_computePropensities,  5, sizeof(cl_int)*nStates, NULL);
	//err |= clSetKernelArg(_computePropensities,  6, sizeof(cl_mem), &_D_mem); this changes based on inp
	err |= clSetKernelArg(_computePropensities,  7, sizeof(cl_int)*totRxnSets, NULL);
	
	// _boundState args
	err |= clSetKernelArg(_boundState,  0, sizeof(cl_mem), &_state_mem);
	// deltaS (depends on low/high computation)
	// leap
	// where answer is stored
	
	// boundCellState args
	int nRxnSets = nCells + 1;
	err |= clSetKernelArg(_boundCellState,  0, sizeof(cl_mem), &_state_mem);
	err |= clSetKernelArg(_boundCellState,  1, sizeof(cl_int)*nStates, NULL);
	err |= clSetKernelArg(_boundCellState,  2, sizeof(cl_mem), &_rxnEvents_mem);
	err |= clSetKernelArg(_boundCellState,  3, sizeof(cl_int)*nRxnSets, NULL);
	err |= clSetKernelArg(_boundCellState,  4, sizeof(cl_int), &nRxnSets);
	// DeltaCell
	// nDeltas 
	// BoundS 
	// answer storage location
	
	// computePCoarse args
	err |= clSetKernelArg(_computePCoarse,  0, sizeof(cl_mem), &_DboundHigh_mem);
	err |= clSetKernelArg(_computePCoarse,  1, sizeof(cl_mem), &_DboundCellHigh_mem);
	err |= clSetKernelArg(_computePCoarse,  2, sizeof(cl_mem), &_DboundLow_mem);
	err |= clSetKernelArg(_computePCoarse,  3, sizeof(cl_mem), &_DboundCellLow_mem);
	err |= clSetKernelArg(_computePCoarse,  4, sizeof(cl_mem), &_rxnEvents_mem);
	err |= clSetKernelArg(_computePCoarse,  5, sizeof(cl_mem), &_taus_mem);
	// (6) size depends on L
	err |= clSetKernelArg(_computePCoarse,  7, sizeof(cl_mem), &_Pcoarse_mem);
	
	
	assert(err == CL_SUCCESS);
	
	// Args to SampleCoarseEvents
	err |= clSetKernelArg(_sampleCoarseEvents, 0, sizeof(cl_mem), &_DboundHigh_mem); 
	err |= clSetKernelArg(_sampleCoarseEvents, 1, sizeof(cl_int), &nCells);
	err |= clSetKernelArg(_sampleCoarseEvents, 2, sizeof(cl_mem), &_rxnEvents_mem);
	err |= clSetKernelArg(_sampleCoarseEvents, 3, sizeof(cl_int)*(nCells+1), NULL);
	// (4) is leap
	err |= clSetKernelArg(_sampleCoarseEvents, 5, sizeof(cl_mem), &_randomUint_mem);
	// (6) is randNumOffset
	
	
	// Args to fineSample
	err |= clSetKernelArg(_fineSample, 0, sizeof(cl_mem), &_cellStartIdx_mem);
	err |= clSetKernelArg(_fineSample, 1, sizeof(cl_mem), &_DboundCellHigh_mem);
	err |= clSetKernelArg(_fineSample, 2, sizeof(cl_float)*totRxnSets, NULL);
	err |= clSetKernelArg(_fineSample, 3, sizeof(cl_mem), &_rxnEvents_mem);
	err |= clSetKernelArg(_fineSample, 4, sizeof(cl_int)*totRxnSets, NULL);
	err |= clSetKernelArg(_fineSample, 5, sizeof(cl_mem), &_randomUint_mem);
	assert(err == CL_SUCCESS);
	
	// Args tausSample
	// (0) leap
	err |= clSetKernelArg(_tausSample, 1, sizeof(cl_mem), &_DboundLow_mem);
	err |= clSetKernelArg(_tausSample, 2, sizeof(cl_mem), &_taus_mem);
	err |= clSetKernelArg(_tausSample, 3, sizeof(cl_mem), &_randomUint_mem);
	
	// Args reactionPermutation
	err |= clSetKernelArg(_reactionPermutation, 0, sizeof(cl_mem), &_rxnEvents_mem);
	err |= clSetKernelArg(_reactionPermutation, 1, sizeof(cl_mem), &_cellStartIdx_mem);
	// (2) of size leap
	err |= clSetKernelArg(_reactionPermutation, 3, sizeof(cl_mem), &_orderedRxnEvents_mem);
	err |= clSetKernelArg(_reactionPermutation, 4, sizeof(cl_mem), &_randomUint_mem);
	assert(err == CL_SUCCESS);
	

	// computeFinalAccept args
	err |= clSetKernelArg(_computeFinalAccept,  0, sizeof(cl_mem), &_state_mem);
	err |= clSetKernelArg(_computeFinalAccept,  1, sizeof(cl_int)*nStates, NULL);
	err |= clSetKernelArg(_computeFinalAccept,  2, sizeof(cl_int), &nStates);
	err |= clSetKernelArg(_computeFinalAccept,  3, sizeof(cl_mem), &_time_mem);
	err |= clSetKernelArg(_computeFinalAccept,  4, sizeof(cl_mem), &_rates_mem);
	err |= clSetKernelArg(_computeFinalAccept,  5, sizeof(cl_mem), &_DboundCellHigh_mem);
	err |= clSetKernelArg(_computeFinalAccept,  6, sizeof(cl_mem), &_taus_mem);
	err |= clSetKernelArg(_computeFinalAccept, 7, sizeof(cl_mem), &_orderedRxnEvents_mem); 
	// local _orderedRxnEvents_mem
	// nEvents
	err |= clSetKernelArg(_computeFinalAccept, 10, sizeof(cl_mem), &_D_mem);
	err |= clSetKernelArg(_computeFinalAccept, 11, sizeof(cl_mem), &_DboundCellLow_mem);
	err |= clSetKernelArg(_computeFinalAccept,  12, sizeof(cl_mem), &_inpSpec_mem);
	err |= clSetKernelArg(_computeFinalAccept,  13, sizeof(cl_mem), &_outSpec_mem);
	// urand
	err |= clSetKernelArg(_computeFinalAccept,  15, sizeof(cl_mem), &_Pcoarse_mem);
	cl_int rxnOffset = _nCells + 1;
	err |= clSetKernelArg(_computeFinalAccept,  17, sizeof(cl_int), &rxnOffset);
	err |= clSetKernelArg(_computeFinalAccept,  18, sizeof(cl_mem), &_Paccept_mem);
	
	assert(err == CL_SUCCESS);
}

void GPUHierLeapSim::_createKernels(GpuInterface *gpuInterface) {
	_computePropensities = gpuInterface->extractKernel("computePropensities");	
	_boundState = gpuInterface->extractKernel("boundState");
	_boundCellState = gpuInterface->extractKernel("boundCellState");
	_sampleCoarseEvents = gpuInterface->extractKernel("sampleCoarseEvents");
	_computePCoarse = gpuInterface->extractKernel("computePCoarse");
	_fineSample = gpuInterface->extractKernel("fineSample");
	_tausSample = gpuInterface->extractKernel("tausSample");
	_reactionPermutation = gpuInterface->extractKernel("reactionPermutation");
	_computeFinalAccept = gpuInterface->extractKernel("computeFinalAccept");
}

// post: _prepareRandNums returns the offset that is followed by nRandNumsNeeded RNs which
//		are promised to not be used by any other process.
int GPUHierLeapSim::_prepareRandNums(int nRandNumsNeeded) {
	assert(nRandNumsNeeded < _totalRandNums);
	if (_currentRnIndex + nRandNumsNeeded > _totalRandNums || _currentRnIndex < 0) {
		// Re-generate random numbers.
		_gpuRandgen->generate_rands_in_gpu(_randomUint_mem, _totalRandNums);
		_currentRnIndex = nRandNumsNeeded;
		return 0;
	}
	int t = _currentRnIndex;
	_currentRnIndex += nRandNumsNeeded;
	return t;
}

// Return a vector of delta X such that for every cell/species non-zero interaction 
// there will be an entry in the returned vector for the largest of these for each (cell, species) pair.
vector<DeltaX> GPUHierLeapSim::calculateDeltaX(Reactions *root, int numStates, bool maximize) {
	vector<DeltaX> deltaXs;
	//int * deltaForCell = new int [numStates];
	vector<int> deltaForCell(numStates);
	clearVector(deltaForCell);
	
	BOOST_FOREACH(ReactionSet *cell, root->getDependents()) {
		BOOST_FOREACH(ReactionSet *rxn, cell->getDependents()) {
			Reaction *r = rxn->asReaction();
			for (int i = 0; i < r->numReactants(); i++) {
				int s = r->getReactantId(i);
				assert(s < numStates);
				deltaForCell[s] = maximize ? max(deltaForCell[s], (int) r->delta(s)) : min(deltaForCell[s], (int) r->delta(s));
			}
			for (int i = 0; i < r->numProducts(); i++) {
				int s = r->getProductId(i);
				assert(s < numStates);
				deltaForCell[s] = maximize ? max(deltaForCell[s], (int) r->delta(s)) : min(deltaForCell[s], (int) r->delta(s));
			}
		}
		
		for (int i = 0; i < numStates; i++) {
			if (deltaForCell[i] != 0) {
				DeltaX dx;
				dx.species = i;
				dx.cell = cell->asReactions()->getUniqueId(); // hopefully this is the right logic.
				dx.delta = deltaForCell[i];
				deltaXs.push_back(dx);
			}
		}
		
		clearVector(deltaForCell);
	}
	
	return deltaXs;
}



