/*
 *  gpuhier_tau_leap.h
 *  gpu_hierleap
 *
 *  Created by David Orendorff on 6/3/10.
 *  Copyright 2010 UC Irvine. All rights reserved.
 *
 */

#include <OpenCL/OpenCL.h>

#include <assert.h>
#include <math.h>
#include <vector.h>
//#include <boost/assign/insert.hpp>
#include <boost/assign/std/vector.hpp>
#include <algorithm>

#include "../utils/common.h"
#include "../utils/opencl_utils.h"
#include "historian.h"
//#include "tau_leap_ssa.h"
#include "hierleap_calculations.h"
#include "../reactions/reaction_permutation.h"

using namespace std;
using namespace boost::assign;

#ifndef _GPUHIER_TAU_LEAP_ 
#define _GPUHIER_TAU_LEAP_

// note: this may not have to be hard coded later on
#define MAX_LEAP 50 

struct DeltaX{
	unsigned short species;
	unsigned short cell;
	short delta;
	short pad;
};

class Reactions;
class State;

namespace hierleap {

}

using namespace hierleap;

class GpuInterface;
class GpuRandgen;

class GPUHierLeapSim {

public:
	GPUHierLeapSim() {}
	
	GPUHierLeapSim(State *s, Reactions *root, GpuInterface *gpuInterface, GpuRandgen *gpuRandgen);
	
	virtual ~GPUHierLeapSim() {
		printf("In ~GPUHierLeapSim destructor.\n");
		vector<cl_mem> mem;
		mem += _randomUint_mem, _state_mem, _cellStartIdx_mem, _inpSpec_mem, _rates_mem, _D_mem, _highDeltaCell_mem,
			_highDeltaS_mem, _highBoundS_mem, _lowDeltaS_mem, _lowDeltaCell_mem, _lowBoundS_mem, _lowCellBoundS_mem, _DboundLow_mem, _DboundCellLow_mem,
			 _highCellBoundS_mem, _DboundHigh_mem, _DboundCellHigh_mem, _rxnEvents_mem, _Pcoarse_mem, 
			_orderedRxnEvents_mem, _time_mem, _taus_mem, _outSpec_mem, _Paccept_mem;
		
		vector<cl_kernel> kernels;
		kernels += _computePropensities, _boundState, _boundCellState, _sampleCoarseEvents, _computePCoarse,
		_fineSample, _tausSample, _reactionPermutation, _computeFinalAccept;
		
		BOOST_FOREACH(cl_mem m, mem) {
			cl_int err = clReleaseMemObject(m);
			assert(err == CL_SUCCESS);
		}
		
		BOOST_FOREACH(cl_kernel k, kernels) {
			cl_int err = clReleaseKernel(k);
			assert(err == CL_SUCCESS);
		}
		
		delete _gpuRandgen;
		
		//delete _ssaSim; 
		delete _calculationDoublechecker;
	}
	
	
	void clearVector(vector<int> & v){
		for (int i = 0; i < v.size(); i++) v[i] = 0;
	}
	
	// Return a vector of delta X such that for every cell/species non-zero interaction 
	// there will be an entry in the returned vector for the largest of these for each (cell, species) pair.
	vector<DeltaX> calculateDeltaX(Reactions *root, int numStates, bool maximize);
	
	void startComputePropensities(cl_mem state_mem, cl_mem result_mem, cl_event *event);
	
	void startHighBoundState(int leap, cl_event *event);
	
	void startLowBoundState(int leap, cl_event *event);
	
	void startHighBoundCellState(cl_event *event);
	
	void startLowBoundCellState( cl_event *event);
	
	void startComputePCoarse(int leap, cl_event *event);
	
	void sampleCoarseEvents(int leap, cl_event *event);
	
	void sampleFineEvents(int leap, cl_event *event);
	
	// Sample leap taus from the simplex
	// ie: |leap| exponentially distribution RV's with lambda=rate[0].
	void startTausSample(int leap, cl_event *event);
	
	void startReactionPermutation(int leap, cl_event *event);
	
	void startComputeFinalAccept(int leap, cl_event *event);
	
	// Fills v with the contents of mem. Note that v needs to be of the correct
	// size and type. As a by product v will also be printed out.
	void gpuMemStdout(cl_mem mem, size_t nElems, size_t elemSize, char* name, char* printfString);
	
	//void dumpGpuMemoryToStdout();
	void dumpGpuMemoryToStdout(int leap) {
		int totRxnSets = _nCells + _nBaseRxn + 1;
		
		printf("\n-------- GPU MEMORY DUMP -----------\nDYNAMIC MEMORY...\n");
		vector<float> z(1);
		gpuMemStdout(_Paccept_mem, z,  "Paccept", "%f, ");
		
		gpuMemStdout(_Pcoarse_mem, z,  "Pcoarse", "%f, ");
		
		gpuMemStdout(_state_mem, vector<int>(_nSpecies),  "State", "%d, ");
		
		gpuMemStdout(_D_mem, vector<float>(totRxnSets),  "D", "%f, ");
		
		gpuMemStdout(_highBoundS_mem, vector<int>(_nSpecies),  "highBoundS", "%d, ");
		gpuMemStdout(_highCellBoundS_mem, vector<int>(_nSpecies),  "highCellBoundS", "%d, ");
	
		gpuMemStdout(_DboundHigh_mem, vector<float>(totRxnSets),  "_DboundHigh_mem", "%f, ");
		gpuMemStdout(_DboundCellHigh_mem, vector<float>(totRxnSets),  "_DboundCellHigh", "%f, ");
	
		
		gpuMemStdout(_lowBoundS_mem, vector<int>(_nSpecies),  "lowBoundS", "%d, ");
		gpuMemStdout(_lowCellBoundS_mem, vector<int>(_nSpecies),  "lowCellBoundS", "%d, ");
		
		gpuMemStdout(_DboundLow_mem, vector<float>(totRxnSets),  "_DboundLow_mem", "%f, ");
		gpuMemStdout(_DboundCellLow_mem, vector<float>(totRxnSets),  "_DboundCelllow", "%f, ");
		
		gpuMemStdout(_rxnEvents_mem, vector<int>(totRxnSets),  "_rxnEvents", "%d, ");
		
		gpuMemStdout(_orderedRxnEvents_mem, vector<int>(leap),  "_orderedRxnEvents", "%d, ");
		
		gpuMemStdout(_taus_mem, vector<float>(leap),  "_taus_mem", "%f, ");
		
		gpuMemStdout(_time_mem, vector<float>(1),  "_time_mem", "%f, ");

		//gpuMemStdout(_randomUint_mem, vector<float>(_totalRandNums),  "_randomUint_mem", "%u, ");
		gpuMemStdout(_randomUint_mem, vector<unsigned int>(min(_totalRandNums, 8)),  "_randomUint_mem", "%X, ");
		
		printf("\nSTATIC MEMORY...\n");
		
		gpuMemStdout(_cellStartIdx_mem, vector<int>(_nCells+1), "_cellStartIdx_mem", "%d, ");
		
		gpuMemStdout(_rates_mem, vector<float>(_nBaseRxn), "_rates_mem", "%f, ");
		
		gpuMemStdout(_inpSpec_mem, vector<int>(2*_nBaseRxn), "_inpSpec_mem", "%d, ");
		
		gpuMemStdout(_outSpec_mem, vector<int>(2*_nBaseRxn), "_outSpec_mem", "%d, ");
		
		gpuMemStdout(_highDeltaS_mem, vector<int>(_nSpecies), "_highDeltaS_mem", "%d, ");
		
		gpuMemStdout(_lowDeltaS_mem, vector<int>(_nSpecies), "_lowDeltaS_mem", "%d, ");
		
		gpuMemStdout(_highDeltaCell_mem, vector<unsigned short>(_nHighDeltaCell * sizeof(DeltaX) / sizeof(unsigned short)), "_highDeltaCell_mem", "%d, ");
		
		gpuMemStdout(_lowDeltaCell_mem, vector<short>(_nLowDeltaCell * sizeof(DeltaX) / sizeof(short)), "_lowDeltaCell_mem", "%d, ");
	}
	
	// Set to these values in GPU.
	void setGpuTime(float time);
	
	void setGpuState(State *s){
		size_t bufferSize = s->X().size() * sizeof(cl_int);
		int err = clEnqueueWriteBuffer(_cmd_queue, _state_mem, CL_TRUE, 0, bufferSize,
									   (void*)&(s->X()[0]), 0, NULL, NULL);	
		err |= clFinish(_cmd_queue);
		assert(err == CL_SUCCESS);
	}
	
	void assertProfilingEnabled(){
		//clSetCommandQueueProperty(_cmd_queue, CL_QUEUE_PROFILING_ENABLE, CL_TRUE, NULL);
		cl_command_queue_properties p;
		cl_int err = clGetCommandQueueInfo(_cmd_queue, CL_QUEUE_PROPERTIES, sizeof(p), &p, NULL);
		assert(err == CL_SUCCESS);
		assert(p & CL_QUEUE_PROFILING_ENABLE); // assert profiling is enabled in GPU.
	}
	
	void assertFeasibleBounds(int leap) {
		int totRxnSets = _nCells + _nBaseRxn + 1;
		bool pass = _calculationDoublechecker->calculateAreBoundsFeasible(leap,
															  gpuMemStdout(_highBoundS_mem, vector<int>(_nSpecies),  NULL, NULL),
															  gpuMemStdout(_highCellBoundS_mem, vector<int>(_nSpecies),  NULL, NULL),
															  gpuMemStdout(_state_mem, vector<int>(_nSpecies), NULL, NULL),
															  gpuMemStdout(_lowCellBoundS_mem, vector<int>(_nSpecies), NULL, NULL), 
															  gpuMemStdout(_lowBoundS_mem, vector<int>(_nSpecies), NULL, NULL), 
															  gpuMemStdout(_DboundHigh_mem, vector<float>(totRxnSets),  NULL, NULL), 
															  gpuMemStdout(_DboundCellHigh_mem, vector<float>(totRxnSets), NULL, NULL), 
															  gpuMemStdout(_D_mem, vector<float>(totRxnSets), NULL, NULL), 
															  gpuMemStdout(_DboundCellLow_mem, vector<float>(totRxnSets), NULL, NULL), 
															  gpuMemStdout(_DboundLow_mem, vector<float>(totRxnSets), NULL, NULL),
															  gpuMemStdout(_cellStartIdx_mem, vector<int>(_nCells+1), NULL, NULL),
															  gpuMemStdout(_inpSpec_mem, vector<int>(2*_nBaseRxn), NULL, NULL),
															  gpuMemStdout(_rates_mem, vector<float>(_nBaseRxn), NULL, NULL));		  
		assert(pass);
	}
	
	// Calculates the value of Paccept in CPU, but only for instances when accept=false
	void assertPacceptCorrect(int leap) {
		int totRxnSets = _nCells + _nBaseRxn + 1;
		vector<int> state = gpuMemStdout(_state_mem, vector<int>(_nSpecies), NULL, NULL);
		float calculatedValue = 
			_calculationDoublechecker->computeFinalAccept(state, 
														  gpuMemStdout(_rates_mem, vector<float>(_nBaseRxn), NULL, NULL), 
														  gpuMemStdout(_DboundCellHigh_mem, vector<float>(totRxnSets),  NULL, NULL), 
														  gpuMemStdout(_taus_mem, vector<float>(leap), NULL, NULL), 
														  gpuMemStdout(_orderedRxnEvents_mem, vector<int>(leap), NULL, NULL), 
														  gpuMemStdout(_DboundCellLow_mem, vector<float>(totRxnSets), NULL, NULL), 
														  gpuMemStdout(_inpSpec_mem, vector<int>(2*_nBaseRxn), NULL, NULL), 
														  gpuMemStdout(_outSpec_mem, vector<int>(2*_nBaseRxn), NULL, NULL), 
														  gpuMemStdout(_Pcoarse_mem, vector<float>(1), NULL, NULL)[0]);
		
		float gpuAnswer = gpuMemStdout(_Paccept_mem, vector<float>(1), NULL, NULL)[0];
		
		double eps = 0.000001;
		//assert(calculatedValue == gpuAnswer);
		if(fabs(calculatedValue - gpuAnswer) > eps) {
			printf("Pcoarse calculated incorrectly!\n");
			dumpGpuMemoryToStdout(leap);
			assert(false);
		}
	}
	
	void assertPcoarseCorrect(int leap) {
		int totRxnSets = _nCells + _nBaseRxn + 1;
		float calculatedValue = 
			_calculationDoublechecker->computePCoarse(gpuMemStdout(_DboundHigh_mem, vector<float>(totRxnSets), NULL, NULL),
													  gpuMemStdout(_DboundCellHigh_mem, vector<float>(totRxnSets),  NULL, NULL),
													  gpuMemStdout(_DboundLow_mem, vector<float>(totRxnSets), NULL, NULL),
													  gpuMemStdout(_DboundCellLow_mem, vector<float>(totRxnSets), NULL, NULL),
													  gpuMemStdout(_rxnEvents_mem, vector<int>(totRxnSets),  NULL, NULL),
													  gpuMemStdout(_taus_mem, vector<float>(leap),  NULL, NULL));
		
		float gpuAnswer = gpuMemStdout(_Pcoarse_mem, vector<float>(1), NULL, NULL)[0];
		double eps = 0.000001;

		if(fabs(calculatedValue - gpuAnswer) > eps) {
			printf("Pcoarse calculated incorrectly!\n");
			dumpGpuMemoryToStdout(leap);
			assert(false);
		}
	}
	
	// Samples taus from CPU and writes result to GPU.
	void cpuTausSample(int leap) {
		int totRxnSets = _nCells + _nBaseRxn + 1; 
		vector<float> taus = _calculationDoublechecker->tausSample(leap, 
										gpuMemStdout(_DboundLow_mem, vector<float>(totRxnSets), NULL, NULL)[0]);
		
		int err = clEnqueueWriteBuffer(_cmd_queue, _taus_mem, CL_TRUE, 0, sizeof(float) * leap,
									   (void*)&(taus[0]), 0, NULL, NULL);
		clFinish(_cmd_queue);
		assert(err==CL_SUCCESS);
	}
	
	void cpuCoarseSample(int leap){
		int totRxnSets = _nCells + _nBaseRxn + 1;
		vector<int> rxnEvents = 
			_calculationDoublechecker->coarseSample(gpuMemStdout(_DboundHigh_mem, vector<float>(totRxnSets), NULL, NULL),
													_nCells, leap);
		int err = clEnqueueWriteBuffer(_cmd_queue, _rxnEvents_mem, CL_TRUE, 0, sizeof(int) * rxnEvents.size(),
									   (void*)&(rxnEvents[0]), 0, NULL, NULL);
		clFinish(_cmd_queue);
		assert(err==CL_SUCCESS);
	}
	
	void cpuReactionPermutation(int leap) {
		int totRxnSets = _nCells + _nBaseRxn + 1;

		vector<int> orderedRxns = _calculationDoublechecker->reactionPermutation(gpuMemStdout(_rxnEvents_mem, vector<int>(totRxnSets), NULL, NULL),
													   gpuMemStdout(_cellStartIdx_mem, vector<int>(_nCells+1), NULL, NULL));
		assert(orderedRxns.size() == leap);
		
		// Write results to GPU memory
		int err = clEnqueueWriteBuffer(_cmd_queue, _orderedRxnEvents_mem, CL_TRUE, 0, sizeof(int) * orderedRxns.size(),
									   (void*)&(orderedRxns[0]), 0, NULL, NULL);
		err |= clFinish(_cmd_queue);
		assert(err==CL_SUCCESS);
	}
	
	void assertIsValidReactionPermutation(int leap) {
		int totRxnSets = _nCells + _nBaseRxn + 1;
		vector<int> validOrderedRxns = _calculationDoublechecker->reactionPermutation(gpuMemStdout(_rxnEvents_mem, vector<int>(totRxnSets), NULL, NULL),
																					  gpuMemStdout(_cellStartIdx_mem, vector<int>(_nCells+1), NULL, NULL));
		
		vector<int> gpuOrderedRxns = gpuMemStdout(_orderedRxnEvents_mem, vector<int>(leap), NULL, NULL);
		sort(validOrderedRxns.begin(), validOrderedRxns.end());
		sort(gpuOrderedRxns.begin(), gpuOrderedRxns.end());
		
		for (int i = 0; i < validOrderedRxns.size(); i++){
			assert(validOrderedRxns[i] == gpuOrderedRxns[i]);
		}
	}
	
	void cpuFineSmaple() {
		int totRxnSets = _nCells + _nBaseRxn + 1;
		vector<int> newRxnEvents = 
			_calculationDoublechecker->fineSample(gpuMemStdout(_cellStartIdx_mem, vector<int>(_nCells+1), NULL, NULL),
												  gpuMemStdout(_DboundCellHigh_mem, vector<float>(totRxnSets), NULL, NULL),
												  gpuMemStdout(_rxnEvents_mem, vector<int>(totRxnSets), NULL, NULL));
		int err = clEnqueueWriteBuffer(_cmd_queue, _rxnEvents_mem, CL_TRUE, 0, sizeof(int) * newRxnEvents.size(),
									   (void*)&(newRxnEvents[0]), 0, NULL, NULL);
		clFinish(_cmd_queue);
		assert(err==CL_SUCCESS);
	}
	
	void assertGpuFineSamplePlausible(int leap, vector<int> & rxnEventsBeforeCalled){
		int totRxnSets = _nCells + _nBaseRxn + 1;
		
		// Asserts cellRxns do not change during fineSample kernel call. 
		vector<int> rxnEvents = gpuMemStdout(_rxnEvents_mem, vector<int>(totRxnSets), NULL, NULL);
		vector<int> cellStardIdx = gpuMemStdout(_cellStartIdx_mem, vector<int>(_nCells+1), NULL, NULL);
		
		assert(rxnEvents[0] == leap);
		int sum = 0;
		for (int i = 1; i <= _nCells; i++){
			sum += rxnEvents[i];
			assert(rxnEventsBeforeCalled[i] == rxnEvents[i]);
		}
		
		// Assert number of rxns Per Cells is correct.
		for (int c = 0; c < _nCells; c++){
			int cellSum = 0;
			for (int i = cellStardIdx[c]; i < cellStardIdx[c+1]; i++) {
				cellSum += rxnEvents[i];
			}
			assert(cellSum == rxnEvents[c+1]);
		}
		
	}
	
	void updateRxnEventsHistory() {
		int totRxnSets = _nCells + _nBaseRxn + 1;
		vector<int> rxnEvents = gpuMemStdout(_rxnEvents_mem, vector<int>(totRxnSets), NULL, NULL);
		for (int i = 0; i < rxnEvents.size(); i++) {
			rxnEventsHistory[i] += rxnEvents[i];
		}
	}
	
	virtual void run(State *s, double simTime, int initialLeap, Historian *hist){
		//assertProfilingEnabled();

		for (int i = 0; i < rxnEventsHistory.size(); i++){
			rxnEventsHistory[i] = 0;
		}		
		
		double t = 0;
		int leap=initialLeap;
		
		vector<string> eventNames;
		// This syntax works courtesy of BOOST, events will come in this order.
		eventNames += "computePropensities", "highBoundState", "lowBoundState", "lowBoundCellState", 
			"ComputeLBSPropensities", "ComputeHBSPropensities", "sampleCoarseEvents", 
			"HighBoundCellState", " ComputeHBCSPropensities", "ComputeLBCSPropensities", "sampleFineEvents", 
			"TausSample", "ReactionPermutation", "ComputePCoarse", "ComputeFinalAccept";

		vector<unsigned long> eventCounters(eventNames.size());
		vector<cl_event*> events(eventNames.size());
		for (int i = 0; i < events.size(); i++) {
			cl_event v;
			//e = v;
			events[i] = NULL;
		}
		
		// todoSetupGpuMem;
		setGpuTime(0.0);
		setGpuState(s);
		int totRxnSets = _nCells + _nBaseRxn + 1;
		
		_time = 0;

		int iter = -1;
		int err = CL_SUCCESS;
		while (t < simTime){
		//while (_time < simTime) {
			iter++;
			// Too few reaction events to efficiently run GpuHierLeap (?)
			//if (s->x(0) < 7){ // TODO: this is a hack to see if the system cannot evolve, generalize later.
			int minElem = *min_element(s->X().begin(), s->X().end());
			if (minElem < leap){
				double dt = 0;
				while(t <= simTime){// && !hist->jot(s, t, dt)) { // if we want to approach times which we record 'slowly'
					// also check for D=0
					//ReactionPermutation<Reaction*> *ssaRxns = _ssaSim->ReactionTauLeap(1, s, dt); //_ssaSim
					ReactionPermutation<Reaction*> *ssaRxns = NULL; // todo: fix this (not returns rxn permutation) if func used later.
					assert(false); 
					t += dt;
					hist->jot(s, t, dt);
					delete ssaRxns;
				}
				break;
			}
			
			assert(leap < MAX_LEAP); 
			int eventIdx = 0;
			startComputePropensities(_state_mem, _D_mem, events[eventIdx++]);
			//clWaitForEvents(eventIdx, *events[0]);
			clFinish(_cmd_queue);
			
			startHighBoundState(leap, events[eventIdx++]); // upper bound on state if nothing is known except L
			clFinish(_cmd_queue);
			
			startLowBoundState(leap, events[eventIdx++]);
			clFinish(_cmd_queue);			

			
			startComputePropensities(_lowBoundS_mem, _DboundLow_mem, events[eventIdx++]);
			clFinish(_cmd_queue);
						
			// clFinish(boundState);
			startComputePropensities(_highBoundS_mem, _DboundHigh_mem, events[eventIdx++]);
			clFinish(_cmd_queue);
			
			
			
//			gpuMemStdout(_rxnEvents_mem, vector<int>(totRxnSets),  "_rxnEvents", "%d, ");
//			printf("sampleCoarseEvents...\n");
			//cpuCoarseSample(leap);
			sampleCoarseEvents(leap, events[eventIdx++]);// Sample multinomial from Dcell(coarse high bound n)
			clFinish(_cmd_queue);
			
			startHighBoundCellState(events[eventIdx++]); // given u, what is high bound on state
			clFinish(_cmd_queue);
		
			startLowBoundCellState(events[eventIdx++]);
			clFinish(_cmd_queue);
			
			startComputePropensities(_highCellBoundS_mem, _DboundCellHigh_mem, events[eventIdx++]);
			clFinish(_cmd_queue);
			
			startComputePropensities(_lowCellBoundS_mem, _DboundCellLow_mem, events[eventIdx++]); 
			clFinish(_cmd_queue);
			// clFinish(computePropensities(sch));
		
			assertFeasibleBounds(leap);

			
			//printf("cpuFineSmaple...\n");
			
//			vector<int> priorRxnEvents = gpuMemStdout(_rxnEvents_mem, vector<int>(totRxnSets), NULL, NULL);
//			sampleFineEvents(leap, events[eventIdx++]); // multinomials for each cell
//			clFinish(_cmd_queue);
//			assertGpuFineSamplePlausible(leap, priorRxnEvents);
			
			cpuFineSmaple(); 
			
			//updateRxnEventsHistory();
			
			startTausSample(leap, events[eventIdx++]); 
			err |= clFinish(_cmd_queue);
			assert(err == CL_SUCCESS);
			//cpuTausSample(leap);
			
			
			startReactionPermutation(leap, events[eventIdx++]); // In the future this will be optional (depending on early acceptance or not)
			err |= clFinish(_cmd_queue);
			assert(err == CL_SUCCESS);
			
			cpuReactionPermutation(leap);
			
			//assertIsValidReactionPermutation(leap);
			
			// clFinish(sample_coarseEvents, boundCellState);
			startComputePCoarse(leap, events[eventIdx++]); 
			err |= clFinish(_cmd_queue);
			assert(err == CL_SUCCESS);
			
//			if (iter % 100 == 0)
//				dumpGpuMemoryToStdout(leap);
			
			// Compute Pearly.
			
			// Compute Paccept from CPU.

//			float cpuCalculatedPaccept = _calculationDoublechecker->computeFinalAccept(gpuMemStdout(_state_mem, vector<int>(_nSpecies), NULL, NULL), 
//														  gpuMemStdout(_rates_mem, vector<float>(_nBaseRxn), NULL, NULL), 
//														  gpuMemStdout(_DboundCellHigh_mem, vector<float>(totRxnSets),  NULL, NULL), 
//														  gpuMemStdout(_taus_mem, vector<float>(leap), NULL, NULL), 
//														  gpuMemStdout(_orderedRxnEvents_mem, vector<int>(leap), NULL, NULL), 
//														  gpuMemStdout(_DboundCellLow_mem, vector<float>(totRxnSets), NULL, NULL), 
//														  gpuMemStdout(_inpSpec_mem, vector<int>(2*_nBaseRxn), NULL, NULL), 
//														  gpuMemStdout(_outSpec_mem, vector<int>(2*_nBaseRxn), NULL, NULL), 
//														  gpuMemStdout(_Pcoarse_mem, vector<float>(1), NULL, NULL)[0]);
//			
			
			
			//clFinish(all);
			startComputeFinalAccept(leap, events[eventIdx++]);
			err |= clFinish(_cmd_queue);
			assert(err == CL_SUCCESS);

//			float gpuAnswer = gpuMemStdout(_Paccept_mem, vector<float>(1), NULL, NULL)[0];
//			if(fabs(cpuCalculatedPaccept - gpuAnswer) > 0.001){
//				dumpGpuMemoryToStdout(leap);
//				assert(false);
//			}
			
			// Read Paccept _Pcoarse_mem
			//assert(err==CL_SUCCESS);
			//dumpGpuMemoryToStdout(leap);
			
			//clFlush(_cmd_queue);
			//clFinish(_cmd_queue);

			hist->jot(s,t, _time - t);
			//float newT = getTimeFromGpu();
			
			//assertPcoarseCorrect(leap);
			//assertPacceptCorrect(leap);
			
			updateStateFromGpu(s);
			t = _time;
			
			//clFlush(_cmd_queue);
			//clFinish(_cmd_queue);
			
			assert(events.size() == eventNames.size());
			for (int i = 0; i < events.size(); i++) {
				//eventCounters[i] += kernelRunTime(events[i]);
				//clReleaseEvent(events[i]);	
				events[i] = NULL;
			}
			//events.resize(0);
			
			//clReleaseEvent(compProp); clReleaseEvent(hbs); clReleaseEvent(lbs); clReleaseEvent(lbcs);
//			clReleaseEvent(lbsD); clReleaseEvent(hbsD); clReleaseEvent(coarseE); clReleaseEvent(hbcs); 
//			clReleaseEvent(hbcsD); clReleaseEvent(lbcsD); clReleaseEvent(fineE); clReleaseEvent(tausE); 
//			clReleaseEvent(rxnPerm); clReleaseEvent(compPCoarse);  clReleaseEvent(compFinalE); 
			//clFlush(_cmd_queue);
		}
		hist->jot(s,t, _time - t);
		
		reportOnRxnEventsHistoryStats();
		//dumpGpuMemoryToStdout(leap);
		
		
//		long sum = 0;
//		for (int i = 0; i < eventNames.size(); i++) {
//			long x = eventCounters[i];
//			printf("\t%lld nanoseconds total for %s.\n", x, eventNames[i].c_str());
//			sum += x;
//		}
//		printf("Total runtime for run %f seconds.\n", sum/1e9);
		
	}
	
	void reportOnRxnEventsHistoryStats() {
		printf("Analyzing rxnEventsHistory...\n");
		
		long nBirthEvents = 0;
		long nDeathEvents = 0;
		// TODO this is fragile. (depends on each cell having birth/death rxn channels 
		int start = _nCells + 1;
		int end = _nCells + _nBaseRxn + 1;
		for (int r = start; r < end; r++){
			if (r % 2 == start % 2) 
				nBirthEvents += rxnEventsHistory[r];
			else 
				nDeathEvents += rxnEventsHistory[r];
		}
		
		long n = nBirthEvents + nDeathEvents;
		double pctBirth = (double) nBirthEvents / n;
		double expectedPctBirth = .1 / (.1 + .11); // TODO: don't hard code rates
		double var = n * expectedPctBirth * (1 - expectedPctBirth);
		printf("birth(%d)/death(%d)=%f pft birth events\nExpected %f\n", nBirthEvents, nDeathEvents, pctBirth, expectedPctBirth);
	}
	
	void updateStateFromGpu(State *s){
		vector<int> x = gpuMemStdout(_state_mem, vector<int>(_nSpecies), NULL, NULL);
		s->setXdirectly(x);
	}
	
	float getTimeFromGpu(){
		vector<float> t = gpuMemStdout(_time_mem, vector<float>(1), NULL, NULL);
		return t[0];
	}
	
	// name: "human friendly" name of memory in GPU. May be NULL.
	// printfString: each element e will be printed out using printf(printfString, e). May be NULL
	template <typename T> vector<T> gpuMemStdout(cl_mem mem, vector<T> v, char* name, char* printfString){
		size_t memSize = v.size() * sizeof(T);
		
		cl_int err = clEnqueueReadBuffer(_cmd_queue, mem, CL_TRUE, 0, memSize,
										 (void*)&v[0], 0, NULL, NULL);
		assert(err == CL_SUCCESS);
		// TODO: may need to wait for job to finish.
		
		if (name)
			printf("%s: ", name);
		if (printfString)
			for (int i = 0; i < v.size() ; i++) {
				printf(printfString, v[i]);
			}
		if (name || printfString)
			printf("\n");
		return v;
	}
	
protected:
	vector<int> rxnEventsHistory;
	
	//TauLeapSSA *_ssaSim; // used when it is inneficient to hierLeap
	HierleapCalculations *_calculationDoublechecker; // make sure steps of GPU simulation are correct.
	float _time;
	
	// Used for Random Number Generation.
	GpuRandgen *_gpuRandgen;
	cl_mem _randomUint_mem;
	int _currentRnIndex;
	int _totalRandNums;
	
	// post: _prepareRandNums returns the offset in _randomUint_mem that is followed by nRandNumsNeeded RNs which
	//		are promised to not be used by any other process.
	int _prepareRandNums(int nRandNumsNeeded);
	
	void _createKernels(GpuInterface *gpuInterface);
	
	size_t _nCells, _nBaseRxn, _nSpecies;
	
	size_t _nHighDeltaCell, _nLowDeltaCell;
	
	// ASSUME: list of reactions from root->allReactions() are "topologically consistent"
	// ASSUME: descendantReactionList() results are in order!
	// Setup *most* memory for GPU, excluding mem blocks of var size (eg related to L).
	void _setupInputGPUMem(cl_context context, State *s, Reactions *root); 
	
	// These arguments to the kernels do not change throughout simulation
	void _inputConstKernelArgs(int nStates, int nCells, int nDeltas, int nRates, int totRxnsSets);
	
	cl_command_queue _cmd_queue;
	
	// OPENCL 'functions'
	cl_kernel _computePropensities, _boundState, _boundCellState, _sampleCoarseEvents, _computePCoarse,
		 _fineSample, _tausSample, _reactionPermutation, _computeFinalAccept;
	
	
	// OPENCL MEMORY //
	
	
	// Used for computePropensities
	cl_mem _state_mem, _cellStartIdx_mem, _inpSpec_mem, _rates_mem, _D_mem;
	
	// Used for boundState: state, highDeltaS, highBoundS
	cl_mem _highDeltaS_mem, _highBoundS_mem;
	
	// For lowerbounding 
	cl_mem _lowDeltaS_mem, _lowDeltaCell_mem, _lowBoundS_mem, _lowCellBoundS_mem, _DboundLow_mem, _DboundCellLow_mem; 
	
	// Used for boundCellState: state, cellEvents, delta(struct Delta[]) , stateBound(see highBoundS), cellBoundS (store result)
	cl_mem _highDeltaCell_mem, _highCellBoundS_mem;
	
	// Used for _sampleCoarseEvents: unormalizedProbs (D rates),numProbs,int *successes,*locSuccesses,  *randomInts\
	
	// Used for computePCoarse:  DBoundHigh, DboundCellHigh, rxnEvents, Pcoarse
	cl_mem _DboundHigh_mem, _DboundCellHigh_mem, _rxnEvents_mem, _Pcoarse_mem;
	
	// Used for _fineSample: cellStartIdx, Dprobs (DcellUpper), rxnEvents, randomUints, randUintsStart (all ready present)
	
	// Used for _tausSample: leap, rate (DBoundLow), taus, tauSum, randNums, randNumStartIdx 
	
	// Used for _reactionPermutation: rxnEvents, idxRxnStart, numRxn, locOrderedRxns, orderedRxns, randNums, startRandIdx
	cl_mem _orderedRxnEvents_mem;
	
	// Used for computeFinalAccept: state, time, rates, DboundCellHigh, taus, inpSpec, outSpec, Pcoarse
	cl_mem _time_mem, _taus_mem, _outSpec_mem, _Paccept_mem;
	
};

#endif