#include "hier_leap.h"

#include <vector.h>
#include <queue>

#include "../reactions/reaction_set.h"
#include "../reactions/reactions.h"
#include "../state/state.h"
#include "../state/reaction_cache.cpp"
#include "../reactions/reaction_permutation.h"
#include "../reactions/propensity_bounds.h"

#include "ssa.h"

using namespace hierleap;

HierLeap::HierLeap(State *s, Reactions *root): ReactionLeaper() {
	_bounder = new PropensityBounds(root, s->speciesTypes(), true);
}


// Returns an ordered list of reactions of length leap :
//	- sampled from SSA equivalent process
//	- this is guaranteed to be recursive with bottom call to ERLeap or SSA
//	- state will NOT be changed
int recursion_level = 0; // used for debugging
ReactionPermutation<Reaction*> *HierLeap::ReactionLeap(State *state, Reactions *root){
	// sample the high-bound of D for all children of root
	vector<ReactionSet*> & children = root->getDependents();
	int numChildren = children.size();

	vector<double> & upperBoundD = upperBoundChildrenD(root, state);
	vector<unsigned int> & numChildReactions = multinomialSelect(upperBoundD, state->numEvents(root));

	// set numEvents for each reaction
	// setup children blocks' reaction permutation
	ReactionPermutation<int> *metaReactions = new ReactionPermutation<int>(); // index by child ID which fires

	for (int i = 0; i < numChildren; i++){
		state->setEvents(children[i], numChildReactions[i]);
		metaReactions->addReaction(i, numChildReactions[i]);
	}

	// calculate I_0 bar (state assuming full outside diffusion) for children reaction sets
	// 		assumes numEvents is set properly for children of root
	State *assumeChildFullDiffusion = diffuseChildrenTogether(root, state);

	SSA *branchSampler = new SSA();

	// sample reaction permutations from children reaction sets
	//		we may use HierLeap OR ERLeap OR SSA for this, depending on children size
	vector<ReactionPermutation<Reaction*>*> rps;
	for (unsigned int i=0; i < root->getDependents().size(); i++){
			ReactionPermutation<Reaction*> *rp = branchSampler->ReactionLeap(state, root->getDependents()[i]->asReactions()); // TODO: choose which method
			rps.push_back(rp);
	}
	
	// create Sigma ordering of reactions
	metaReactions->permute();
	ReactionPermutation<Reaction*> *fullOrdering = sigmaOrdering(metaReactions, rps);

	// free memory from rps
	delete metaReactions;
	BOOST_FOREACH(ReactionPermutation<Reaction*> *x, rps){
		delete x;
	}
	
	
	// see if approximate sample is a substitute for the real thing
	double probAccept = acceptanceProbability(state, assumeChildFullDiffusion, fullOrdering, root);
	
	delete branchSampler; delete assumeChildFullDiffusion; 

	if (recursion_level > 15){
		recursion_level = recursion_level + 1 - 1;
	}
	
	if (urand() < probAccept) {
		recursion_level = 0;
		return fullOrdering;
	}
	
	recursion_level++;
	delete fullOrdering;
	return ReactionLeap(state, root); // sample rejected, retry...

}



// Calculate probability of acceptance as :
//		first calculating probability the events were distributed to the branches properly (according to P')
//		second return the probability that our sample form P' is acceptable
double HierLeap::acceptanceProbability(State *initial, State *assumeChildFullDiffusion,
		ReactionPermutation<Reaction*> * sigmaOrdering, Reactions *root){
	
	
	
	ReactionPermutation<Reaction*> *fullOrdering = new ReactionPermutation<Reaction*>(sigmaOrdering);
	fullOrdering->lockOrder(); // no need to permute here?

	// we are actually computing two acceptance probabilities here
	double probAcceptApprox = 1.0;
	double probApproxSampledCorrectly = 1.0;
	
	vector<double> & uppD = initial->cache(root)->upperDchildren();
	double sumDTilde = sum(&uppD[0], uppD.size());

	double Msample = 1.5; // TODO XXX this is arbitrary, check it later on
	double Maccept = 7.0;
	
	State *stateCopy = new State(initial);
	while(fullOrdering->hasNext()){
		Reaction* rxn = fullOrdering->next();
		ReactionSet* rxnForefatherAndRootsChild = rxn->highestParentUnder(root);

		double D = rxn->D(stateCopy);
		if (D==0){ probAcceptApprox == 0; break;}
		
		probAcceptApprox *= D / rxn->D( assumeChildFullDiffusion); 
		probAcceptApprox *= root->D(assumeChildFullDiffusion) / root->D(stateCopy);// same as D^bar / D 
		
		
		// This is it for seeing if Dbar was sampled correctly
		//		P *= (Drbar/Drtilde * Dtilde/Dbar)
		probApproxSampledCorrectly *= rxnForefatherAndRootsChild->D(assumeChildFullDiffusion)/
							uppD[rxnForefatherAndRootsChild->indexInParent()];
		probApproxSampledCorrectly *= sumDTilde / root->D(assumeChildFullDiffusion);
		
		

		//applyReaction(root, stateCopy, rxn);
		rxn->fire(stateCopy);
		if (rxn->selfFullyContainedIn(root)){ 
			rxn->fire(assumeChildFullDiffusion);
		}else{
			// only impact the source, not destination for diffusion
			rxn->fireReactantsOnly(assumeChildFullDiffusion);
		}
	}
	
	
	probApproxSampledCorrectly *= 1/Msample;
	probAcceptApprox *= 1/Maccept;
	
	
	if(0 > probApproxSampledCorrectly || probApproxSampledCorrectly > 1.0001){
		printf("\npre-normalized probApproxSampledCorrectly: %f", probApproxSampledCorrectly * Msample);
		assert(false);
	}
	
	// reject sample, D~ was bad
	if (urand() > probApproxSampledCorrectly){
		probAcceptApprox = 0;
	}
	
	if(0 > probAcceptApprox || probAcceptApprox > 1.0001){
		printf("\npre-normalized probAcceptApprox: %f", probAcceptApprox * Maccept);
		assert(false);
	}
	
	
	delete fullOrdering; delete stateCopy;
	
	
	return probAcceptApprox;
}

// Calculate probability of acceptance as :
//		Psig = PROD[ F(r) / F(r)^bar * D(children with r)^bar / D(children with r)^~ ]
//double HierLeap::acceptanceProbability(State *initial, State *assumeChildFullDiffusion,
//		ReactionPermutation<Reaction*> * sigmaOrdering, Reactions *root){
//	
//	
//	
//	ReactionPermutation<Reaction*> *fullOrdering = new ReactionPermutation<Reaction*>(sigmaOrdering);
//	fullOrdering->lockOrder(); // no need to permute here?
//
//	// we are actually computing two acceptance probabilities here
//	double probAcceptApprox = 1.0;
//	double probApproxSampledCorrectly = 1.0;
//	
//	vector<double> & uppD = initial->cache(root)->upperDchildren();
//	double sumDTilde = sum(&uppD[0], uppD.size());
//
//	double M = 5.0; // TODO XXX this is arbitrary, check it later on
//	
//	State *stateCopy = new State(initial);
//	while(fullOrdering->hasNext()){
//		Reaction* rxn = fullOrdering->next();
//		ReactionSet* rxnForefatherAndRootsChild = rxn->highestParentUnder(root);
//
//		double D = rxn->D(stateCopy);
//		if (D==0){ probAccept == 0; break;}
//		
//		double stepProb = D / rxn->D( assumeChildFullDiffusion); 
////		if(0 > probAccept || probAccept > 1.0001){
////			assert(false);
////		}
//		
//		// TODO unremove
//		stepProb *= rxnForefatherAndRootsChild->D(assumeChildFullDiffusion) /
//			initial->DtildeCached(rxnForefatherAndRootsChild);// same as D^bar / D^~  
//		
//		//  mult F / F_bar * Dm(r)_bar / Dm(r)_higher * (Droot^tilde / D(root))
//		stepProb *= sumDTilde / root->D(stateCopy);
//		
//		probAccept *= stepProb;
//		
//
//		//applyReaction(root, stateCopy, rxn);
//		rxn->fire(stateCopy);
//		if (rxn->selfFullyContainedIn(root)){ 
//			rxn->fire(assumeChildFullDiffusion);
//		}else{
//			// only impact the source, not destination for diffusion
//			rxn->fireReactantsOnly(assumeChildFullDiffusion);
//		}
//	}
//	
//	probAccept *= 1/M;
//	if(0 > probAccept || probAccept > 1.0001){
//		printf("\npre-normalized probAccept: %f", probAccept * M);
//		assert(false);
//	}
//	
//	
//	delete fullOrdering; delete stateCopy;
//	return probAccept;
//}

// modify state to make assumption that all leaves diffuse into one another
int x;
State *HierLeap::diffuseChildrenTogether(Reactions *root, State* state){

	// if this is one base cell just return
	// TODO Q: do we actually need to create a new state?
	if (root->onlyContainReaction()) return state; 

	State *newState = new State(state);
	// XXX: could definitely cache the results if need to optimize

	// for every child, see which base *cells* under root are nbrs
	// diffuse from child into every neighboring cell
	vector<ReactionSet*> & children = root->getDependents();
	for (int i=0; i < children.size(); i++){
		if (state->numEvents(children[i]) == 0) continue;
		vector<Reactions*> baseNbrs = children[i]->asReactions()->baseNeighbors(); // all the base cells that this Reactions* borders which are also under root
		for (int j=0; j < baseNbrs.size(); j++){
			baseNbrs[j]->assumeDiffusionInto(newState, state->numEvents(children[i])); // this does modify newState
		}
	}
	return newState;
}

// Assume reactions only occur inside of this cell
vector<double> & HierLeap::upperBoundChildrenD(Reactions* root, State *state){
	int leap = state->numEvents(root);
	vector<double> & upperD = state->cache(root)->upperDchildren();

	// do this for all dependents
	vector<ReactionSet*> & dep = root->getDependents();
	if (upperD.size() != dep.size()) upperD.resize(dep.size());
	
	for (int i=0; i < dep.size(); i++){
		if(dep[i]->isReaction()){
			upperD[i] = _bounder->upperBoundDrxn(dep[i]->asReaction(), state, leap);
		}else{
			upperD[i] = _bounder->upperBoundDrxns(dep[i]->asReactions(), state, leap);
		}
	}

	return upperD;
}