/*
 *  hierleap_calculations.h
 *  gpu_hierleap
 *
 *  Created by David Orendorff on 1/10/11.
 *  Copyright 2011 UC Irvine. All rights reserved.
 *
 *
 *  This class performs calculations relevant to hierleap using only primitives. Functionallity here
 *  is duplicated in other places, for example gpuhier_tau_leap.h. There primary purpose of this class is 
 *  to provide checking of calculations done in GPU.
 */

#include <boost/assign/std/vector.hpp>
#include <boost/assign/std/vector.hpp>

#include <cmath>
#include <assert.h>
#include <algo.h>

#include "../utils/common.h"

using namespace std;
using namespace boost::assign;

#ifndef HIERLEAP_CALCULATIONS_H
#define HIERLEAP_CALCULATIONS_H

class HierleapCalculations {
public:
	HierleapCalculations(int nSpecies, int nCells) : _nSpecies(nSpecies), _nCells(nCells) {}
	
	float computePCoarse(vector<float> DboundHigh, vector<float> DboundCellHigh, vector<float> DboundLow, 
						 vector<float> DboundCellLow, vector<int> rxnEvents, vector<float> taus);
	
	// Returns true if these bounds are feasible.
	// This function is not especially quick.
	// False is returned or an assert may be thrown.
	bool calculateAreBoundsFeasible(int leap, vector<int> highBoundS, vector<int> highCellBoundS, vector<int> state, vector<int> lowCellBoundS, vector<int> lowBoundS,
									vector<float> DboundHigh, vector<float> DboundCellHigh, vector<float> D, vector<float> DboundCellLow, vector<float> DboundLow,
									vector<int> cellStartIdx, vector<int> inpSpec, vector<float> rates){
		int nCells = cellStartIdx[0] - 1;
		int S = state.size();
		int R = D.size();
		assert(S==highBoundS.size()&&S==highCellBoundS.size()&&S==lowCellBoundS.size()&&S==lowBoundS.size());
		assert(R==DboundHigh.size()&&R==DboundCellHigh.size()&&R==DboundCellLow.size()&&R==DboundLow.size());
		
		for (int s = 0; s < state.size(); s++) {
			vector<int> v;
			v += lowBoundS[s], lowCellBoundS[s], state[s], highCellBoundS[s], highBoundS[s];
			assert(v[0] >= 0);
			if (lowBoundS[s] > 0) // TODO fragile asserts
				assert(lowBoundS[s] == state[s] - leap + 1);
			assert(state[s] + leap - 1 == highBoundS[s]);
			
			if (!is_sorted(v.begin(), v.end())){
				assert(false);
				return false;
			}
		}
		
		// Make sure baseRxn propensities are calculated correctly.
		vector<float> Ds[] = {DboundHigh, DboundCellHigh, D, DboundCellLow, DboundLow};
		// vector<int> states[] =  todo: fill this in for neatness.
		int sizeDs = 5;
		
		for (int rr = cellStartIdx[0]; rr < R; rr++){
			float p = propensity(rr, highBoundS, rates, inpSpec, cellStartIdx[0]);
			float eps = p / 100000.0;
			float d = fabs(DboundHigh[rr] - p);
			assert(d < eps);
			
			assert(propensity(rr, highCellBoundS, rates, inpSpec, cellStartIdx[0]) == DboundCellHigh[rr]);
			assert(propensity(rr, state, rates, inpSpec, cellStartIdx[0]) == D[rr]);
			assert(propensity(rr, lowCellBoundS, rates, inpSpec, cellStartIdx[0]) == DboundCellLow[rr]);
			assert(propensity(rr, lowBoundS, rates, inpSpec, cellStartIdx[0]) == DboundLow[rr]);
		}
		
		// now assert cell propensities add up to the correct amount.
		for (int c = 0; c < _nCells; c++) {
			for (int i = 0; i < sizeDs; i++){
				float sum = 0;
				for (int d = cellStartIdx[c]; d < cellStartIdx[c+1]; d++){
					sum += Ds[i][d];
				}
				assert(Ds[i][c+1] == sum);
			}
		}
		
		for (int r = 0; r < D.size(); r++) {
			vector<int> vv;
			vv += DboundLow[r], DboundCellLow[r], D[r], DboundCellHigh[r], DboundHigh[r];
//			int size = sizeof( vv ) / sizeof( vv[0] );
//			assert(vv[0] >= 0);
			if (!is_sorted(vv.begin(), vv.end())){
				assert(false);
				return false;
			}
		}
		
		return true;
	}
	
	float computeFinalAccept(vector<int> state, vector<float> rates, vector<float> DboundCellHigh, 
							 vector<float> taus, vector<int> orderedRxnEvents, vector<float> DboundCellLow,
							 vector<int> inpSpec, vector<int> outSpec, float Pcoarse) {
		
		// For each reaction in orderedRxnEvents we compute: Pfine = PROD D[r] / DhighCell[r]
		// We also evolve D so that: Pfine *= exp(-tau_i * (D(i) - DlowCell))
		// Finally, Pfine *= Pcoarse;
		double expSum = 0.0; 
		double Pfine = 1.0;
		int rxnStartIdx = 1 + _nCells;
		
		for (int i = 0; i < orderedRxnEvents.size(); i++) {
			int rxnId = orderedRxnEvents[i];

			int rxnNum = rxnId - rxnStartIdx;
			double p = propensity(rxnId, state, rates, inpSpec, rxnStartIdx);
			double ratio = p / DboundCellHigh[rxnId];
			//printf("%f / %f * ", p, DboundCellHigh[rxnId]);
			assert(ratio <= 1.0 && ratio >= 0);
			Pfine *= ratio;
			
			double D = 0.0;
			for (int r = rxnStartIdx; r < DboundCellHigh.size(); r++) {
				D += propensity(r, state, rates, inpSpec, rxnStartIdx);
			}
			
			expSum += taus[i] * (D - DboundCellLow[0]);

			updateState(rxnId-rxnStartIdx, state, inpSpec, outSpec);
		}
		//printf("\n");
		Pfine *= exp(-expSum); 
		return Pfine * Pcoarse; 
	}
	
	// Returns a list of Taus sampled with the given properties
	vector<float> tausSample(int leap, float DboundLow) {
		vector<float> taus;
		for (int i = 0; i < leap; i++){
			taus.push_back(-log(urand()) / DboundLow);
		}
		return taus;
	}
	
	// Returns a vector of size |all rxn sets|, that is symanticaly equivalent to rxnEvents
	vector<int> coarseSample(vector<float> DboundHigh, int nCells, int leap){
		vector<double> probs;
		for (int i = 1; i <= nCells; i++){
			probs.push_back(DboundHigh[i]);
		}
		
		vector<unsigned int> sample = multinomial(probs, (unsigned int) leap);
		
		// sample formated in a way congenial to gpuHierleap
		vector<int> retSample(DboundHigh.size(), 0);
		for (int i = 1; i <= nCells; i++){
			retSample[i] = sample[i-1];
		}
		retSample[0] = leap;
		return retSample;
	}
	
	// Returns rxnEvents with intercellular events filled in.
	vector<int> fineSample(vector<int> cellStartIdx, vector<float> DboundCellHigh, vector<int> rxnEvents) {
		int nCells = cellStartIdx[0] - 1;
		
		// For each cell sample a multinomial
		vector<int> fineRxnEvents(rxnEvents.size(), 0);
		fineRxnEvents[0] = rxnEvents[0];
		for (int cell = 0; cell < nCells; cell++){
			int e = rxnEvents[cell+1];
			fineRxnEvents[cell+1] = e;
			int i0 = cellStartIdx[cell];
			int i1 = cellStartIdx[cell+1];
			if (e == 0) continue;
			
			vector<double> p;
			for (int i = i0; i < i1; i++){
				p.push_back(DboundCellHigh[i]);
			}
			
			vector<unsigned int> sample = multinomial(p, (unsigned int) e);
			
			for (int i = i0; i < i1; i++){
				fineRxnEvents[i] = sample[i - i0];
			}
		}
		return fineRxnEvents;
	}
	
	// Results (orderedRxns) is returned by this function
	vector<int> reactionPermutation(vector<int> rxnEvents, 
									vector<int> cellStartIdx)
	{ 
		vector<int> orderedRxns;
		
		// create ordered rxns
		for (int i = cellStartIdx[0]; i < cellStartIdx.back(); ){
			if (rxnEvents[i] > 0){
				rxnEvents[i]--;
				orderedRxns.push_back(i);
				i = cellStartIdx[0]; 
			} else {
				i++;
			}
		}
		randperm(&orderedRxns[0], orderedRxns.size());
		
		return orderedRxns;
	}
	
private:
	void updateState(int rxnId, vector<int> & state, vector<int> & inputSpecies, vector<int> & outputSpecies){
		int ins[2] = {inputSpecies[2*rxnId], inputSpecies[2*rxnId+1]};
		int outs[2] = {outputSpecies[2*rxnId], outputSpecies[2*rxnId+1]};
		
		if (ins[0] >= 0) state[ins[0]]--;
		if (ins[1] >= 0) state[ins[1]]--;
		if (outs[0] >= 0) state[outs[0]]++;
		if (outs[1] >= 0) state[outs[1]]++;
	}
	
	double propensity(int rxnId, vector<int> & state, vector<float> & rates, vector<int> & inpSpec, int rxnStartIdx){
		assert(rxnId >= rxnStartIdx);
		
		int rxnNum = rxnId - rxnStartIdx;
		float prop = rates[rxnNum];
		
		int inp[2] = {inpSpec[rxnNum*2], inpSpec[rxnNum*2+1]};
		if (inp[0] >= 0) 
			prop *= state[inp[0]];
		if (inp[1] == inp[0])
			prop *= state[inp[1]] - 1;
		else if(inp[1] >=0 )
			prop *= state[inp[1]];
		
		return prop;
	}
	
	int _nSpecies, _nCells;
};

#endif