/*
 * simulator.h
 *
 *  Created on: Sep 3, 2009
 *      Author: Dave Orendorff
 */

#include <cmath>
#include <string.h>

#include "../state/state.h"
#include "../reactions/reactions.h"
#include "reaction_leaper.h"
#include "ssa.h"
#include "er_leap.h"
#include "hier_leap.h"
#include "reaction_tally.h"
#include "historian.h"
#include "../globals.h"

#ifndef SIMULATOR_H_
#define SIMULATOR_H_

using namespace hierleap;
using namespace std;

// Assumes the simulator engine passed into run samples
//	 from reaction events, and this class samples from tau.
class Simulator {
	friend class FullSimulator;
public:
	Simulator(){ assert(false);}
	
	Simulator(Reactions *root, State *state);

	// updates state to reflect SSA equivalent simulation
	//		of reactions locate under *root* for duration *time*
	void run(ReactionLeaper *sim, Historian *hist, Reactions *root, double time, int leap, State *state){
		assert(leap > 0 && leap < MAX_EVENTS);
		
		int iter = 0;
		double t = 0.0;
		while (t <= time){
			
			if (root->D(state) == 0){
				double dt = INFINITY; 
				hist->jot(state, t, dt);
				t += dt;
				break;
			}
			
			// sample reactions
			state->setEvents(root, leap);
			ReactionPermutation<Reaction*> *rxns = sim->ReactionLeap(state, root);
			
			// sample tau
			
			// for debugging
//			if (iter==1742){
//				iter = iter + 1 - 1; // for debugging
//				_tauSample(root, state, rxns) ;
//			}
			//double dt = _tauSample(root, state, rxns);
			
			double dt = _tauSampleSSA(root, state, rxns);
			rxns->lockOrder();
			
			// update time and state
			hist->jot(state, t, dt); // write to file
			updateState(state, rxns);
			t += dt;
			
			// TODO write state to file(?)

			iter++;
			delete rxns;
		}
		printf("%d, ", iter);
	}

	// Calculates the probability of reactions occurring over nSamples worth
	//		of reaction leaps. rxnEventsPerSample is the number of steps each
	//		reactionLeap(leap) will at least simulate.
	void calcProbReactions(ReactionLeaper *sim, Reactions *root, State *initialState, int leap, int nSamples, int rxnEventsPerSample){
		assert(rxnEventsPerSample % leap == 0); // ascertain that we can achieve exactly rxnEventsPerSample

		//ReactionLeaper *sim = new SSA();
		//ReactionLeaper *s = new ERLeap(); // TODO make root/state calling order consistent
		//ReactionLeaper *sim = new HierLeap(initialState, root); // change this line to change leaper method
		
		ReactionTally *rxnHist = new ReactionTally(root);

		for (int sample=0; sample < nSamples; sample++){
			State *state = new State(initialState);
			for (int nEvents = 0; nEvents < rxnEventsPerSample; nEvents += leap){
				state->setEvents(root, leap);
				ReactionPermutation<Reaction*> *rxns = sim->ReactionLeap(state, root);
				rxnHist->tally(rxns);
				updateState(state, rxns);
				delete rxns;
			}
			delete state;
		}
		
		rxnHist->printResults();
		delete sim;
	}

	// Applies all of the reactions in the ReactionPermutation to state
	void updateState(State *state, ReactionPermutation<Reaction*> *rxns){
		while (rxns->hasNext()){
			Reaction *r = rxns->next();
			r->fire(state);
		}
	}

	~Simulator(){
		delete _pb;
	}

private:
	
	// Given the ordered list of reactions, samples a dT equivalent to SSA
	double _tauSample(Reactions *root, State *initialState, ReactionPermutation<Reaction*> *orderedReactions);
	
	double _tauSampleSSA(Reactions *root, State *initialState, ReactionPermutation<Reaction*> *orderedReactions);
protected:
	PropensityBounds *_pb;
};

#endif /* SIMULATOR_H_ */
