/*
 *  tau_leap_independent_cells.cpp
 *  gpu_hierleap
 *
 *  Created by David Orendorff on 6/8/11.
 *  Copyright 2011 UC Irvine. All rights reserved.
 *
 */

#include "tau_leap_independent_cells.h"

#include <boost/foreach.hpp>
#include <math.h>
#include <tr1/unordered_map>

#include "reaction_tau_leaper.h"

#include "../reactions/delta_propensity_bounds.h"
#include "../reactions/reaction_set.h"
#include "../reactions/reaction.h"
#include "../reactions/reactions.h"
#include "../reactions/reaction_permutation.h"
#include "../reactions/propensity_bounds.h"
#include "../state/state.h"
#include "../utils/common.h"

using namespace hierleap;
using namespace std;

TauLeapIndependentCells::TauLeapIndependentCells(Reactions *root, State *s):_root(root){
	int n = s->speciesSize();
	_propensityBoundsCalc[_root] = new PropensityBoundsNoCells(_root, n, false); // todo this may not work (??)
	vector<ReactionSet*> & dependents = _root->getDependents();
	assert(_propensityBoundsCalc.count(_root) == 1);
	assert(_propensityBoundsCalc.count(dependents[0]->asReactions()) == 0);
	BOOST_FOREACH(ReactionSet *rs, dependents) {
		Reactions *r = rs->asReactions();
		_propensityBoundsCalc[r] = new PropensityBoundsNoCells(r, n, false);
		_deltaPropensityBoundsCalc[r] = new DeltaPropensityBounds(r);
	}
	_deltaPropensityBoundsCalc[_root] = new DeltaPropensityBounds(_root);
	
	_cellD = vector<double>(dependents.size());
	_cellHighBoundD = vector<double>(dependents.size());
	_cellLowBoundD = vector<double>(dependents.size());
}


double TauLeapIndependentCells::ReactionTauLeap(int leap, State *s, double &dt) {
	vector<int> updatedState = s->X(); // copy, we will be modifying
	
	// calculate upper/lower bounds on propensities for each group.
	vector<ReactionSet*> & dependents =  _root->getDependents();

	for (int i = 0; i < dependents.size(); i++) {
		_cellD[i] = dependents[i]->DbyUniqueId(updatedState);
	}
	
	vector<int> & highBoundState = _propensityBoundsCalc[_root]->_highBoundState(s->X(), leap); 
	
	vector<double> & cellUpperBoundD = _cellUpperBoundD(s, _cellD, leap); // get delta instead of upper bound?
	vector<double> & cellLowerBoundD = _cellLowerBoundD(s, _cellD, leap);
	
	// calculate lower bound on total propensities for tau sampling.
	double rootUpperBoundD = _rootUpperBoundD(s, _cellD, cellUpperBoundD, leap);
	//double rootLowerBoundD = _rootLowerBoundD(s, _cellD, cellLowerBoundD, leap);
	//double quickRootLowerBoundD = _rootLowerBoundD(s, _cellD, cellLowerBoundD, leap);
	//double rootLowerBoundD = _tightLowerBoundD(_root, s, leap); // todo: changed this line
	//double rootLowerBoundD = getDeltaBounder(_root)->lowerBoundDeltaD(highBoundState);
	double rootLowerBoundD = sum(cellLowerBoundD); // todo: remove this line
	//assertEpsilonLessThan(quickRootLowerBoundD, rootLowerBoundD);
		
	// sample cell events from multinomial.
	vector<unsigned int> cellEvents = multinomial(cellUpperBoundD, leap);
	dt = erlang_sample(rootLowerBoundD, leap);
	
	double acceptanceProbability = urand();
	
	
	// Calculate and check for early acceptance.
	double earlyAcceptance = 1.0;
	for (int i = 0; i < cellUpperBoundD.size(); i++) {
		if (cellEvents[i]) {
			assert(cellEvents[i] > 0 && cellLowerBoundD[i] >= 0 && cellUpperBoundD[i] >= 0);
			earlyAcceptance *= ipow(cellLowerBoundD[i] / cellUpperBoundD[i], cellEvents[i]);
		}
		assert(earlyAcceptance >= 0 && earlyAcceptance <= 1.0);
	}
	earlyAcceptance *= exp(-dt * (rootUpperBoundD - rootLowerBoundD));
	//earlyAcceptance = 0; // todo XXX remove this
	
	// simulate each group (in parallel, when possible). If early acceptance is true then update state accordingly.
	//ReactionPermutation<Reaction*> *rp = new ReactionPermutation<Reaction*>();
	vector<ReactionPermutation<Reaction*>*> rps(dependents.size());
	
	for (int cellIdx = 0; cellIdx < cellUpperBoundD.size(); cellIdx++) {
		if (cellEvents[cellIdx]) {
			rps[cellIdx] = new ReactionPermutation<Reaction*>();
			_ssaSteps(_root->getDependents()[cellIdx]->asReactions(), cellEvents[cellIdx], updatedState, rps[cellIdx]);
		}
	}
	
	//ReactionPermutation<Reaction*> *rp2 = new ReactionPermutation<Reaction*>(rp);
	//delete rp;
	//rp = rp2;
	
	// Do full P(accept) calculation if needed.
	if (!(acceptanceProbability <= earlyAcceptance)) {
		vector<double> taus = _uniformSimplex(leap, dt);
		vector<int> tempState = s->X();
		ReactionPermutation<Reaction*> *rp = new ReactionPermutation<Reaction*>();
		rp->joinPermutations(rps);
		ReactionPermutation<Reaction*> *rpHist = new ReactionPermutation<Reaction*>(rp);

		double probAccept = 1.0;
		double expSum = 0.0;
		while (rp->hasNext()){
			Reaction *r = rp->nextFromBack();
			Reactions *cell = r->getParent();
			assert(cell);
			int cellIdx = cell->indexInParent();
			assert(cellIdx >= 0 && cellIdx < _root->getDependents().size());
			assert(_root->getDependents()[cellIdx] == cell);
			double tau = taus.back();
			taus.pop_back();
			
			double cellD = cell->DbyUniqueId(tempState);
			double rootD = _root->DbyUniqueId(tempState);
			
			assertEpsilonLessThan(cellD, cellUpperBoundD[cellIdx]); 
			assertEpsilonLessThan(cellLowerBoundD[cellIdx], cellD);
			assertEpsilonLessThan(rootLowerBoundD, rootD);
			
			probAccept *= cellD / cellUpperBoundD[cellIdx]; 
			
			expSum += -tau * (rootD - rootLowerBoundD);
			//if (probAccept==0 || !isfinite(probAccept)) break; isinf
			if (probAccept==0) break;
			r->fire(tempState);
		}
		probAccept *= exp(expSum);
		delete rpHist;
		delete rp;
		
		// Reject sample, try again.
		if (probAccept > acceptanceProbability) {
			delete_non_null_objs(rps);
			return ReactionTauLeap(leap, s, dt);
		}
	}
	delete_non_null_objs(rps);
	
	s->setXdirectly(updatedState);
	return _root->DbyUniqueId(s->X());
}

// Completes <events> reaction and modifies s in the process.
void TauLeapIndependentCells::_ssaSteps(Reactions *parent, int events, vector<int> & s, ReactionPermutation<Reaction*> *rp) {
	vector<Reaction*> & rxn = parent->descendantReactionList();
	
	while (events--) {
		double targetSum = parent->DbyUniqueId(s) * urand();
		
		int ridx = 0;
		double sum=0;
		for (  ;; ridx++) {
			sum += rxn[ridx]->DbyUniqueId(s);
			if (sum >= targetSum) break;
		}
		
		assert(ridx < rxn.size());
		Reaction *r = rxn[ridx];
		
		// add r to reaction permutation, fire and adjust accordingly
		rp->addBack(r);
		r->fire(s);
	}
}

vector<double> & TauLeapIndependentCells::_cellUpperBoundD(State *s, vector<double> & cellD, int leap) {
	vector<ReactionSet*> & dependents =  _root->getDependents(); 
	
	// TODO(XXX): assuming cellular indepencence this is ok, otherwise i need to think about it...
	vector<int> & highBoundState = _propensityBoundsCalc[_root]->_highBoundState(s->X(), leap); 
	assert(dependents.size() == _cellHighBoundD.size());
	for (int i = 0; i < dependents.size(); i++) {
		//_cellHighBoundD[i] = _quickTightUpperBoundD(dependents[i]->asReactions(), highBoundState, s, D, leap);
		_cellHighBoundD[i] = cellD[i] + getDeltaBounder(dependents[i]->asReactions())->upperBoundDeltaD(highBoundState) * (leap - 1);
	}
	return _cellHighBoundD;
}

vector<double> & TauLeapIndependentCells::_cellLowerBoundD(State *s, vector<double> & cellD, int leap) {
	vector<ReactionSet*> & dependents =  _root->getDependents(); 
	vector<int> & highBoundState = _propensityBoundsCalc[_root]->_highBoundState(s->X(), leap); 
	assert(dependents.size() == _cellLowBoundD.size());
	for (int i = 0; i < dependents.size(); i++) {
		//_cellLowBoundD[i] = _tightLowerBoundD(dependents[i]->asReactions(), s, leap);
		//_cellHighBoundD[i] = _quickTightLowerBoundD(dependents[i]->asReactions(), s, D, leap);
		double delta = getDeltaBounder(dependents[i]->asReactions())->lowerBoundDeltaD(highBoundState);
		assert(delta <= 0);
		//_cellLowBoundD[i] = cellD[i] + delta * (leap - 1);
		_cellLowBoundD[i] = cellD[i] + delta * (leap );
		_cellLowBoundD[i] = max(0.0, _cellLowBoundD[i]);
	}
	return _cellLowBoundD;
}

double TauLeapIndependentCells::_rootUpperBoundD(State *s, vector<double> & D, vector<double> & cellUpperBoundD, int leap) {
	// take the max difference between a cell and upper bound and add it to total D.
	double maxDiff = 0.0;
	for (int i = 0; i < D.size(); i++) { 
		maxDiff = max(cellUpperBoundD[i] - D[i], maxDiff);
	}
	
	return sum(D) + maxDiff;
}

// Calculates the lower bound on root after leap-1 reaction events.
// NOTE: ASSUMES cells are independent!!
// TODO(dweo): if D[i] == 0 this will be a bug XXX, fix. 
//     we really need to see max delta given 1 event and interpolate from there (even if it means D[i] would be negative) 
double TauLeapIndependentCells::_rootLowerBoundD(State *s, vector<double> & D, vector<double> & cellLowerBoundD, int leap) {
	// take the max difference between a cell and upper bound and add it to total D.
	double maxDiff = 0.0;
	for (int i = 0; i < D.size(); i++) { 
		maxDiff = max(D[i] - cellLowerBoundD[i], maxDiff);
	}
	
	return sum(D) - maxDiff;
}

// Provides a close to actual upper bound on D (sum of propensities) but at the cost of more computational effort.
// Cell has at least one reaction
double TauLeapIndependentCells::_tightUpperBoundD(Reactions *cell, State *s, int leap) {
	// TODO: for now do a brute force search of the max deltaD, but later optimize.
	vector<int> highBoundState = _propensityBoundsCalc[cell]->_highBoundState(s->X(), leap);
	
	// see which rxn causes the greatest increase in D.
	vector<int> tempState = s->X();
	double maxDelta = 0.0;
	double D = cell->DbyUniqueId(tempState);
	BOOST_FOREACH(Reaction *r, cell->descendantReactionList()) {
		r->fire(tempState);
		double newD = cell->DbyUniqueId(tempState);
		r->unFire(tempState);
		maxDelta = max(maxDelta, newD - D);
	}
	
	return D + maxDelta * (leap - 1);
}

// Provides a close to actual upper bound on D (sum of propensities) but at the cost of more computational effort.
// Cell has at least one reaction
//double TauLeapIndependentCells::_quickTightUpperBoundD(Reactions *cell, vector<int> & highBoundState, State *s, double cellD, int leap) {
//	vector<int> & highBoundState = _propensityBoundsCalc[cell]->_highBoundState(s->X(), leap);
//	return cellD + getDeltaBounder(cell)->upperBoundDeltaD(highBoundState) * (leap - 1);
//}

// Note: this actually isn't as optimal as the quickUpperBoundVersion, could be tighter
//double TauLeapIndependentCells::_quickTightLowerBoundD(Reactions *cell, State *s, double cellD, int leap) {
//	vector<int> highBoundState = _propensityBoundsCalc[cell]->_highBoundState(s->X(), leap);
//	return max(0.0, cellD + getDeltaBounder(cell)->lowerBoundDeltaD(highBoundState) * (leap - 1));
//}

// Provides a close to actual lower bound on D (sum of propensities) but at the cost of more computational effort.
// Pre: Cell has at least one reaction
double TauLeapIndependentCells::_tightLowerBoundD(Reactions *cell, State *s, int leap) {
	// TODO: for now do a brute force search of the min deltaD, but later optimize.
	vector<int> highBoundState = _propensityBoundsCalc[cell]->_highBoundState(s->X(), leap);
	
	// see which rxn causes the greatest increase in D.
	vector<int> tempState = s->X();
	double minDelta = 0.0;
	double D = cell->DbyUniqueId(tempState);
	BOOST_FOREACH(Reaction *r, cell->descendantReactionList()) {
		r->fire(tempState);
		double newD = cell->DbyUniqueId(tempState);
		r->unFire(tempState);
		minDelta = min(minDelta, newD - D);
	}
	
	return D + minDelta * (leap - 1);
}

// Samples uniformly from the simplex such that
//		x = x1 + x2 + ... + x(dimensions)
vector<double> TauLeapIndependentCells::_uniformSimplex(int dimensions, double sumsToThis) {
	vector<double> sideLens(dimensions);
	double sum = 0.0;
	for (int i = 0; i < dimensions; i++) {
		sideLens[i] = erand();
		sum += sideLens[i];
	}
	double normalize = sumsToThis / sum;
	for (int i = 0; i < dimensions; i++) {
		sideLens[i] *= normalize;
	}
	return sideLens;
}