/*
 *  tau_leap_independent_cells.h
 *  gpu_hierleap
 *
 *  Created by David Orendorff on 6/6/11.
 *  Copyright 2011 UC Irvine. All rights reserved.
 *
 */

#ifndef TAU_LEAP_INDEPENDENT_CELLS_H_
#define TAU_LEAP_INDEPENDENT_CELLS_H_

#include <boost/foreach.hpp>
//#include <map>
#include <tr1/unordered_map>

#include "reaction_tau_leaper.h"
#include "../state/state.h"
#include "../reactions/reaction_set.h"
#include "../reactions/reaction.h"
#include "../reactions/reaction_permutation.h"
#include "../reactions/propensity_bounds.h"
#include "../utils/common.h"

class DeltaPropensityBounds;
class State;
class Reactions;
class PropensityBounds;

namespace hierleap{
	class Reaction;
	class Species;
}
using namespace hierleap;
using namespace std;

/*
 * Assumes each child reactions of root is independent of all other children of root. 
 * This class may take advantage of parallel architecture and uses rejection sampling to do so.
 */
class TauLeapIndependentCells : public ReactionTauLeaper {
	
public:
	TauLeapIndependentCells(Reactions *root, State *s);
	
	
	virtual char* name(){ return "TauLeapIndependentCells";}
	
	
	// State is updated upon return after simulation an equivelent to leap SSA rxn events.
	// dt is the amount of time that has elapsed during the <leap> reaction events.
	// TODO: if needed this function could be sped up by caching Ds
	virtual double ReactionTauLeap(int leap, State *s, double &dt);

private:
	tr1::unordered_map <Reactions*, PropensityBounds *> _propensityBoundsCalc;
	tr1::unordered_map <Reactions*, DeltaPropensityBounds*> _deltaPropensityBoundsCalc;
	
	vector<double> _cellD;  //propensity for each cell
	vector<double> _cellHighBoundD; // So we do not have to reallocate every time we call cell bound
	vector<double> _cellLowBoundD; 
	
	DeltaPropensityBounds* getDeltaBounder(Reactions *cell) {
		assert(cell);
		DeltaPropensityBounds *db = _deltaPropensityBoundsCalc[cell];
		assert(db);
		return db;
	}
	
	// Completes <events> reaction and modifies s in the process.
	void _ssaSteps(Reactions *parent, int events, vector<int> & s, ReactionPermutation<Reaction*> *rp);
	
	vector<double> & _cellUpperBoundD(State *s, vector<double> & cellD, int leap);	
	
	vector<double> & _cellLowerBoundD(State *s, vector<double> & cellD, int leap);
	
	double _rootUpperBoundD(State *s, vector<double> & D, vector<double> & cellUpperBoundD, int leap);
	
	// Calculates the lower bound on root after leap-1 reaction events.
	double _rootLowerBoundD(State *s, vector<double> & D, vector<double> & cellLowerBoundD, int leap);
		
	//double _quickTightUpperBoundD(Reactions *cell, State *s, double cellD, int leap);
	
	//double _quickTightLowerBoundD(Reactions *cell, State *s, double cellD, int leap); 
	
	// Provides a close to actual upper bound on D (sum of propensities) but at the cost of more computational effort.
	// Cell has at least one reaction
	double _tightUpperBoundD(Reactions *cell, State *s, int leap);
	
	// Provides a close to actual lower bound on D (sum of propensities) but at the cost of more computational effort.
	// Pre: Cell has at least one reaction
	double _tightLowerBoundD(Reactions *cell, State *s, int leap);
	
	// Samples uniformly from the simplex such that
	//		x = x1 + x2 + ... + x(dimensions)
	vector<double> _uniformSimplex(int dimensions, double sumsToThis);
	
	Reactions *_root;
};

#endif