#include "../reactions/reaction.h"
#include "../state/state.h"

#include "reaction_tau_leaper.h"


#ifndef TIME_HIER_SSA_H_
#define TIME_HIER_SSA_H_

using namespace hierleap;

struct CompareReactions {
	CompareReactions(State *state){ s=state;}
	State *s;
	//bool operator() (Reaction *r1,Reaction * r2) { assert(s); return (r1->D(s)<r2->D(s));}
	bool operator() (Reaction *r1,Reaction * r2) { assert(s); return (r1->D(s) > r2->D(s));}
} ; 


class TimeHierarchySSA : public ReactionTauLeaper {
public:
	TimeHierarchySSA(){ assert(false);}
	
	TimeHierarchySSA(Reactions *root, State *s):_fastReactions(NULL),_slowReactions(NULL){
		_resetSlowRxnCounters();
		_partitionReactions(root, s);
		_printROClikeStatus(s);
	}
	
	virtual char* name(){ return "TimeHierarchySSA";}
	
	virtual void changeRuntimeParameters(State *s, int leap) {}
	
	// assumes we update state when selecting one reaction event
	virtual double ReactionTauLeap(int leap, State *s, double &dt){
		assert(leap==1);
		
		double Dfast = _fastReactions->D(s);
		Reaction *fastRxn = _reactionSelect(s, _fastReactions->descendantReactionList(), Dfast);
		dt = erand() / Dfast;
		
		fastRxn->fire(s);
		
		double newSRDDTS = _slowRxnsDdtSum + _slowReactions->D(s) * dt;
		if (newSRDDTS >= _targetLogUniformVal){
			// perform one slow reaction
			fastRxn->unFire(s);
			Reaction *slowRxn = _reactionSelect(s, _slowReactions->descendantReactionList(), _slowReactions->D(s));
			
			// solve for dt s.t. SUM(dt * As) = -log(u)
			//	OR equivalently:		_slowRxnsDdtSum + dt * _slowReactions->D(s) = _targetLogUniformVal
			//		.:.					dt = (_targetLogUniformVal - _slowRxnsDdtSum) / _slowReactions->D(s)
			dt = (_targetLogUniformVal - _slowRxnsDdtSum) / _slowReactions->D(s);
			
			_resetSlowRxnCounters();
			
			slowRxn->fire(s);
		}else{
			_slowRxnsDdtSum = newSRDDTS;
		}
		assert(false); // fix return rxn perm if we ever need this class.
		return NULL;
	}
	
private:
	
	// Prints an ordered list of tuples
	//	of the form (lambda_i, Dslow / Dtot: assuming all rxns slower than i are in slow rxn)
	// NOTE: this could be optimized by O(n^2) -> O(n) if it mattered
	void _printROClikeStatus(State *s){
		vector<Reaction*> rxns = _fastReactions->descendantReactionList() ;
		//rxns.append(_slowReactions->descendantReactionList());
		rxns.insert(rxns.end(), _slowReactions->descendantReactionList().begin(), _slowReactions->descendantReactionList().end());
		
		// calc D, and get fastest rxn
		double D = 0.0;
		double fastestD = 0;
		for (int i=0; i < rxns.size(); i++){
			double Dr = rxns[i]->D(s);
			D += Dr;
			fastestD = max(fastestD, Dr);
		}
		
		// sort by lambda
		CompareReactions comparer(s);
		sort(rxns.begin(), rxns.end(), comparer);
		printf("\n");
		
		for (int i=0; i < rxns.size(); i++){
			assert(fastestD >= rxns[i]->D(s));
			double lambda = rxns[i]->D(s) / fastestD;
			double slowD = 0.0;
			for (int j=i; j < rxns.size(); j++){
				slowD += rxns[j]->D(s);
			}
			printf("%f, %f\n", slowD / D, rxns[i]->D(s) / fastestD);
		}
		
		// Print species numbers
		vector<int> ss = s->X();
		sort(ss.begin(), ss.end());
		BOOST_FOREACH(int x, ss){
			printf("%d\n", x);
		}
		
	}
	
	void _resetSlowRxnCounters(){
		_slowRxnsDdtSum = 0.0;
		_targetLogUniformVal = -log(urand());
	}
	
	// Puts reactions in _fastReactions/_slowReactions 
	//		according to current reaction rate properties
	void _partitionReactions(Reactions *root, State *s){
		if (_fastReactions) delete _fastReactions;
		if (_slowReactions) delete _slowReactions;
		_fastReactions = new Reactions(NULL);
		_slowReactions = new Reactions(NULL);
		
		// reset counters relevant to partition
		//assert(_slowRxnsDdtSum == 0.0);
		_resetSlowRxnCounters();
		
		
		// find 'knee' in efficiency ROC-like curve
		// see which reactions occur efficiencyConst times slower than the fastest reaction  Di / Dfastest
		double targetRatio = 1.0 / 9;
		
		// we'd like some slow reactions to occur ~every leap * c reaction events, with c=10 for now (?)
		// in other words: Dfast >= leap*c*Dslow
		//
		// IDEA: calculate D sum, sort by D, add to fastReactions until one before Dfast / Dslow >= leap*c
		
		//double efficiencyConst = 10; // sort of arbitrary right now
		//int leap = 3; // TODO XXX set this properly
		
		vector<Reaction*> & rxns = root->descendantReactionList();
		
		CompareReactions comparer(s);
		sort(rxns.begin(), rxns.end(), comparer);
		
		double D = 0.0;
		double fastestD = 0;
		for (int i=0; i < rxns.size(); i++){
			double Dr = rxns[i]->D(s);
			D += Dr;
			fastestD = max(fastestD, Dr);
		}
		
		// now add up D values of reactions until we get to our target D value for fast reactions
		int i=0;
		double Dfast = rxns[i]->D(s);
		double slowestFastRxnD = targetRatio * fastestD;
		
		for (int i=0; i < rxns.size(); i++){
			if (slowestFastRxnD < rxns[i]->D(s)){
				_fastReactions->addDependent(rxns[i]);
			}else{
				_slowReactions->addDependent(rxns[i]); 
			}
		}
//		while (i < rxns.size() && slowestFastRxnD < rxns[i]->D(s) ){ // Dfast + Dfast*l*c < l*c*D
//			_fastReactions->addDependent(rxns[i]);
//			i++;
//			Dfast += rxns[i]->D(s); 
//		}
//		// put the rest of reactions in slow reactions
//		while (i < rxns.size()){
//			assert(rxns[i]->D(s) < slowestFastRxnD);
//			_slowReactions->addDependent(rxns[i]); 
//			i++;
//		}
	}
	
	// could probably have this inlined
	Reaction *_reactionSelect(State *s, vector<Reaction*> & reactions, double D){
		double r = urand() * D;
		double propensitySum = reactions[0]->D(s);
		int i=0;
		while (r > propensitySum){
			i++;
			assert(i < reactions.size());
			propensitySum += reactions[i]->D(s);
		}
		return reactions[i];
	}
	
	double _slowRxnsDdtSum;
	double _targetLogUniformVal; //  -log(u) | u ~ uniform(0,1)
	
	Reactions* _fastReactions;
	Reactions* _slowReactions;
};



#endif /*TIME_HIER_SSA_H_*/
