/*
 *  time_hierleap_separate_clean.cpp
 *  gpu_hierleap
 *
 *  Created by David Orendorff on 5/4/12.
 *  Copyright 2012 UC Irvine. All rights reserved.
 *
 */


#include "time_hierleap_separate_clean.h"

#include <cmath>
#include <numeric>
#include <functional>

#include "../globals.h"

#include "tbb/blocked_range.h"
#include "tbb/parallel_for.h"

#include "../reactions/reaction_permutation.h"
#include "../reactions/propensity_bounds.h"
#include "../reactions/reaction_set.h"
#include "../reactions/reactions.h"
#include "../reactions/reaction.h"
#include "../reactions/setup_topology/build_reaction_topology.h"
#include "../reactions/delta_propensity_bounds.h"

#include "common/bounding_arrays.h"
#include "common/functors.h"
#include "common/functor_add_reaction_calculate_low_bound.h"
#include "../sim/cell_computation_task.h"

#include "../state/state.h"
#include "../state/reaction_cache.h"

#include "../utils/common.h"

using namespace tbb;
using namespace std;

#define INTEL_THREADS_UPDATE_BOUNDS 21
#define SERIAL_UPDATE_BOUNDS 77
#define DIY_THREADS_UPDATE_BOUNDS 4

#define IDEAL_GRAIN_SIZE 4


double TimeHierLeapSepClean::ReactionTauLeap(int leap, State *s, double &dt) {
	// it would be nice if we could assert that changeRuntimeParameters has been called first.
	assert(_state && _leap >= 1);
	assert(leap == _leap && s == _state);
	_recursionLevel = 0; // note: this may be incorrect.
	ReactionTauLeapPreComputedBounds(_leap, _state, dt);
	return _cellD->x[_root->getUniqueId()];
}

TimeHierLeapSepClean::TimeHierLeapSepClean(Reactions *root, State *s, int maxThreads) : 
		_debugState(0), _leap(-1), _state(NULL), _rtlPreComputedBoundsIterations(0), 
		_cellUpdatePovBounds(NULL), _sampleBlocks(NULL), _maxThreads(maxThreads), _updatePropensitiesBound(NULL), _calculateD(NULL) {
	root->assertIdsTopological(); 
	
	_root = root;
	int nRxs = root->numRxnSets() +1;
	int n = s->X().size();
	
	_deltaPropensityCompute = new ReactionSetMap<DeltaPropensityBounds*>(nRxs);
	_propensityBoundsCalc = new ReactionSetMap<PropensityBounds*>(nRxs);
	_deltaCompute = new ReactionSetMap<DeltaEquation*>(nRxs);
	
	(*_propensityBoundsCalc)[_root] = new PropensityBoundsNoCells(_root, n, false); // todo this may not work (??)
	
	// assert(exactly two levels of reactions (and then root));
	BOOST_FOREACH(ReactionSet *cellPrnt, root->getDependents()){
		assert(!cellPrnt->isReaction());
		Reactions *cell = cellPrnt->asReactions();
		(*_propensityBoundsCalc)[cellPrnt] = new PropensityBoundsNoCells(cell, n, false);
		BOOST_FOREACH(ReactionSet *baseChild, cellPrnt->getDependents()){
			assert(baseChild->isReaction());
		}
		(*_deltaPropensityCompute)[cell] = new DeltaPropensityBounds(cell);
	}
	
	// Delta computers
	vector<Reaction*> & rxn = _root->descendantReactionList();
	BOOST_FOREACH(Reaction *r, rxn) {
		(*_deltaCompute)[r] = new DeltaEquation(_root, r);
	}
	_deltaCompute->rehash(_deltaCompute->size()); // (maybe) makes lookup more effecient
	
	_pb = new PropensityBoundsNoCells(root, s->species().size(), false);	
	_cellD = new BoundingArrays<double>(_root->numRxnSets());
	_states = new BoundingArrays<int>(s->X().size());
	_metaCellEvents = vector<unsigned int>(_root->numDependents() + 1, NULL);
	
	_lastDUpdateTimestamp.resize(_root->numRxnSets(), -1);
	
	_recursionLevel = 0;
			
	int preferedMethod = DIY_THREADS_UPDATE_BOUNDS;
	_calculateCellBoundsMethod = preferedMethod;
	_blockSampleMethod = preferedMethod;
	_updateDMethod = preferedMethod;
	_updateBoundsMethod = preferedMethod;Exact and Hierarchical Reaction Leaping: Asymptotic Improvements to the Stochastic Simulation Algorithm
}

// sample from multinomial and write num events to the children of _root in state
// uses values in _cellD->highBound for probabilities
void TimeHierLeapSepClean::sampleCoarseEvents(State *s, vector<int> & highState, int leap){
	vector<ReactionSet*> & deps = _root->getDependents();
	assert(_metaCellEvents.size() == _root->numDependents() + 1);
	_metaCellEvents[_root->getUniqueId()] = leap;
	multirnd(leap, &_cellD->highBound[_root->smallestChildId()], _root->getDependents().size(), &_metaCellEvents[_root->smallestChildId()]);
	
	// set coarse events
	for (unsigned int i=0; i < deps.size(); i++){
		s->setEvents(deps[i], _metaCellEvents[deps[i]->getUniqueId()]);
	}
}

double TimeHierLeapSepClean::_computePCoarse() {
	double Pcoarse = 1.0;

	vector<ReactionSet*> & deps = _root->getDependents();
	
	for (int i=0; i < deps.size(); i++){
		Reactions *r = deps[i]->asReactions();
		int cidx = r->getUniqueId();
		int cevents = _metaCellEvents[deps[i]->getUniqueId()];
		
		double Dhigh = _cellD->highBound[cidx];
		double DhighCell = _cellD->highCellBound[cidx];
		
		double Dlow =  _cellD->lowBound[cidx]; 
		double DlowCell = _cellD->lowCellBound[cidx]; //r->DbyUniqueId(_states->lowCellBound);
		assert(Dlow <= DlowCell);
		
		
		if(cevents>0){
			Pcoarse *= ipow(DhighCell / Dhigh, cevents);
		} 
		
		// Runtime checks
		assert(DhighCell <= Dhigh);
		assert(DlowCell >= Dlow);
		assert(Pcoarse <= 1.0);
	}
	assert(_metaCellEvents[_root->getUniqueId()] > 0); 
	
	assert(Pcoarse >= 0 && Pcoarse <= 1.0);
	
	return Pcoarse;
}

// for (r : rp->previous()) r->unfire(state);
// post: rp->hasPrevious() = false, state will be as if no reactions in rp ever occured.
void TimeHierLeapSepClean::_unFire(ReactionPermutation<Reaction*> * rp, vector<int> & state) {
	while (rp->hasPrevious()) {
		Reaction *r = rp->previous();
		r->unFire(state);
	}
}

double TimeHierLeapSepClean::_acceptanceBlockProb(ReactionPermutation<Reaction*> * rp, vector<int> & state, vector<double> & highCellBound) {
	if (!rp->isLocked()) {
		rp->permute();
		rp->lockOrder();
	}
	
	double p = 1.0;
	while (rp->hasNext()) {
		Reaction *r = rp->next();
		int ridx = r->getUniqueId();
		double ds = r->DbyUniqueId(state);
		double d = ds / highCellBound[ridx];
		if (ds == 0) d = 0;
		assert(d >= 0 && d <= 1.0);
		p *= d;
		bool viable = r->isViable(state);
		if (!rp->hasNext() || !viable) {
			if (!viable)
				p = 0;
			// don't fire final reaction in order to save work when  unfiring
			rp->previous();
			break;
		}
		r->fire(state);
	}
	_unFire(rp, state);
	return p;
}

// Provides a close to actual upper bound on D (sum of propensities) but at the cost of more computational effort.
// Cell has at least one reaction
// Note: this is only valid if cells are independent!!! XXX: change from descendantReactionList to include neighboring cell rxns
double TimeHierLeapSepClean::_tightUpperBoundD(Reactions *cell, State *s, int leap) {
	// TODO: for now do a brute force search of the max deltaD, but later optimize.

	// see which rxn causes the greatest increase in D.
	double maxDelta = _getDeltaPropensityComputer(cell)->upperBoundDeltaD(_states->highBound);
	double D = _cellD->x[cell->getUniqueId()];
	
	return D + maxDelta * (leap - 1);
}

void TimeHierLeapSepClean::_calculateCellBoundsNoPov(State *s) {
	assert(_metaCellEvents[0] > 0); // assert leap > 0
	s->setEvents(_root, _metaCellEvents[_root->getUniqueId()]);
	BOOST_FOREACH(ReactionSet *cell, _root->getDependents()) {
		s->setEvents(cell, _metaCellEvents[cell->getUniqueId()]);
	}
	
	//calculate bounds on state
	_pb->upperBoundSpeciesParentEventsGiven(s, _states->highCellBound);
	_pb->lowerBoundSpeciesParentEventsGiven(s, _states->lowCellBound);
	
	// calculate propensity bounds
	BOOST_FOREACH(ReactionSet *cell, _root->getDependents()) {
		double highCellD = 0;
		double lowCellD = 0;
		int cid = cell->getUniqueId();
		assert(_metaCellEvents[cid] == s->numEvents(cell));
		
		// we'd like to do this more efficiently, but not sure how.
		BOOST_FOREACH(Reaction *r, cell->asReactions()->descendantReactionList()) {
			int rid = r->getUniqueId();
			_cellD->highCellBound[rid] = r->DbyUniqueId(_states->highCellBound);
			highCellD += _cellD->highCellBound[rid];
			
			_cellD->lowCellBound[rid] = r->DbyUniqueId(_states->lowCellBound);
			lowCellD += _cellD->lowCellBound[rid];
		}
		
		_cellD->highCellBound[cid] = min(highCellD, _cellD->highBound[cid]);
		_cellD->lowCellBound[cid] = max(lowCellD, _cellD->lowBound[cid]);
	}
	_cellD->highCellBound[_root->getUniqueId()] = sum(&_cellD->highCellBound[_root->smallestChildId()], _root->numDependents());
	_cellD->lowCellBound[_root->getUniqueId()] = _cellD->lowBound[_root->getUniqueId()]; // XXX: TODO: changed definition of _cellD->lowCellBound[_root->getUniqueId()] here.
	
}

// Uses variables belonging to *this* and stores results in _cellD and _states
// specifically, updates lowCellBound/highCellBound in _states and _cellD
void TimeHierLeapSepClean::_calculateCellBounds(State *s) {
	// set cell events, note: we'd eventually like to NOT use s->numEvents(...), etc.
	assert(_metaCellEvents[0] > 0); // assert leap > 0
	s->setEvents(_root, _metaCellEvents[_root->getUniqueId()]);
	BOOST_FOREACH(ReactionSet *cell, _root->getDependents()) {
		s->setEvents(cell, _metaCellEvents[cell->getUniqueId()]);
	}

	// calculate propensity bounds
	
	vector<ReactionSet*> & deps = _root->getDependents();
	//#pragma omp parallel for
	for (int i = 0; i < deps.size(); i++) {
		Reactions *cell = deps[i]->asReactions();
		if (s->numEvents(cell) == 0) {
			_cellD->highCellBound[cell->getUniqueId()] = _cellD->x[cell->getUniqueId()];
			_cellD->lowCellBound[cell->getUniqueId()] = _cellD->x[cell->getUniqueId()];
			
			continue;
		}
		
		// Get multinomial coefficients for children
		double cellDhighCellSum = 0.0;
		double cellDlowCellSum = 0.0;
		
		vector<int> & exactUpperBound = cell->speciesSandbox(s->speciesSize());
		_pb->upperBoundSpeciesParentEventsGivenWithCellPov(cell, s, exactUpperBound);
		BOOST_FOREACH(Reaction *r, cell->descendantReactionList()){
			_cellD->highCellBound[r->getUniqueId()] = r->DbyUniqueId(exactUpperBound);
			//_cellD->highCellBound[r->getUniqueId()] = r->DbyUniqueId(_states->highCellBound);
			cellDhighCellSum += _cellD->highCellBound[r->getUniqueId()];
		}
		
		vector<int> & exactLowerBound = exactUpperBound;
		_pb->lowerBoundSpeciesParentEventsGivenWithCellPov(cell, s, exactLowerBound);
		BOOST_FOREACH(Reaction *r, cell->descendantReactionList()){
			_cellD->lowCellBound[r->getUniqueId()] = r->DbyUniqueId(exactLowerBound);
			cellDlowCellSum += _cellD->lowCellBound[r->getUniqueId()];
		}
		
		_cellD->highCellBound[cell->getUniqueId()] = min(cellDhighCellSum, _cellD->highBound[cell->getUniqueId()]);
		_cellD->lowCellBound[cell->getUniqueId()] = max(cellDlowCellSum, _cellD->lowBound[cell->getUniqueId()]);
	}
	
	// todo: optimize this later
	_cellD->highCellBound[_root->getUniqueId()] = sum(&_cellD->highCellBound[_root->smallestChildId()], _root->numDependents());
	_cellD->lowCellBound[_root->getUniqueId()] = sum(&_cellD->lowCellBound[_root->smallestChildId()], _root->numDependents());
}

// Assumes _states->highBound, _states->lowBound, _cellD->highBound, _cellD->lowBound are pre-computed releative to s.
// otherwise, the functionallity is identical to ReactionTauLeap(...); though optimizations may abound
bool TimeHierLeapSepClean::ReactionTauLeapPreComputedBounds(int leap, State *s, double &tauSum) {
	assert(_cellD->x[0] >= _cellD->lowBound[0]);
	
	_rtlPreComputedBoundsIterations++;
	double standardUniformRand = urand();
	vector<ReactionSet*> & deps = _root->getDependents();
	
	tauSum = erlang_sample(_cellD->lowBound[_root->getUniqueId()], leap);
	if (!isfinite(tauSum)) { // 'infinite' amount of time lapses. 
		_rejectedSample(NULL, tauSum); 
		return false;
	}
	assert(tauSum >= 0);
	
	sampleCoarseEvents(s, _states->highBound, leap);
	
	//	_calculateCellBoundsNoPov(s);
	//_calculateCellBounds(s);
	int blockSampleMethod = DIY_THREADS_UPDATE_BOUNDS; //INTEL_THREADS_UPDATE_BOUNDS;
	int idealGrainSize = IDEAL_GRAIN_SIZE;// deps.size(); // 1;
	
	// Cell Bounds block POV.
	bool calcCellPovBoundsSeperatly = true;
	if (calcCellPovBoundsSeperatly){
		switch (_calculateCellBoundsMethod) {
			case INTEL_THREADS_UPDATE_BOUNDS:
				parallel_for(blocked_range<std::vector<ReactionSet*>::iterator>(deps.begin(), deps.end(), idealGrainSize), 
							 ApplyCalculateCellBoundsWithPov(this));
				break;
			case DIY_THREADS_UPDATE_BOUNDS:
				if (!_cellUpdatePovBounds) {
					vector<Reactions*> rxnsChild;
					BOOST_FOREACH(ReactionSet *r, deps) {
						rxnsChild.push_back(r->asReactions());
					}
					_cellUpdatePovBounds = new ComputationTask<ApplyCalculateCellBoundsWithPov, Reactions*>(
												new ApplyCalculateCellBoundsWithPov(this), rxnsChild, _maxThreads);
				}
				_cellUpdatePovBounds->run();
				break;
			case SERIAL_UPDATE_BOUNDS:
				_calculateCellBounds(s);
				break;
			default: assert(false);
		};
	}
	
	_cellD->lowCellBound[_root->getUniqueId()] = _cellD->lowBound[_root->getUniqueId()];
	
	floor(_states->lowCellBound, _states->lowBound);
	ceil(_states->highCellBound, _states->highBound);
	
	//assertArrayEpsilonLessThan(_cellD->lowCellBound, _cellD->x);
	
	ReactionPermutation<Reaction*> *rxns = new ReactionPermutation<Reaction*>();
	vector<ReactionPermutation<Reaction*> *> rpToJoin;
	BOOST_FOREACH(ReactionSet *cell, _root->getDependents()) {
		int cid = cell->getUniqueId();
		_state->setEvents(cell, _metaCellEvents[cid]);
		if (_metaCellEvents[cid] > 0)
			rpToJoin.push_back(cell->asReactions()->reactionPermutationSandbox());
	}
	
	// double PfineLowBound = exp(-tauSum * (_cellD->highCellBound[_root->getUniqueId()] - _cellD->lowCellBound[_root->getUniqueId()]));
	double PfineLowBound = exp(-tauSum * (_cellD->highBound[_root->getUniqueId()] - _cellD->lowBound[_root->getUniqueId()]));
	bool needPermutation = PfineLowBound <= standardUniformRand;
	
	// Sample blocks.
	bool tbbSampleFail = false;
	switch (_blockSampleMethod) {
		case INTEL_THREADS_UPDATE_BOUNDS:
			parallel_for(blocked_range<std::vector<ReactionSet*>::iterator>(deps.begin(), deps.end(), idealGrainSize), 
						 ApplyCalculateLowBoundsSampleBlock(this, needPermutation, tbbSampleFail));
			if (tbbSampleFail) {
				return _recurseOrReject(NULL, tauSum, leap, s);
			}
			break;
		case DIY_THREADS_UPDATE_BOUNDS:
			if (!_sampleBlocks) {
				vector<Reactions*> rxnsChild;
				BOOST_FOREACH(ReactionSet *r, deps) {
					rxnsChild.push_back(r->asReactions());
				}
				_sampleBlocks = new ComputationTask<ApplyCalculateLowBoundsSampleBlock, Reactions*>(
						new ApplyCalculateLowBoundsSampleBlock(this, needPermutation, tbbSampleFail), rxnsChild, _maxThreads);
			}
			_sampleBlocks->run(); 
			if (tbbSampleFail) {
				return _recurseOrReject(NULL, tauSum, leap, s);
			}
			
			break;
		case SERIAL_UPDATE_BOUNDS:
			BOOST_FOREACH(ReactionSet *cell, _root->getDependents()) {
				int cellEvents = _metaCellEvents[cell->getUniqueId()];
				if (!cellEvents) {
					continue;
				}
				if (_cellD->highCellBound[cell->getUniqueId()] == 0){
					return _recurseOrReject(NULL, tauSum, leap, s);
				}
				
				ReactionPermutation<Reaction*> *rp = cell->asReactions()->reactionPermutationSandbox();
				ApplyAddReactionCalculateLowBound apar(rp, _root, _cellD); // allows for efficiency and thread safety
				while (true) {
					rp->clear();
					apar.setParentCell(cell->asReactions());
					multirnd_callback(cellEvents, &_cellD->highCellBound[cell->smallestChildId()], cell->numDependents(), apar);
					double u = urand();
					if (u <= apar.getLowerAcceptanceBound())
						break;
					rp->permute();
					if (u <= _acceptanceBlockProb(rp, s->X(), _cellD->highCellBound))
						break;
					apar.resetLowerAcceptanceBound();
				}
				if (needPermutation && !rp->isLocked())
					rp->permute();
			}
			break;
		default : assert(false);
	};
	
	
	// compute p coarse from _metaCellEvents
	double Pcoarse = _computePCoarse();
	bool earlyRejectWithPcoarse = true;
	if (earlyRejectWithPcoarse){ 
		if (urand() > Pcoarse) {
			return _recurseOrReject(NULL, tauSum, leap, s);
		}
		Pcoarse = 1.0;
	}
	
	rxns->joinPermutations(rpToJoin);
	assert(rxns->numEvents()==leap);
	if (!needPermutation) {
		//rxns->joinPermutations(rpToJoin); // todo: need 'dumb/fast' join permutation.
		return _acceptedSample(rxns, tauSum, s, leap);
	}
	
	//rxns->joinPermutations(rpToJoin);
	assert(rxns->numEvents()==leap);
	
	_Paccept = Pcoarse * PfineLowBound;
	
	// sample taus on cell level, only if not early accept
	double tsum = 0.0;
	for (int i=0; i < leap; i++) {
		double r = erand();
		tsum += r;
		_taus.push_back(r);
	}

	
	// Compute Pfine if no early acceptance
	vector<int> initialState = s->X(); 
	double lowD = _cellD->lowBound[_root->getUniqueId()];
	double Pfine = 1.0;
	double expSumFine = 0.0;
	double rootD = _cellD->x[_root->getUniqueId()];
	while (rxns->hasNext() && Pfine > 0){
		
		assertEpsilonLessThan(lowD, rootD);
		for (int i = 0; i < _state->X().size(); i++) {
			assert(_state->x(i) >= _states->lowBound[i]);
		}
		
		
		Reaction *r = rxns->next();
		if (!r->isViable(s)) {
			rxns->previous(); // put rxn back into permutation so we don't have to 'undo' the work.
			Pfine = 0.0;
			break;
		}
		
		int rid = r->getUniqueId();
		double dHighCell = _cellD->highCellBound[rid]; 
		double d = r->D(s);
		Pfine *= d / dHighCell; 
		
		if (Pfine==0) {
			rxns->previous();
			break;
		}
		
		double tau = _taus.back(); _taus.pop_back();
		expSumFine += -tau*(rootD - lowD); 
		
		rootD += _getDeltaComputer(r)->eval(s->X()); 
		
		r->fire(s); // changes state
		
		assert(d <= dHighCell);
		assert(d >= _cellD->lowCellBound[rid]);
		assert(Pfine >= 0 && Pfine <= 1);
	}
	
	
	// Pfine *= exp(expSumFine * tauSum / tsum); TODO
	Pfine = exp(expSumFine * tauSum / tsum);	
	assert(Pfine >= 0 && Pfine <= 1);
	
	_Paccept = Pcoarse * Pfine;
	if (Pcoarse * Pfine < standardUniformRand) {
		// reject sample
		//rxns->resumeCheckpoint();
		while (rxns->hasPrevious()){
			Reaction *r = rxns->previous();
			r->unFire(s);
		}
		// also rootD needs to be set
		delete rxns;
		rxns = NULL;
		assert(std::equal(s->X().begin(), s->X().end(), initialState.begin()));
		return _recurseOrReject(rxns, tauSum, leap, s);
//		if (_recursionLevel > 9){
//			//printf("Hierleap Recursion: %d, running SSA instead: \n", _recursionLevel);
//			printf("NoGo ");
//			_recursionLevel = 0;
//			tauSum = 0;
//			return _rejectedSample(rxns, tauSum);
//		}
//		// no need to recalculate bounds
//		return ReactionTauLeapPreComputedBounds(leap, s, tauSum);
	}
	
	// update _cellD and state (highBound, lowBound) based on rxns
	return _acceptedSample(rxns, tauSum, s, leap);
}

// Returns the delta computer for r wrt the entire system propensity.
DeltaEquation* TimeHierLeapSepClean::_getDeltaComputer(Reaction *r) {
	return checkNotNull((*_deltaCompute)[r]);
}

DeltaPropensityBounds* TimeHierLeapSepClean::_getDeltaPropensityComputer(Reactions *r) {
	return checkNotNull((*_deltaPropensityCompute)[r]);
}

// Updates _cellD->x; if rxns != NULL assumes _cellD->x was correct prior to rxns being exectued. changes D for rxn 'related' to rxns. 
//
// @arg: UpdateTimestamp is used to keep track of what reactions fired between the previous timeStamp.
//     The only requirement of updateTimestamp is that it be a different value when called with different s and the 
//     value should not be too periodic.
//
// makes changes to _cellD->x[_root], _cellD->x[_cells] and _cellD->x[rxns<..>()->effectedBy()]
void TimeHierLeapSepClean::_updateD(State *s, ReactionPermutation<Reaction*> *rxns, int updateTimestamp) {
	int updateDMethod = DIY_THREADS_UPDATE_BOUNDS;
	vector<double> & D = _cellD->x;
	double & rootD = _cellD->x[_root->getUniqueId()];
	if (rxns && false) { // TODO, added for testing
		list<Reactions*> cellsToRecompute;
		BOOST_FOREACH(Reaction *r, rxns->_myReactionList()) {
			BOOST_FOREACH(Reaction *r2, *r->changesPropensityFor()) {
				int rid = r2->getUniqueId();
				if (_lastDUpdateTimestamp[rid] == updateTimestamp) 
					continue; // q: is this actually faster than re-doing the calculation?
				
				D[rid] = r2->DbyUniqueId(s->X());
				_lastDUpdateTimestamp[rid] = updateTimestamp;
				
				// put cell on recalculate queue if it's not already there.
				int cid = r2->getParent()->getUniqueId();
				if (_lastDUpdateTimestamp[cid] != updateTimestamp) {
					_lastDUpdateTimestamp[cid] = updateTimestamp;
					cellsToRecompute.push_front(r2->getParent());;
				}
			}
			BOOST_FOREACH(Reactions *cell, cellsToRecompute) {
				D[cell->getUniqueId()] = sum(&D[cell->smallestChildId()], cell->numDependents());
			}
		}
	} else {
		
		switch (updateDMethod) {
			case SERIAL_UPDATE_BOUNDS:  
				// compute all D from scratch
				rootD = 0.0;
				BOOST_FOREACH(Reaction *r, _root->descendantReactionList()) {
					double d = r->DbyUniqueId(s->X());
					int rid = r->getUniqueId();
					D[rid] = d;
					rootD += d;
					_lastDUpdateTimestamp[rid] = updateTimestamp;
				}
				BOOST_FOREACH(ReactionSet *cell, _root->getDependents()) {
					int cid = cell->getUniqueId();
					D[cid] = sum(&D[cell->smallestChildId()], cell->numDependents());
				}
				break;
			case DIY_THREADS_UPDATE_BOUNDS:
				if (!_calculateD) {
					vector<Reactions*> rxnsChild;
					BOOST_FOREACH(ReactionSet *r, _root->getDependents()) {
						rxnsChild.push_back(r->asReactions());
					}
					_calculateD = new ComputationTask<ApplyCalculateD, Reactions*>(new ApplyCalculateD(&_state->X(), &_cellD->x), rxnsChild, _maxThreads);
				}
				_calculateD->run();
		}
	}
	
	// always recompute D[root] from cell summation.
	D[_root->getUniqueId()] = sum(&D[_root->smallestChildId()], _root->numDependents());
	assertNearlyEqual(D[0], _root->DbyUniqueId(s->X()));
}

// can update bounds in parallel
void TimeHierLeapSepClean::_updateBounds( State *s, ReactionPermutation<Reaction*> *rxns, int leap) { 
	//void TimeHierLeapSepClean::_serialUpdateBounds( State *s, ReactionPermutation<Reaction*> *rxns, int leap) { 
	_pb->highBoundState(s->X(), leap, _states->highBound);
	_pb->lowBoundState(s->X(), leap, _states->lowBound);
	
	_finalReactions = rxns;
	int idealGrainSize = IDEAL_GRAIN_SIZE;
	vector<ReactionSet*> & deps = _root->getDependents();
	
	switch (_updateBoundsMethod) {
		case INTEL_THREADS_UPDATE_BOUNDS:
			parallel_for(blocked_range<std::vector<ReactionSet*>::iterator>(deps.begin(), deps.end(), idealGrainSize), 
						 ApplyCellBoundUpdate(_getDeltaPropensityComputerAccesor(), 
											  _states, _cellD, &_finalReactions, leap));
			break;
		case DIY_THREADS_UPDATE_BOUNDS:
			if (!_updatePropensitiesBound) {
				vector<Reactions*> rxnsChild;
				BOOST_FOREACH(ReactionSet *r, deps) {
					rxnsChild.push_back(r->asReactions());
				}
				_updatePropensitiesBound = new ComputationTask<ApplyCellBoundUpdate, Reactions*>(new ApplyCellBoundUpdate(_getDeltaPropensityComputerAccesor(), 
																														 _states, _cellD, &_finalReactions, leap), 
																								rxnsChild, _maxThreads) ;
			}
			
			_updatePropensitiesBound->run(); 
			break;
		case SERIAL_UPDATE_BOUNDS:
			//#pragma omp parallel for
			for (int i = 0; i < deps.size(); i++) {
				ReactionSet *cell = deps[i];
				int cid = cell->getUniqueId();
				_getDeltaPropensityComputer(cell->asReactions())->updateUpperBoundDeltaDs(rxns, _states->highBound);
				double maxPosDelta =_getDeltaPropensityComputer(cell->asReactions())->upperBoundDeltaD(_states->highBound);
				double maxNegDelta = _getDeltaPropensityComputer(cell->asReactions())->lowerBoundDeltaD(_states->highBound);
				assert(maxNegDelta <= 0);
				
				_cellD->highBound[cell->getUniqueId()] = _cellD->x[cid] + maxPosDelta * (leap - 1); 
				_cellD->lowBound[cell->getUniqueId()] = max(_cellD->x[cid] + maxNegDelta * (leap - 1), 0.0); 
			}
			break;
		default:
			assert(false);
	} 
	
	_finalReactions = NULL;
	
	
	_cellD->highBound[_root->getUniqueId()] = sum(&_cellD->highBound[_root->smallestChildId()], _root->getDependents().size());
	_cellD->lowBound[_root->getUniqueId()] = sum(&_cellD->lowBound[_root->smallestChildId()], _root->getDependents().size());
}

// assuming rxns occur, which lead to s, compute bound on _cellD and _state (highBound, lowBound)
void TimeHierLeapSepClean::_serialUpdateBounds(State *s, ReactionPermutation<Reaction*> *rxns, int leap) {
	
	// not optimal, but good enough for now
	_pb->highBoundState(s->X(), leap, _states->highBound);
	_pb->lowBoundState(s->X(), leap, _states->lowBound);
	
	BOOST_FOREACH(ReactionSet *cell, _root->getDependents()) {
		_cellD->highBound[cell->getUniqueId()] = _tightUpperBoundD(cell->asReactions(), s, leap); // todo: not tight bounds when confirming correctness
		_cellD->lowBound[cell->getUniqueId()] = cell->DbyUniqueId(_states->lowBound);
	}
	_cellD->highBound[_root->getUniqueId()] = sum(&_cellD->highBound[_root->smallestChildId()], _root->getDependents().size());
	_cellD->lowBound[_root->getUniqueId()] = sum(&_cellD->lowBound[_root->smallestChildId()], _root->getDependents().size());
}

bool TimeHierLeapSepClean::_recurseOrReject(ReactionPermutation<Reaction*> *rxns, double &dt, int leap, State *s) {
	_taus.clear();
	if (rxns) 
		delete rxns;
	_recursionLevel++;
	if (_recursionLevel > 9){
		//printf("Hierleap Recursion: %d, running SSA instead: \n", _recursionLevel);
		printf("NoGo ");
		return _rejectedSample(rxns, dt);
	}
	return ReactionTauLeapPreComputedBounds(leap, s, dt);
}

bool TimeHierLeapSepClean::_rejectedSample(ReactionPermutation<Reaction*> *rxns, double &dt) {
	_taus.clear();
	if (rxns) 
		delete rxns;
	_recursionLevel = 0;
	dt = 0;
	return false;
}

bool TimeHierLeapSepClean::_acceptedSample(ReactionPermutation<Reaction*> *rxns, double &dt, State *s, int leap) {
	_recursionLevel = 0;
	
	if (rxns->hasNext()) {
		// apply rxns, then update bounds
		//foobar;
		while (rxns->hasNext()) {
			rxns->next()->fire(s);
		}
	}
	_updateD(s, rxns, _rtlPreComputedBoundsIterations++);
	_updateBounds(s, rxns, leap);
	
	
	
	if (rxns) delete rxns;
	assert(dt != 0);
	assert(isfinite(dt));
	return true;
}
