/*
 *  time_hierleap_separate_clean.h
 *  gpu_hierleap
 *
 *  Created by David Orendorff on 5/4/12.
 *  Copyright 2012 UC Irvine. All rights reserved.
 *
 */

#ifndef TIME_HIERLEAP_SEPARATE_CLEAN_H_
#define TIME_HIERLEAP_SEPARATE_CLEAN_H_

#include <vector>
#include <functional>

#include "reaction_tau_leaper.h"
#include "common.h"

using namespace std;

class PropensityBoundsNoCells;
class PropensityBounds;
class ApplyCellBoundUpdate;
class ApplyCalculateLowBoundsSampleBlock;
class ApplyCalculateCellBoundsWithPov;
class ApplyCalculateD;

class Reactions;
class ReactionSet;
template <class CellExecutor, class ExecutorOnMeObj>
class ComputationTask;
class State;
template <class ValueType> class BoundingArrays;
template <class D> class ReactionSetMap;

class DeltaPropensityBounds;
class DeltaEquation;

/*
 * ER-leap on meta-reactions, refined to reactions and 
 * then calculate probability of acceptance.
 * 
 * As of 10/27/11 this class is based on V5 of the MS.
 */
class TimeHierLeapSepClean : public ReactionTauLeaper {
	
public:
	
	TimeHierLeapSepClean(Reactions *root, State *s, int maxThreads);
	
	bool ReactionTauLeapPreComputedBounds(int leap, State *s, double &dt);
	
	virtual double ReactionTauLeap(int leap, State *s, double &dt);
	
	virtual char* name(){ return "TimeHierLeapSepClean"; }
	
	// If changes to leap or s are made (not by this object); call this function before (resuming) simulation.
	virtual void changeRuntimeParameters(State *s, int leap) {
		_state = checkNotNull(s);
		_leap = leap;
		_updateD(s, NULL, _rtlPreComputedBoundsIterations++);
		_updateBounds(s, NULL, leap);	
	}
	
	void sampleCoarseEvents(State *s, vector<int> & highState, int leap);
	
	double getPreviousAcceptanceProbability() {
		return _Paccept;
	}
	
	friend class ApplyCellBoundUpdate;
	friend class ApplyCalculateCellBoundsWithPov;
	friend class ApplyCalculateLowBoundsSampleBlock;
private:
	State *_state;
	int _leap;
	int _rtlPreComputedBoundsIterations;
	vector<int> _lastDUpdateTimestamp;
	ReactionPermutation<Reaction*> * _finalReactions;
	ComputationTask<ApplyCalculateD, Reactions*> * _calculateD;
	ComputationTask<ApplyCellBoundUpdate, Reactions*> * _updatePropensitiesBound;
	ComputationTask<ApplyCalculateCellBoundsWithPov, Reactions*> * _cellUpdatePovBounds;
	ComputationTask<ApplyCalculateLowBoundsSampleBlock, Reactions*> * _sampleBlocks;
	double _Paccept;
	
	boost::function<DeltaPropensityBounds* (Reactions*)> _getDeltaPropensityComputerAccesor() {
		return std::bind1st(std::mem_fun(&TimeHierLeapSepClean::_getDeltaPropensityComputer), this);
	}
	
	void _changeLeap(State *s, int leap) {
		_leap = leap;
		_updateBounds(s, NULL, leap);
	}
	
	// updates bounds in parallel
	void _updateBounds( State *s, ReactionPermutation<Reaction*> *rxns, int leap);
	
	void _serialUpdateBounds(State *s, ReactionPermutation<Reaction*> *rxns, int leap); 
	
	void _updateD(State *s, ReactionPermutation<Reaction*> *rxns, int updateTimestamp);
	
	void _unFire(ReactionPermutation<Reaction*> * rp, vector<int> & state);
	
	double _computePCoarse();
	
	double _acceptanceBlockProb(ReactionPermutation<Reaction*> * rp, vector<int> & state, vector<double> & highCellBound);
	
	void _calculateCellBounds(State *s);
	
	void _calculateCellBoundsNoPov(State *s);
	
	DeltaPropensityBounds* _getDeltaPropensityComputer(Reactions *r);
	
	DeltaEquation* _getDeltaComputer(Reaction *r);
	
	// Provides a close to actual upper bound on D (sum of propensities) but at the cost of more computational effort.
	// Cell has at least one reaction
	double _tightUpperBoundD(Reactions *cell, State *s, int leap);
	
	bool _acceptedSample(ReactionPermutation<Reaction*> *rxns, double &dt, State *s, int leap);
	
	bool _recurseOrReject(ReactionPermutation<Reaction*> *rxns, double &dt, int leap, State *s);
	
	// Reject the sample and do the appropriate 'book keeping'
	bool _rejectedSample(ReactionPermutation<Reaction*> *rxns, double &dt);
	
	ReactionSetMap<DeltaPropensityBounds *> *_deltaPropensityCompute;
	ReactionSetMap<PropensityBounds *> *_propensityBoundsCalc;
	ReactionSetMap<DeltaEquation*> *_deltaCompute;
	
	// avoid allocations during execution
	BoundingArrays<double> * _cellD;
	
	BoundingArrays<int> * _states;
	vector<unsigned int> _metaCellEvents; // of size 1 + root->getDependents().size()
	
	vector<double> _taus;
	
	PropensityBoundsNoCells *_pb;
	Reactions *_root;
	int _recursionLevel; 
	int _debugState; // For debugging, so we can stop at a certain instance in time
	int _maxThreads;
	
	int _calculateCellBoundsMethod;
	int _blockSampleMethod;
	int _updateDMethod;
	int _updateBoundsMethod;
	
};

#endif /*TIME_HIERLEAP_H_*/