/*
 *  time_hierleap_seperate_accept.cpp
 *  gpu_hierleap
 *
 *  Created by David Orendorff on 2/21/12.
 *  Copyright 2012 UC Irvine. All rights reserved.
 *
 */

#include "time_hierleap_seperate_accept.h"

#include <math.h>
#include <float.h>
#include <numeric>
#include <functional>

#include "../globals.h"

#include "tbb/blocked_range.h"
#include "tbb/parallel_for.h"

#include "../reactions/reaction_permutation.h"
#include "../reactions/propensity_bounds.h"
#include "../reactions/reaction_set.h"
#include "../reactions/reactions.h"
#include "../reactions/reaction.h"
#include "../reactions/setup_topology/build_reaction_topology.h"
#include "../reactions/delta_propensity_bounds.h"

#include "common/bounding_arrays.h"
#include "common/functors.h"
#include "../sim/cell_computation_task.h"

#include "../state/state.h"
#include "../state/reaction_cache.h"

#include "../utils/common.h"

using namespace tbb;

double TimeHierLeapSeperateAccept::ReactionTauLeap(int leap, State *s, double &dt) {
	// it would be nice if we could assert that changeRuntimeParameters has been called first.
	assert(_state && _leap >= 1);
	assert(leap == _leap && s == _state);
	_recursionLevel = 0; 
	if (!ReactionTauLeapPreComputedBounds(_leap, _state, dt)) {
		// no sample; D may have in fact be 0, but not seen as such due to rounding error. Recalculate.
		_updateD(_state, NULL, _rtlPreComputedBoundsIterations++);
		assert(dt == 0);
	}
	
	return _cellD->x[_root->getUniqueId()];
}

TimeHierLeapSeperateAccept::TimeHierLeapSeperateAccept(Reactions *root, State *s) : _debugState(0), _leap(-1), _state(NULL), _rtlPreComputedBoundsIterations(0) {
	root->assertIdsTopological(); 
	_getDeltaPropensityComputerPtr = std::bind1st(std::mem_fun(&TimeHierLeapSeperateAccept::_getDeltaPropensityComputer), this);
	
	_root = root;
	int nRxs = root->numRxnSets() +1;
	int n = s->X().size();
	
	_deltaPropensityCompute = new ReactionSetMap<DeltaPropensityBounds*>(nRxs);
	_propensityBoundsCalc = new ReactionSetMap<PropensityBounds*>(nRxs);
	_deltaCompute = new ReactionSetMap<DeltaEquation*>(nRxs);
	
	(*_propensityBoundsCalc)[_root] = new PropensityBoundsNoCells(_root, n, false); // todo this may not work (??)
	
	
	// assert(exactly two levels of reactions (and then root));
	BOOST_FOREACH(ReactionSet *cellPrnt, root->getDependents()){
		assert(!cellPrnt->isReaction());
		Reactions *cell = cellPrnt->asReactions();
		(*_propensityBoundsCalc)[cellPrnt] = new PropensityBoundsNoCells(cell, n, false);
		BOOST_FOREACH(ReactionSet *baseChild, cellPrnt->getDependents()){
			assert(baseChild->isReaction());
		}
		(*_deltaPropensityCompute)[cell] = new DeltaPropensityBounds(cell);
	}
	
	// Delta computers
	vector<Reaction*> & rxn = _root->descendantReactionList();
	BOOST_FOREACH(Reaction *r, rxn) {
		(*_deltaCompute)[r] = new DeltaEquation(_root, r);
	}
	_deltaCompute->rehash(_deltaCompute->size()); // (maybe) makes lookup more effecient
	
	_pb = new PropensityBoundsNoCells(root, s->species().size(), false);	
	_cellD = new BoundingArrays<double>(_root->numRxnSets());
	_states = new BoundingArrays<int>(s->X().size());
	_metaCellEvents = vector<unsigned int>(_root->numDependents() + 1, NULL);
	
	_lastDUpdateTimestamp.resize(_root->numRxnSets(), -1);
	
	_recursionLevel = 0;
	
	_updateCellPropensities = NULL; 
	
	/////////// speed tests to determine efficiency of propensity calculations
	//	int nSamples = 1000;
	//	copy(s->X().begin(), s->X().end(), _states->x.begin());
	//	vector<int> * st = &_states->x;
	//	vector<double> * d = &_cellD->x;
	//	vector<ReactionSet*> & deps = _root->getDependents();
	//	
	//	int method = 1;
	//	for (int idealGrainSize = 1; idealGrainSize < _root->numDependents(); idealGrainSize *= 2) {
	//	ApplyCalculateD acd = ApplyCalculateD(st, d);
	//	blocked_range<std::vector<ReactionSet*>::iterator> range(deps.begin(), deps.end(), idealGrainSize);
	//	
	//	struct timeval start, end;
	//	gettimeofday(&start, NULL);
	//
	//	for (int i = 0; i < nSamples; i++) {
	//		switch (method) {
	//			case 0:
	//				_updateD(s, NULL, _rtlPreComputedBoundsIterations++);
	//				break;
	//			case 1:
	//				parallel_for(range, acd);
	//				(*d)[_root->getUniqueId()] = sum(&((*d)[_root->smallestChildId()]), _root->numDependents()); 
	//				break;
	//			case 2:
	////				if (!_updateCellPropensities) {
	////					vector<Reactions*> rxnsChild;
	////					BOOST_FOREACH(ReactionSet *r, deps) {
	////						rxnsChild.push_back(r->asReactions());
	////					}
	////					_updateCellPropensities = new ComputationTask<ApplyCellBoundUpdate, Reactions*>(new ApplyCellBoundUpdate(this, leap), rxnsChild) ;
	////				}
	////				
	////				_updateCellPropensities->run(); 
	//				break;
	//		}
	//	}
	//	gettimeofday(&end, NULL);
	//	double deltaTime = elapsedSeconds(start, end);
	//	printf("Average time for grainsize %d to calculate propensity: %f sec. \n", idealGrainSize, deltaTime / nSamples);
	//	}
	//	exit(11);
	
	/////////// end speed tests.
}

// sample from multinomial and write num events to the children of _root in state
// uses values in _cellD->highBound for probabilities
void TimeHierLeapSeperateAccept::sampleCoarseEvents(State *s, vector<int> & highState, int leap){
	vector<ReactionSet*> & deps = _root->getDependents();
	
	//vector<unsigned int> cellEvents = multinomial(cellProbs, leap);
	// multirnd(unsigned int num_samples, double *probs, unsigned int num_probs, unsigned int *results);
	assert(_metaCellEvents.size() == _root->numDependents() + 1);
	_metaCellEvents[_root->getUniqueId()] = leap;
	multirnd(leap, &_cellD->highBound[_root->smallestChildId()], _root->getDependents().size(), &_metaCellEvents[_root->smallestChildId()]);
	
	// set coarse events
	for (unsigned int i=0; i < deps.size(); i++){
		s->setEvents(deps[i], _metaCellEvents[deps[i]->getUniqueId()]);
	}
}

double TimeHierLeapSeperateAccept::_computePCoarse(double tauSum) {
	double Pcoarse = 1.0;
	vector<ReactionSet*> & deps = _root->getDependents();
	
	for (int i=0; i < deps.size(); i++){
		Reactions *r = deps[i]->asReactions();
		int cidx = r->getUniqueId();
		int cevents = _metaCellEvents[cidx];
		
		if (!cevents) 
			continue;
		
		double DhighCell = _cellD->highCellBound[cidx];
		double Dhigh = _cellD->highBound[cidx];
		if(cevents>0){
			Pcoarse *= ipow(DhighCell / Dhigh, cevents);
		} 
		
		// Runtime checks
		assert(DhighCell <= Dhigh);
		assert(Pcoarse <= 1.0);
	}
	assert(_metaCellEvents[_root->getUniqueId()] > 0); 
	assert(Pcoarse >= 0 && Pcoarse <= 1.0);
	
	return Pcoarse;
}

// Provides a close to actual upper bound on D (sum of propensities) but at the cost of more computational effort.
// Cell has at least one reaction
// Note: this is only valid if cells are independent!!! XXX: change from descendantReactionList to include neighboring cell rxns
double TimeHierLeapSeperateAccept::_tightUpperBoundD(Reactions *cell, State *s, int leap) {
	// see which rxn causes the greatest increase in D.
	double maxDelta = _getDeltaPropensityComputer(cell)->upperBoundDeltaD(_states->highBound);
	double D = _cellD->x[cell->getUniqueId()];
	
	return D + maxDelta * (leap - 1);
}

// Uses variables belonging to *this* and stores results in _cellD and _states
// specifically, updates lowCellBound/highCellBound in _states and _cellD
void TimeHierLeapSeperateAccept::_calculateCellBounds(State *s) {
	// set cell events, note: we'd eventually like to NOT use s->numEvents(...), etc.
	assert(_metaCellEvents[0] > 0); // assert leap > 0
	s->setEvents(_root, _metaCellEvents[_root->getUniqueId()]);
	BOOST_FOREACH(ReactionSet *cell, _root->getDependents()) {
		s->setEvents(cell, _metaCellEvents[cell->getUniqueId()]);
	}

	// calculate propensity bounds
	
	vector<ReactionSet*> & deps = _root->getDependents();
	//BOOST_FOREACH(ReactionSet *c, _root->getDependents()) {
	//#pragma omp parallel for
	for (int i = 0; i < deps.size(); i++) {
		Reactions *cell = deps[i]->asReactions();
		if (s->numEvents(cell) == 0) {
			_cellD->highCellBound[cell->getUniqueId()] = _cellD->x[cell->getUniqueId()];
			_cellD->lowCellBound[cell->getUniqueId()] = _cellD->x[cell->getUniqueId()];
			continue;
		}
		
		// Get multinomial coefficients for children
		double cellDhighCellSum = 0.0;
		double cellDlowCellSum = 0.0;
		
		vector<int> & exactUpperBound = cell->speciesSandbox(s->speciesSize());
		_pb->upperBoundSpeciesParentEventsGivenWithCellPov(cell, s, exactUpperBound);
		BOOST_FOREACH(Reaction *r, cell->descendantReactionList()){
			_cellD->highCellBound[r->getUniqueId()] = r->DbyUniqueId(exactUpperBound);
			cellDhighCellSum += _cellD->highCellBound[r->getUniqueId()];
		}
		
		vector<int> & exactLowerBound = exactUpperBound;
		_pb->lowerBoundSpeciesParentEventsGivenWithCellPov(cell, s, exactLowerBound);
		BOOST_FOREACH(Reaction *r, cell->descendantReactionList()){
			_cellD->lowCellBound[r->getUniqueId()] = r->DbyUniqueId(exactLowerBound);
			cellDlowCellSum += _cellD->lowCellBound[r->getUniqueId()];
		}
		
		// XXX: this may cause bias (?)
		_cellD->highCellBound[cell->getUniqueId()] = min(cellDhighCellSum, _cellD->highBound[cell->getUniqueId()]);
		_cellD->lowCellBound[cell->getUniqueId()] = max(cellDlowCellSum, _cellD->lowBound[cell->getUniqueId()]);
	}
	
	// todo: optimize this later
	_cellD->highCellBound[_root->getUniqueId()] = sum(&_cellD->highCellBound[_root->smallestChildId()], _root->numDependents());
	_cellD->lowCellBound[_root->getUniqueId()] = sum(&_cellD->lowCellBound[_root->smallestChildId()], _root->numDependents());
}

// for (r : rp->previous()) r->unfire(state);
// post: rp->hasPrevious() = false, state will be as if no reactions in rp ever occured.
void TimeHierLeapSeperateAccept::_unFire(ReactionPermutation<Reaction*> * rp, vector<int> & state) {
	while (rp->hasPrevious()) {
		Reaction *r = rp->previous();
		r->unFire(state);
	}
}

// Assumes _metaCellEvents is properly initionalized wrt u.
// Note: for efficiency the returned object is reused later, so DO NOT delete it.
// states->x upon return will be the value of states->x upon entering this method.
// The returned ReactionPermutation may or may not have locked order.
// pAccept will have the acceptance probability (or a lower bound) upon return.
ReactionPermutation<Reaction*> * TimeHierLeapSeperateAccept::_sampleBlock(Reactions *cell, double & pAccept) {
	int cellEvents = _metaCellEvents[cell->getUniqueId()];
	if (!cellEvents) {
		pAccept = 1.0;
		return NULL;
	} 
	ReactionPermutation<Reaction*> *rp = cell->reactionPermutationSandbox();
	vector<int> & s = _state->X();
	
	bool useSSA = false;
	
#ifdef DEBUG
	vector<int> stateBackup = s; 
	assert(stateBackup == _state->X()); 
#endif
	
	if (useSSA) {
		rp->clear();
		_ssaSampleBlock(cell, cellEvents, _state->X(), rp);
		// todo: not sure what to do with pAccept here...
		_unFire(rp, _state->X());
#ifdef DEBUG
		assert(stateBackup == s);
		assert(stateBackup == _state->X());
#endif
	}
	
	// this condition is actually ok, but it means we reject the sample. if 
	// this assert crops up too often then code in the appropriate behavior.
	assert(_cellD->highCellBound[cell->getUniqueId()] > 0);
	
	// Note: we may want to add obj to Reactions cache so we don't have to make
	//	 a new one of these upon every block sample.
	ApplyAddReactionCalculateLowBound aarclb(rp, _root, _cellD); 
	aarclb.setParentCell(cell);
	
	int MAX_ITER = 8;
	for (int i = 0; i < MAX_ITER; i++){
		rp->clear();
		if (cellEvents == 1) {
			// optimization (boring)
			int ridx = select(&_cellD->x[cell->smallestChildId()], cell->numDependents(), _cellD->x[cell->getUniqueId()]);
			aarclb(ridx, 1);
		} else {
			multirnd_callback(cellEvents, &_cellD->highCellBound[cell->smallestChildId()], cell->numDependents(), aarclb);
		}
		rp->permute(); // not needed for early accept.
		
		if (aarclb.getLowerAcceptanceBound() == 1.0 ) {
			pAccept = 1.0;
			return rp;
		}
		double u = urand();
		pAccept = aarclb.getLowerAcceptanceBound();
		if (u <= aarclb.getLowerAcceptanceBound()) 
			return rp;
		
		// calculate acceptance as PROD(fi/fi_bar)
		pAccept = _acceptanceBlockProb(rp, s, _cellD->highCellBound);
		
		assert(pAccept >= aarclb.getLowerAcceptanceBound());
		assert(pAccept >= 0 && pAccept <= 1.0);
#ifdef DEBUG 
		assert(stateBackup == s);
		assert(stateBackup == _state->X());
		
#endif

		if (u <= pAccept)
			return rp;

		
		aarclb.resetLowerAcceptanceBound();
	}
	// assert(false); // we rejected too many times, it is inefficient, and there may be a more sinister problem.
	return NULL;
}

double TimeHierLeapSeperateAccept::_acceptanceBlockProb(ReactionPermutation<Reaction*> * rp, vector<int> & state, vector<double> & highCellBound) {
	double p = 1.0;
	while (rp->hasNext()) {
		Reaction *r = rp->next();
		int ridx = r->getUniqueId();
		double d = r->DbyUniqueId(state) / highCellBound[ridx];
		assert(d >= 0 && d <= 1.0);
		p *= d;
		bool viable = r->isViable(state);
		if (!rp->hasNext() || !viable) {
			if (!viable)
				p = 0;
			// don't fire final reaction in order to save work when  unfiring
			rp->previous();
			break;
		}
		r->fire(state);
	}
	_unFire(rp, state);
	return p;
}

// Completes <events> reaction in parent according to probabilities as calculated by SSA.
//
// Pre:
//	- event > 0
//  - rp NOT NULL and rp->clear() has been called
//
// Post: 
//	- rp will all be unfired and locked
//	- does NOT modify s
//  - rp will contain events Reaction*
void TimeHierLeapSeperateAccept::_ssaSampleBlock(Reactions *parent, int events, vector<int> & s, ReactionPermutation<Reaction*> *rp) {
	assert(parent); assert(events > 0); assert(rp); assert(!rp->numEvents());
	
	vector<Reaction*> & rxn = parent->descendantReactionList();
	
	rp->lockOrder();
	rp->beginInsertFiredReactionMode();
	while (events--) {
		double targetSum = parent->DbyUniqueId(s) * urand();
		
		int ridx = 0;
		double sum=0;
		for (  ;; ridx++) {
			sum += rxn[ridx]->DbyUniqueId(s);
			if (sum >= targetSum) break;
		}
		
		assert(ridx < rxn.size());
		Reaction *r = rxn[ridx];
		
		// add r to reaction permutation, fire and adjust accordingly
		rp->addFiredReaction(r); // better to have ~ rp->addFiredRxn(r);
		if (events == 0) {
			break;
		}
		r->fire(s);
	}
	rp->endInsertFiredReactionMode(); // all rxn in RP are mared as fired.
	rp->previous(); // we didn't actually fire the final rxn, so skip it.
	while (rp->hasPrevious()) {
		rp->previous()->unFire(s);
	}
}

/*
 * Pre: Assumes states->upperBound and states->lowerBound are filled out correctly.
 * Post: cellD->upperBound and cellD->lowerBound are filled out s.t.
 *		cellD->upperBound[r] = r->DbyUniqueId(states->upperBound), for r={root...allRxn}
 */
void TimeHierLeapSeperateAccept::_calculateDBoundsFromSpecies() {
	int idealGrainSize = 20;
	
	vector<double> * dHigh = &_cellD->highBound;
	vector<double> * dLow = &_cellD->lowBound;
	ApplyCalculateD acdHigh = ApplyCalculateD(&_states->highBound, dHigh);
	ApplyCalculateD acdLow = ApplyCalculateD(&_states->lowBound, dLow);
	vector<ReactionSet*> & deps = _root->getDependents();
	
	blocked_range<std::vector<ReactionSet*>::iterator> range(deps.begin(), deps.end(), idealGrainSize);
	
	parallel_for(range, acdHigh);
	parallel_for(range, acdLow);
	
	(*dHigh)[_root->getUniqueId()] = sum(&((*dHigh)[_root->smallestChildId()]), _root->numDependents()); 
	(*dLow)[_root->getUniqueId()] = sum(&((*dLow)[_root->smallestChildId()]), _root->numDependents()); 
	assertNearlyEqual((*dHigh)[_root->getUniqueId()], _root->DbyUniqueId(_states->highBound));
}

// Assumes _states->highBound, _states->lowBound, _cellD->highBound, _cellD->lowBound are pre-computed releative to s.
// otherwise, the functionallity is identical to ReactionTauLeap(...); though optimizations may abound
bool TimeHierLeapSeperateAccept::ReactionTauLeapPreComputedBounds(int leap, State *s, double &tauSum) {
	_recursionLevel++;
	_rtlPreComputedBoundsIterations++;
	
	//_calculateDBoundsFromSpecies(); // this *may* not be necessary, can do faster and tighter method later on.
	
	tauSum = erlang_sample(_cellD->lowBound[_root->getUniqueId()], leap);
	if (!isfinite(tauSum)) { // 'infinite' amount of time lapses. 
		_rejectedSample(NULL, tauSum); 
		return false;
	}
	
	sampleCoarseEvents(s, _states->highBound, leap);
	
	_calculateCellBounds(s);
	
	// this is so comparison to rootD is still bounded.
	_cellD->lowCellBound[_root->getUniqueId()] = _cellD->lowBound[_root->getUniqueId()]; 
	
	// compute p coarse from _metaCellEvents
	double Pcoarse = _computePCoarse(tauSum); 
	assert(Pcoarse >= 0);
	if (Pcoarse < urand()) {
		if (_maxRecursionReached())
			return _rejectedSample(NULL, tauSum);
		
		// no need to recalculate bounds
		return ReactionTauLeapPreComputedBounds(leap, s, tauSum);
	}
	Pcoarse = 1.0;
	
	// Sample fine events
	ReactionPermutation<Reaction*> *rxns = new ReactionPermutation<Reaction*>();
	vector<ReactionPermutation<Reaction*> *> rpToJoin;
	double pBlocks = 1.0;
	
	BOOST_FOREACH(ReactionSet *cell, _root->getDependents()) {
		int cellEvents = _metaCellEvents[cell->getUniqueId()];
		double blockAcceptP;
		if (!cellEvents) {
			continue;
		}
		if (_cellD->highCellBound[cell->getUniqueId()] == 0){
			return _rejectedSample(rxns, tauSum);
		}
		
		ReactionPermutation<Reaction*> *rp = _sampleBlock(cell->asReactions(), blockAcceptP);
		assert(blockAcceptP >= 0);
		pBlocks *= blockAcceptP;
		assert(pBlocks >= 0 && pBlocks <= 1.0);
		if (!rp) {
			_Paccept = Pcoarse * pBlocks;
			return _rejectedSample(rxns, tauSum);
		}
		assert(rp);
		rpToJoin.push_back(rp);
	}
	rxns->joinPermutations(rpToJoin);
	pBlocks = 1.0; // accept/reject occured inside of blocks sample.
	
	// Compute PfineLowerBound
	double PfineLowBound = exp(-tauSum * (_cellD->highBound[_root->getUniqueId()] - _cellD->lowBound[_root->getUniqueId()]));
	
	_Paccept = Pcoarse * PfineLowBound * pBlocks;
	double u = urand();
#ifndef CALCULATE_P_ACCEPT_ONLY
	if (PfineLowBound >= u) { 
		// early acceptance
		_acceptedSample(rxns, tauSum, s, leap);
		return true;
	}
#endif
	
	// sample taus on cell level, only if not early accept
	double tsum = 0.0;
	_taus.clear();
	for (int i=0; i < leap; i++) {
		double r = erand();
		tsum += r;
		_taus.push_back(r);
	}
	normalize(_taus, tauSum); 
	assert(tauSum >= 0);
	
	// Compute Pfine if no early acceptance
	// this is just PROD(exp(-tau *(D - D_~)))
	double lowDcell = _cellD->lowCellBound[_root->getUniqueId()]; //_root->DbyUniqueId(_states->lowCellBound);
	double expSumFine = 0.0;
	double rootD = _cellD->x[_root->getUniqueId()];
	while (rxns->hasNext()){
		assertEpsilonLessThan(lowDcell, rootD);
		
		Reaction *r = rxns->next();
		if (!r->isViable(s)) {
			rxns->previous(); // put rxn back into permutation so we don't have to 'undo' the work.
			expSumFine = -DBL_MAX;
			break;
		}
		
		double tau = _taus.back(); _taus.pop_back();
		assert(tau > 0);
		expSumFine += -tau*(rootD - lowDcell); 
		
		rootD += _getDeltaComputer(r)->eval(s->X()); 
		
		r->fire(s); // changes state
	}
	
	double Pfine = exp(expSumFine);
	assert(Pfine >= 0 && Pfine <= 1);
	
	_Paccept = Pcoarse * Pfine * pBlocks;
#ifdef CALCULATE_P_ACCEPT_ONLY
	_Paccept = Pfine;
	return _rejectedSample(rxns, tauSum);
#endif

	if (_Paccept < u) {
		// reject sample
		while (rxns->hasPrevious()){
			Reaction *r = rxns->previous();
			r->unFire(s);
		}
		// also rootD needs to be set
		delete rxns;
		rxns = NULL;
		if (_maxRecursionReached()) {
			return _rejectedSample(NULL, tauSum);
		}
		// no need to recalculate bounds
		return ReactionTauLeapPreComputedBounds(leap, s, tauSum);
	}
	
	// update _cellD and state (highBound, lowBound) based on rxns
	return _acceptedSample(rxns, tauSum, s, leap);
}

// Returns the delta computer for r wrt the entire system propensity.
DeltaEquation* TimeHierLeapSeperateAccept::_getDeltaComputer(Reaction *r) {
	return checkNotNull((*_deltaCompute)[r]);
}

DeltaPropensityBounds* TimeHierLeapSeperateAccept::_getDeltaPropensityComputer(Reactions *r) {
	return checkNotNull((*_deltaPropensityCompute)[r]);
}

// Updates _cellD->x; if rxns != NULL assumes _cellD->x was correct prior to rxns being exectued. changes D for rxn 'related' to rxns. 
//
// @arg: UpdateTimestamp is used to keep track of what reactions fired between the previous timeStamp.
//     The only requirement of updateTimestamp is that it be a different value when called with different s and the 
//     value should not be too periodic.
//
// makes changes to _cellD->x[_root], _cellD->x[_cells] and _cellD->x[rxns<..>()->effectedBy()]
void TimeHierLeapSeperateAccept::_updateD(State *s, ReactionPermutation<Reaction*> *rxns, int updateTimestamp) {
	vector<double> & D = _cellD->x;
	double & rootD = _cellD->x[_root->getUniqueId()];
	if (rxns) {
		list<Reactions*> cellsToRecompute;
		BOOST_FOREACH(Reaction *r, rxns->_myReactionList()) {
			BOOST_FOREACH(Reaction *r2, *r->changesPropensityFor()) {
				int rid = r2->getUniqueId();
				if (_lastDUpdateTimestamp[rid] == updateTimestamp) 
					continue; // q: is this actually faster than re-doing the calculation?
				
				D[rid] = r2->DbyUniqueId(s->X());
				_lastDUpdateTimestamp[rid] = updateTimestamp;
				
				// put cell on recalculate queue if it's not already there.
				int cid = r2->getParent()->getUniqueId();
				if (_lastDUpdateTimestamp[cid] != updateTimestamp) {
					_lastDUpdateTimestamp[cid] = updateTimestamp;
					cellsToRecompute.push_front(r2->getParent());;
				}
			}
			BOOST_FOREACH(Reactions *cell, cellsToRecompute) {
				D[cell->getUniqueId()] = sum(&D[cell->smallestChildId()], cell->numDependents());
			}
		}
	} else {
		// compute all D from scratch
		rootD = 0.0;
		BOOST_FOREACH(Reaction *r, _root->descendantReactionList()) {
			double d = r->DbyUniqueId(s->X());
			int rid = r->getUniqueId();
			D[rid] = d;
			rootD += d;
			_lastDUpdateTimestamp[rid] = updateTimestamp;
		}
		BOOST_FOREACH(ReactionSet *cell, _root->getDependents()) {
			int cid = cell->getUniqueId();
			D[cid] = sum(&D[cell->smallestChildId()], cell->numDependents());
		}
	}
	
	// always recompute D[root] from cell summation.
	D[_root->getUniqueId()] = sum(&D[_root->smallestChildId()], _root->numDependents());
	assertNearlyEqual(D[0], _root->DbyUniqueId(s->X()));
}


#define INTEL_THREADS_UPDATE_BOUNDS 21
#define SERIAL_UPDATE_BOUNDS 77
#define DIY_THREADS_UPDATE_BOUNDS 4

// can update bounds in parallel
void TimeHierLeapSeperateAccept::_updateBounds( State *s, ReactionPermutation<Reaction*> *rxns, int leap) { 
	_pb->highBoundState(s->X(), leap, _states->highBound);
	_pb->lowBoundState(s->X(), leap, _states->lowBound);
	
	_finalReactions = rxns;
	int idealGrainSize = 8;
	vector<ReactionSet*> & deps = _root->getDependents();
	
	int updateBoundsMethod = INTEL_THREADS_UPDATE_BOUNDS;
	switch (updateBoundsMethod) {
		case INTEL_THREADS_UPDATE_BOUNDS:
			parallel_for(blocked_range<std::vector<ReactionSet*>::iterator>(deps.begin(), deps.end(), idealGrainSize), 
						 ApplyCellBoundUpdate(_getDeltaPropensityComputerAccesor(), _states, _cellD, &_finalReactions, leap));
			break;
		case DIY_THREADS_UPDATE_BOUNDS:
			if (!_updateCellPropensities) {
				vector<Reactions*> rxnsChild;
				BOOST_FOREACH(ReactionSet *r, deps) {
					rxnsChild.push_back(r->asReactions());
				}
				_updateCellPropensities = 
					new ComputationTask<ApplyCellBoundUpdate, Reactions*>(new ApplyCellBoundUpdate(_getDeltaPropensityComputerAccesor(), 
																								   _states, _cellD, &_finalReactions, leap), 
																		  rxnsChild) ;
			}
			
			_updateCellPropensities->run(); 
			break;
		case SERIAL_UPDATE_BOUNDS:
			//#pragma omp parallel for
			for (int i = 0; i < deps.size(); i++) {
				ReactionSet *cell = deps[i];
				int cid = cell->getUniqueId();
				_getDeltaPropensityComputer(cell->asReactions())->updateUpperBoundDeltaDs(rxns, _states->highBound);
				double maxPosDelta =_getDeltaPropensityComputer(cell->asReactions())->upperBoundDeltaD(_states->highBound);
				double maxNegDelta = _getDeltaPropensityComputer(cell->asReactions())->lowerBoundDeltaD(_states->highBound);
				assert(maxNegDelta <= 0);
				
				_cellD->highBound[cell->getUniqueId()] = _cellD->x[cid] + maxPosDelta * (leap - 1); 
				//_cellD->highBound[cell->getUniqueId()] = cell->DbyUniqueId(_states->highBound);
				//assertNearlyEqual(cell->DbyUniqueId(s->X()), _cellD->x[cid]);
				
				//_cellD->lowBound[cell->getUniqueId()] = cell->DbyUniqueId(_states->lowBound);
				_cellD->lowBound[cell->getUniqueId()] = max(0.0, _cellD->x[cid] + maxNegDelta * (leap - 1)); 
			}
			break;
		default:
			assert(false);
	} 
	
	_finalReactions = NULL;
	
	
	_cellD->highBound[_root->getUniqueId()] = sum(&_cellD->highBound[_root->smallestChildId()], _root->getDependents().size());
	_cellD->lowBound[_root->getUniqueId()] = sum(&_cellD->lowBound[_root->smallestChildId()], _root->getDependents().size());
}

bool TimeHierLeapSeperateAccept::_rejectedSample(ReactionPermutation<Reaction*> *rxns, double &dt) {
	if (rxns) 
		delete rxns;
	_recursionLevel = 0;
	dt = 0;
	return false;
}

bool TimeHierLeapSeperateAccept::_acceptedSample(ReactionPermutation<Reaction*> *rxns, double &dt, State *s, int leap) {
	_recursionLevel = 0;
	if (rxns->hasNext()) {
		while (rxns->hasNext()) {
			rxns->next()->fire(s);
		}
	}
	_updateD(s, rxns, _rtlPreComputedBoundsIterations++);
	_updateBounds(s, rxns, leap);
	
	if (rxns) delete rxns;
	assert(dt != 0);
	assert(isfinite(dt));
	return true;
}
