// Run SSA in OpenCl
// David Orendorff
// 4/1/11

#define MY_CONSTANT global


// todo: only if we are running on AMD
#pragma OPENCL EXTENSION cl_amd_printf : enable

// #define DEBUG

#ifdef DEBUG
#define assert(x) \
		if (! (x)) \
		{ \
			printf("Assert(%s) failed in %s:%d", #x, __FUNCTION__, __LINE__); \
		}
		// todo: add 	"		exit(-1); \           "    back into if statement.
#else
	#define assert(x)
	#define printf(fmt, ...)
#endif


void printfArray(local int *a, int len) {
		for (int i = 0; i < len; i++) {
			printf("%d, ", a[i]);
		}
		printf("\n");
}


void printfInt2Array(local int2 *a, int len) {
		for (int i = 0; i < len; i++) {
			printf("(%d, %d), ", a[i].x, a[i].y);
		}
		printf("\n");
}

void printFloatArray(local float *a, int len) {
		for (int i = 0; i < len; i++) {
			printf("%f, ", a[i]);
		}
		printf("\n");
}

void constMemPrintf(local float  *rates, local int2 *inpSpec, 
					local int2 *outSpec, int stateSize, int totRxn, int gid, int nThreads) {
	if (gid != 0) return;
	
	printf("InpSpecies...\n");
	printfInt2Array(inpSpec, totRxn);
	
	printf("OutSpecies...\n");
	printfInt2Array(outSpec, totRxn);
	
	printf("Rates... \n");
	printFloatArray(rates, totRxn);

}

///////// END PRINTF ////////

// Returns a uniform float (0, 1) based on the random uniform integer.
// This will never return 0 or 1.
private float randomFloat(uint randomInt) {
	float f = (float) randomInt / UINT_MAX;
	if (f >= 1.0 || f <= 0) {
		// this should occur very rarely, but this is a source of small biase: todo fix
		f = nextafter(f, 0.5f); // should be slightly less than 1.0 or greater than 0
	}
	
	if (!(0 < f && f < 1)) {
		printf("warning, uniform float generated (%f) from value %u \ %u \n", f, randomInt, UINT_MAX);
		assert(false);
	}
	return f;
}

void rxnPropensity(local int *state, int xSpecies, int ySpecies, float rate, private float *D){
	assert(rate >= 0);
	float d = rate;

	if (xSpecies >= 0){
		d *= (float) state[xSpecies];
		if (state[xSpecies] < 0) d = 0.0;
	}
	
	if (ySpecies >= 0)  {
		d *= xSpecies != ySpecies ? state[ySpecies] : state[ySpecies] - 1;
		if (state[ySpecies] < 0) d = 0.0;
	}

	assert(d >= 0);
	*D = d;
}

// pre:
//	- temp is of size at least nThreads
//	- inpSpec has entries which are valid (smaller than nSpecies)
void propensity(local int *state, local int2 *inpSpec, local float *rates, local float *D, local float *Dsum, int nRxn, local float *temp, int gid, int nThreads) {
	temp[gid] = 0;
	private float d;

	for (int i = gid; i < nRxn; i += nThreads) {
		rxnPropensity(state, inpSpec[i].x, inpSpec[i].y, rates[i], &d);
//		d = rates[i];
//		if (inpSpec[i].x != -1)  d *= inpSpec[i].x == inpSpec[i].y ? state[inpSpec[i].x] - 1: state[inpSpec[i].x];
//		if (inpSpec[i].y != -1) d *= state[inpSpec[i].y];
		
		
		//printf("\n%d+%d->NA, r: %f\t-->D(%f)", inpSpec[i].x, inpSpec[i].y, rates[i], d);
		temp[gid] += d;
		D[i] = d;
		//printf("D[%d]=(%f)\t", i, d);
	}

	barrier(CLK_LOCAL_MEM_FENCE);
	
	if (gid == 0){ 
		*Dsum = 0;
		for (int i = 0; i < nThreads; i++){
			*Dsum += temp[i];
		}
	}
}

// Execute multiple steps of SSA in a row. This is performed in one workgroup.
// This kernel will exit if either:
//		1. The simulated time will exceed totalSimtime by exactly one reaction event.
//		2. There are not enough pre-generated random numbers to continue (lastValidRandIdx < randIdx). If this is the case then *randIdx is set to -1.
//
// This kernel should run on any number of threads: 1 to MAX_LOCAL_THREADS
// todo: (possible) move randUints from global to MY_CONSTANT space.
// important: tempSumSpace should be of size get_global_size(0)
__kernel
void ssa(global int *globalState, int stateSize,
			  global float *totalSimtime, // we will run the simulator for slightly longer (within L reaction events of totalSimTime)
			  float MY_CONSTANT *gRates,	
			  int2 MY_CONSTANT *gInpSpec, // Assume each rxn has at most N=2 inp species
			  int2 MY_CONSTANT *gOutSpec,
			  global unsigned int *randomInts, global int *randIdx,
			  int totalRandomNums,
			  int totRxn,
			  local int *state, local float *D, local float  *rates, local int2 *inpSpec, 
			  local int2 *outSpec, local float *DrxnSubgroups){

	local float endTime; endTime = *totalSimtime;
	local float time; time = 0.0f;
	
	local int randUIntsOffset; randUIntsOffset = *randIdx;;
	__private int gid = get_global_id(0);
	__private int nThreads = get_global_size(0);
	local int randIntIdx; randIntIdx = *randIdx;
	printf("GID(%d) / %d, LocalId(%d) \n", gid, nThreads, get_local_id(0));

	async_work_group_copy(state, globalState, stateSize, 0);
	async_work_group_copy(rates, gRates, totRxn, 0);
	async_work_group_copy(inpSpec, gInpSpec, totRxn, 0);
	async_work_group_copy(outSpec, gOutSpec, totRxn, 0);
	
	barrier(CLK_LOCAL_MEM_FENCE);
	
	if (gid == 0)
		printf("\n>");
	
	local float dt; dt = 0.0f;
	local int rxnToFire; rxnToFire = -1;
	local float Dsum;
	
	//constMemPrintf(rates, inpSpec, outSpec, stateSize, totRxn, gid, nThreads);
	
	int nRandPerIteration = 2;
	private int iter = 0;
	private bool isIter5 = state[0]==8646; // for debugging only.
	int iterationWithError = 20018;
	bool outputNow;
	local bool abortLoop; abortLoop = false;
	while (time < endTime) {
		// outputNow = gid == 0 && isIter5 && iter > iterationWithError;
		outputNow = isIter5 && iter > iterationWithError;
		if (outputNow) printf(".(GID(%d))", gid);
		if (abortLoop) break;
		
		if (outputNow) printf("calculating propensity GID(%d)...\n", gid);
		// calculate propensity and propensity sum (Dsum).
		propensity(state, inpSpec, rates, D, &Dsum, totRxn, DrxnSubgroups, gid, nThreads);
		
		#ifdef DEBUG
			// sanity check
			if (gid == 0) {
				float s = 0;
				for (int i = 0; i < totRxn; i++) 
					s += D[i];
				assert(fabs(s - Dsum) <= Dsum / 10000.0);
			}
		#endif
		
		// sample rxnToFire, fire reaction.
		if (gid == 0) { 
			
			float targetSum = Dsum * randomFloat(randomInts[randUIntsOffset++]);
			int cidx = 0;
			float sum = 0;
		
			if (outputNow) printf("sampling reaction group  GID(%d)...\n", gid);
			
			// Determine the subgroup (cidx) which contains the target rxn
			for ( ;sum + DrxnSubgroups[cidx] < targetSum ;) {
				sum += DrxnSubgroups[cidx];
				cidx++;
			}
			assert(cidx < totRxn);
			
			// Determine rxn in subgroup
			if (outputNow) printf("sampling reaction in group...\n");
			int ridx = cidx;
			for (  ;; ridx += nThreads) {
				sum += D[ridx];
				
				// Note: rounding error will very occasionally allow us to 
				// go beyond targetSum and cause a memory invalid access error.
				// That the reason for 2nd part of the conditional.
				if (sum >= targetSum || ridx + nThreads >= totRxn) break;
			}
			assert(ridx < totRxn);

			// Apply ridx to state.
			int2 ins = inpSpec[ridx];
			int2 outs = outSpec[ridx];

			if (outputNow) printf("Sampled reaction %d (%d+%d --> %d + %d).\n", 
									ridx, ins.x, ins.y, outs.x, outs.y);

			if (ins.x >= 0) {
				state[ins.x]--; 
				assert(ins.x < stateSize);
			}
			if (ins.y >= 0) {
				state[ins.y]--;
				assert(ins.y < stateSize);
			}
			if (outs.x >= 0) {
				state[outs.x]++;
				assert(outs.x < stateSize);
			}
			if (outs.y >= 0) {
				state[outs.y]++;
				assert(outs.y < stateSize);
			}
			
			dt = -log((float) randomFloat(randomInts[randUIntsOffset++])) / Dsum;
			time += dt;
			if (outputNow) printf("Sampled tau(%f), time is now %f.\n", dt, time);
			if (outputNow) printf("randUIntsOffset(%d), totalRandomNums(%d).\n", 
							randUIntsOffset, totalRandomNums);
			
			//printf("%d, dt(%f), Dsum(%f) ", ridx, dt, Dsum);
			iter++;
			
			abortLoop = randUIntsOffset + nRandPerIteration > totalRandomNums - 1;
			if (abortLoop) randUIntsOffset = -1;
		}
		
		barrier(CLK_LOCAL_MEM_FENCE);
	}
	
	//if (outputNow) printf("Finished iteration. Saving to global memory.\n");
	
	if (gid == 0) {
		*randIdx = randUIntsOffset;
		*totalSimtime = time;
		//	printf("\nsimulated time(%f)", time);
	}
	async_work_group_copy(globalState, state, stateSize, 0);
	//if (isIter5) printf("GID(%d) finished ssa.cl:ssa.\n", gid);
}