/*
 * State.cpp
 *
 *  Created on: Sep 7, 2009
 *      Author: dweo
 */

#include "state.h"

#include "../reactions/reactions.h"
#include "../reactions/reaction_set.h"
#include "../reactions/reaction.h"
#include "species.h"
#include "reaction_cache.h"

using namespace hierleap;

State::State() {
	assert(false);
}

// TODO Maybe do a lazy copy (?)
State::State(const State *s){
	_x = s->_x;
	_reactionCache = s->_reactionCache; // do a shallow copy
	_species = s->_species;
	_root = s->_root;
	_numSpeciesTypes = s->_numSpeciesTypes;
	BOOST_FOREACH(Species *s, _species) {
		if (s->getParent() == NULL)
			s->setParent(_root);
	}
}


State::State(vector<Species*> species, Reactions* root){
	_root = root;
	_species = species;
	// get num reactions and num species
	int numRxn = root->numRxnSets();
	int numSpecies = species.size();
	_x.resize(numSpecies);
	_reactionCache.resize(numRxn);

	// determine the number of species types
	_numSpeciesTypes = -1;
	int maxSpeciesId = -1;
	// TODO this is a little fragile, may want to make sure all 0...(n-1) species types are included
	for (unsigned int i=0; i < species.size(); i++){
		assert(species[i]->getTypeId() >= 0); // safety check, make sure everyone has a type
		maxSpeciesId = max(maxSpeciesId, species[i]->getTypeId());
	}
	_numSpeciesTypes = maxSpeciesId +1; // Assume counting starts at 0
	
	BOOST_FOREACH(Species *s, _species) {
		//if (s->getParent() == NULL)
		s->setParent(root);
	}
	
	_setupReactionCache(root);
	for (int i=0; i < numRxn; i++){
		assert(_reactionCache[i]);
	}
	// need to see how many 
}

void State::changeX(int speciesId, int deltaX){
	_x[speciesId] += deltaX;
	if(_x[speciesId] < 0){
		assert(false);
	}
}

// traverse the reactions tree and create new reaction cache for each reaction
void State::_setupReactionCache(ReactionSet *root){
	int idx = root->getUniqueId(); // make sure this works ok
	vector<ReactionSet*> & deps = root->getDependents();
	_reactionCache[idx] = new ReactionCache(_numSpeciesTypes, deps.size()); 
	for (int i=0; i < deps.size() ; i++){
		_setupReactionCache(deps[i]);
	}
}

State::~State() {
	// TODO Auto-generated destructor stub
}

vector<Reaction*> & State::touchingRxn(int speciesId) {
	if (_touchingRxn.size() == 0) {
		_touchingRxn.resize(speciesSize(), NULL);
		for (int i = 0; i < _touchingRxn.size(); i++)
			_touchingRxn[i] = new vector<Reaction*>();
		
		BOOST_FOREACH(Reaction *r, _root->descendantReactionList()) {
			for (int i = 0; i < r->numReactants(); i++) {
				_touchingRxn[r->getReactantId(i)]->push_back(r);
			}
			for (int i = 0; i < r->numProducts(); i++) {
				_touchingRxn[r->getProductId(i)]->push_back(r);
			}
		}
	}
	assert(speciesId < _touchingRxn.size() && speciesId >= 0);
	return *_touchingRxn[speciesId];
}

// initialXbySpecies should be a vector of size |species types|
void State::setXbySpeciesId(vector<int> & initialXbySpecies){
	assert(speciesTypes() == initialXbySpecies.size());
	for (unsigned int i=0; i < _species.size(); i++){
		int sidx = _species[i]->getUniqueId();
		int type = _species[i]->getTypeId();
		_x[sidx] = initialXbySpecies[type];
		assert(_x[sidx] >= 0);
	}
}

void State::setXdirectly(vector<int> x){
	if(x.size() != _x.size()){
		assert(false);
	}
	for (int i=0; i < x.size(); i++){
		assert(x[i] >= 0);
		_x[i] = x[i];
 	}
}

vector<int> & State::maximalSpeciesByType(ReactionSet *r){
	if (r->isReaction())
		return _maximalSpeciesByType(r->asReaction());
	return _maximalSpeciesByType((Reactions*)r);
}

// The maximum of each species of all children for every TYPE is returned
vector<int> & State::_maximalSpeciesByType(Reactions *r){
	// TODO: cache these answers
	vector<int> & msbt = _reactionCache[r->getUniqueId()]->msbt();
	BOOST_FOREACH( ReactionSet *rChild, r->getDependents()){
		vector<int> & childMsbt = maximalSpeciesByType(rChild);
		for (unsigned int i=0; i < msbt.size(); i++){
			msbt[i] = max(msbt[i], childMsbt[i]);
		}
	}
	return msbt;
}

// Base case for above maximalSpeciesByType
vector<int> & State::_maximalSpeciesByType(Reaction *r){
	// just return a vector of the species in this cell
	return r->getParent()->speciesXByType(this);
}


ReactionCache* State::cache(ReactionSet *r){
	int idx = r->getUniqueId();
	assert(_reactionCache[idx]);
	return _reactionCache[idx];
}

vector<int> & State::cacheMSBT(ReactionSet *r){ 
	int idx = r->getUniqueId();
	int sz = _reactionCache.size();
	ReactionCache *rc = _reactionCache[idx];
	return rc->msbt(); 
}

double State::DtildeCached(ReactionSet *r){
	int idx = r->indexInParent();
	vector<double> & dtildecache = cache((ReactionSet*)r->getParent())->upperDchildren();
	assert(dtildecache.size() > idx);
	return dtildecache[idx];
}

void State::clearEvents(){
	BOOST_FOREACH(ReactionCache *rc, _reactionCache){
		if (rc) rc->setNumEvents(0);
	}
}

void State::setEvents(ReactionSet *r, int events){ _reactionCache[r->getUniqueId()]->setNumEvents(events); }
int State::numEvents(ReactionSet *r){ 
		assert(r->getUniqueId() >=0 && r->getUniqueId() < _reactionCache.size());
		return _reactionCache[r->getUniqueId()]->numEvents(); }

int State::x(Species *s){ return x(s->getUniqueId()); }

