//	Copyright 2007 David Orendorff, Eric Mjolsness
//
//  Algorithm By Eric M., code primarily by David O.
//  Reference Manuscript by: Eric Mjolsness, David Orendorff,
//		Philippe Chatelain and Petros Koumoutsakos.
//
//	This file is part of ER_Leap.
//
//	ER_Leap is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	ER_Leap is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with ER_Leap.  If not, see <http://www.gnu.org/licenses/>.

/*
 *  common.cpp
 *  erleap
 *
 *  Created by David  Orendorff on 6/27/07.
 *  Implements common utility functions.
 *
 */

#define USE_GSL

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

#ifdef USE_GSL
#include <gsl/gsl_randist.h>
#endif

// For boost random
#include <boost/random.hpp>
#include <boost/random/linear_congruential.hpp>
#include <boost/random/uniform_real.hpp>
#include <boost/random/variate_generator.hpp>

#include "common.h"

#include "random_number_provider_thread.h"

// This is a typedef for a random number generator.
// Try boost::mt19937 or boost::ecuyer1988 instead of boost::minstd_rand
typedef boost::minstd_rand base_generator_type;
//typedef RandomNumberProviderThread<boost::minstd_rand> base_generator_type;

base_generator_type *generator = NULL;
boost::uniform_real<> *uni_dist;
boost::variate_generator<base_generator_type&, boost::uniform_real<> > * uni; // uniform random generator
boost::random_number_generator<base_generator_type> * rand_ints; // random ints generator

void set_boost_seed(void* state, unsigned long int seed) {
	generator->seed(seed);
}

double get_boost_urand(void* state) {
	return urand();
}

unsigned long int boost_urand_int_max(void* state) {
	return (*rand_ints)(UINT_MAX);
}

static const gsl_rng_type ran_type = {
	"boost_rng",                   /* name */
	generator->max(),                           /* RAND_MAX */
	generator->min(),                            /* RAND_MIN */
	sizeof (1), // we don't actually need to store any information in state, we're accessing static variables instead
	&set_boost_seed,
	&boost_urand_int_max, // rand uint
	&get_boost_urand // rand double
};

const gsl_rng_type *gsl_rng_boost = &ran_type;

void init_gsl_if_available(unsigned int seed);

//void init_random_with_threads(unsigned int seed) {
//	RandomNumberProviderThread<boost::minstd_rand> *generator = new RandomNumberProviderThread<boost::minstd_rand>();
//	
//}

void init_random(unsigned int seed){
	
	// clean up after self if we've already setup
	if (generator) {
		delete generator; delete uni_dist; delete uni; delete rand_ints;
		generator=NULL;
		return init_random(seed);
	}

	generator = new base_generator_type(seed); // boost::minstd_rand
	uni_dist = new boost::uniform_real<>(0,1);
	uni = new boost::variate_generator<base_generator_type&, boost::uniform_real<> >(*generator, *uni_dist);
	rand_ints = new boost::random_number_generator<base_generator_type>( *generator );
	
	init_gsl_if_available(seed);
}

//StochasticLib1 *_stoc;

//void init_random(int seed){
//	_stoc = new StochasticLib1(seed);
//}

int sum(int *arr, int len){
	int sum = 0;
	for (int i=0; i < len; i++) sum += arr[i];
	return sum;
}
double sum(double *arr, int len){
	double sum = 0;
	for (int i=0; i < len; i++) sum += arr[i];
	return sum;
}

void printf_array(double *arr, int len){
	for (int i=0; i < len; i++){
		printf("%f, ", arr[i]);
	}
	printf("\n");
}
void printf_array(int *arr, int len){
	for (int i=0; i < len; i++){
		printf("%d, ", arr[i]);
	}
	printf("\n");
}

void printf_array(unsigned int *arr, int len){
	for (int i=0; i < len; i++){
		printf("%d, ", arr[i]);
	}
	printf("\n");
}

// computes power in log(p) time recursivly
double _ripow(double x, int p){
	if (p==0) return 1;
	if (p==1) return x;
	double half_pow = _ripow(x, p/2);
	half_pow *= half_pow;
	return p % 2 ? x * half_pow : half_pow;
}

// random double between (0,1] exclusive
double urand(){
	return (*uni)();
}

//// returns normal random variable sample
//double nrand(double mean, double sd){
//	//return RNOR;
//	return _stoc->Normal(mean, sd);
//}

void swap(int &a, int &b){
	int t = a; a = b; b = t;
}

// random uniform number (low, high]. 
unsigned int urand_int(unsigned int low, unsigned int high){
	unsigned int rng = high - low;
	return (*rand_ints)(rng) + low;
}

// go through the array, at each index swap element with 
// uniformly random chosen element.
// shuffled idx

void randperm(int *arr, int n){
	int j;
	int swap;
	for (int i=0; i<n-1; i++) {
		// item number i has n-i numbers to choose between
		j = urand_int(i,n-1);
		// swap items i and j
		swap = arr[j];  arr[j] = arr[i];  arr[i] = swap;
	}
}

#define LARGE_MULTIPLIER 100000
#define INV_LARGE_MULTIPLIER 0.00001
#define LOG_LARGE_MULTIPLIER 11.5129255

double erlang_sample(double a, int m){
	assert(a >= -0.00000000001);
	if (a < 0.00000000001) 
		return INFINITY;
	//return gsl_ran_gamma_mt(_gsl_rand_eng, float(m), 1.0/a); // note: there is a bug with gsl_ran_gamma
//}
	// TODO: not efficient but correct
	double prod = 1.0;
	double logOffset = 0; // for large m we need this to avoid underflow.
	for (int i=0; i < m; i++) {
		if (prod < INV_LARGE_MULTIPLIER) {
			prod *= LARGE_MULTIPLIER;
			logOffset += LOG_LARGE_MULTIPLIER;
		}
		prod *= urand();
	}
	return -1/a * (log(prod) - logOffset);
}


//// returns the log-probability of this multinomial sample
//float logprob_multinomial(int num_samples, double *probs, int num_probs, int *sample){
//	double probsum = 0;
//	for (int i=0; i < num_probs; i++) probsum += probs[i];
//
//	// prob = ns! / (PRDO(results!)) * PROD(probs.^results)
//	double logprob =logfact(num_samples);
//	for (int i=0; i< num_probs; i++){
//		logprob -= logfact(sample[i]); // denominator
//		logprob += log(probs[i] / probsum) * float(sample[i]);
//	}
//	assert(logprob <= 0);
//	return logprob;
//}
#ifdef USE_GSL
gsl_rng * _gsl_rand_eng;

void init_gsl_if_available(unsigned int seed) {
	//_gsl_rand_eng = gsl_rng_alloc (gsl_rng_taus);
	assert(generator); // only if we are using the boost generator
	_gsl_rand_eng = gsl_rng_alloc (gsl_rng_boost);
	gsl_rng_set(_gsl_rand_eng, seed);
}

// Samples index from i=0...len-1, such that 
//		Prob(i) = arr[i]/arrSum
//		where: arr[k] >= 0, arrSum > 0
int select(double *arr, int len, double arrSum) {
	assert(len >= 1);
	if (len==1) return 0;
	
	double targetSum = arrSum * urand();
	int idx = 0;
	for (double sum=0;; idx++) {
		sum += arr[idx];
		if (sum >= targetSum) break;
	}
	assert(idx >= 0 && idx < len);
	return idx;
}

vector<unsigned int> multinomial(vector<double> & probs, unsigned int num_samples){
	assert(sum(probs)>0);
	vector<unsigned int> hits;
	hits.resize(probs.size());
	multirnd(num_samples, &probs[0], probs.size(), &hits[0]);
	return hits;
}

//(leap, NUM_RXN, rxn_probs, rxn_firings);
// todo: problem if sizeof(int) ~= 32bits
void multirnd(unsigned int num_samples, double *probs, unsigned int num_probs, unsigned int *results){
//Multinomial (int32 * destination, double * source, int32 n, int colors)
	assert(sizeof(num_samples)==4);
	//_stoc->Multinomial((int32*)results, probs, (int32)num_samples, num_probs);
	gsl_ran_multinomial(_gsl_rand_eng , num_probs, num_samples, probs, results);
	return;
	
	// this code not used
	/*
	double psum = sum(probs, num_probs);
	
	assert((psum - 1)*(psum - 1) < 0.0005);
	double *cumsum = (double*)calloc(num_probs, sizeof(double));
	
	cumsum[0]=probs[0];
	results[0] = 0;
	for(int i=1; i < num_probs; i++){
		cumsum[i] = cumsum[i-1] + probs[i];
		results[i] = 0;
	}
	
	for (int i=0; i < num_samples; i++){
		double r = urand();
		int idx = 0;
		while (cumsum[idx] < r) {idx++;}
		results[idx]++;
	}
	free(cumsum);
	return results;
	*/
}

// multirnd, but probs_idx is a (best guess) approximation to how list is sorted
// assumes probabilities are normalized
// return: how many random Binomials were computed
int idx_multirnd(int num_samples, double *probs, int *probs_idx, int num_probs, int *results){
/*
   This function generates a vector of random variates, each with the
   binomial distribution.

   The multinomial distribution is the distribution you get when drawing
   balls from an urn with more than two colors, with replacement.

   Parameters:
   destination:    An output array to receive the number of balls of each 
                   color. Must have space for at least 'colors' elements.
   source:         An input array containing the probability or fraction 
                   of each color in the urn. Must have 'colors' elements.
                   All elements must be non-negative. The sum doesn't have
                   to be 1, but the sum must be positive.
   n:              The number of balls drawn from the urn.                   
   colors:         The number of possible colors. 
*/
  assert(num_samples >=1 && num_probs >= 1);
  double s, sum=1;
  int x;
  int i;
  assert(num_samples > 0 && num_probs > 0) ;
  for (i=0; i < num_probs; i++) results[i] = 0; // pre-set everything to 0

  for (i=0; i<num_probs-1; i++) { 
    // generate output by calling binomial (colors-1) times
	if (num_samples==0) return i; // no need to move down the list
	
	int idx = probs_idx[i];
    s = probs[idx];
    if (sum <= s) {
      // this fixes two problems:
      // 1. prevent division by 0 when sum = 0
      // 2. prevent s/sum getting bigger than 1 in case of rounding errors
      x = num_samples;}
    else {    
      x = gsl_ran_binomial(_gsl_rand_eng, s/sum, num_samples);
	}
    num_samples -= x; sum -= s;
    results[idx] = x;
   }
  // get the last one
  results[probs_idx[i]] = num_samples;
  return i+1;
}

int random_select_one(double *probs, int probs_len, double probs_sum) {
	double target = probs_sum * urand();	
	int idx = 0;
	double sum = 0;
	
	for (  ;idx < probs_len; idx++) {
		sum += probs[idx];
		if (sum >= target) break;
	}
	return idx;
}

// Samples from a multinomial. For each sampled value, the callback function is called
// of the form f(idx, k), where k >=1 and idx is in {0, num_probs]. f(idx,...) may be called
// multiple times for the same idx or once for each idx when k is the sum of number of times idx 
// has been chosen and in any order.
//
// Probabilities do not need to be normalized.
// Returns the number of random number calls.
int multirnd_callback(int num_samples, double *probs, int num_probs, ApplyMultirndCallbackFunc & f) {
	assert(num_samples >=1 && num_probs >= 1);
	double s;
	double sum=0.0;
	for (int i = 0; i < num_probs; i++) sum += probs[i];
	int x;
	int i;
	if (num_samples == 1) {
		int k = random_select_one(probs, num_probs, sum);
		f(k, 1);
		return 1;
	}
	
	assert(num_samples > 0 && num_probs > 0);
	
	for (i=0; i<num_probs-1; i++) { 

		if (num_samples==0) return i; // no need to move down the list
		
		int idx = i;
		s = probs[idx];
		if (sum <= s) {
			// this fixes two problems:
			// 1. prevent division by 0 when sum = 0
			// 2. prevent s/sum getting bigger than 1 in case of rounding errors
			x = num_samples;}
		else {    
			x = gsl_ran_binomial(_gsl_rand_eng, s/sum, num_samples);
		}
		if (x) {
			f(i, x);
		}
		num_samples -= x; sum -= s;
	}
	// get the last one
	if (num_samples) f(i, num_samples);
	return i+1;
}

// gamma random variable
double grand(double a, double b){
	return gsl_ran_gamma(_gsl_rand_eng, a, b);
}

//double grand(double a, double b){
//	//return sgamma(a) / b;
//	return rgamma(a) / b;
//	//return earlang_sample(b, a);
//}

// normal exponential rand
double erand(){
	double e = gsl_ran_exponential(_gsl_rand_eng, 1);
	return e;
	//return (*expo)();
}
#else
// GSL library not used or unavailable.

void init_gsl_if_available(unsigned int seed) {}

vector<unsigned int> multinomial(vector<double> & probs, unsigned int num_samples){
	printf("GLS not being used");
	assert(false);
}

void multirnd(unsigned int num_samples, double *probs, unsigned int num_probs, unsigned int *results){
	printf("GLS not being used");
	assert(false);
}
int idx_multirnd(int num_samples, double *probs, int *probs_idx, int num_probs, int *results){
	printf("GLS not being used");
	assert(false);
}

double grand(double a, double b){
	printf("GLS not being used");
	assert(false);
}
double erand(){
	printf("GLS not being used");
	assert(false);
}
#endif

// float raised to positive integer power
//static unsigned int *ipow_hist;//[100][200];
//static bool initialized = false;
//static int p_max = 550;
//#define HIST(x,p) ipow_hist[x*p_max+p]

// log factorial
float logfact(int n){
	double logprod = 0;
	while (n > 0){
		logprod *= log(n); n--;
	}
	return logprod;
}

double ipow(double x, int p){

	assert(p >= 0);
	if (p==0) return 1;
	return _ripow(x, p);
}

//double ipow(double x, int p){
//
//	assert(p >= 0);
//	if (p==0) return 1;
//	return _ripow(x, p);
///*
//	if (x > .995){
//		double eps = 1 - x;
//		double peps = p*eps/2; // todo: this may not be accurate enough
//		return 1 - peps + peps*(peps-eps)*.667; // use binomial expansion
//	}
//
//	if (p > 4) return _ripow(x, p);
//	double xt = x;
//	while (p > 1) {xt *= x; p--;}
//	return xt; */
//}

