//	Copyright 2007 David Orendorff, Eric Mjolsness
//
//  Algorithm By Eric M., code primarily by David O.
//  Reference Manuscript by: Eric Mjolsness, David Orendorff,
//		Philippe Chatelain and Petros Koumoutsakos.
//
//	This file is part of ER_Leap.
//
//	ER_Leap is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	ER_Leap is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with ER_Leap.  If not, see <http://www.gnu.org/licenses/>.

/*
 *  common.h
 *
 *  Created by David  Orendorff on 6/27/07.
 *
 */

 
#ifndef COMMON_H
#define COMMON_H

#include <queue>
#include <numeric>
#include <boost/foreach.hpp>
#include <boost/thread.hpp>
#include <math.h>
#include <sys/time.h>

using namespace std;

#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define MAX(a,b) ((a) > (b) ? (a) : (b))

// change this if you are running Windows
#define OS_SEP '/'

class ApplyMultirndCallbackFunc {
public:
	virtual void operator () (int idx, int k) = 0;
};

// Returns a vector of elements that a1 and a2 have in common. This method is 
// most efficient if a1 << a2. Behavior is undefined for repeated elements.
// Returns NULL if no elements are in common.
template <typename T> vector<T> * intersection(vector<T> & a1, vector<T> & a2) {
	vector<T> * r = NULL;
	BOOST_FOREACH(T x1, a1) {
		BOOST_FOREACH(T x2, a2) {
			if (x1 == x2) {
				if (!r)
					r = new vector<T>();
				r->push_back(x1);
				break;
			}
		}
	}
	assert(!r || r->size() <= min(a1.size(), a2.size()));
	return r;
}

template <class T>class LessThan {
public:
	bool operator()(T * a, T * b) {
		return a->operator<(*b);// < *b;
	}
};

template <class T>class Equals {
public:
	bool operator()(T a, T b) {
		return a == b;
	}
};

// Sort vector of pointers by using object comparison operator>.
template <typename T> void sort_pointers(vector<T*> & a) {
	sort( a.begin(), a.end(), LessThan<T>() );
}

//template <typename T> vector<T*>::iterator unique_pointers(vector<T*> & a) {
//	// return std::unique(a.begin(), a.end(), Equals<T*>());
//	return std::unique(a.begin(), a.end());
//}

template <class ForwardIterator, class T> ForwardIterator unique_pointers(ForwardIterator first, ForwardIterator last) {
	return std::unique(first, last, Equals<T>());
}

template <typename T> void remove_duplicates_and_sort_pointers(vector<T*> & a) {
	// code shamelessly taken from online forums.
	sort_pointers(a);
	//a.erase(unique_pointers(a), a.end());
	a.erase(std::unique(a.begin(), a.end()), a.end());
}

template <typename T> void extend_vector(vector<T> & a1, vector<T> & a2){
	for (unsigned int i=0; i < a2.size(); i++){
		a1.push_back(a2[i]);
	}
}

template <typename T> bool contains(vector<T> & a1, T x){
	BOOST_FOREACH(T b, a1){
		if (b==x) return true; 
	} return false;
}

template <typename T> T sum(vector<T> &a){
	return std::accumulate(a.begin(), a.end(), 0.0);
}

// post: no value in array will be larger than the corresponding value in ceiling
// array[i] = min(array[i],ceiling[i]) 
template <typename T> void ceil(vector<T> &array, vector<T> &ceiling) {
	for (int i = 0; i < array.size(); i++) {
		if (array[i] > ceiling[i])
			array[i] = ceiling[i];
	}
}

// post: no value in array will be smaller than the corresponding value in floor
// array[i] = max(array[i],ceiling[i]) 
template <typename T> void floor(vector<T> &array, vector<T> &floor){
	for (int i = 0; i < array.size(); i++) {
		if (array[i] < floor[i])
			array[i] = floor[i];
	}
}


// eg: f([4,7,0,10,0]) --> [4,7,10,0,0]
template <typename T> void move_non_null_to_front(vector<T> & a) {
	int back = a.size() - 1;
	int front = 0;
	while (front <= back) {
		if (a[front]) {
			front++;
		} else if(!a[back]) {
			back--;
		} else {
			swap(a[front], a[back]);
		}
	}
}

template <typename T> void delete_non_null_objs(vector<T> & a){
	for (unsigned int i=0; i < a.size(); i++){
		if (a[i]) delete a[i];
		a[i] = NULL;
	}
}

template <typename T> T * checkNotNull(T * a){
	assert(a != NULL);
	return a;
}

// prints to stdout, adding newline at end
void printf_array(int *arr, int len);
void printf_array(unsigned int *arr, int len);
void printf_array(double *arr, int len);

// computes power in log(p) time recursivly
//  recursive helper to ipow
double _ripow(double x, int p);

// float raised to positive integer power
 double ipow(double x, int p);

// throws assert if not smallerNumber - eps <= biggerNumber
// ie: less than within precision
inline void assertEpsilonLessThan(double smallerNumber, double biggerNumber) {
	if (smallerNumber - smallerNumber / 10000.0 > biggerNumber)
		assert(false);
}

inline void assertArrayEpsilonLessThan(vector<double> & smaller, vector<double> & larger){
	assert(smaller.size() == larger.size());
	for (int i = 0; i < smaller.size(); i++) {
		assertEpsilonLessThan(smaller[i], larger[i]);
	}
}

inline bool isNearlyEqual(double a, double b, double eps = 0.00001) {
	assert(a >= 0); assert(b >= 0); // will need to rethink logic if dealing with (-)
	double m = max(a, b);
	return (m - (a+b)/(2.0) <= m*eps) || m < eps;
}

// Asserts a is within (a * eps) of b.
inline bool assertNearlyEqual(double a, double b, double eps = 0.00001) {
	assert(isNearlyEqual(a, b, eps));
	return true;
}

// f(n,k) = n! / (n-k)!
inline double combinations(int n,int k){
	assert(n >= 0);
	if (n<1 || k < 1 || (n-k) < 0) return 0;
	if (k==1) return (double)n;
	double prod = 1;
	for (int i=n; i > (n-k); i--)
		prod *= (double)i;
	return prod;
}

// Returns maximum number of simultanous computations.
inline unsigned int maxConcurrentThreads() {
	return max(boost::thread::hardware_concurrency(), (unsigned int) 1);
}

inline double elapsedSeconds(struct timeval start, struct timeval end) {
	double usec_in_sec = 1e6;
	double stime = (end.tv_sec  - start.tv_sec) + (end.tv_usec - start.tv_usec) / usec_in_sec;
	return stime;
}

int sum(int *arr, int len);

double sum(double *arr, int len);

inline void normalize(vector<double> & arr, double desiredSum) {
	int n = arr.size();
	assert(n >= 1);
	
	double mFact = desiredSum / sum(&arr[0], n);
	if (mFact == 0) return;
	
	double s = 0;
	for (int i = 0; i < n-1; i++) {
		arr[i] *= mFact;
		s += arr[i];
	}
	arr[n-1] = desiredSum - s; // so reformated array sums exactly to desiredSum
}

void init_random(unsigned int seed);

// random double between [0,1] exclusive
double urand();

//// returns normal random variable
//double nrand(double mean, double sd);

void swap(int &a, int &b);

// random uniform number (low, high]. 
unsigned int urand_int(unsigned int low, unsigned int high);

// go through the array, at each index swap element with 
// uniformly random chosen element.
void randperm(int *arr, int size);

// draws one random sample from the earlang distibution
double erlang_sample(double a, int m);

// Samples index from 0...len-1, such that 
//		Prob(i) = arr[i]/arrSum
int select(double *arr, int len, double arrSum);

// Multinomial select from unnormalized probabilities
void multirnd(unsigned int num_samples, double *probs, unsigned int num_probs, unsigned int *results);

vector<unsigned int> multinomial(vector<double> & probs, unsigned int num_samples);

// Samples from a multinomial. For each sampled value, the callback function is called
// of the form f(idx, k), where k >=1 and idx is in {0, num_probs]. f(idx,...) may be called
// multiple times for the same idx or once for each idx when k is the sum of number of times idx 
// has been chosen and in any order.
//
// Probabilities do not need to be normalized.
// Returns the number of random_binomial calls.
int multirnd_callback(int num_samples, double *probs, int num_probs, ApplyMultirndCallbackFunc & f);

//float logfact(int n);

// returns the log-probability of this multinomial sample 
//float logprob_multinomial(int num_samples, double *probs, int num_probs, int *sample);

// multirnd, but probs_idx is a (best guess) approximation to how list is sorted
// assumes probabilities are normalized
// return: how many random Binomials were computed
int idx_multirnd(int num_samples, double *probs, int *probs_idx, int num_probs, int *results);


// gamma random variable
double grand(double a, double b);

// normal exponential random variable sample
double erand();

#endif
