/*
 *  dynamic_priority_queue.h
 *  gpu_hierleap
 *
 *  Created by David Orendorff on 11/17/11.
 *  Copyright 2011 UC Irvine. All rights reserved.
 *
 *  A constant size priority queue that allows us to increase/decrease a key 
 *  in logN time and view the top value in O(1) time.
 */

#include <tr1/unordered_map>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <vector>
#include <algorithm>

#include "../reactions/reaction_set_map.h"

using namespace std;
using namespace boost::lambda;

template <class V, class K>
struct Comparer {
	const bool _firstHigher;
	Comparer(bool firstHigher) : _firstHigher(firstHigher) {}
	
	bool operator () (pair<V, K>* const &p1, pair<V, K>* const &p2) {
		return compareKeys(p1->second, p2->second);
	}
	bool compareKeys(const K &k1, const K &k2) {
		return _firstHigher ? k1 > k2 : k2 > k1;
	}
};

template <class T, typename C = vector<int> >
class Test {
public:
	//C *c;
	Test() {
	//	c = new C();
	}
};



//template <class Value, class Key, class Mymap = tr1::unordered_map<Value, int> >
class DynamicPriorityQueue {
	typedef Reaction* Value;
	typedef double Key;
	typedef ReactionSetMap<int> Mymap;
	//typedef tr1::unordered_map<Value, int> Mymap;
	Test<int> t;
	//_Map m;
public:
//	static DynamicPriorityQueue<Value, Key, Mymap> *buildMaxPriorityQueue(vector<Value> & vals, vector<Key> & keys){
//		return new DynamicPriorityQueue<Value, Key, Mymap>(vals, keys, true);
//	}
//	
//	static DynamicPriorityQueue<Value, Key, Mymap> *buildMinPriorityQueue(vector<Value> & vals, vector<Key> & keys){
//		assert(false); // build in functinoality.
//		return new DynamicPriorityQueue<Value, Key, Mymap>(vals, keys, false);
//	}
	
	void changePriority(Value v, Key newKey) {
		assert(_heap.size() > 0);
		//assert(_valueIndex.count(v) == 1); // not the most efficient..
		int idx = _valueIndex[v];
		if (newKey == _heap[idx]->second) return;

		bool floatUp = _isHigher(newKey, _heap[idx]->second);
		_heap[idx]->second = newKey;
		if (floatUp) {
			_floatUp(idx);
		} else {
			_sink(idx);
		}
	}
	
	Value top() {
		assert(_heap.size() > 0);
		return _heap[0]->first;
	}
	
	Key topKey() {
		return _heap.size() > 0 ?_heap[0]->second : 0;
	}
	
	DynamicPriorityQueue(vector<Value> & vals, vector<Key> & keys, bool isMaxHeap) : 
			_comparer(new Comparer<Value, Key>(isMaxHeap)) {
//		assert(vals.size() > 0);
//		assert(vals.size() == keys.size());
		for (int i = 0; i < vals.size(); i++) {
			_heap.push_back(new pair<Value, Key>(vals[i], keys[i]));
		}
		sort(_heap.begin(), _heap.end(), *_comparer); // lazy mans way to heapify, could be done in O(n) if it mattered.
		//sort(_heap.begin(), _heap.end(), *boost::lambda::_1.first); 
		for (int i = 0; i < _heap.size(); i++) {
			_valueIndex[_heap[i]->first] = i;
		}
		_valueIndex.rehash(_valueIndex.size()); // i like to pretend this makes things faster
	}
	
	~DynamicPriorityQueue() {
		for (int i = 0; i < _heap.size(); i++) {
			if (_heap[i])
				delete _heap[i];
			_heap[i] = NULL;
		}
		delete _comparer;
	}
	
private:
	const Comparer<Value, Key> *_comparer;

	
	vector<pair<Value, Key> *> _heap;
	Mymap _valueIndex;
	
	// swaps the values at _heap:idx1 and idx2, updates _valueIndex accordingly
	void _swap(int idx1, int idx2) {
		assert(idx1 >= 0 && idx1 < _heap.size());
		assert(idx2 >= 0 && idx2 < _heap.size());
		
		// set _valueIndex
		int t = _valueIndex[_heap[idx1]->first];
		_valueIndex[_heap[idx1]->first] = _valueIndex[_heap[idx2]->first];
		_valueIndex[_heap[idx2]->first] = t;
		
		pair<Value, Key> *p = _heap[idx1]; 
		_heap[idx1] = _heap[idx2];
		_heap[idx2] = p;
	}
	
	// The value at idx may be too large and not satisfy the heap criteria; fix this.
	void _sink(int idx) {
		assert(idx >= 0 && idx < _heap.size());
		
		while (idx * 2 < _heap.size()) {
			int j = 2*idx;
			if (j + 1 < _heap.size() && _isHigherPair(_heap[j+1], _heap[j])) {
				j++;
			}
			
			// j is now idx of larger of children
			if (_isHigherPair(_heap[idx], _heap[j])) {
				break; // idx:value is larger than children
			}
			_swap(idx, j);
			idx = j;
		} 
	}
	
	void _floatUp(int idx) {
		assert(idx >= 0 && idx < _heap.size());
		
		while (idx > 0 && _isHigherPair(_heap[idx], _heap[idx/2])) {
			// bubble up value at idx.
			_swap(idx, idx / 2);
			idx = idx / 2;
		}
	}
	
	/* 
	 * Returns true if p1 should be higher up in queue than p2.
	 */
	static bool _isHigherPair(pair<Value, Key> *p1, pair<Value, Key> *p2) {
		assert(p1); 
		assert(p2);
		return _isHigher(p1->second, p2->second);
	}
	
	static bool _isHigher(Key k1, Key k2) {
		return k1 >= k2;
	}
};