/*
 *  gpu_interface.h
 *  gpu_hierleap
 *
 *  Created by David Orendorff on 10/12/10.
 *  Copyright 2010 UC Irvine. All rights reserved.
 *
 */

#ifdef WIN32
#include <CL/OpenCl.h>
#else
#include <OpenCl/OpenCL.h>
#endif

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>

#ifndef GPU_INTERFACE_
#define GPU_INTERFACE_

#define MAX_LOG_SIZE 100000

// Builds a device/context/cmd_queue and loads the program.
//
// Kernels are extracted via extractKernel(...) function.
class GpuInterface {
public:
	cl_context _context;
	cl_program _program;
	cl_device_id _device;
	cl_command_queue _cmd_queue;
	
	// prefered_cl_device is the prefered device to use, eg
	//		CL_DEVICE_TYPE_GPU or CL_DEVICE_TYPE_CPU
	// It may be null in which case CL_DEVICE_TYPE_GPU is used.
	// Note: CL_QUEUE_PROFILING_ENABLE is enabled.
	GpuInterface(cl_int prefered_cl_device, char* filename, bool enableProfiling) {
		cl_int err;
		if (!prefered_cl_device) prefered_cl_device = CL_DEVICE_TYPE_GPU;
		
		_device = getClDevice(prefered_cl_device);		
		
		_context = clCreateContext(0, 1, &_device, NULL, NULL, &err); 
		assert(err == CL_SUCCESS);
		
		_program = buildProgram(filename, _context, _device);
		_cmd_queue = clCreateCommandQueue(_context, _device, enableProfiling ? CL_QUEUE_PROFILING_ENABLE : 0, NULL);
	}
	
	virtual ~GpuInterface() {
		char *errorString = {"Warning, problem releasing an OpenCl resource, error code %d.\n"};
		
		cl_int err = clReleaseContext(_context);
		if (err != CL_SUCCESS) printf(errorString, err);
		err = clReleaseProgram(_program);
		if (err != CL_SUCCESS) printf(errorString, err);
		err = clReleaseCommandQueue(_cmd_queue);
		if (err != CL_SUCCESS) printf(errorString, err);
	}
	
	GpuInterface(cl_command_queue cmd_queue, char* filename) {
		_cmd_queue = cmd_queue;
		
		size_t retSize;
		float err = clGetCommandQueueInfo(_cmd_queue, CL_QUEUE_CONTEXT, sizeof(_context) , &_context, &retSize);
		assert(err == CL_SUCCESS);	
		assert(retSize == sizeof(_context));	
		
		err = clGetContextInfo(_context, CL_CONTEXT_DEVICES, sizeof(_device), &_device, &retSize);
		assert(err == CL_SUCCESS);
		assert(retSize == sizeof(_device));	
		
		_program = buildProgram(filename, _context, _device);
	}
	
	int maxWorkgroupSize() {
		size_t maxGroupSize;
		clGetDeviceInfo(_device, CL_DEVICE_MAX_WORK_GROUP_SIZE, sizeof(maxGroupSize), &maxGroupSize, NULL); 
		return maxGroupSize;										   
	}

	cl_context getContext() { return _context; }
	
	cl_command_queue getCmdQueue() { return _cmd_queue; }
	
	cl_kernel extractKernel(char *kernelName) {
		cl_int err;
		cl_kernel kernel = clCreateKernel(_program, kernelName, &err);
		assert(err == CL_SUCCESS);
		return kernel;
	}
	
private:
	
	// Tries to get a cl device specified by prefered_cl_device.
	// If unsuccesful CPU device is returned.
	cl_device_id getClDevice(cl_int prefered_cl_device) {
		cl_int err = 0;
		cl_device_id device;
		cl_platform_id platforms[5];
		cl_uint nPlatforms = 5;

		err |= clGetPlatformIDs(0 , NULL, &nPlatforms);
		assert(err == CL_SUCCESS);
		err |= clGetPlatformIDs(nPlatforms , platforms, &nPlatforms);
		assert(err == CL_SUCCESS);
		printf("%d platforms found.\n", nPlatforms);
		for (int i = 0; i < nPlatforms; i++) {
			char info[2000];
			cl_platform_info infos[5];
			infos[0] = CL_PLATFORM_PROFILE; infos[1] = CL_PLATFORM_VERSION; infos[2] = CL_PLATFORM_NAME;
			infos[3] = CL_PLATFORM_VENDOR; infos[4] = CL_PLATFORM_EXTENSIONS;
			for (int j = 0; j < 5; j++){
				err |= clGetPlatformInfo(platforms[i], infos[j], sizeof(char) * 2000, info, NULL);
				printf(info);
				assert(err == CL_SUCCESS);
			}
			printf("\n");
		}

		assert(err == CL_SUCCESS);
		err |= clGetDeviceIDs(platforms[0], prefered_cl_device, 1, &device, NULL);
		
		if (err != CL_SUCCESS) {
			err = clGetDeviceIDs(platforms[0], CL_DEVICE_TYPE_CPU, 1, &device, NULL);
			printf("Unable to load prefered Opencl device\n.");
		}
		assert(err == CL_SUCCESS);
		
		// Print out information about the device.
		// Get some information about the returned device
		cl_char vendor_name[1024] = {0};
		cl_char device_name[1024] = {0};
		//cl_char comp_units[1024] = {0};
		cl_int units[100] = {0};
		cl_int max_param_size[20] = {0};
		cl_int max_const_args[20] = {0};
		cl_int local_mem_size[20] = {0};
		cl_char extension_abilities[2056] = {0};
		cl_char driver_version[120] = {0};
		cl_int max_buffer_size[20] = {0};
		
		size_t returned_size;
		err = clGetDeviceInfo(device, CL_DEVICE_VENDOR, sizeof(vendor_name), 
							  vendor_name, &returned_size);
		err |= clGetDeviceInfo(device, CL_DEVICE_NAME, sizeof(device_name), 
							   device_name, &returned_size);
		err |= clGetDeviceInfo(device, CL_DEVICE_MAX_COMPUTE_UNITS, sizeof(units), 
							   units, &returned_size);
		err |= clGetDeviceInfo(device, CL_DEVICE_MAX_PARAMETER_SIZE, sizeof(max_param_size), 
							   max_param_size, &returned_size);
		err |= clGetDeviceInfo(device, CL_DEVICE_MAX_CONSTANT_ARGS, sizeof(max_const_args), 
							   max_const_args, &returned_size);
		err |= clGetDeviceInfo(device, CL_DEVICE_LOCAL_MEM_SIZE, sizeof(local_mem_size), 
							   local_mem_size, &returned_size);
		err |= clGetDeviceInfo(device, CL_DEVICE_EXTENSIONS, sizeof(extension_abilities), 
							   extension_abilities, &returned_size);
		err |= clGetDeviceInfo(device, CL_DRIVER_VERSION, sizeof(driver_version), 
							   driver_version, &returned_size);
		err |= clGetDeviceInfo(device, CL_DEVICE_MAX_MEM_ALLOC_SIZE, sizeof(max_buffer_size), 
							   max_buffer_size, &returned_size);
		

		
		assert(err == CL_SUCCESS);
		printf("Connecting to %s %s (%d cores)...\n", vendor_name, device_name, units[0]);
		printf("MaxParamSize(%d), MaxConstArgs(%d), LocalMemSize(%d), MaxBufferSize(%d), Extension abilities...\n%s\nDriver: %s\n", 
			   max_param_size[0], max_const_args[0], local_mem_size[0], max_buffer_size[0], extension_abilities, driver_version);
		
		return device;
	}
	
	cl_program buildProgram(char *filename, cl_context context, cl_device_id device) {
		char *program_source = load_program_source(filename);
		
		cl_int err;
		cl_program program = clCreateProgramWithSource(context, 1, (const char**)&program_source,
													   NULL, &err);
		free(program_source);
		assert(err == CL_SUCCESS);
		
		err = clBuildProgram(program, 0, NULL, NULL, NULL, NULL);
		
		if (err != CL_SUCCESS){
			// query build about the problem
			size_t logSize;
			
			char log[MAX_LOG_SIZE];
			
			printf("Error: Failed to build program executable from %s\n", filename);
			
			// get the details on the error, and store it in buffer
			clGetProgramBuildInfo(
								  program,              // the program object being queried
								  device,            // the device for which the OpenCL code was built
								  CL_PROGRAM_BUILD_LOG, // specifies that we want the build log
								  sizeof(log),       // the size of the buffer
								  log,               // on return, holds the build log
								  &logSize);                // on return, the actual size in bytes of the
			assert(logSize < MAX_LOG_SIZE);
			
			// write to stdout an exit.
			printf("%s\n", log);			
			exit(21);
		}
		return program;
	}
	
	
	char* load_program_source(const char *filename) { 		
		FILE *fh = fopen(filename, "r");
		if (fh == 0) {
			assert(false); // unable to load file.
			return 0; 
		}
		
		struct stat statbuf;
		stat(filename, &statbuf);
		size_t filesize = statbuf.st_size;
		char *source = (char *) malloc(filesize + 1);
		fread(source, filesize, 1, fh);
		source[filesize] = '\0'; 
		
		return source; 
	} 
	
};

#endif