#ifndef RANDOM_NUMBER_PROVIDER_THREAD_H_
#define RANDOM_NUMBER_PROVIDER_THREAD_H_

/*
 *  random_number_provider_thread.h
 *  gpu_hierleap
 *
 *  Created by David Orendorff on 10/19/11.
 *  Copyright 2011 UC Irvine. All rights reserved.
 *
 * 
 * Creates an independent thread to generate random numbers.
 * operator() will return the next random number. If a random number has
 * not yet been generated, the function will block until it can return a random number.
 *
 * This class is a drop in replacement for the boost random number generators.
 *
 * The thread will exist as long as the class instantiation exists.
 */

#include <boost/thread.hpp>

template<class Engine>
class RandomNumberProviderThread {
public:
	// at least this many threads may concurrently ask for random numbers; 
	// this enables us to safely bypass several annoying locking mechanisms
	static const int DEFAULT_BUFFER_SIZE = 65536;
	
	// if buffer is too small (more threads than buffer size) we may experience deadlock
	static const int MIN_BUFFER_SIZE = 1024; 
	
	typedef typename Engine::result_type result_type;

//	typedef Engine engine_type;
//	
//	bool has_fixed_range = Engine::has_fixed_range;
//	result_type min_value = Engine::min_value;
//	result_type max_value = Engine::max_value;
	Engine _engine;
//	
	RandomNumberProviderThread() {
		_engine = Engine();
		_initThreadAndQueue(DEFAULT_BUFFER_SIZE);
	}
	
	explicit RandomNumberProviderThread(unsigned int seed) {
		_engine = Engine(seed);
		_initThreadAndQueue(DEFAULT_BUFFER_SIZE);
	}
	
	explicit RandomNumberProviderThread(unsigned int seed, unsigned int bufferSize) {
		_engine = Engine(seed);
		_initThreadAndQueue(bufferSize);
	}
	
	// Return (and "consume") next generated random number
	result_type operator()() {
		__sync_add_and_fetch(&_consumptionsInProgress, 1);
		//_consumptionsInProgress++;
		int rnIdx = __sync_add_and_fetch(&_consumerIdx, 1);
		//int rnIdx = _consumerIdx; _consumerIdx++;
		result_type val = _retrieveRandomNumber(rnIdx);
		__sync_fetch_and_sub(&_approxNumRns, 1);
		//_approxNumRns--;
		__sync_sub_and_fetch(&_consumptionsInProgress, 1);
		//_consumptionsInProgress--;
		return val;
	}
	
	// note: this may not be thread safe, it 
	// depends on the random number generating engine 
	void seed(long x) {
		_engine.seed(x);
	}
	
	result_type min() const {
		return _engine.min();
	}
	
	result_type max() const {
		return _engine.max();
	}
	
	static bool validation(result_type x) {
		return Engine::validation(x);
	}
	
private:
	boost::thread _generatingThread;
	boost::mutex _uselessMutex;
	boost::condition_variable _randomNumbersProduced;
	boost::condition_variable _randomNumbersConsumed;
	int _producerIdx;
	int _consumerIdx;
	int _approxNumRns; // lower bound on random number available
	int _consumptionsInProgress; // upper bound on number of consumers simultanously aquiring a random number
	vector<result_type> _random_numbers;
	
	void _initThreadAndQueue(int bufferSize) {
		assert(bufferSize > MIN_BUFFER_SIZE); 
		
		_approxNumRns = 0;
		_consumerIdx = 0;
		_consumptionsInProgress = 0;
		_random_numbers = vector<result_type>(bufferSize, NULL);
		
		_producerIdx = 0;
		_random_numbers[_producerIdx] = _engine();
		
		//_generating_thread = boost::thread(&_generateRandomNumbers, boost::ref(_engine), _random_numbers);
		_generatingThread = boost::thread(&_generateRandomNumbers, this);
	}
	
	int _rnIndexInRange(int idx) {
		return _rnIndexInRange(idx, _random_numbers.size());
	}
	
	// if idx is out of range, return idx % arrayLen
	static int _rnIndexInRange(int idx, int arrayLen) {
		return idx % arrayLen;
	}
	
	// Returns true if for every _consumptionsInProgress there
	// exists a randomly generated number.
	// 
	// This uses upper bounds to determine if another random 
	// number is available. It will never return true if there
	// are not enough random numbers available. However, it may 
	// return false even if there are random numbers available
	bool _isRandomNumbersAvailable() {
		return _approxNumRns > _consumptionsInProgress;	
	}
	
	// pre: this thread is uniquely accessing the random number at pos; 0 <= pos < n
	// post: returns a random number
	// note: if the rn at pos has not yet been generated we need to wait on that condition
	result_type _retrieveRandomNumber(int pos) {
		if (_isRandomNumbersAvailable()) {
			// if about 1/3 of rns queue is left we're starting to run low on random numbers
			// notify generating thread as such
			if (_random_numbers.size() * 1 / 3 > (_approxNumRns - _consumptionsInProgress)) {
				_randomNumbersConsumed.notify_one();
			}
			
			return _random_numbers[_rnIndexInRange(pos)];
		} else {
			// we need to wait for random numbers to be generated before continuing
			{
				boost::mutex::scoped_lock uselessLock(_uselessMutex);
				// I don't see a reason for this lock other than the api for conditional_var requirement
				_randomNumbersProduced.wait(uselessLock);
			}
			return _retrieveRandomNumber(pos);
		}
	}
	
	/**
	 * This function is run continuously in another thread until *this* class obj is destroyed.
	 * Unfortionatly, threads may only be executed from a static function, so we pass in *this*.
	 *
	 * Attributes:
	 *		- this function may ONLY write to: randomNumbers, producerIdx, lastConsumedIdx, approxNumRns
	 */
	static void _generateRandomNumbers(RandomNumberProviderThread *thiis) {
		//, Engine engine, vector<result_type> & randomNumbers, 
//									   int & lastConsumedIdx, int & producerIdx, int & consumptionsInProgress,
//									   int & approxNumRns, 
//									   boost::condition_variable new_random_numbers_exist, 
//									   boost::condition_variable reduced_number_of_rns) 
		int n = thiis->_random_numbers.size();
		boost::mutex uselessMutex;
		
		printf("\n----in generating rng thread\n----\n");
		
		// for *this* function we can assume: 0 <= producerIdx - (consumerIdx+consumptionsInProgress) < n
		while (true) {
			assert(thiis->_approxNumRns < n);
			assert(thiis->_approxNumRns >= 0);
			
			int nextProducerIdx = thiis->_producerIdx + 1;
			if (nextProducerIdx - (thiis->_consumerIdx - thiis->_consumptionsInProgress) >= n) {
				{
					// the queue is full and so we wait until some random numbers have been used.
					boost::mutex::scoped_lock uselessLock(uselessMutex);
					thiis->_randomNumbersConsumed.wait(uselessLock);
				}
			} else {
				thiis->_random_numbers[_rnIndexInRange(nextProducerIdx, n)] = thiis->_engine();

				thiis->_producerIdx = nextProducerIdx;
				__sync_fetch_and_add(&thiis->_approxNumRns, 1);
				thiis->_randomNumbersProduced.notify_all(); // maybe don't call this on every iteration? or notify_one ?
			}
			
			// keep indecies from away integer overflow
			if (thiis->_consumerIdx > INT_MAX / 2) {
				int subN = (thiis->_consumerIdx / n) * n;
				{
					// i wish we could easily do this block in completely atomic execution
					boost::this_thread::disable_interruption di;
					__sync_sub_and_fetch(&thiis->_consumerIdx, subN);
					__sync_sub_and_fetch(&thiis->_producerIdx, subN);
				}
			}
			//_generating_thread.inturuption_point();
		}
	}
};

#endif